#ifndef __AfhFM_H__
#define __AfhFM_H__
/*
 *   AfhFM.h
 *
 *   Copyright 1997-2014 AFH Systems
 *
 *   FilterMeister is free software. You can redistribute and modify it
 *   under the terms of the MPL2 License.
 *   See: https://www.mozilla.org/en-US/MPL/2.0/
 *   Any changes to the FilterMeister files have to be published or sent
 *   to harry@thepluginsite.com. The filter plugin code itself does not
 *   need to be disclosed.
 *
 *  22AUG2013  Alex Hunter	1. Redefine 'class' as 'klass' to avoid a problem
 *							   with the VS IDE editor.
 *  08NOV2013  Alex Hunter  1. Add X64GEN and USE_SSE controls (delete OPTBEV).
 *  30DEC2013  Alex Hunter  1. Add MAX_SCALE_FACTOR to define max proxy reduction.
 *  03JAN2014  Alex Hunter  1. Add copyRect32, scaleRect32, shrinkRect32.
 *  05JAN2014  Alex Hunter  1. Add missing prototypes for roundUpScaleFactor and 
 *							   roundDownScaleFactor.
 *	08JAN2014  Alex Hunter	1. Add a modifed version of FilterBigDocument.h
 *							   from the googlecode WinGP project.
 *  11JAN2014  Alex Hunter  1. Make changes for 32-bit image coordinates.
 *							2. Add prototypes from FilterBigDocument.h.
 *	20JAN2014  Alex Hunter  1. Move formatted print routine macros to AfhFM.h.
 *	22MAR2014  Alex HUnter  1. Double the size of CB_initCode, CB_onFilterStart,
 *							   and CB_onFilterEnd (since x64 generates more code
 *							   bytes than x32 did).
 *	13JUN2014  Alex Hunter  1. Change association of user_vars_init_flags
 *							   to 1 bit per byte for X64.
 */

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0400 //need Windows 4.0 or greater (for TryEnterCriticalSection, etc.)
#endif

#include <windows.h>
#include <commctrl.h>       //for UDACCEL
//#include "PITypes.h"
//#include "PIGeneral.h"
#include "PIFilter.h"
#include <setjmp.h>         //for jmp_buf

#ifdef X64
///typedef __int64				INT64;
typedef unsigned __int64		QWORD;
#endif

#include "EmitX86.h"        //for CB


/**************************** Utility Macros *********************************/

// The Visual Studio IDE treats 'class' as a reserved word, even
// when compiling pure C code.  This confuses the editor and spoils
// popup hovering to see the definition of ids in the editor.  The
// following kludge mitigates this problem (but can be turned off
// later if it causes other unanticipated problems; e.g., in the
// debugger.

#define class klass

/******************* Version Number Macros ***********************************/

// An FM verson number is a positive int comprised of two short int's,
// with the (positive) major value in the high-order 16 bits and the
// (unsigned) minor value in the low-order 16 bits of the int.
#define VERSION(major, minor) ( (major & 0x7fff) << 16 | (minor & 0xffff) )

#define FML_VERSION VERSION(nProgRevMajor, nProgRevMinor)

/******************* Formatted print routines for FM *************************/

void FMprint(const char *buff);

#define PRINTF(f) FMprint(f)

#ifdef FULL

static char buff[1024];		//POSSIBLE OVERFLOW??

#if 1
#define PRINTF2(f,a) ( sprintf(buff, f, a), FMprint(buff) )
#else
#define PRINTF2(f,a) \
    do {\
        sprintf(buff, f, a);\
        FMprint(buff);\
    } while(0)
#endif

#define PRINTF3(f,a,b) \
    do {\
        sprintf(buff, f, a, b);\
        FMprint(buff);\
    } while(0)

#define PRINTF4(f,a,b,c) \
    do {\
        sprintf(buff, f, a, b, c);\
        FMprint(buff);\
    } while(0)

#define EMIT0(f) \
    do {\
        if ((gParams->flags & (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) == \
                              (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) {\
            int n=0;\
            n += sprintf(&buff[n], (f));\
            n += sprintf(&buff[n], "\r\n");\
            FMprint(buff);\
        }\
    } while(0)

#define EMIT(f) \
    do {\
        if ((gParams->flags & (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) == \
                              (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) {\
            int n;\
            n = sprintf(buff, "  ");\
            n += sprintf(&buff[n], (f));\
            n += sprintf(&buff[n], "\r\n");\
            FMprint(buff);\
        }\
    } while(0)

#define EMIT1 EMIT

#define EMIT2(f,a) \
    do {\
        if ((gParams->flags & (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) == \
                              (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) {\
            int n;\
            n = sprintf(buff, "  ");\
            n += sprintf(&buff[n], (f), (a));\
            n += sprintf(&buff[n], "\r\n");\
            FMprint(buff);\
        }\
    } while(0)

#define EMIT2L(f,a) \
    do {\
        if ((gParams->flags & (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) == \
                              (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) {\
            int n;\
            n = sprintf(buff, (f), (a));\
            n += sprintf(&buff[n], "\r\n");\
            FMprint(buff);\
        }\
    } while(0)

#define EMIT3(f,a,b) \
    do {\
        if ((gParams->flags & (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) == \
                              (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) {\
            int n;\
            n = sprintf(buff, "  ");\
            n += sprintf(&buff[n], (f), (a), (b));\
            n += sprintf(&buff[n], "\r\n");\
            FMprint(buff);\
        }\
    } while(0)

#define EMIT3L(f,a,b) \
    do {\
        if ((gParams->flags & (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) == \
                              (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) {\
            int n;\
            n = sprintf(buff, (f), (a), (b));\
            n += sprintf(&buff[n], "\r\n");\
            FMprint(buff);\
        }\
    } while(0)

#define EMIT4(f,a,b,c) \
    do {\
        if ((gParams->flags & (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) == \
                              (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) {\
            int n;\
            n = sprintf(buff, "  ");\
            n += sprintf(&buff[n], (f), (a), (b), (c));\
            n += sprintf(&buff[n], "\r\n");\
            FMprint(buff);\
        }\
    } while(0)

#define EMIT5(f,a,b,c,d) \
    do {\
        if ((gParams->flags & (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) == \
                              (AFH_PSEUDOCODE|AFH_EDITMODE|AFH_ADVANCED)) {\
            int n;\
            n = sprintf(buff, "  ");\
            n += sprintf(&buff[n], (f), (a), (b), (c), (d));\
            n += sprintf(&buff[n], "\r\n");\
            FMprint(buff);\
        }\
    } while(0)

#else
#define EMIT(f)
#define EMIT2(f,a)
#define EMIT2L(f,a)
#define EMIT3(f,a,b)
#define EMIT3L(f,a,b)
#define EMIT4(f,a,b,c)
#define EMIT5(f,a,b,c,d)
#endif //ifdef FULL

/*****************************************************************************/

#if 1	//thread-safe
#define THREAD_LOCAL __declspec(thread)
#else
#define THREAD_LOCAL
#endif

/*****************************************************************************/
	// Code generation configuration.

#define isFalse(x) (!(x))
#define isTrue(x) (!!(x))

// Macros to determine what machine we are generating code for.
#define X86GEN isFalse(gParams->flags & AFH_X64GEN)
#define X64GEN isTrue(gParams->flags & AFH_X64GEN)

/*****************************************************************************/

// Flag bit definitions for gParams->flags.

#define AFH_TRACE                   0x00000001
#define AFH_USE_MMX                 0x00000002
#define AFH_USE_ADVANCESTATE        0x00000004
#define AFH_DRAGGING_PROXY          0x00000008
#define AFH_USE_HOST_PADDING        0x00000010
#define AFH_INIT_OPTIMAL_ZOOM       0x00000020
#define AFH_PROXY_BIG_GULP          0x00000040
#define AFH_DRAG_FILTERED_OUTPUT    0x00000080
#define AFH_FLUSH_ADVANCESTATE      0x00000100
#define AFH_ZERO_OUTPUT_DRAG        0x00000200
#define AFH_SUPPORT_RGB             0x00000400
#define AFH_SUPPORT_GRAY            0x00000800
#define AFH_SUPPORT_CMYK            0x00001000
#define AFH_SUPPORT_LAB             0x00002000
#define AFH_SUPPORT_INDEXED         0x00004000
#define AFH_SUPPORT_BITMAP          0x00008000
#define AFH_SUPPORT_DUOTONE         0x00010000
#define AFH_SUPPORT_MULTI           0x00020000
///#define AFH_OPTIMIZE_BEVEL          0x00040000	//Obsolete
#define AFH_DISASSEMBLE             0x00080000
#define AFH_DUMP_SDL                0x00100000
#define AFH_PSEUDOCODE              0x00200000
#define AFH_USE_CMOV                0x00400000
#define AFH_USE_INLINES             0x00800000
#define AFH_IS_TILEABLE             0x01000000
#define AFH_EDITMODE                0x02000000
#define AFH_ADVANCED                0x04000000
//#define AFH_QUICK_TOGGLE          0x08000000
#define AFH_USE_SSE					0x10000000
#define AFH_X64GEN					0x20000000

// Flag bit definitions for control properties.

#define CTP_TRACK                   0x00000001
#define CTP_READONLY                0x00000002
#define CTP_HORZ                    0x00000004
#define CTP_VERT                    0x00000008
#define CTP_ORIENT_MASK             (CTP_HORZ|CTP_VERT)
#define CTP_TOP                     0x00000010
#define CTP_BOTTOM                  0x00000020
#define CTP_LEFT                    0x00000040
#define CTP_RIGHT                   0x00000080
#define CTP_SIDE_MASK               (CTP_TOP|CTP_BOTTOM|CTP_LEFT|CTP_RIGHT)
#define CTP_EXTENDEDUI              0x00000100
#define CTP_MOUSEOVER               0x00000200
#define CTP_CONTEXTMENU             0x00000400
#define CTP_DRAWITEM                0x00000800
#define CTP_MOUSEMOVE               0x00001000
#define CTP_PREVIEWDRAG             0x00002000


// Should add a version # to Parameters, and ignore old params if version #
// mismatch!!!

#define MAX_LABEL_SIZE 1023
#define MAX_TOOLTIP_SIZE 127

#if 0
#define MAX_SCALE_FACTOR 10000000	//max proxy reduction to reduce 2Gpx to 200px 
#else
		///////////////// OOPS. For now scaleFactor can't exceed 30,000 (or 32,767).
		/////////////////       This limit is imposed because inputRate and MaskRate
		/////////////////		are fixed point 16.16 numbers, and scaleFactor must
		/////////////////		fit in the high (signed) 16 bits!  This is a PS limitation,
		/////////////////		even in CS5.1 X64.  So if scaleFactor is <= 30000 and
		/////////////////		we want to scale an image to fit entirely within a 300 x 300 px
		/////////////////		frame, then no image dimension should exceeed 300px * 30000 =
		/////////////////		9,000,000 px, or approxmately 10Mpx.  Of course, this is well
		/////////////////		within the (published) CS5 limit of 300,000 x 300,000 px.
#define MAX_SCALE_FACTOR 30000	//max proxy reduction for now
#endif

// Number of FM Arrays
#define N_ARRAYS 100
#define IS_OK_ARRAY_NR(nr) ((UINT_PTR)(nr) < N_ARRAYS)			//note uint also catches < 0
#define IS_ALLOCATED_ARRAY(nr) (IS_OK_ARRAY_NR(nr) && Array[nr])

//Test for a power of 2, for use in index masking...
#define IS_POWER_OF_TWO(n) (((n) & -(n)) == (n))

//number of anonymous put/get cells
#define N_CELLS 1024
#if !IS_POWER_OF_TWO(N_CELLS)
#error N_CELLS must be a power of two!
#endif
#define IS_OK_CELL_NR(n) ((UINT_PTR)(n) < N_CELLS)				//note uint also catches < 0
#define IS_VALID_CELL(n) IS_OK_CELL_NR(n)

//number of dialog controls
#define N_CTLS 256
#if !IS_POWER_OF_TWO(N_CTLS)
#error N_CTLS must be a power of two!
#endif
#define IS_OK_CTL_NR(n) ((UINT_PTR)(n) < N_CTLS)					//note uint also catches < 0
#define IS_INUSE_CTL(n) (IS_OK_CTL_NR(n) && gParams->ctl[n].inuse)

//predefined control indices
#define CTL_OK          (N_CTLS - 1)
#define CTL_CANCEL      (N_CTLS - 2)
#define CTL_EDIT        (N_CTLS - 3)
#define CTL_LOGO        (N_CTLS - 4)
#define CTL_PREVIEW     (N_CTLS - 5)
#define CTL_PROGRESS    (N_CTLS - 6)
#define CTL_HOST        (N_CTLS - 7)
#define CTL_ZOOM        (N_CTLS - 8)
#define CTL_FRAME       (N_CTLS - 9)
#define CTL_BACKGROUND  (N_CTLS - 10)

//last available user control index (with room to spare...)
#define CTL_LAST_USER   (N_CTLS - 11)

//predefined control IDs
#define IDX_OK          (IDC_CTLBASE + CTL_OK)
#define IDX_CANCEL      (IDC_CTLBASE + CTL_CANCEL)
#define IDX_EDIT        (IDC_CTLBASE + CTL_EDIT)
#define IDX_LOGO        (IDC_CTLBASE + CTL_LOGO)
#define IDX_PREVIEW     (IDC_CTLBASE + CTL_PREVIEW)
#define IDX_PROGRESS    (IDC_CTLBASE + CTL_PROGRESS)
#define IDX_HOST        (IDC_CTLBASE + CTL_HOST)
#define IDX_ZOOM        (IDC_CTLBASE + CTL_ZOOM)
#define IDX_FRAME        (IDC_CTLBASE + CTL_FRAME)

#ifndef BS_TYPEMASK
#define BS_TYPEMASK 0x0000000F  //missing from some versions of <Winuser.h>!!!
#endif
#define CBS_TYPEMASK 0x00000003

#ifndef TCS_FLATBUTTONS
#define TCS_FLATBUTTONS 0x0008  //missing from some versions of <commctrl.h>
#endif

// Control classes...
typedef enum {
    CC_UNUSED,
    CC_STANDARD,
    CC_SCROLLBAR,
    CC_TRACKBAR,
    CC_SPINNER,
    CC_UPDOWN,
    CC_PUSHBUTTON,
    CC_CHECKBOX,
    CC_RADIOBUTTON,
    CC_GROUPBOX,
    CC_OWNERDRAW,
    CC_LISTBOX,
    CC_COMBOBOX,
    CC_SLIDER,
    CC_PROGRESSBAR,
    CC_EDIT,
    CC_STATICTEXT,
    CC_FRAME,
    CC_RECT,
    CC_IMAGE,
    CC_BITMAP,
    CC_ICON,
    CC_METAFILE,
    CC_TOOLTIP,
    CC_ANIMATION,
    CC_PREVIEW,
    CC_ZOOM,
	CC_TAB
} CTLCLASS;

// Default control actions...
typedef enum {
    CA_NONE,
    CA_CANCEL,
    CA_APPLY,
    CA_PREVIEW,
    CA_EDIT,
    CA_ABOUT,
    CA_RESET,
    CA_SIZE
} CTLACTION;

// FM Event codes...
typedef enum {
    FME_UNKNOWN,
    FME_ZERODIVIDE,
    FME_DIVIDEOVERFLOW,
    FME_CLICKED,
    FME_DBLCLK,
    FME_PAGEUP,
    FME_PAGEDOWN,
    FME_LINEUP,
    FME_LINEDOWN,
    FME_MOUSEOVER,
    FME_MOUSEOUT,
    FME_MOUSEMOVE,
    FME_LEFTCLICKED_DOWN,
    FME_LEFTCLICKED_UP,
    FME_RIGHTCLICKED_DOWN,
    FME_RIGHTCLICKED_UP,
    FME_TIMER,
    FME_ZOOMCHANGED,
    FME_KEYDOWN,
    FME_KEYUP,
    FME_CHANGED,//FME_VALUECHANGED,
    FME_SIZE,
    FME_ENTERSIZE,
    FME_EXITSIZE,
    FME_DRAWITEM,
    FME_CUSTOMEVENT,
    FME_COMBO_DROPDOWN,
    FME_COMBO_CLOSEUP,
    FME_CONTEXTMENU,
    FME_SETFOCUS,//FME_SETEDITFOCUS,
    FME_KILLFOCUS,//FME_KILLEDITFOCUS,
    FME_PREVIEWDRAG,
	FME_CANCEL,
	FME_INIT,
    FME_READSCRIPT,
    FME_WRITESCRIPT

} FME;


// Dialog background image modes...
typedef enum {
    DIM_EXACT=0,
    DIM_TILED=1,
    DIM_STRETCHED=2
} DLGIMAGEMODE;


// Size of double literal pool (also contains string literal pool for now)
#define MAX_DLITS 8192//4096

// Maximum number of (4-byte) locals in an FM function (for now)
#define MAX_LOCALS 1000//400

// Maximum number of (4- or 8-byte) temps (pseudo-registers) in an FM function
// (=depth of expression evaluation stack).
#define MAX_TEMPS 20

// Maximum number of case labels in a switch statement
#define MAX_CASE_LABELS 256

// Size of Semantic Descriptor List (SDL)
#define NSDL 300000//90000

#ifdef USE_EDIT_CONTROL
#define MAX_SOURCE_CODE_SIZE 31999 //more than enough?--default hard limit in Edit
                                   //control seems to be exactly 30,000.
#else
#define MAX_SOURCE_CODE_SIZE 799999//399999//199999 //for now
#endif //USE_EDIT_CONTROL

// Maximum number of user-declared global/static (4-byte) variables.
#define MAX_GLOBAL_VARS 10000

#define ANCHOR_LEFT		1
#define ANCHOR_RIGHT	2
#define ANCHOR_TOP		4
#define ANCHOR_BOTTOM	8


    /* Predefined global variables, saved across filter invocations */
typedef struct tagPREDEF_GLOBALS
{
    int i0, i1, i2, i3, i4, i5, i6, i7, i8, i9;    //WARNING: i0, i1 are Premiere vars!
    int j0, j1, j2, j3, j4, j5, j6, j7, j8, j9;
    int k0, k1, k2, k3, k4, k5, k6, k7, k8, k9;
    double x0, x1, x2, x3, x4, x5, x6, x7, x8, x9;
    double y0, y1, y2, y3, y4, y5, y6, y7, y8, y9;
    double z0, z1, z2, z3, z4, z5, z6, z7, z8, z9;
    char str0[256];
    char str1[256];
    char str2[256];
    char str3[256];
    char str4[256];
    char str5[256];
    char str6[256];
    char str7[256];
    char str8[256];
    char str9[256];
    char str10[256];
    char str11[256];
    char str12[256];
    char str13[256];
    char str14[256];
    char str15[256];
    char str16[256];
    char str17[256];
    char str18[256];
    char str19[256];
	//space for user-declared global (and static) variables, assuming 4 bytes per variable
    int user_vars[MAX_GLOBAL_VARS];
	//one flag bit per global/static byte; bit=1 if associated var has been initialized	
    BYTE user_vars_init_flags[(MAX_GLOBAL_VARS*sizeof(int)+7) / 8];
} PREDEF_GLOBALS;

typedef struct TParameters
{
    short version_major, version_minor;
    unsigned long flags;
    ///short bandWidth;
    short queryForParameters;
    struct
    {
        CTLCLASS class;
        HWND hCtl;
        HWND hBuddy1, hBuddy2;
        int xPos, yPos;
        int width, height;
        int editwidth, editheight;
        int initVal, val, defval, minval, maxval, lineSize, pageSize, thumbSize;
        int selStart, selEnd;
        int style, styleEx;
		int buddy1Style, buddy1StyleEx;
		int buddy2Style, buddy2StyleEx;
        int properties;

        int divisor;
        int gamma;

        CTLACTION action;
        HBRUSH hBrush;     //background brush
        COLORREF bkColor;   //background color; -1=transparent
        COLORREF textColor;
        HFONT hFont;
        int fontSize;
        int iParent, leftBuddy, rightBuddy;
        //int id;
        char label[MAX_LABEL_SIZE+1];
        char tooltip[MAX_TOOLTIP_SIZE+1];
        char image[_MAX_PATH+1];
        int  imageType; // 'B' = bitmap, 'W' = Windows metafile, etc.
        char enabled;   //0=invisible/disabled, 1=visible/disabled, 3=visible/enabled
        char inuse;

		int theme;
		int tabControl;
		int tabSheet;
		int anchor;
		int distWidth;
		int distHeight;

        int scripted;

        union { //class-specific variants
            struct {    //standard scrollbar-specific
                int w2, h2;     //width and height of static text buddy (in pixels)
            } sb;
            struct {    //pushbutton-specific
                char pushState; //pushed...
            } pb;
            struct {    //trackbar-specific
                int ticFreq;
                int tics[8];        //trackbar
            } tb;
            struct {    //up-down-specific
                int nBase;      // for buddy
                int nInc;
                UDACCEL aAccel[4];  //up-down
            } ud;
            struct {    //tooltip-specific
                int iDelay;     //tooltip
            } tt;
            struct {    //image-specific
                        BITMAP bmInfo;
                        HDC hdcMem;
                        HDC hdcAnd;
                        HBITMAP hbm;
                        HBITMAP hbmOld;
                        HBITMAP hbmAnd;
                        HBITMAP hbmOldAnd;
            } im;
        };
    } ctl[N_CTLS];
    BOOL fValidParse;
    BOOL fValidCompile;
#ifndef STANDALONE
    BOOL fGotSourceCode;
    BOOL fNeedSourceSave;
    unsigned int SDL[NSDL];        /* The semantic descriptor list (sized elsewhere). */
    char szSrcCode[MAX_SOURCE_CODE_SIZE+1];
    CB cbCodeBuffer;                //Save these CB's across invocations,
    CB cbInitCode;                  //for use by Make.
    CB cbDivCheck;
    CB cbZeroDivide;
    CB cbDivideOverflow;
    CB cbOnCtl;
    CB cbOnFilterStart;
    CB cbForEveryTile;
    CB cbOnFilterEnd;
    CB cbUserFunctions;
    int gUserFunctions;
#endif //ifndef STANDALONE				//YIKES! Standalone allocates FULL size code buffers???!!
    BYTE codeBuffer[256000];			//<doOneRow_part1>,<ChannelCode>,<doOneRow_part2>
    BYTE CB_initCode[256000];			//was 128000
    BYTE CB_onZeroDivide[32000];
    BYTE CB_onDivideOverflow[32000];
    BYTE CB_divCheck[128];              //invariant????
    BYTE CB_onCtl[512000];
    BYTE CB_onFilterStart[128000];		//was 64000
    BYTE CB_forEveryTile[512000];
    BYTE CB_onFilterEnd[128000];		//was 64000
    BYTE CB_UserFunctions[256000];
#if 0
    //not used at present...
    BYTE CB_onInitFilter[64000];
#endif
    // Preserve dialog configuration
    BOOL     gInitialInvocation;
    BOOL     gCtlImagesSet;
    int      gImagesEmbedded;
    BOOL     gWavesSpecified;
    int      gWavesEmbedded;
    HRGN     gDialogRegion;
    int      gDialogStyle;
    int      gDialogInitialStyle;
    int      gDialogStyleEx;
    int      gDialogInitialStyleEx;
    int      gDialogDragMode;
    RECT     gDialogRect;
    COLORREF gDialogGradientColor1;
    COLORREF gDialogGradientColor2;
    int      gDialogGradientDirection;
    int      gDialogImageMode;
    int      gDialogImageStretchMode;
    char    *gDialogImage_save;
    char     gDialogImageFileName[_MAX_PATH+1];
    char     gDialogTitle[MAX_LABEL_SIZE+1];
	int      gDialogTheme;
	int      gDialogEvent;
    RECT     gEditorRect;
    HKEY     gRegRoot;
    char     gRegPath[MAX_PATH+1];
    // Preserve the following initialized storage from the context record...
    int need_iuv;
    int need_d;
    int need_m;
    int need_tbuf;
    int need_t2buf;
	int need_t3buf;
	int need_t4buf;
    int needPremiere;
    int needPadding;
    int bandWidth;
#ifndef DONT_SAVE_GLOBALS
    /* Save the following predeclared global and user-declared global/static variables */
    PREDEF_GLOBALS globvars;     //renamed from 'globals' to 'globvars' to avoid lexical ambiguity with other 'globals'!!!
#endif //DONT_SAVE_GLOBALS
    /* Public FF+ built-in strings and arrays */
    char filterAuthorText[256];
    char filterCategoryText[256];
    char filterTitleText[256];
    char filterFilenameText[256];
    char filterCopyrightText[256];
    char filterDescriptionText[256];
    char filterOrganizationText[256];
    char filterURLText[256];
    char filterVersionText[256];
    char filterAboutText[256];      //temporary
    char filterUniqueID[37];
    /* Double-precision floating-point literal pool. */
    int ndlits;
    double dlitPool[MAX_DLITS];
#ifndef STANDALONE
    // Array to hold resource embedding info...
    int cbEmbeds;       //Number of chars used in embeds array
    char embeds[1024];
    // PiPL contents for AfhFMmake...
    FilterCaseInfo fici[7];
    int SupportedModes;
    BOOL EI_specified;      //true iff enable info was specified
    int  EI_start_index;    //starting index of enable info string
    int  EI_count;          //number of chars in enable info string
#endif
} TParameters, *PParameters, **HParameters;

/*****************************************************************************/

typedef struct Globals
{
    short result;
    Rect    proxyRect;
    FilterRecord *filterParamBlock;
} Globals, *GPtr, **GHdl;
    
#define gResult ((*MyGlobals).result)
#define gStuff  ((*MyGlobals).filterParamBlock)
#define gProxyRect  ((*MyGlobals).proxyRect)
#define gParams  ((PParameters) *gStuff->parameters)

/*****************************************************************************/

typedef struct {
    WORD   wLength;
    WORD   wValueLength;
    WORD   wType; 
    WCHAR  szKey[1];
    //WORD  Padding1[];
    //VS_FIXEDFILEINFO Value;
    //WORD  Padding2[]; 
    //WORD  Children[];
} VS_BLOCK_HEADER;

/*****************************************************************************/

/* Global variables */
GPtr const MyGlobals;
void *MyAdjustedMaskData;
void *MyAdjustedInData;
HWND MyHDlg;    // For use with MessageBox() -- will default to NULL before dialog is init'ed.
HWND ghTT;      //tooltip control handle
HANDLE hDllInstance;
char *MDP_looking_for;

char    *gDialogImage;	//THREAD-SAFE??

char szTitleName[_MAX_FNAME + _MAX_EXT + 1];

#ifndef STANDALONE
extern char *defaultAboutText;
#endif

extern BOOL bAuthorized;
extern DWORD dwAuthorized2;
extern DWORD dwAuthorized3;
extern DWORD RegCodeFingerPrint[4];

extern int FM_is_busy;

extern BOOL fScaling;

extern BOOL scaleLock;

#ifdef SCRIPTABLE
    extern int ScriptParamRead;
    extern int ScriptParam[256];
    extern BOOL ScriptParamUsed[256];
#endif

extern int EditBoxUpdate;
extern int MouseOverWhenInvisible;
extern int ClickDrag;
extern POINT imageOffset;

MINMAXINFO MaxMinInfo;
extern int sizeGrip;

extern HHOOK KeyHook;

typedef union UIF 
{
	unsigned int	i;
	float		f;
} UIF; 

//extern int cancelEventPerformed;

/* Global variables for code generation */
int X64gen;			//nonzero when generating x64 code
int X86gen;			//nonzero when generating x86 code
int nTmps;			//number of temp registers
int nVolatileTmps;	//number of Volatile temp registers
int nProgRevMajor;	//FML program major revision level
int nProgRevMinor;	//FML program minor revision level

/* Mnemonics for opcodes that depend on native machine word size.   */
/* (The _V suffix means 'variable'; i.e., '_D' or '_Q', depending   */
/* on native machine word size -- 32-bit or 64-bit, resp.)          */
int LEA_V;
int ADDR_V;
int ADD_V;
int ADDI_V;
int ORR_V;
int OR_V;
int ADCR_V;
int ADC_V;
int SBBR_V;
int SBB_V;
int ANDR_V;
int AND_V;
int ANDI_V;
int SUBR_V;
int SUB_V;
int SUBI_V;
int XORR_V;
int XOR_V;
int CMPR_V;
int CMP_V;
int MOVR_V;
int MOV_V;
int XCHG_V;
int TEST_V;
int INC_V;
int DEC_V;
int IMUL32_V;
int IMULI_V;
int IDIV_V;
int CDQ_V;
int ROL1_V;
int ROR1_V;
int SHL1_V;
int SHLI_V;
int SAR1_V;
int SARI_V;
int FILD_V;
int FISTP_V;
int NEG_V;
int NOT_V;
int MOVSXD_V;	//MOV_D (32) or MOVSXD_Q (64)
int MOVSXB_V;
int MOVSXW_V;
int SHL_V;
int SHR_V;
int CMPI_V;

/* Parser interface routines */
int Mdp_parse(char *in, unsigned int flags, unsigned int *sdlp, unsigned int nsdl);
void Mdp_dumpSDL(const char line[]);
char *Mdp_syntaxErr(void);
void FMprint(const char *buff);

/* Message Box routines */
int msgBox(UINT uType, LPCTSTR lpCaption, LPCTSTR lpFmt,...);
int fm_printf(LPCTSTR lpFmt,...);
int Info(LPCTSTR lpFmt,...);
int Warn(LPCTSTR lpFmt,...);
int Error(LPCTSTR lpFmt,...);
int ErrorOk(LPCTSTR lpFmt,...);
int YesNo(LPCTSTR lpFmt,...);
int YesNoCancel(LPCTSTR lpFmt,...);

/* Function prototypes */
void DoAbout (GPtr globals, short selector);
void scaleRect(Rect *l, long n, long d);
void scaleRect32(VRect *l, long n, long d);
void shrinkRect(Rect *l, long x, long y);
void shrinkRect32(VRect *l, long x, long y);
void copyRect(Rect *l, const Rect *r);
void copyRect32(VRect *l, const VRect *r);
void PromptUserForInput (GPtr globals);
void PromptUserForInputNoProxy (GPtr globals);
void CalcProxyScaleFactor (const VRect *filterRect, Rect *proxyRect, long *scaleFactor);
void DoFilterRect (GPtr globals);
void UpdateProxy (HWND hDlg, BOOL bEraseBackground);
void ShowOutputBuffer( GPtr globals, HWND hDlg, BOOL fDragging);
void SetupFilterRecordForProxy (GPtr globals, long *scaleFactor, long prevScaleFactor,
                                POINT *offset);   
#ifndef DEMO
void StartWithAdvanceState (GPtr globals);
#endif

void StartNoAdvanceState (GPtr globals);
void DoInitialRect (GPtr globals);

#ifndef DEMO
void DoFilterRectNoAdvanceState (GPtr globals);
#endif

void createPredefinedCtls(void);
int  DoAction(CTLACTION action);
BOOL DoAuthorize(BOOL forceReauthorize);
void EditorThread(void *arg);
void StartEditor(void);


int setZoom (HWND hDlg, int n, int m);
int roundUpScaleFactor(int sf);
int roundDownScaleFactor(int sf);

#ifndef STANDALONE
    int openSourceFile ();
#endif
void GetProxyItemRect (HWND hDlg, Rect *proxyItemRect);
INT_PTR scrollPreview (HWND hDlg, INT_PTR mode, INT_PTR offsetX, INT_PTR offsetY);
INT_PTR setPreviewCursor (INT_PTR res);
INT_PTR getPreviewCursor (void);
HRESULT setDefaultWindowTheme(HWND hwnd);

INT_PTR fm_PixelsToHDBUs(INT_PTR h);
INT_PTR fm_PixelsToVDBUs(INT_PTR v);
INT_PTR fm_terminateThread(INT_PTR hThread);

INT_PTR fm_scaleCtls(INT_PTR dialogWidth, INT_PTR dialogHeight);

INT_PTR fm_updateProgress(INT_PTR n, INT_PTR d);
INT_PTR fm_abort(void);

INT_PTR fm_ctlEnabledAs(INT_PTR n);

const char *FmSearchFile(const char *szFileName);
void *FmLoadFile(const char *szFileName, int *pcbSize);
void EncryptFmCode(char *p, int iAlgorithm, int iKey);
int LoadFmCode(void);

/* AfhFMmake routines */
BOOL FmMakeStandalone(HWND hDlg);
void FmInitPiPLContents(void);

#ifdef _WIN64
/* Assembler routines from AfhFMml64.asm */
void X86_fninit(void);
void X86_fnclex(void);
void X86_fnsave(XMM_SAVE_AREA32 *envp);
void X64_set_nargs(int n);
INT64 X64_get_nargs(void);
INT64 X64_call_fmf1(BYTE *code, INT_PTR p1);
INT_PTR fm_callLibFmc(INT_PTR fn, ...);
INT_PTR fm_callLib(INT_PTR fn, ...);
INT_PTR fm_getCpuReg(INT_PTR nr);
void fm_finit(void);
void fm_fninit(void);
INT_PTR fm_fstsw(void);
INT_PTR fm_fstcw(void);
void fm_fldcw(INT_PTR cw);
#endif //_WIN64

/* Windows Registry routines */
HKEY ROK(HKEY hKey, LPCTSTR lpSubKeyName);
HKEY RCK(HKEY hKey, LPCTSTR lpSubKeyName);
LONG RQV(HKEY hKey, LPCTSTR lpValueName, LPBYTE dataBuffer, DWORD cbData);
LONG RQB(HKEY hKey, LPCTSTR lpValueName, VOID *dataBuffer, DWORD cbData);
LONG RQD(HKEY hKey, LPCTSTR lpValueName, DWORD *pdwValue);
LONG RQS(HKEY hKey, LPCTSTR lpValueName, CHAR *szBuffer, DWORD cbMaxLen);
LONG RSB(HKEY hKey, LPCTSTR lpValueName, CONST VOID *dataBuffer, DWORD cbData);
LONG RSD(HKEY hKey, LPCTSTR lpValueName, DWORD dwValue);
LONG RSS(HKEY hKey, LPCTSTR lpValueName, LPCTSTR szBuffer);
LONG REK(HKEY hKey, DWORD dwIndex, LPTSTR lpName, DWORD cbName);
LONG REV(HKEY hKey, DWORD dwIndex, LPTSTR lpValueName, LPDWORD lpcbValueName,
         LPDWORD lpType, LPBYTE lpData, LPDWORD lpcbData);
LONG RDK(HKEY hKey, LPCTSTR lpSubKey);
LONG RDV(HKEY hKey, LPCTSTR lpValueName);

/* PiPL resource routines */
BOOL CreatePipl(void);
BOOL AddPiplProp(PIType propertyKey, int32  propertyID, int32  propertyLength,
                 const void *propertyData);
char *FinishPipl(int *pcbPiplSize);

/* RT_VERSION resource routines */
BOOL VS_init(void);
BOOL VS_begin_binary_block(const char *szKey, const void *pValue, int cbValueSize);
BOOL VS_begin_string_block(const char *szKey, const char *szValue);
void VS_end_block(void);
VS_BLOCK_HEADER *VS_finish(void);

/* Resource update routines */

HANDLE afhBeginUpdateResource(
    LPCTSTR pFileName,              // pointer to file in which to update resources  
    void *pRamTemplate              // deletion option or pointer to RAM template
   );

BOOL afhUpdateResource(
    HANDLE pResDir,         // root of resource tree to update 
    LPCTSTR lpType,         // address of resource type to update 
    LPCTSTR lpName,         // address of resource name to update  
    WORD wLanguage,         // language identifier of resource 
    LPVOID lpData,          // address of resource data 
    DWORD cbData            // length of resource data, in bytes 
   );

BOOL afhEndUpdateResource(
    HANDLE hUpdate, // update-file handle  
    BOOL fDiscard   // write flag
   );

void Trace (GPtr globals, const char *szFormat, ...);

long long2fixed(long value);

Boolean CheckProxiesSupported (DisplayPixelsProc displayPixels);
Boolean DoNextRect (GPtr globals);

INT_PTR fm_setDialogPos(INT_PTR fAbs, INT_PTR x, INT_PTR y, INT_PTR w, INT_PTR h);
INT_PTR fm_resetAllCtls(void);
INT_PTR fm_createCtl(INT_PTR n, INT_PTR c, INT_PTR t, INT_PTR x, INT_PTR y,
					 INT_PTR w, INT_PTR h, INT_PTR s, INT_PTR sx, INT_PTR p,
					 INT_PTR e);
INT_PTR fm_setCtlText(INT_PTR n, INT_PTR iText);
INT_PTR fm_setCtlTextv(INT_PTR n, INT_PTR iFmt,...);
INT_PTR fm_setCtlVal(INT_PTR n, INT_PTR val);
INT_PTR fm_setCtlRange(INT_PTR n, INT_PTR lo, INT_PTR hi);
INT_PTR fm_setCtlLineSize(INT_PTR n, INT_PTR a);
INT_PTR fm_setCtlPageSize(INT_PTR n, INT_PTR b);
INT_PTR fm_setCtlThumbSize(INT_PTR n, INT_PTR a);
INT_PTR fm_setCtlFontColor(INT_PTR n, INT_PTR color);
INT_PTR fm_setCtlColor(INT_PTR n, INT_PTR color);
INT_PTR fm_getCtlColor(INT_PTR n);
INT_PTR fm_setCtlAction(INT_PTR n, INT_PTR a);
INT_PTR fm_setCtlPos(INT_PTR n, INT_PTR x, INT_PTR y, INT_PTR w, INT_PTR h);
INT_PTR fm_setDialogText(INT_PTR title);
INT_PTR fm_setDialogTextv(INT_PTR iFmt,...);
INT_PTR fm_enableToolTipBalloon(INT_PTR enable);
INT_PTR fm_setCtlToolTip(INT_PTR n, INT_PTR iText, INT_PTR s);
INT_PTR fm_setCtlImage(INT_PTR n, INT_PTR iName, INT_PTR iType);
INT_PTR fm_setCtlDivisor(INT_PTR n, INT_PTR div);
INT_PTR fm_getCtlDivisor(INT_PTR n);
INT_PTR fm_setCtlTicFreq(INT_PTR n, INT_PTR m);
INT_PTR fm_setCtlTheme(INT_PTR n, INT_PTR state);
INT_PTR fm_setDialogTheme (INT_PTR state);
INT_PTR fm_setCtlBuddyStyle(INT_PTR n, INT_PTR buddy, INT_PTR flags);
INT_PTR fm_clearCtlBuddyStyle(INT_PTR n, INT_PTR buddy, INT_PTR flags);
INT_PTR fm_setCtlBuddyStyleEx(INT_PTR n, INT_PTR buddy, INT_PTR flags);
INT_PTR fm_clearCtlBuddyStyleEx(INT_PTR n, INT_PTR buddy, INT_PTR flags);
INT_PTR fm_enableCtl(INT_PTR n, INT_PTR level);
int enableCtl(int n);
INT_PTR fm_setCtlTab(INT_PTR n, INT_PTR t, INT_PTR s);
INT_PTR fm_setCtlAnchor(INT_PTR n, INT_PTR flags);
INT_PTR fm_setCtlScripting(INT_PTR n, INT_PTR state);

INT_PTR fm_playSoundWave(INT_PTR szWaveFile);
INT_PTR fm_playSoundWaveLoop(INT_PTR szWaveFile);
INT_PTR fm_playSoundWaveSync(INT_PTR szWaveFile);

char *fm_getFilterInstallDir(void);
char *fm_appendEllipsis(const char *s);
char *fm_stripEllipsis(const char *s);
char *fm_formatString(const char *s);

INT_PTR fm_RGB(INT_PTR r, INT_PTR g, INT_PTR b);
INT_PTR fm_RGBA(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR a);
INT_PTR fm_Rval(INT_PTR rgba);
INT_PTR fm_Gval(INT_PTR rgba);
INT_PTR fm_Bval(INT_PTR rgba);
INT_PTR fm_Aval(INT_PTR rgba);
int fm_getOpenFileName(int flags,
                     LPSTR lpstrFile, int nMaxFile,
                     int *pnFileOffset, int *pnFileExtension,
                     LPSTR lpstrFileTitle, int nMaxFileTitle,
                     LPCSTR lpstrFilter,
                     LPSTR lpstrCustomFilter, int nMaxCustFilter,
                     int *pnFilterIndex,
                     LPCSTR lpstrInitialDir,
                     LPCSTR lpstrDialogTitle,
                     LPCSTR lpstrDefExt,
                     int *pOflags
                   );
int fm_getSaveFileName(int flags,
                     LPSTR lpstrFile, int nMaxFile,
                     int *pnFileOffset, int *pnFileExtension,
                     LPSTR lpstrFileTitle, int nMaxFileTitle,
                     LPCSTR lpstrFilter,
                     LPSTR lpstrCustomFilter, int nMaxCustFilter,
                     int *pnFilterIndex,
                     LPCSTR lpstrInitialDir,
                     LPCSTR lpstrDialogTitle,
                     LPCSTR lpstrDefExt,
                     int *pOflags
                   );
INT_PTR fm_chooseColor(INT_PTR defaultColor, const char *szPrompt, ...);

INT_PTR fm_HDBUsToPixels(INT_PTR h);
INT_PTR fm_VDBUsToPixels(INT_PTR v);
INT_PTR fm_setGamma(double gamma);
INT_PTR fm_gamma(INT_PTR i);

//Registry access functions
int fm_setRegRoot(HKEY hkey);
int fm_getRegRoot(HKEY *phkey);
int fm_setRegPath(LPCSTR szPath, ...);
int fm_getRegPath(LPSTR szPath, int maxPathLen);
int fm_putRegInt(int iValue, LPCSTR szValueName, ...);
int fm_getRegInt(int *iValue, LPCSTR szValueName, ...);
int fm_putRegString(LPCSTR szString, LPCSTR szValueName, ...);
int fm_getRegString(LPSTR szString, int iMaxlen, LPCSTR szValueName, ...);
int fm_putRegData(const void *pData, int dataLen, LPCSTR szValueName, ...);
int fm_getRegData(void *pData, int dataLen, LPCSTR szValueName, ...);
int fm_enumRegSubKey(int index, LPSTR szSubKey, int maxSubKeyLen);
int fm_enumRegValue(int index, LPSTR szValueName, int maxValueNameLen, int *iType, int *cbData);
int fm_deleteRegSubKey(LPCSTR szSubKey, ...);
int fm_deleteRegValue(LPCSTR szValueName, ...);

INT_PTR fm_triggerThread(INT_PTR n, INT_PTR event, INT_PTR previous);
INT_PTR fm_triggerEvent(INT_PTR n, INT_PTR event, INT_PTR previous);
INT_PTR fm_doEvents(void);

LRESULT CALLBACK KeyHookProc( int code, WPARAM wParam, LPARAM lParam);


extern jmp_buf Mde_error_exit_jb;

//Code buffer headers
extern CB cbCodeBuffer;
extern CB cbInitCode;
extern CB cbDivCheck;
extern CB cbZeroDivide;
extern CB cbDivideOverflow;
extern CB cbOnCtl;
extern CB cbOnFilterStart;
extern CB cbForEveryTile;
extern CB cbOnFilterEnd;
extern CB cbUserFunctions;

//Code buffer section kinds
typedef enum {
    CBK_FOREVERYPIXEL=1,
    CBK_INITCODE,
    CBK_DIVCHECK,
    CBK_ONZERODIVIDE,
    CBK_ONDIVIDEOVERFLOW,
    CBK_ONCTL,
    CBK_ONFILTERSTART,
    CBK_FOREVERYTILE,
    CBK_ONFILTEREND,
    CBK_USERFUNCTIONS,
    CBK_LITPOOL,
    CBK_INITFMC,
} CBSECTIONKIND;

//Code buffer section header
typedef struct {
    int   nBytes;       //# bytes in section to follow
                        //-- will get padded out to a multiple of 4
                        //   so next header is dword-aligned.
    int   iKind;        //what kind of section
    int   iEncrypt;     //kind of encryption
    int   iKey;         //encryption key;
    int   iChecksum;    //checksum of data that follows
} CB_SECTION_HEADER;


/*****************************************************************************/
/*  Following is a modifed version of FilterBigDocument.h
/*  from the googlecode WinGP project at:
/*	http://wingp.googlecode.com/svn/photogp/common/includes/
/*****************************************************************************/
//#ifndef __FilterBigDocument_H__
//#define __FilterBigDocument_H__

//#include "PIDefines.h"
//#include "PITypes.h"
//#include "PIFilter.h"

VPoint GetImageSize(void);
VRect GetFilterRect(void);
VRect GetInRect(void);
VRect GetOutRect(void);
VRect GetMaskRect(void);
VPoint GetFloatCoord(void);
VPoint GetWholeSize(void);

void SetInRect(VRect inRect);
void SetOutRect(VRect inRect);
void SetMaskRect(VRect inRect);

/* Globals */
FilterRecord *gFilterRecord;
BigDocumentStruct *gBigDocumentData;

//#endif
/*****************************************************************************/
/* End of modified FilterBigDocument.h
/*****************************************************************************/

#endif

