; Written by: Fibergeek, http://www.codegurus.be/
; Date:       2004-10-07
; Subject:    Testing RIP relative addressing
;
; Updates:
;   2006-10-22
;     Added "public Main", this is needed from now on
;     Updated the command line you need to run
;     Turned "END Main" into a simple "END"
;
;   2012-0703
;     AFH mods.
;
; To Compile & Link: 
;   ml64 /Fl file.asm /link /subsystem:windows /defaultlib:kernel32.lib /defaultlib:user32.lib /entry:Main
; N.B. x64-based ML64 8.0 crashes when listing is generated -- use x86-based ML64 8.0 instead!
;
;	31Mar2014  Alex Hunter  1. Added X64_fmod_fix trampoline code to fix up 
;							   a bug in the X64 CRT fmod routine.
;	24Apr2014  Alex Hunter  1. Correct comments in fm_callLibFmc regarding
;							   location of return value for floating-point
;							   functions (should be xmm0, not st(0) as was
;							   true in x86).
;	21Jun2014  Alex Hunter  1. New faster version of fm_getCpuReg.
;	03Jul2014  Alex Hunter	1. Delete unnecessary code by Fibergeek.
;							2. Move LM64 test code to ML64TestCode.asm.
;	04Jul2014  Alex Hunter  1. Add X64_ipow_core() in lieu of inline _asm code.

; This include file is in the "Microsoft Visual Studio 8\VC\PlatformSDK\include" folder
include ksamd64.inc

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; FM64 Utility Functions (in lieu of inline _asm, which is not available in ML64)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;	X86_fninit	-  initialize FPU with FNINIT instruction
;
LEAF_ENTRY	X86_fninit, _text
	fninit
	ret
LEAF_END	X86_fninit, _text
;
;	X86_fnclex	-  clear FPU exceptions with FNCLEX instruction
;
LEAF_ENTRY	X86_fnclex, _text
	fnclex
	ret
LEAF_END	X86_fnclex, _text
;
;	X86_fnsave(envp) -  Save FPU/MMX context with FNSAVE intruction
;	envp is a ptr to the context save area
;
LEAF_ENTRY	X86_fnsave, _text
	fnsave	[rcx]
	ret
LEAF_END	X86_fnsave, _text
;
;  X64_set_nargs(n) - set nargs = n for callee
;
LEAF_ENTRY	X64_set_nargs, _text
	mov		rax, rcx	; nargs is passed in RAX
	; but will the call itself clobber RAX???
	ret
LEAF_END	X64_set_nargs, _text
;
;  nargs = X64_get_nargs() - get nargs passed by caller
;
LEAF_ENTRY	X64_get_nargs, _text
	;nargs is in RAX, and we just return it in RAX
	;--but we have to hope RAX was not clobbered by
	;the caller's prolog!
	db 40h	; dummy REX byte for 2-byte hot-patching
	ret
LEAF_END	X64_get_nargs, _text

;;;;;;;;;;;;;; Following are not really leaf functions, ;;;;;;;;;	//FIX!!!
;;;;;;;;;;;;;; since they may call or jump to other		;;;;;;;;;
;;;;;;;;;;;;;; functions.								;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;; X64_call_fmf1 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;	X64_call_fmf1(BYTE *code, INT_PTR p1)
;	call fmf1 code ptr with one parameter p1, and setting
;	nargs = 1
;
LEAF_ENTRY	X64_call_fmf1, _text
	xchg	rcx, rdx	;swap 1st and 2nd parameters
	; could save the xchg by reversing the parameters
	; in the call, but what the heck...
	mov		rax, 1		; nargs=1 is passed in RAX
	call	rdx		; call code(p1) 
	ret
	; could do a jmp rdx to save a call/ret, but that
	; might confuse the exception handling mechanism!
LEAF_END	X64_call_fmf1, _text

;;;;;;;;;;;;;;;;;;;;;;;;;; fm_callLibFmc ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; //Will it work with fp args??
;
; X64 version of 'fm_callLibFmc', using explicit
; assembler code since MSVC X64 does not support
; inline _asm code.
;
;	void    fm_callLibFmc(INT_PTR fn, ...);
; or
;	INT_PTR fm_callLibFmc(INT_PTR fn, ...);
; or
;	double  fm_callLibFmc(INT_PTR fn, ...);
;
; Calls a function fn that takes exactly the same parameters as used
; to call this function, except the first argument (fn) is replaced by
; a pointer to the FM Context record (fmcp).
;
; This function is agnostic about the return value, which may be void,
; INT_PTR (returned in rAX), double (returned in xmm0),
; etc., depending on an appropriate cast of the function signature.
;
; N.B.	These functions assume the called function fn uses the FASTCALL
;		calling convention (which is the only standard WIN64 ABI).
;
; N.B.#2	This function will be deprecated when FM has the ability to
;			call a function indirectly through a pointer, in which case
;			simply replace:
;				ival = callLibFmc(fn, a, b, c,...);
;			with:
;				ival = fn(fmcp, a, b, c,...);	//THREAD PROBLEM: must be task-local fmcp!!! (e.g. fmcp->fmcp??? or fn(eBX,a,b,c...)???)
;			etc.
;
; Example of called functions in a DLL: 
;
;      __declspec(dllexport)
;      INT_PTR MyForEveryTile(FMcontext *fmcp, INT_PTR X, INT_PTR Y) { return false; }
;
;      __declspec(dllexport)
;      double MyFloat(FMcontext *fmcp, INT_PTR a, double x, double y) {
;          return a*(x - y);
;      }
;
; To invoke these functions from FM:
;
;      INT_PTR g_hMyDll = loadLib("MyDll");
;      if (!g_hMyDll) ERROR...
;
;      INT_PTR g_pfnMyForEveryTile = getLibFn(g_hMyDll, "MyForEveryTile");
;      if (!g_pfnMyForEveryTile) ERROR...
;
;      INT_PTR g_pfnMyFloat = getLibFn(g_hMyDll, "MyFloat");
;      if (!g_pfnMyFloat) ERROR...
;
;      INT_PTR iRes = callLibFmc(g_pfnMyForEveryTile, 100, 200);
;
;      double fRes = fcallLibFmc(g_pfnMyFloat, 10, 1.2, 3.14);
;
; where FM effectively defines callLibFmc and fcallLibFmc as:
;
;      typedef INT_PTR (*ifn1v) (INT_PTR, ...);
;      typedef double (*dfn1v) (INT_PTR, ...);
;      #define callLibFmc    ((ifn1v)callLibFmc)
;      #define fcallLibFmc   ((dfn1v)callLibFmc)

LEAF_ENTRY	fm_callLibFmc, _text
; upon entry, the following registers are assumed to contain:
;	rbx	=	address of FM context record (FMcontext64 *)
;	rcx =	address of function to be called (fn)
;	rdx =	2nd argument to fn (if any)
;	r8	=	3rd argument to fn (if any)
;	r9	=	4th argument to fn (if any)
; any higher arguments to fn are already on the call stack,
; above the 4-qword shadow space for the 1st 4 arguments

		mov rax, rcx			; rax <- fn
		mov rcx, rbx			; replace 1st arg (rcx) with rbx (fmcp)
		jmp rax					; jump to fn, which returns to caller
        ; return result will be in rax (for callLibFmc) or
        ; in xmm0 (for fcallLibFmc) --
        ; only our caller cares which.
LEAF_END	fm_callLibFmc, _text

;;;;;;;;;;;;;;;;;;;;;;;;;; fm_callLib ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; X64 version of 'fm_callLib', using explicit
; assembler code since MSVC X64 does not support
; inline _asm code.
;
;	void    fm_callLib(INT_PTR fn, ...);
; or
;	INT_PTR fm_callLib(INT_PTR fn, ...);
; or
;	double  fm_callLib(INT_PTR fn, ...);
;
; Calls a function that takes the same parameters as used
; to call this function with the exception of the first one, fn.
;
; Example of called function in a DLL:
;
;		INT_PTR myFunction(double a, double b) { return a + b; }
;
; N.B.	This function will be deprecated when FM has the ability to
;		call a function indirectly through a pointer, in which case
;		simply replace:
;			ival = callLib(fn, a, b, c,...);
;		with:
;			ival = fn(a, b, c,...);
;		etc.
;
; Florian Xhumari's original method:
; Use all data in previous function's frame (including all arguments passed
; to the current function) as arguments to the function fn. Actually,
; not all data, but all data excluding last argument (fn).
;
; Then, call function fn and remove added arguments.
;
; It is assumed that called function fn follows FASTCALL conventions,
; in particular it preserves register rbx.
;
; NFG!! To preserve 16-byte stack alignment, copy arg2 to arg1 slot,
;		arg3 to arg2 slot, etc., for nargs-1 args (=AX-1)
;
if 1
LEAF_ENTRY	fm_callLib, _text	; shifts args down, high to low
		xor r10d, r10d			;;make sure we don't generate a NaN??
L1:		cmp rax, 5				; rax has # of actual args
		jl L2					; args 1 thru 4 are in registers
if 0
		xchg r10, [rsp+rax*8]	;;asserts LOCK# -- NFG??
else
		mov r11, [rsp+rax*8]	; fake an xchg to shift args down
		mov [rsp+rax*8], r10	; the stack
		mov r10, r11
endif
		sub rax, 1
		jmp L1
L2:								;<==use an indexed jump on rax to skip unneeded copies???
		xchg r10, r9			; move arg5 down to arg4
		movd xmm3, r9			; copy arg4 to float arg reg
		xchg r10, r8			; move arg4 down to arg3
		movd xmm2, r8			; copy arg3 to float arg reg
		xchg r10, rdx			; move arg3 down to arg2
		movd xmm1, rdx			; copy arg2 to float arg reg
		xchg r10, rcx			; move arg2 down to arg1
		movd xmm0, rcx			; copy arg1 to float arg reg
		jmp	r10				; r10 gets original arg1, which is the fn address.
							; jump to fn.  We already removed one "arg" from stack,
							; so the callee will see arg2, arg3, arg4, arg5,... in
							; the correspondng locations for arg1, arg2, arg3, arg4,...
LEAF_END	fm_callLib, _text
elseif 1
LEAF_ENTRY	fm_callLib, _text	; shifts args down, low to high
		mov	r11, rcx		; r11 = fn (arg1)
		mov rcx, rdx		; move arg2 to arg1
		movd xmm0, rcx
		mov rdx, r8			; move arg3 to arg2
		movd xmm1, rdx
		mov r8, r9			; move arg4 to arg3
		movd xmm2, r8
		mov r9, [rsp+40]	; load arg4 from arg5 on stack (if it exists, otherwise harmless)
		movd xmm3, r9
		
		sub rax, 5			; nargs - 5
		jle L3				; br unless more than 5 actual args
if 1
		push rcx
		push rsi
		push rdi
		mov rcx, rax		; nargs - 5
		lea rsi, [rsp+48]
		lea rdi, [rsp+40]
		rep movsq
		pop rdi
		pop rsi
		pop rcx
else
		mov r10, [rsp+		;<==unfinished (abandoned?) code?
endif
L3:
		jmp	r11				; jump to fn.  We already removed one "arg" from stack,
							; so the callee will see arg2, arg3, arg4, arg5,... in
							; the correspondng locations for arg1, arg2, arg3, arg4,...
LEAF_END	fm_callLib, _text
else
LEAF_ENTRY	fm_callLib, _text	; old -- misaligns the stack!!
if 0
		pop	rax				; rax = return ip
		mov [rsp], rax		; return ip overwrites arg1 shadow slot
else
		pop [rsp]			; arg1 = return ip?? depends if rsp is before or after pop
endif
		mov	rax, rcx		; rax = fn (arg1)
		mov rcx, rdx		; move arg2 to arg1
		mov rdx, r8			; move arg3 to arg2
		mov r8, r9			; move arg4 to arg3
		mov r9, [rsp+32]	; load arg4 from arg5 on stack (if it exists, otherwise harmless)
							; any remaining stack args are in correct position on the stack
		jmp	rax				; jump to fn.  We already removed one "arg" from stack,
							; so the callee will see arg2, arg3, arg4, arg5,... in
							; the correspondng locations for arg1, arg2, arg3, arg4,...
							; but NOTE: the stack is mis-aligned at this point!! <=== NFG in some cases!!
LEAF_END	fm_callLib, _text
endif

;;;;;;;;;;;;;;;;;;;;;;;;;; fm_getCpuReg ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;	_int64 fm_getCpuReg(int nr) - Get value of specified GPR
;
; nr: 0 = rax, 1 = rcx, 2 = rdx,  3 = rbx,  4 = rsp,  5 = rbp,  6 = rsi,  7 = rdi,
;	  8 = r8,  9 = r9, 10 = r10, 11 = r11, 12 = r12, 13 = r13, 14 = r14, 15 = r15,
;    16 = rip
;
LEAF_ENTRY	fm_getCpuReg, _text
; nr is arg1 (in rcx)
if 0 ;-------------------------------------------------------
; default is to return rax in rax (nr = 0)
		cmp rcx, 1
		cmove rax, rcx		; if nr = 1, return rcx
		cmp rcx, 2
		cmove rax, rdx		; if nr = 2, return rdx
		cmp rcx, 3
		cmove rax, rbx		; if nr = 3, return rbx
		cmp rcx, 4
		cmove rax, rsp		; if nr = 4, return rsp
		cmp rcx, 5
		cmove rax, rbp		; if nr = 5, return rbp
		cmp rcx, 6
		cmove rax, rsi		; if nr = 6, return rsi
		cmp rcx, 7
		cmove rax, rdi		; if nr = 7, return rdi
		cmp rcx, 8
		cmove rax, r8		; if nr = 8, return r8
		cmp rcx, 9
		cmove rax, r9		; if nr = 9, return r9
		cmp rcx, 10
		cmove rax, r10		; if nr = 10, return r10
		cmp rcx, 11
		cmove rax, r11		; if nr = 11, return r11
		cmp rcx, 12
		cmove rax, r12		; if nr = 12, return r12
		cmp rcx, 13
		cmove rax, r13		; if nr = 13, return r13
		cmp rcx, 14
		cmove rax, r14		; if nr = 14, return r14
		cmp rcx, 15
		cmove rax, r15		; if nr = 15, return r15
		cmp rcx, 16
		cmove rax, [rsp]	; if nr = 16, return rip from top of stack
		ret
elseif 0 ;--------------------------------------------------------
							; return rip will be at [rsp+8*16]
							; push remaining GPR values onto stack
		push r15			; r15 will be at [rsp+8*15]
		push r14			; r14 will be at [rsp+8*14]
		push r13			; r13 will be at [rsp+8*13]
		push r12			; r12 will be at [rsp+8*12]
		push r11			; r11 will be at [rsp+8*11]
		push r10			; r10 will be at [rsp+8*10]
		push r9				; r9  will be at [rsp+8*9]
		push r8				; r8  will be at [rsp+8*8]
		push rdi			; rdi will be at [rsp+8*7]
		push rsi			; rsi will be at [rsp+8*6]
		push rbp			; rbp will be at [rsp+8*5]
		push rsp			; current rsp will be at [rsp+8*4]
		push rbx			; rbx will be at [rsp+8*3]
		push rdx			; rdx will be at [rsp+8*2]
		push rcx			; rcx (nr) will be at [rsp+8*1]
		push rax			; rax will be at [rsp+8*0]
		and ecx, 1fh		; avoid a possible access violation if nr (far) out of range
		mov rax, [rsp+8*rcx]	; load selected GPR or return RIP from stack
		;;add rsp, 8*16		; pop 16 cpu regs from stack
		sub rsp, -8*16		; (sub rsp,-128 is 3 bytes shorter than add rsp,128)
		ret					; return selected reg in rax
else ;-------------------------------------------------------------------------
		test rcx, rcx		; if rcx == 0, go directly to case rax:
		jz L1				; since otherwise the switch logic clobbers rax
		cmp rcx, 16			; use a switch-like construct to return desired reg
		ja L1				; return rax if nr out of range
							; assumes all 'cases' (except last) occupy 4 bytes!!
		lea rax, L1			; can't use an SIB with rip-relative addressing, so
		lea rax, [rax+rcx*4]; break it down into 2 stages
		jmp rax				; rax has been destroyed at this point!!
		align 4
L1:
		mov rax, rax		; nr 0 - return rax
		ret
		align 4
		mov rax, rcx		; nr 1 - return rcx
		ret
		align 4
		mov rax, rdx		; nr 2 - return rdx
		ret
		align 4
		mov rax, rbx		; nr 3 - return rbx
		ret
		align 4
		mov rax, rsp		; nr 4 - return rsp
		ret
		align 4
		mov rax, rbp		; nr 5 - return rbp
		ret
		align 4
		mov rax, rsi		; nr 6 - return rsi
		ret
		align 4
		mov rax, rdi		; nr 7 - return rdi
		ret
		align 4
		mov rax, r8			; nr 8 - return r8
		ret
		align 4
		mov rax, r9			; nr 9 - return r9
		ret
		align 4
		mov rax, r10		; nr 10 - return r10
		ret
		align 4
		mov rax, r11		; nr 11 - return r11
		ret
		align 4
		mov rax, r12		; nr 12 - return r12
		ret
		align 4
		mov rax, r13		; nr 13 - return r13
		ret
		align 4
		mov rax, r14		; nr 14 - return r14
		ret
		align 4
		mov rax, r15		; nr 15 - return r15
		ret
		align 4
L2:
		mov rax, [rsp]		; nr 16 - return rip
		ret
	
; Consistency check: L2 - L1 must equal exactly 64 bytes...
		
		.errnz (L2 - L1) - 64
		
endif
LEAF_END	fm_getCpuReg, _text

;
;	fm_finit(void)	-  initialize FPU with FINIT instruction
;
LEAF_ENTRY	fm_finit, _text
	finit
	ret
LEAF_END	fm_finit, _text

;
;	fm_fninit(void)	-  initialize FPU with FNINIT instruction
;
LEAF_ENTRY	fm_fninit, _text
	fninit
	ret
LEAF_END	fm_fninit, _text

;
;	_int64 fm_fstsw(void)	-  return 16-bit FPU status word in rax
;
LEAF_ENTRY	fm_fstsw, _text
        xor eax, eax        ; clear 64-bit rax
        fstsw ax            ; return 16-bit status word in rax
        ret
LEAF_END	fm_fstsw, _text

;
;	_int64 fm_fstcw(void)	-  return 16-bit FPU control word in rax
;
LEAF_ENTRY	fm_fstcw, _text
        push 0              ; clear a qword on the stack
        fstcw word ptr[rsp] ; store control word in low 16 bits of qword
        pop rax             ; pop zero-extended control word into rax
        ret
LEAF_END	fm_fstcw, _text

;
;	void fm_fldcw(int cw)	-  load FPU control word from low 16 bits of argument cw
;
LEAF_ENTRY	fm_fldcw, _text
		mov [rsp+8], rcx		; home the cw argument in shadow stack frame
        fldcw word ptr[rsp+8]   ; load FPU control word from low 16 bits of cw
        ret
LEAF_END	fm_fldcw, _text

;
;	double X64_fmod_fix(double x, double y) - Fix the broken MVSCRT fmod proc.
;	Added 31MAR2014 by AFH to fix up bug in the X64 CRT fmod routine.
;
extrn fmod: PROC
LEAF_ENTRY	X64_fmod_fix, _text
		;xmm0=x, xmm1=y
		fnclex					;clear any sticky exceptions
		jmp	fmod				;now execute and  return from the CRT fmod...
LEAF_END	X64_fmod_fix, _text

;;;;;;;;;;;;;;;;;;;;;;;;;; X64_ipow_core ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;	INT64 X64_ipow_core(INT64 i, INT64 n)
;	- Gets value of i**n, assuming n > 0.
;	- Sets errno=ERANGE and returns +/-huge_value if overflow occurs.
;
EXTRN	_errno:PROC
LEAF_ENTRY	X64_ipow_core, _text
; i is arg1 (in rcx)
; n is arg2 (in rdx)
; register mapping:
; rcx = i, rdx = n, rax = iRes, r8b = isNeg
;
; From <limits.h>:
; minimum signed 64 bit value:
I64_MIN  equ  (-9223372036854775807 - 1)
; maximum signed 64 bit value:
I64_MAX  equ    9223372036854775807
; From <errno.h>
ERANGE	equ	34				; Range error

		; bool isNeg = (i < 0) && (n & 1);	//sign of result
		xor		r8b, r8b	;isNeg = false
		cmp		rcx, 0
		jge		LL01		;if (i >=0) goto LL01
		test	dl, 1		;i is < 0; test for n odd or even
		jz		LL01		;if (n is even) goto LL01
		inc		r8b			;isNeg = true if (i < 0) && (n is odd)
LL01:
		mov		rax, 1		;iRes = 1
LP01:						;do {
		test	dl, 1		;	n & 1
		jz		LP02		;	if (n & 1) {
		imul	rax, rcx	;		iRes *= i
		jo		oflo1		;		if (overflow) goto oflo1
LP02:						;	}
		imul	rcx, rcx	;	i *= i
		jo		oflo2		;	if (overflow) goto oflo2
		shr		rdx, 1		;	n >>= 1
		jnz		LP01		;} while (n)
done:
		ret					;return iRes in rax
;
;	Handle (possible) overflow
;
oflo2:	; Overflow here only if n >> 1 !=0
		shr		rdx, 1
		jz		done			;if no overflow, take normal return
oflo1:  ; Overflow => Range error: return approximation to +/-infinity
        ; and set errno = ERANGE...
        push	r8				;save isNeg across call to _error
        push	rbp				;set up a stack frame
        mov		rbp, rsp
        and		rsp, -16		;align stack to 16 bytes
        sub		rsp, 32			;allocate minimum arglist
        call	_errno			;rax <- address of errno in TLS
        mov		dword ptr [rax], ERANGE	;set errno = ERANGE
		mov		rsp, rbp		;pop the stack frame
		pop		rbp				;restore rbp
		pop		r8				;restore isNeg
        mov		rax, I64_MAX	;return approximate +infinity
        test	r8b, 1			;unless isNeg,
        jz		done			;in which case
        mov		rax, I64_MIN	;return approximate -infinity
        ret						;and return +/-infinity
LEAF_END	X64_ipow_core, _text
END
