#ifndef FMDIB_ANTIALIAS_H
#define FMDIB_ANTIALIAS_H

/*
 *   fmdib_antialias.h
 *
 *   Copyright 1997-2014 AFH Systems
 *
 *   FilterMeister is free software. You can redistribute and modify it
 *   under the terms of the MPL2 License.
 *   See: https://www.mozilla.org/en-US/MPL/2.0/
 *   Any changes to the FilterMeister files have to be published or sent
 *   to harry@thepluginsite.com. The filter plugin code itself does not
 *   need to be disclosed.
 *
 */

//---------------------------------------------------------------------------

static INT_PTR fm_setThinLineAA(INT_PTR start_x, INT_PTR start_y,
								INT_PTR end_x, INT_PTR end_y, UINT_PTR color)
{
	double xx, yy, slope, alfa;
	int x, y;
 
	int left = (int)min(start_x, end_x), right  = (int)max(start_x, end_x);
	int top =  (int)min(start_y, end_y), bottom = (int)max(start_y, end_y);
 
	if (modeGDI) {
		BitBlt(hdcDIB, 0, 0, GDIcanvasWidth, GDIcanvasHeight, hdcGDI, 0, 0, SRCCOPY);
		GdiFlush();
		modeGDI = false;
	}
 
	if (left == right && top == bottom) {
		GDI_dibSetPixel(left, top, (UINT)color);
	} else if (bottom - top > right - left) {
		color = ((color >> 16) & 0x000000FF) | (color & 0x0000FF00) | ((color << 16) & 0x00FF0000);
		slope = 1.0 / (double)(end_y - start_y);
		for (y = top; y <= bottom; ++y) {
			xx = slope * ((end_y - y) * start_x + (y - start_y) * end_x);
			x = (int)floor(xx);
			alfa = x + 1 - xx;
			GDI_dibMixPixel_BGR(x, y, (UINT)color, alfa);
			GDI_dibMixPixel_BGR(x + 1, y, (UINT)color, 1. - alfa);
		}
	} else {
		color = ((color >> 16) & 0x000000FF) | (color & 0x0000FF00) | ((color << 16) & 0x00FF0000); 
		slope = 1.0 / (double)(end_x - start_x);
		for (x = left; x <= right; ++x) {
			yy = slope * ((end_x - x) * start_y + (x - start_x) * end_y);
			y = (int)floor(yy);
			alfa = y + 1 - yy;
			GDI_dibMixPixel_BGR(x, y, (UINT)color, alfa);
			GDI_dibMixPixel_BGR(x, y + 1, (UINT)color, 1. - alfa);
		}
	}
 
	return true;
}

//---------------------------------------------------------------------------

static INT_PTR fm_setThinEllipseAA(INT_PTR start_x, INT_PTR start_y,
								   INT_PTR end_x,   INT_PTR end_y,   UINT_PTR color)
{ //NEW!
	int left 	= (int)min(start_x, end_x);
	int right 	= (int)max(start_x, end_x);
	int top 	= (int)min(start_y, end_y);
	int bottom 	= (int)max(start_y, end_y);

	if (modeGDI) {
		BitBlt(hdcDIB, 0, 0, GDIcanvasWidth, GDIcanvasHeight, hdcGDI, 0, 0, SRCCOPY);
		GdiFlush();
		modeGDI = false;
	}

	if (left == right && top == bottom) {
		GDI_dibSetPixel(left, top, (UINT)color);
	} else {
		int x, y;
		double xx, yy, slope, alfa, delta;
		double aa = .5 * (right - left), bb = .5 * (bottom - top); // semi axes;
		double aa2 = aa * aa, bb2 = bb * bb;
		double xxc = .5 * (left + right), yyc = .5 * (top + bottom); // center coordinates
		double xx45 = aa2 / sqrt(aa2 + bb2), yy45 = bb2 / sqrt(aa2 + bb2);

		color = ((color >> 16) & 0x000000FF) | (color & 0x0000FF00) | ((color << 16) & 0x00FF0000);

		if (bb > 0) {
			slope = aa / bb;
			for (y = (int)floor(yyc - yy45); y < yyc + yy45 + 1; ++y) {
				delta = slope * sqrt(bb2 - (y - yyc) * (y - yyc));

				xx = xxc + delta; // right arc
				x = (int)floor(xx);
				alfa = x + 1 - xx;
				GDI_dibMixPixel_BGR(x, y, (UINT)color, alfa);
				GDI_dibMixPixel_BGR(x + 1, y, (UINT)color, 1. - alfa);

				xx = xxc - delta; // left arc
				x = (int)floor(xx);
				alfa = x + 1 - xx;
				GDI_dibMixPixel_BGR(x, y, (UINT)color, alfa);
				GDI_dibMixPixel_BGR(x + 1, y, (UINT)color, 1. - alfa);
			} // next y
		} // end if bb > 0

		if (aa > 0) {
			slope = bb / aa;
			for (x = (int)floor(xxc - xx45); x < xxc + xx45 + 1; ++x) {
				delta = slope * sqrt(aa2 - (x - xxc) * (x - xxc));

				yy = yyc + delta; // bottom arc
				y = (int)floor(yy);
				alfa = y + 1 - yy;
				GDI_dibMixPixel_BGR(x, y, (UINT)color, alfa);
				GDI_dibMixPixel_BGR(x, y + 1, (UINT)color, 1. - alfa);

				yy = yyc - delta; // top arc
				y = (int)floor(yy);
				alfa = y + 1 - yy;
				GDI_dibMixPixel_BGR(x, y, (UINT)color, alfa);
				GDI_dibMixPixel_BGR(x, y + 1, (UINT)color, 1. - alfa);
			} // next x
		} // end if aa > 0
	} // end draw ellipse

	return true;
} // end draw antialiased

//---------------------------------------------------------------------------

#endif
// FMDIB_ANTIALIAS_H