#if defined(VAXC) || defined(__DECC)
/* Following pragma packs data structure members on byte  */
/* boundaries while aligning the base of the structure on */
/* a longword (as required, e.g., for SLCNET buffers).    */
#pragma nomember_alignment LONGWORD
/* Following pragma kills some spurious error messages    */
/* caused by the preceding pragma.                        */
#pragma message disable (NONMULTALIGN)
#endif

///#define DEBUG_OP_CALL 1

/*************************************************************************/
/*                                                                       */
/*                          AfhFMcompiler.c                              */
/*                                                                       */
/*  This module of the FilterMeister contains the SDL expression stack   */
/*  native code generator (compiler).                                    */
/*                                                                       */
/*  Copyright 1997-2014 AFH Systems                                      */
/*                                                                       */
/*  FilterMeister is free software. You can redistribute and modify it   */
/*  under the terms of the MPL2 License.                                 */
/*  See: https://www.mozilla.org/en-US/MPL/2.0/                          */
/*  Any changes to the FilterMeister files have to be published or sent  */
/*  to harry@thepluginsite.com. The filter plugin code itself does not   */
/*  need to be disclosed.                                                */
/*                                                                       */
/*************************************************************************/
/*                                                                       */
/*              R E V I S I O N   H I S T O R Y                          */
/*                                                                       */
/*  From MDY.C:                                                          */
/*                                                                       */
/*    13Feb97  Alex Hunter  1. Preliminary support for unions.           */
/*    20Feb97  Alex Hunter  1. MD386__NOUNIQUE (and MD386__AMBIGLAB)     */
/*                             were being issued in inappropriate cases  */
/*                             (e.g., in some cases when the symbol was  */
/*                             actually unambiguous, or when the symbol  */
/*                             was not found in the specified scope).    */
/*    18Apr97  Alex Hunter  1. Emulate Lib$addx/Lib$subx for WIN32.      */
/*                          2. No sys$gettim (and hence %TIME is NYI)    */
/*                             for WIN32.                                */
/*    20Apr97  Alex Hunter  1. Adjust IEEE/VAX conversions for WIN32.    */
/*    30Nov97  Alex Hunter  1. Add operations to support C language.     */
/*    02Dec97  Alex Hunter  1. Allow mixed integer/real math ops.        */
/*    03Dec97  Alex Hunter  1. Implement OP_POP.                         */
/*    04Dec97  Alex Hunter  1. Implement OP_STORE.                       */
/*    05Dec97  Alex Hunter  1. Don't recurse indefinitely if default     */
/*                             type index is null.                       */
/*                          2. Per ANSI C, value and type of a store op  */
/*                             is value and type of lhs, not value and   */
/*                             type of rhs.  (Affects simple and         */
/*                             compound assignments, and the post/pre-   */
/*                             increment/decrement operations.)          */
/*                          3. Implement OP_XCHG.                        */
/*    06Dec97  Alex Hunter  1. Allow assignment of integer value to a    */
/*                             pointer or selector, but warn if the      */
/*                             value is non-zero.                        */
/*                          2. Dereferencing an array in C yields a      */
/*                             pointer to the first element of the       */
/*                             array.                                    */
/*                          3. Dereferencing a (pointer) to a function   */
/*                             is a no-op in C.                          */
/*                          4. Make gettypelength (somewhat) more        */
/*                             robust.                                   */
/*                          5. Implement scaled pointer addition and     */
/*                             subtraction for C.                        */
/*                                                                       */
/*  For AfhFMinterpreter.c:                                              */
/*                                                                       */
/*  12Jan98  Alex Hunter  1. Derived from MDY.C.                         */
/*  26Feb98  Alex Hunter  1. Implement REGION operators.                 */
/*                                                                       */
/*  For AfhFMcompiler.c:                                                 */
/*                                                                       */
/*  17Mar98  Alex Hunter  1. Derived from AfhFMcompiler.c.               */
/*  05Apr98  Alex Hunter  1. Implement FFP inline operators: OP_GET,...  */
/*  01Sep12  Alex Hunter  1. Add minimal support for X64.				 */
/*	06JUL13	 Alex Hunter  1. Add rip-relative code for x64 switch stmt	 */
/*							 implementation.							 */
/*  15JUL13	 Alex Hunter  1. Modify for WIN64.							 */
/*	03AUG13	 Alex Hunter  1. Changes for "chameleon" oontext record.	 */
/*	23AUG13	 Alex Hunter  1. Add quickFillArray.						 */
/*  28AUG13  Alex Hunter  1. Prune out code not used by FM.				 */
/*	04SEP13	 Alex Hunter  1. Use MOVSDX instead of XMOVQ, per VC++.		 */
/*	07SEP13	 Alex Hunter  1. Change FLD_IV to FILD_V, FSTP_IV to FISTP_V.*/
/*  09SEP13  Alex Hunter  1. Add calls to check_ndp_stack_room (NYI).	 */
/*  10SEP13  Alex Hunter  1. Use MOVAPS for float xmm to xmm copy.		 */
/*						  2. Add X86_error().							 */
/*	13SEP13  Alex Hunter  1. Don't allow cross-platform code execution.	 */
/*						  2. Recode op_ctl() using MOVSXD opcode.		 */
/*	19SEP13  Alex Hunter  1. Use X86_emit_op_lab_disp8 instead of        */
/*							 X86_emit_op_lab in op_min and op_max.		 */
/*						  2. Should also check for other examples of 1.	 */
/*						  3. Bug fix: change FST_STN to FSTP_STN in cmov */
/*							 branch op op_min and op_max.				 */
/*  30SEP13  Alex Hunter  1. Use a static arglist for level 1 calls.	 */
/*	03JAN14	 Alex Hunter  1. Add floatCoordHeight and floatCoordWidth to */
/*							 the dictionary.							 */
/*	12JAN14	 Alex Hunter  1. Temporary workaround for no RIP-relative	 */
/*							 addressing in AfhEmitX86.c. (Remove this    */
/*							 kludge when AfhEmitX86.c is fixed!!!)		 */
/*  20JAN14  Alex Hunter  1. No longer necessary to round static arglist */
/*							 size up to a multiple of 16 (see delayed    */
/*							 prolog).									 */
/*						  2. OOPS. Back this out until I've had a chance */
/*							 to rethink it!!!							 */
/*	20JAN14  Alex Hunter  1. Move formatted print routine macros to		 */
/*							 AfhFM.h.									 */
/*						  2. Change format "+%d" to "%+d" to handle      */
/*							 signed values.								 */
/*	26JAN14  Alex Hunter  1. OP_LOAD now sign-extends int4 to int8.      */
/*						     NOTE: Much work remains for OP_LOAD,		 */
/*							 OP_STORE, etc., to handle all the new       */
/*							 FM 2.0 types.								 */
/*						  2. Add some missing chameleon ops: NEG_V, 	 */
/*							 NOT_V, SHL_V, SHR_V, CMPI_V, etc.			 */
/*						  3. Change op_D to op_V in many cases.			 */
/*						  4. Introduce MOVSXD_V for MOV_D/MOVSXD_Q.		 */
/*	29JAN14	 Alex Hunter  1. Change default filterCategoryText to		 */
/*							 "My Category".								 */
/*	31JAN14	 Alex Hunter  1. Refine the "Cannot execute X32/64" alerts.  */
/*	19FEB14  Alex Hunter  1. Display error msgs in Mdc_execute_init_code */
/*							 only if in Advanced Edit mode.				 */
/*	23FEB14  Alex Hunter  1. Add X86_abort2().							 */
/*						  2. Expand some error messages with X86_abort2. */
/*	24FEB14	 Alex Hunter  1. Reset the ndp stack slot counter at the     */
/*							 beginning of each compilation.				 */
/*	25FEB14	 Alex Hunter  1. Use special logic in dereference() because  */
/*							 there is no MR form of MOVSXD_Q.			 */
/*	03MAR14	 Alex Hunter  1. Add SD_PROG for program type and revision.	 */
/*	08MAR14	 Alex Hunter  1. Fix emission of "Floating point stack		 */
/*							 overflow" message when ndp_slots > 1.		 */
/*						  2. Fix erroneous emission of "Floating point	 */
/*							 stack overflow" message at end of true part */
/*							 of a ternary conditional expression (fix is */
/*							 in dec_stack()).							 */
/*	10MAR14  Alex Hunter  1. Add built-in symbols abort_mode, FMC_TARGET,*/
/*							 FMC_COMPILE_PLATFORM, FMC_MAJOR_VERSION,    */
/*							 FMC_MINOR_VERSION.							 */
/*					**NOTE** FMC_COMPILE_PLATFORM not correct in		 */
/*							 STANDALONE!!								 */
/*	20MAR14	 Alex Hunter  1. Don't copy real arg into corresp. int reg	 */
/*							 unless fml version is <= 1.5 or the arg	 */
/*							 is a vararg.								 */
/*					**NOTE** Should consolidate cases and eliminate both */
/*							 switches within the for-loop.				 */
/*	09MAY14  Alex Hunter  1. Weed out some '#if[n]def FML' sections.	 */
/*						  2. Weed out '#if COMPILE/INTERPRET'.			 */
/*						  3. Delete most PLMC32-generated declarations	 */
/*							 for file-level variables.					 */
/*	14MAY14	 Alex Hunter  1. Begin modifying op_load and op_store for	 */
/*							 various scalar sizes (also need to modify	 */
/*							 op_preinc et al.).							 */
/*	15MAY14	 Alex Hunter  1. Continue modifying op_load and op_store,	 */
/*							 for unsigned cases.						 */
/*						  2. Fix bug when valuep is dereferencd in		 */
/*							 dpvalue.									 */
/*	25MAY14	 Alex Hunter  1. Add MOVSXB_V, MOVSXW_V.					 */
/*			What about intrinsics such as i64 = get(v)?					 */
/*	30May14  Alex Hunter  1. Fix OP_FLOAT and OP_DOUBLE for X64.		 */
/*	02Jun14	 Alex Hunter  1. Revise logic for suppressing ndp stack      */
/*							 overflow messages.							 */
/*	03Jun14  Alex Hunter  1. Modify op_postinc for X64.					 */
/*	05Jun14	 Alex Hunter  1. Modify op_postdec for X64.					 */
/*						  2. Modify op_sizeof for X64.					 */
/*	06Jun14  Alex Hunter  1. Delete member(), concatenate(), buildptr(). */
/*						  2. Delete index().							 */
/*						  3. Disable qualify() for now.					 */
/*						  4. Delete typesareequivalent().				 */
/*						  5. Delete symbolsareequivalent().				 */
/*						  6. Delete lookupgrp(),lookuplin(),lookupseg(). */
/*						  7. Clean up many "#if 0" sections.			 */
/*  08Jun14  Alex Hunter  1. Add unique code to each NYI msg for compiler*/
/*							 debugging.									 */
/*	11Jun14  Alex Hunter  1. Use new fmfNONNEG flag.					 */
/*						  2. Check for attempt to modify a 'const' var.  */
/*						  3. Use new 'flags' and 'namep' fields in VARs. */
/*	13Jun14	 Alex Hunter  1. Change association of user_vars_init_flags  */
/*							 to 1 bit per byte for X64.					 */
/*  20Jun14  Alex Hunter  1. Add const 'needPremiere' to predefined      */
/*							 dictionary.								 */
/*						  2. Set CONST attribute on appropriate vars	 */
/*							 in the predefined dictionary.				 */
/*	24Jun14  Alex Hunter  1. Minor code cleanups.						 */
/*						  2. Fix bug in X32 inline pow() in 3 places --  */
/*							 pseudocode was correct, but generated code  */
/*							 was wrong!!								 */
/*						  3. Reduce size of arglist to powi in X32 from  */
/*							 16 to 12.									 */
/*	27Jun14	 Alex Hunter  1. Definitions for FILD_V and FISTP_V were     */
/*							 reversed for X32 and X64.					 */
/*	28Jun14  Alex Hunter  1. Fixes for errno with the new VAR format.	 */
/*	13Jul14	 Alex Hunter  1. Temporary kludge in pushlititem() to extend */
/*							 a 32-bit lit to a mem64; it *should* zero-	 */
/*							 extend the imm32, but currently is sign-    */
/*							 extends instead. (This fix is necessary to  */
/*							 load a 32-bit lit into a non-reg temp.)	 */
/*                                                                       */
/*************************************************************************/

/*************************************************************************/
/*                                                                       */
/*	TODO:                                                                */
/*	1. Optimize layout of for-loops:                                     */
/*		for (e1; e2; e3) s1;                                             */
/*	  currently compiles to:											 */
/*      e1; test: e2; jmpf break; s1; continue: e3; jmp test; break:     */
/*	  better would be:													 */
/*		e1; jmp test; top: s1; continue: e3; test: e2; jmpt top; break:	 */
/*	  advantages:														 */
/*	  -	no useles unconditional jmp at end of loop						 */
/*	  -	conditional jmp is backward, permitting easy optimization of	 */
/*		displacement, as opposed to forward jmp in current plan			 */
/*    disadvantages:                                                     */
/*                                                                       */
/*	2. Gather stats on max ndp st depth reached, output in compilation	 */
/*	   summary.															 */
/*                                                                       */
/*	3. Spill x87 regs to temps before every call. (see lab16.fml)		 */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*************************************************************************/

#include "AfhFM.h"                      /* FM globals.                        */

// We need both 32-bit and 64-bit versions of the FM context record...
#undef TARGET
#define TARGET 32
#include "AfhFMcontext.h"               /* 32-bit FM context record.          */
#undef TARGET
#define TARGET 64
#include "AfhFMcontext.h"               /* 64-bit FM context record.          */
#undef TARGET

#include "AfhFMsymtab.h"                /* FM symbol table definitions.       */
#include "EmitX86.h"                    /* X86 code emitter routines.         */

#include <windows.h>                    /* HWND, HRGN, Create*Rgn.            */
#include "plmcvax.h"
#include <stddef.h>                     /* NULL.                              */
#include <string.h>                     /* memset, memcmp, memcpy.            */
#include <stdio.h>                      /* sprintf.                           */
#include <setjmp.h>                     /* setjmp, jmp_buf.                   */
#include <assert.h>                     /* assert.                            */
#include "md386def.hc"                  /* MD386__ message codes.             */
#include "md386lit.h"                   /* MD386 common definitions.          */
#include "mdg.h"                        /* MD386 global data definitions.     */
#include "mdd.h"                        /* MD386 dictionary data structures.  */
#include "mdp.h"                        /* MD386 parser definitions.          */
#include "mdy.h"                        /* QUALIFIERSTRUCTURE.                */
#include "md386proto.h"                 /* MD386 intermodule prototypes.      */

#include <time.h> 

//#define min(a, b)  (((a) < (b)) ? (a) : (b)) //use intrinsic if avalilable

        /**********************************/
        /*                                */
        /*     configurable parameters    */
        /*                                */
        /**********************************/

//...//

    /* Globals */

GPtr globals;
int X64gen = 0;
int X86gen = 1;			//default is x86 code generation
int nTmps = 2;			//x86 uses 2 temp regs
int nVolatileTmps = 0;	//x86 uses no volatile temp regs
enum PROG prog_type = PROG_UNKNOWN;	// Program type
int nProgRevMajor = INT_MAX;	// Program major revision level
int nProgRevMinor = INT_MAX;	// Program minor revision level


/* Mnemonics for opcodes that depend on native machine word size.   */
/* (The _V suffix means 'variable'; i.e., '_D' or '_Q', depending   */
/* on native machine word size -- 32-bit or 64-bit, resp.)          */
int LEA_V	= LEA;		//belongs in globals???
int ADDR_V	= ADDR_D;
int ADD_V	= ADD_D;
int ADDI_V	= ADDI_D;
int ORR_V	= ORR_D;
int OR_V	= OR_D;
int ADCR_V	= ADCR_D;
int ADC_V	= ADC_D;
int SBBR_V	= SBBR_D;
int SBB_V	= SBB_D;
int ANDR_V	= ANDR_D;
int AND_V	= AND_D;
int ANDI_V	= ANDI_D;
int SUBR_V	= SUBR_D;
int SUB_V	= SUB_D;
int SUBI_V  = SUBI_D;
int XORR_V	= XORR_D;
int XOR_V	= XOR_D;
int CMPR_V	= CMPR_D;
int CMP_V	= CMP_D;
int MOVR_V	= MOVR_D;
int MOV_V	= MOV_D;
int XCHG_V	= XCHG_D;
int TEST_V	= TEST_D;
int INC_V	= INC_D;
int DEC_V	= DEC_D;
int IMUL32_V = IMUL32;
int IMULI_V	= IMULI_D;
int IDIV_V	= IDIV_D;
int CDQ_V	= CDQ;
int ROL1_V	= ROL1_D;
int ROR1_V	= ROR1_D;
int SHL1_V	= SHL1_D;
int SHLI_V	= SHLI_D;
int SAR1_V	= SAR1_D;
int SARI_V	= SARI_D;
int FILD_V	= FILD_I;
int FISTP_V = FISTP_I;
int NEG_V	= NEG_D;
int NOT_V	= NOT_D;
int MOVSXD_V = MOVSXD_D;
int MOVSXB_V = MOVSXB_D;
int MOVSXW_V = MOVSXW_D;
int SHL_V	= SHL_D;
int SHR_V	= SHR_D;
int CMPI_V	= CMPI_D;

/* Stuff for EmitX86... */

static BYTE CB_channelCode[240000];  //temporary code buffer
static BYTE CB_saveChannelCode[240000];

CB cbCodeBuffer;
CB cbChannelCode;
CB cbSaveChannelCode;
CB cbInitCode;
CB cbDivCheck;
CB cbZeroDivide;
CB cbDivideOverflow;
CB cbOnCtl;
CB cbOnFilterStart;
CB cbForEveryTile;
CB cbOnFilterEnd;
CB cbUserFunctions;

void X86_print(const char *msg)
{
    PRINTF(msg);
} /*X86_print*/

void X86_abort(const char *msg)
{
#ifdef FULL
    PRINTF2("\r\nX86 EMIT: Fatal error - %s\r\n", msg);
#endif
    msgBox(MB_ICONERROR, "X86 Fatal Error", msg);
    gParams->fValidCompile = FALSE;
    longjmp(Mde_error_exit_jb, 1);
} /*X86_abort*/

void X86_abort2(const char *fmt, int iarg1)
{
    char msgBuffer[1024];
	//should check result of following to detect buffer overflow!!
	_snprintf(msgBuffer, sizeof(msgBuffer), fmt, iarg1);
	msgBuffer[sizeof(msgBuffer)-1] = '\0';	//ensure NUL-terminated
    X86_abort(msgBuffer);
} /*X86_abort2*/

void X86_error(const char *msg)
{
    int reply;
#ifdef FULL
    PRINTF2("\r\nX86 EMIT: Error - %s\r\n", msg);
#endif
    reply = msgBox(MB_ICONERROR|MB_OKCANCEL, "X86 Error", msg);
    if (reply == IDCANCEL) {
        Mde_abort_command(MD386__USERCAN, 0, 0, 0);
    }
} /*X86_error*/

void X86_error2(const char *fmt, int iarg1)
{
    char msgBuffer[1024];
	//should check result of following to detect buffer overflow!!
	_snprintf(msgBuffer, sizeof(msgBuffer), fmt, iarg1);
	msgBuffer[sizeof(msgBuffer)-1] = '\0';	//ensure NUL-terminated
    X86_error(msgBuffer);
} /*X86_error2*/

void X86_warn(const char *msg)
{
    int reply;
#ifdef FULL
    PRINTF2("\r\nX86 EMIT: Warning - %s\r\n", msg);
#endif
    reply = msgBox(MB_ICONWARNING|MB_OKCANCEL, "X86 Warning", msg);
    if (reply == IDCANCEL) {
        Mde_abort_command(MD386__USERCAN, 0, 0, 0);
    }
} /*X86_warn*/

        /************************************/
        /*                                  */
        /*                                  */
        /*      Structure of SDL items      */
        /*                                  */
        /*                                  */
        /************************************/

 typedef struct
         {
             UINT1          kind;
             UINT1          type;
             UINT2          nbytes;
             UINT4          param[1];
         } SDLITEMSTRUCTURE;


        /************************************/
        /*                                  */
        /*         Global variables         */
        /*                                  */
        /************************************/

#if 1
 //HORRIBLE KLUDGE!!! For now, only arrays and pointers in FM are to CHARs,
 //so temporarily force PLM pointer rules, making scalefactor==1 in all
 //pointer math.  !!!! YECHHHHHH !!!!
 UINT4 Mdg_language = LANG_PLM;
#else
 UINT4 Mdg_language = LANG_C;
#endif
 UINT1 Mdg_current_mode_number_rep = 0; /* IEEE mode */		//UNUSED???
 UINT4 Mdg_current_default_type_index = 0;


        /************************************/
        /*                                  */
        /*         Evaluation stack         */
        /*                                  */
        /************************************/

 GENERALOBJECT         Estack[MAX_TEMPS];

 static INT4           tos; /* index to top of stack */
 static INT4           vos; /* index to last VAR on stack */
 static GENERALOBJECT *tosp; /* pointer to top of stack */
 static GENERALOBJECT *nosp; /* pointer to next object below top of stack */
 static GENERALOBJECT *nos2p; /* pointer to 2nd object below top of stack */
 static VAROBJECT     *vosp; /* pointer to last VAR on stack */

 #define TOSOBJECT (*(GENERALOBJECT *) tosp)
 #define TOSVAR (*(VAROBJECT *) tosp)
 #define TOSID (*(IDOBJECT *) tosp)
 #define TOSLIN (*(LINOBJECT *) tosp)
 #define TOSLIT (*(LITOBJECT *) tosp)
 #define TOSREG (*(REGOBJECT *) tosp)
 #define TOSSEG (*(SEGOBJECT *) tosp)
 #define TOSGRP (*(GRPOBJECT *) tosp)
 #define TOSFIL (*(FILOBJECT *) tosp)
 #define TOSDCL (*(DECLOBJECT *) tosp)
 #define TOSTYP (*(TYPOBJECT *) tosp)

 #define NOSOBJECT (*(GENERALOBJECT *) nosp)
 #define NOSVAR (*(VAROBJECT *) nosp)
 #define NOSID (*(IDOBJECT *) nosp)
 #define NOSLIN (*(LINOBJECT *) nosp)
 #define NOSLIT (*(LITOBJECT *) nosp)
 #define NOSREG (*(REGOBJECT *) nosp)
 #define NOSSEG (*(SEGOBJECT *) nosp)
 #define NOSGRP (*(GRPOBJECT *) nosp)
 #define NOSFIL (*(FILOBJECT *) nosp)
 #define NOSDCL (*(DECLOBJECT *) nosp)
 #define NOSTYP (*(TYPOBJECT *) nosp)

 #define NOS2OBJECT (*(GENERALOBJECT *) nos2p)
 #define NOS2VAR (*(VAROBJECT *) nos2p)
 #define NOS2ID (*(IDOBJECT *) nos2p)
 #define NOS2LIN (*(LINOBJECT *) nos2p)
 #define NOS2LIT (*(LITOBJECT *) nos2p)
 #define NOS2REG (*(REGOBJECT *) nos2p)
 #define NOS2SEG (*(SEGOBJECT *) nos2p)
 #define NOS2GRP (*(GRPOBJECT *) nos2p)
 #define NOS2FIL (*(FILOBJECT *) nos2p)
 #define NOS2DCL (*(DECLOBJECT *) nos2p)
 #define NOS2TYP (*(TYPOBJECT *) nos2p)

 #define VOSOBJECT (*(GENERALOBJECT *) vosp)
 #define VOSVAR (*(VAROBJECT *) vosp)


        /************************************/
        /*                                  */
        /*                                  */
        /*          Literal Space           */
        /*                                  */
        /*                                  */
        /************************************/

 static UINT1          literalspace[4096];	//UNUSED???
 static UINT4          nextavailablebyte;	//UNUSED???


 /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
 /* PLMC32-generated prototypes for local functions (useful     */
 /* for resolving forward references in the original PLM code). */
 /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
 static void add(void);
 static void advance_sdlp(void);
 static void assign_add(void);
 static void assign_and(void);
 static void assign_div(void);
 static void assign_ior(void);
 static void assign_lsh(void);
 static void assign_rem(void);
 static void assign_mul(void);
 static void assign_rsh(void);
 static void assign_sub(void);
 static void assign_xor(void);
 static void bug(UINT4 n);
 static void dec_stack(void);
 static void dereference(void);
 static void divide(void);
 static double dpvalue(void *litobjectp);
 static void duplicate(void);
 static void ensurenoperands(int n);
 static void ensureoneoperand(void);
 static void ensurethreeoperands(void);
 static void ensuretwooperands(void);
 static void fillbytes(int nbytes, int val);
 static void getbytes(UINT4 nbytes, void *bufferp);
 static UINT4 gettypelength(UINT4 typeindex);
 static void init_arglist(void);
 static void *installqualifier(void *lnamep);
 static int integer_is_zero(void *litobjectp);
 static void integer8value(void *litobjectp, void *valuep);
 static INT4 integervalue(void *litobjectp);
 static void logical_and(void);
 static void logical_not(void);
 static void logical_or(void);
 static void lookup(void);
 static void lookupid(void);
 static void markstack(void);
 static void multiply(void);
 static void negate(void);
 static void op_abs(void);
 static void op_addr(void);
 static void op_and(void);
 static void op_arglist(void);
 static void op_arg(void);
 static void op_call(void);
 static void op_cast(void);
 static void op_ceil(void);
 static void op_chop(void);
 static void op_cnvX_Y(int x_or_y);
 static void op_ctl(void);
 static void op_cos(void);
 static void op_dif(void);
 static void op_else(void);
 static void op_eq(void);
 static void op_floor(void);
 static void op_gamma(void);
 static void op_ge(void);
 static void op_get(void);
 static void op_gt(void);
 static void op_if(void);
 static void op_ior(void);
 static void op_le(void);
 static void op_lt(void);
 static void op_max(void);
 static void op_min(void);
 static void op_ne(void);
 static void op_nop(void);
 static void op_not(void);
 //static void op_on_ctl(void);
 //static void op_on_divide_overflow(void);
 //static void op_on_end(void);
 //static void op_on_init_dialog(void);
 //static void op_on_zero_divide(void);
 static void op_postdec(void);
 static void op_postinc(void);
 static void op_pow(void);
 static void op_predec(void);
 static void op_preinc(void);
 static void op_put(void);
 static void op_region(int region_op);
 static void op_round(void);
 static void op_sin(void);
 static void op_sizeof(void);
 static void op_sqr(void);
 static void op_tan(void);
 static void op_then(void);
 static void op_val(void);
 static void op_xchg(void);
 static void op_xor(void);
 static void performoperation(void);
 static void physicaladdress(void);
 static void pop(void);
 static void pop_arglist(void);
 static void push_arglist(void);
 static void push(UINT1 kind);
 static void pushdeclitem(UINT1 type);
 static void pushfilitem(void);
 static void pushidlititem(UINT1 kind);
 static void pushlabitem(void);
 static void pushlastitem(void);
 static void pushlinitem(void);
 static void pushlititem(void);
 static void pushnameitem(UINT1 kind);
 static void pushnextitem(void);
 static void pushpcitem(void);
 static void pushregitem(void);
 static void pushvaritem(UINT4 typeindex, UINT4 offset, UINT2 frame);
 static void putbytes(UINT4 nbytes, void *bufferp);
 //static void qualify(void);
 static void remainder(void);
 static void setmode(void);
 static void setradix(void);
 static void settype(void);
 static void settypefromvar(void);
 static void shiftleft(void);
 static void shiftright(void);
 static void store(void);
 static void *storeliteral(UINT4 nbytes, void *valuep);
 static void subtract(void);


 /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
 /* PLMC32-generated declarations for file-level copies of   */
 /* local variables referenced from "inner" functions.       */
 /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
#define SDLITEM        (*(SDLITEMSTRUCTURE *) M3__sdlp)
 static void           *M3__sdlp;
 static UINT4          regindex;	//unused by FM??
 static int            if_depth;
 static int            eosdl;
 static int            sdlop;

 static int            compile_flag;
 static struct {
     CGL else_lab;
     CGL endif_lab;
 } if_info[128];


        /************************************/
        /*                                  */
        /*                                  */
        /*              bug                 */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void bug(UINT4 n) /* NOTE: last used bug # = 24 */
 {
     Mde_abort_command(MD386__BUG, n, 3, (UINT_PTR) "MDC");
 } /*bug*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              warn                */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void warn(char *msg, UINT_PTR p1, UINT_PTR p2, UINT_PTR p3)
 {
     int reply;

     reply = msgBox(MB_ICONWARNING|MB_OKCANCEL, "FilterMeister Warning", msg, p1, p2, p3);
     if (reply == IDCANCEL) {
         Mde_abort_command(MD386__USERCAN, 0, 0, 0);
     }
 } /*warn*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          store$literal           */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void *storeliteral(UINT4 nbytes, void *valuep)		//UNUSED??? Maybe should use it for FM64??
 {
     if (nextavailablebyte + nbytes > sizeof(literalspace))
         Mde_abort_command(MD386__LITOV, 0, 0, 0);

     memcpy(&literalspace[nextavailablebyte], valuep, nbytes);
     nextavailablebyte += nbytes;
     return &literalspace[nextavailablebyte - nbytes];
 } /*storeliteral*/


        /************************************/
        /*                                  */
        /*                                  */
        /*       check_ndp_stack_room       */
        /*                                  */
        /*                                  */
        /************************************/
#define MAX_NDP_SLOTS 8		//for x87 ndp
 // call check_ndp_stack_room(+n) to allocate n ndp stack slots
 // call check_ndp_stack_room(-n) to release n ndp stack slots
 // call n = check_ndp_stack_room(0) to set n to current # of stack slots allocated
 // call check_ndp_stack_room(MAX_NDP_SLOTS+1) to reset allocated stack slots to 0
 /**procedure**/
 static int check_ndp_stack_room(int ndp_slots)
 {
	 static int ndp_slots_in_use = 0;
	 static int display_overflow_error = true;

	 if (ndp_slots == MAX_NDP_SLOTS+1) {
		 //(re)initialize...
		 ndp_slots_in_use = 0;
		 display_overflow_error = true;
		 EMIT2L(";st=%d.", ndp_slots_in_use);
	 }
	 else if (ndp_slots > 0) {
		 //need to add ndp stack slots...
		 ndp_slots_in_use += ndp_slots;
		 EMIT2L(";st=%d:", ndp_slots_in_use);
		 if (ndp_slots_in_use >= MAX_NDP_SLOTS + 1) {
			 //ERROR -- ndp stack overflow!!
			 if (display_overflow_error) {
				 X86_error2("Floating point stack overflow:\r\n"
							"Expression too complex (%d levels).",
							ndp_slots_in_use);
				 //suppress further messages for now...
				 display_overflow_error = false;
			 }
			 //but keep counting to see if it happens again...
		 }
	 }
	 else if (ndp_slots < 0) {
		 //decrement slots in use...
		 ndp_slots_in_use += ndp_slots;
		 EMIT2L(";st=%d^", ndp_slots_in_use);
		 if (ndp_slots_in_use == 0) {
			 //re-enable display of overflow msgs...
			 display_overflow_error = true;
		 }
		 else if (ndp_slots_in_use < 0) {
			 //ABORT -- ndp stack underflow is a compiler error
			 X86_abort2("Floating point stack underflow (%d.)\r\n"
					    "FM compiler error.", ndp_slots_in_use);
			 //reset the error in case we try to continue...
			 ndp_slots_in_use = 0;
		 }
	 }
	 return ndp_slots_in_use;
 } /*check_ndp_stack_room*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              push                */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void push(UINT1 kind)
 {
     if (tos >= MAX_TEMPS - 1)
         Mde_abort_command(MD386__EVALOV, 0, 0, 0);

     tos++;
	 X86_record_tos(tos);
     tosp = &Estack[tos];
     nosp = &Estack[tos - 1];
     nos2p = &Estack[tos - 2];
     memset(tosp, 0, sizeof(*tosp));
     TOSOBJECT.kind = kind;
 } /*push*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            dec_stack             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void dec_stack(void)
 {
     if (tos < 0) bug(30);
     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL) {
         //decrement the ndp stack
		 check_ndp_stack_room(-1);
     }
     tos--;
     tosp = &Estack[tos];
     nosp = &Estack[tos - 1];
     nos2p = &Estack[tos - 2];
 } /*dec_stack*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              pop                 */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pop(void)
 {
     if (tos < 0) bug(2);
     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL) {
         //pop the ndp stack
         EMIT("fpop");
         X86_emit_op(FPOP);
         //X86_emit_op_R(FSTP_STN, 0);	//alternate encoding
		 check_ndp_stack_room(-1);
     }
     tos--;
     tosp = &Estack[tos];
     nosp = &Estack[tos - 1];
     nos2p = &Estack[tos - 2];
 } /*pop*/


        /************************************/
        /*                                  */
        /*                                  */
        /*        ensure$one$operand        */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void ensureoneoperand(void)
 {
     if (tos < 0) bug(6);
 } /*ensureoneoperand*/


        /************************************/
        /*                                  */
        /*                                  */
        /*        ensure$two$operands       */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void ensuretwooperands(void)
 {
     if (tos < 1) bug(5);
 } /*ensuretwooperands*/


        /************************************/
        /*                                  */
        /*                                  */
        /*      ensure$three$operands       */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void ensurethreeoperands(void)
 {
     if (tos < 2) bug(5);
 } /*ensurethreeoperands*/


        /************************************/
        /*                                  */
        /*                                  */
        /*        ensure$n$operands         */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void ensurenoperands(int n)
 {
     if (tos < n - 1) bug(5);
 } /*ensurenoperands*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             op_xchg              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_xchg(void)
 {
     GENERALOBJECT temp;

     ensuretwooperands();
     //swap at run-time only if actually needed...
     if (NOSOBJECT.kind == VAROBJECTKIND ||
         NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type != LITTYPEREAL) {
         if (TOSOBJECT.kind == VAROBJECTKIND ||
             TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type != LITTYPEREAL) {
             EMIT3("xchg r%d, r%d ;op_xchg", tos-1, tos);
             X86_emit_op_TT(XCHG_V, tos-1, tos);
         }
#ifndef SEPARATE_INTEGER_STACK	//eh???
         else {
			 //real stays put on x87 stack, integral moves up on integer stack(???)
             EMIT3("mov r%d, r%d ;op_xchg", tos, tos-1);
             X86_emit_op_TT(MOV_V, tos, tos-1);
         }
#endif
     }
#ifndef SEPARATE_INTEGER_STACK	//eh???
     else if (TOSOBJECT.kind == VAROBJECTKIND ||
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type != LITTYPEREAL) {
		 //real stays put on x87 stack, integral moves down on integer stack(???)
         EMIT3("mov r%d, r%d ;op_xchg", tos-1, tos);
         X86_emit_op_TT(MOV_V, tos-1, tos);
     }
#endif
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL) {
         EMIT("fxch st(1) ;op_xchg");
         X86_emit_op_R(FXCH_STN, 1);
     }
     //finally, swap the entries on the evaluation stack
     temp = *tosp;
     *tosp = *nosp;
     *nosp = temp;
} /*op_xchg*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          push$name$item          */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pushnameitem(UINT1 kind)
 {
     push(kind);
     TOSID.lnamep = &SDLITEM.param[0];
     TOSID.nbytes = SDLITEM.param[0];
     TOSID.charp = *(char **)&SDLITEM.param[1];
     TOSID.qualp = NULL;
 } /*pushnameitem*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          push$idlit$item         */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pushidlititem(UINT1 kind)
 {
     assert(kind == IDOBJECTKIND);
     push(kind);
     /* create a string descriptor within the item -- ACK! doesn't survive OP_XCHG!!! */
     TOSID.nbytes = SDLITEM.nbytes - 1; //subtract the ASCIZ NUL char
     TOSID.charp = (char *)&SDLITEM.param[0];
     TOSID.lnamep = &TOSID.nbytes;
     TOSID.qualp = NULL;
#if 0
     PRINTF4("nbytes=%d, charp->%.4s, lnamep=%x\r\n",
         TOSID.nbytes, TOSID.charp, TOSID.lnamep);
     PRINTF4("&nbytes=%x, &charp=%x, &lnamep=%x\r\n",
         &TOSID.nbytes, &TOSID.charp, &TOSID.lnamep);
#endif
 } /*pushidlititem*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          push$FIL$item           */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pushfilitem(void)						//UNUSED??
 {
     push(FILOBJECTKIND);
     TOSFIL.len = SDLITEM.param[0];
     TOSFIL.ptr = (char **) &SDLITEM.param[1];
 } /*pushfilitem*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          push$LIT$item           */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pushlititem(void)
 {
     push(LITOBJECTKIND);
     TOSLIT.type = SDLITEM.type;
     TOSLIT.nbytes = (UINT2)SDLITEM.nbytes;
     TOSLIT.valuep = (UINT1 *)SDLITEM.param;
     if (TOSLIT.type == LITTYPEINTEGER || TOSLIT.type == LITTYPELOGICAL) {
         if (TOSLIT.nbytes == 4) {												//X64??
             TOSLIT.ivalue = SDLITEM.param[0];		//X64??
             TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
             //
             //  Peephole opts:
             //
             // value = 0  ==> xor r,r		   if r is a reg	//LEAVES FALSE DEPENDENCY??
             //                and m,sext(0)   if r is mem, but uses r/w cycle 
             //								   and is 7 clks vs 2 for mov m,0 ???
             //
             if (TOSLIT.ivalue == 0 && isReg(tos)) {
                 int reg = tmpReg(tos);
                 EMIT4("xor r%d, r%d ;r%d=0", tos, tos, tos);
                 X86_emit_op_RR(XOR_D, reg, reg);	//X64 zero-extends to 64-bits
             }
             //
             // value = -1  ==> or r,sext(-1)  if r is a reg	//LEAVES FALSE DEPENDENCY??
			 //					or m,sext(-1)  if r is mem, but uses r/w cycle
			 //								   and is 7 clks vs 2 for mov m,-1 ???
             //
             else if (TOSLIT.ivalue == -1 && isReg(tos)) {
                 int reg = tmpReg(tos);
                 EMIT4("or r%d, %d ;r%d=-1", tos, TOSLIT.ivalue, tos);
				 if (X64gen) {
	                 X86_emit_op_R_simm32(Q(ORI_D), reg, TOSLIT.ivalue);
				 } else
				 { //x86
					 X86_emit_op_R_simm32(ORI_D, reg, TOSLIT.ivalue);	//USE ORI_V???
				 }
             }
             else {
				EMIT3("mov r%d, %d", tos, TOSLIT.ivalue);
				if (X64gen) {	//<===STILL NOT RIGHT if bit 31 is set!!!  TEMP KLUDGE!!
					//shouldn't X64 zero-extend here???? (Lits are unsigned at present)
					X86_emit_op_T_imm32(MOVI_Q, tos, TOSLIT.ivalue); //sign-extends imm32 to r/m64
				}
				else { //X32
					X86_emit_op_T_imm32(MOVI_D, tos, TOSLIT.ivalue);
				}
			}
         }
         else bug(5601);	//nbytes != 4
     }
     else if (TOSLIT.type == LITTYPEREAL) {
         if (TOSLIT.nbytes == 8) {
             TOSLIT.rvalue = *(double *)SDLITEM.param;
             TOSLIT.valuep = (UINT1 *)&TOSLIT.rvalue;
            if (TOSLIT.rvalue == 0.0) {
				check_ndp_stack_room(1);		//need one more ndp stack entry
                EMIT("fldz");
                X86_emit_op(FLDZ);
            }
            else if (TOSLIT.rvalue == 1.0) {
				check_ndp_stack_room(1);		//need one more ndp stack entry
                EMIT("fld1");
                X86_emit_op(FLD1);
            }
            // could also approximate PI, etc. (Use OP_PI, OP_E, etc.?)
            else {
                if (gFmcp->ndlits >= MAX_DLITS) {
                    Mde_fatal(MD386__DLITOV, MAX_DLITS, 0, 0);
                }
                gFmcp->dlitPool[gFmcp->ndlits] = TOSLIT.rvalue;
				check_ndp_stack_room(1);		//need one more ndp stack entry
                EMIT3("fld_d dlitPool[%d]; =%#g", gFmcp->ndlits, TOSLIT.rvalue);
                X86_emit_op_M(FLD_D, FMC(dlitPool) + gFmcp->ndlits*8);	//8=sizeof(double)
                gFmcp->ndlits++;
            }
         }
		 else bug(5602);	//nbytes != 8
     }
     else if (TOSLIT.type == LITTYPECHAR) {
         // Store string in the dlit pool for now...
         int qwords_needed = (TOSLIT.nbytes + 7)/8;	//8=sizeof(double)
         if (gFmcp->ndlits + qwords_needed > MAX_DLITS) {
             Mde_fatal(MD386__DLITOV, MAX_DLITS, 0, 0); //for now,,,
         }
         //Use memcpy instead of strcpy, since string may contain embedded nuls ('\0')!!!
         memcpy(&gFmcp->dlitPool[gFmcp->ndlits], TOSLIT.valuep, TOSLIT.nbytes);
         EMIT4("lea r%d, dlitPool[%d]; =\"%s\"", tos, gFmcp->ndlits, TOSLIT.valuep);
         X86_emit_op_TM(LEA_V, tos, FMC(dlitPool) + gFmcp->ndlits*8);
         gFmcp->ndlits += qwords_needed;
     }
 } /*pushlititem*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          push$VAR$item           */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pushvaritem(UINT4 typeindex, UINT4 offset, UINT2 frame)
 {
     push(VAROBJECTKIND);
     TOSVAR.typeindex = typeindex;
     TOSVAR.offset = offset;
     TOSVAR.frame = frame;
	 TOSVAR.namep = NULL;
     TOSVAR.pSym = NULL;
     TOSVAR.pType = NULL;
	 TOSVAR.flags = 0;
     TOSVAR.extradata = 0;
	 TOSVAR.nbytes = 0;
 } /*pushvaritem*/


        /***********************************/
        /*                                 */
        /*          push$PC$item           */
        /*                                 */
        /***********************************/

 /**procedure**/
 static void pushpcitem(void)
 {
     bug(17);
     //Mdt_check_task_registers();
     //pushvaritem(Mdg_label_type_index, Mdg_8086_state.eip, Mdg_8086_state.cs);
 } /*pushpcitem*/


        /************************************/
        /*                                  */
        /*          push$LAST$item          */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pushlastitem(void)						//UNUSED??
 {
         Mde_abort_command(MD386__NOLAST, 0, 0, 0);
 } /*pushlastitem*/


        /************************************/
        /*                                  */
        /*          push$NEXT$item          */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pushnextitem(void)						//UNUSED??
 {
         Mde_abort_command(MD386__NONEXT, 0, 0, 0);
 } /*pushnextitem*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              set$type            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void settype(void)						//UNUSED??
 {
     bug(18);
 } /*settype*/


        /************************************/
        /*                                  */
        /*                                  */
        /*         set$type$from$VAR        */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void settypefromvar(void)						//UNUSED??
 {
         bug(10);
 } /*settypefromvar*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             set$radix            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void setradix(void)						//UNUSED??
 {
 } /*setradix*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             set$mode             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void setmode(void)						//UNUSED??
 {
     bug(11);
 } /*setmode*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          push$REG$item           */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pushregitem(void)						//UNUSED??
 {
         bug(3);
 } /*pushregitem*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          push$LIN$item           */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pushlinitem(void)						//UNUSED??
 {
 } /*pushlinitem*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          push$LAB$item           */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pushlabitem(void)						//UNUSED??
 {
     Mde_abort_command(MD386__NOLABEL, SDLITEM.param[0], 0, 0);
 } /*pushlabitem*/


        /************************************/
        /*                                  */
        /*                                  */
        /*        install$qualifier         */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void *installqualifier(void *lnamep)						//UNUSED??
 {
     return 0;
 } /*installqualifier*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          integer_is_zero         */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static int integer_is_zero(void *litobjectp)						//UNUSED??
 {
     #define LIT (*(LITOBJECT *) litobjectp)
     int     i;

     if (LIT.nbytes == 4) {
         return (LIT.ivalue == 0);
     }

     for (i = LIT.nbytes-1; i >= 0 && LIT.valuep[i] == 0; i--)
         ;
     if (i < 0)
         return 1;
     else
         return 0;
     
     #undef LIT
 } /*integer_is_zero*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          integer$8$value         */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void integer8value(void *litobjectp, void *valuep)						//UNUSED??
 {
     #define LIT (*(LITOBJECT *) litobjectp)
     #define VALUE (*(UINT1 (*)[8]) valuep)   /* integer*8 */

     if (LIT.nbytes == 4) {
         ((INT4 *)valuep)[0] = LIT.ivalue;
         ((INT4 *)valuep)[1] = LIT.ivalue < 0 ? -1 : 0;
     }
     else if (LIT.nbytes < 8)
     {   /* sign-extend to integer*8 */
         memcpy(VALUE, LIT.valuep, LIT.nbytes);
         if ((VALUE[LIT.nbytes - 1] & 0x80) != 0)
             memset(&VALUE[LIT.nbytes], 0xFF, 8 - LIT.nbytes);
         else 
             memset(&VALUE[LIT.nbytes], 0, 8 - LIT.nbytes);
     }
     else 
     {   /* truncate to integer*8 */
         memcpy(VALUE, LIT.valuep, 8);
     }
     
     #undef LIT
     #undef VALUE
} /*integer8value*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          integer$value           */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static INT4 integervalue(void *litobjectp)						//UNUSED??
 {
     #define LIT (*(LITOBJECT *) litobjectp)
     INT4           value;

     // assert(LIT.kind == LITOBJECTKIND);
     // assert(LIT.type == LITTYPEINTEGER || LIT.type == LITTYPELOGICAL);
     // assert(LIT.nbytes > 0);

     if (LIT.nbytes == 4) {
         return LIT.ivalue;
     }

     if (LIT.nbytes <= 4) {
         value = 0;
         memcpy(&value, LIT.valuep, LIT.nbytes);
         /* NON-PORTABLE!! Assumes >> sign-extends a signed value !!! */
         if (LIT.nbytes == 1)
             value = (value << 24) >> 24;
         else if (LIT.nbytes == 2)
             value = (value << 16) >> 16;
         else if (LIT.nbytes == 3)
             value = (value << 8) >> 8;
     }
     else {
         // truncate to 4 byte value
         memcpy(&value, LIT.valuep, 4);
     }

     return value;
     
     #undef LIT
 } /*integervalue*/


        /************************************/
        /*                                  */
        /*              dp$value            */
        /*                                  */
        /************************************/

 /**procedure**/
 static double dpvalue(void *litobjectp)
 {
     #define LIT (*(LITOBJECT *) litobjectp)

     if (LIT.nbytes == 8)
     {
         return LIT.rvalue;
     }
     else if (LIT.nbytes == 4)
     {
         ///return (double)*(float *)LIT.valuep;
         return LIT.rvalue;
     }
     else 
     {
         bug(7);
         return 0.0;
     }
     
     #undef LIT
 } /*dpvalue*/


        /************************************/
        /*                                  */
        /*             op_region            */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_region(int region_op)
 {	 //a proc to emit code for all RGN ops,
	 //where region_op is a binary region operator
	 //NOS = SRC1 and DST (handle to region)
	 //TOS = SRC2 (handle to region)
	 //computes NOS = NOS region_op TOS by calling:
	 //CombineRgn((HRGN)NOSLIT.ivalue, (HRGN)NOSLIT.ivalue, (HRGN)TOSLIT.ivalue, region_op);

#ifdef X64
	 if (X64gen) {
 		 ///EMIT("and RSP, -16 ;align 16");			//not needed if stack always aligned
		 EMIT("sub RSP, 32 ;alloc arglist");		//32 = 4 args * 8 bytes //NOT NEEDED if level-1!!
		 switch (region_op) {
		 case RGN_AND:
			 EMIT2("mov R9, %d ;RGN_AND", region_op);
			 break;
		 case RGN_OR:
			 EMIT2("mov R9, %d ;RGN_OR", region_op);
			 break;
		 case RGN_XOR:
			 EMIT2("mov R9, %d ;RGN_XOR", region_op);
			 break;
		 case RGN_DIFF:
			 EMIT2("mov R9, %d ;RGN_DIFF", region_op);
			 break;
		 case RGN_COPY:
			 EMIT2("mov R9, %d ;RGN_COPY", region_op);
			 break;
		 default:
			 bug(447766);
			 break;
		 } //switch
		 EMIT2("mov R8, r%d", tos);					//operand 2
		 EMIT2("mov RDX, r%d", tos-1);				//operand 1
		 EMIT("mov RCX, RDX");						//operand 1 is also the destination
		 EMIT("call CombineRgn");					//FASTCALL saves rsp
		 EMIT("add RSP, 32 ;pop arglist");			//restore (aligned) stack

		 ///X86_emit_op_R_simm32(ANDI_V, rsp, -16);	//not needed if stack always aligned 16
		 X86_emit_op_R_simm32(SUBI_V, rsp, 32);
		 X86_emit_op_R_simm32(MOVI_D, r9, region_op);// region_op is int32
		 X86_emit_op_RT(MOV_V, r8, tos);			//HRGN is a 64-bit pointer (the upper half 0??)
		 X86_emit_op_RT(MOV_V, rdx, tos-1);
		 X86_emit_op_RR(MOV_V, rcx, rdx);			//faster if tos-1 is a tmp.
		 X86_emit_op_M(CALLI_N, FMC(CombineRgn));   //FASTCALL, saves rsp
		 X86_emit_op_R_simm32(ADDI_V, rsp, 32);
	 } else
#endif X64
	 {	//X86
		 EMIT2("push %d ;RGN_OP", region_op);
		 EMIT2("push r%d", tos);					//operand 2
		 EMIT2("push r%d", tos-1);					//operand 1
		 EMIT2("push r%d", tos-1);					//operand 1 is also the destination
		 EMIT("call CombineRgn");

		 X86_emit_op_simm32(PUSHI_D, region_op);
		 X86_emit_op_T(PUSH_D, tos);
		 X86_emit_op_T(PUSH_D, tos-1);
		 X86_emit_op_T(PUSH_D, tos-1);
		 X86_emit_op_M(CALLI_N, FMC(CombineRgn));   //STDCALL pops args
	 } //X86
	 pop();	// pop operand 2 from TOS
	 // operand 1 (result) is now in TOS
 } /*op_region*/
		 
 
		/************************************/
        /*                                  */
        /*                                  */
        /*              add                 */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void add(void)
 {
     double         dpresult;
     UINT4          scalefactor;

     ensuretwooperands();

     if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == LITOBJECTKIND && 
         TOSLIT.type == LITTYPEINTEGER)
     {   /* ptr + int --> ptr */
         if (Mdg_language == LANG_C)
         {   /* For C-language, need to scale the addition. */
             scalefactor = gettypelength(NOSVAR.typeindex);
             // Should complain if pointer type is void * or a function ptr!!!
             if (scalefactor == 0)
             {   /* Illegal to perform arithmetic on pointer to void. */
                 Mde_warn(MD386__VOIDPTR, 0, 0, 0);
                 scalefactor = 1;
             }
         }
         else
         {
             /* Languages such as PLM don't scale pointer arithmetic. */
             scalefactor = 1;
             warn("Scaled pointer arithmetic is not yet implemented;\n"
                  "unscaled arithmetic will be used instead.", 0, 0, 0);
         }
         if (scalefactor == 2) {
             EMIT2("shl r%d, 1", tos);
             X86_emit_op_T(SHL1_V, tos);
         }
         else if (scalefactor == 4) {
             EMIT2("shl r%d, 2", tos);
             X86_emit_op_T_imm8(SHLI_V, tos, 2);
         }
         else if (scalefactor == 8) {
             EMIT2("shl r%d, 3", tos);
             X86_emit_op_T_imm8(SHLI_V, tos, 3);
         }
         else if (scalefactor == 16) {
             EMIT2("shl r%d, 4", tos);
             X86_emit_op_T_imm8(SHLI_V, tos, 4);
         }
         else if (scalefactor != 1) {
             EMIT3("imul r%d, %d", tos, scalefactor);
             X86_emit_op_T_simm32(IMULI_V, tos, scalefactor);
         }
         EMIT3("add r%d, r%d", tos-1, tos);
         X86_emit_op_TT(ADD_V, tos-1, tos);
         NOSVAR.offset = (UINT4) (NOSVAR.offset + integervalue(tosp)*scalefactor);
         pop();
     }
     else if (TOSOBJECT.kind == VAROBJECTKIND && NOSOBJECT.kind == LITOBJECTKIND && 
              NOSLIT.type == LITTYPEINTEGER)
     {   /* int + ptr --> ptr */
         if (Mdg_language == LANG_C)
         {   /* For C-language, need to scale the addition. */
             scalefactor = gettypelength(TOSVAR.typeindex);
             // Should complain if pointer type is void * or a function ptr!!!
             if (scalefactor == 0)
             {   /* Illegal to perform arithmetic on pointer to void. */
                 Mde_warn(MD386__VOIDPTR, 0, 0, 0);
                 scalefactor = 1;
             }
         }
         else
         {
             /* Languages such as PLM don't scale pointer arithmetic. */
             scalefactor = 1;
             warn("Scaled pointer arithmetic is not yet implemented;\n"
                  "unscaled arithmetic will be used instead.", 0, 0, 0);
         }
         if (scalefactor == 2) {
             EMIT2("shl r%d, 1", tos-1);
             X86_emit_op_T(SHL1_V, tos-1);
         }
         else if (scalefactor == 4) {
             EMIT2("shl r%d, 2", tos-1);
             X86_emit_op_T_imm8(SHLI_V, tos-1, 2);
         }
         else if (scalefactor == 8) {
             EMIT2("shl r%d, 3", tos-1);
             X86_emit_op_T_imm8(SHLI_V, tos-1, 3);
         }
         else if (scalefactor == 16) {
             EMIT2("shl r%d, 4", tos-1);
             X86_emit_op_T_imm8(SHLI_V, tos-1, 4);
         }
         else if (scalefactor != 1) {
             EMIT3("imul r%d, %d", tos-1, scalefactor);
             X86_emit_op_T_simm32(IMULI_V, tos-1, scalefactor);
         }
         EMIT3("add r%d, r%d", tos-1, tos);
         X86_emit_op_TT(ADD_V, tos-1, tos);
         TOSVAR.offset = (UINT4) (TOSVAR.offset + integervalue(nosp)*scalefactor);
         *nosp = *tosp;
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {   /* int + int --> int */
         NOSLIT.nbytes = 4;
         NOSLIT.ivalue = integervalue(nosp) + integervalue(tosp);
         NOSLIT.valuep = (UINT1 *)&NOSLIT.ivalue;
         EMIT3("add r%d, r%d", tos-1, tos);
         X86_emit_op_TT(ADD_V, tos-1, tos);
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {   /* real + real --> real */
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpvalue(nosp) + dpvalue(tosp);
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         //EMIT3("faddp st(%d), st(%d)", tos-1, tos);
         EMIT("fadd");
		 check_ndp_stack_room(-1);		//FADD pops 1 ndp stack entry
		 X86_emit_op(FADD);
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {   /* int + real --> real */
         dpresult = integervalue(nosp) + dpvalue(tosp);
         NOSLIT.type = LITTYPEREAL;
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         EMIT2("fadd_i r%d", tos-1);
         X86_emit_op_Tm(FADD_I, tos-1);
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {   /* real + int --> real */
         dpresult = dpvalue(nosp) + integervalue(tosp);
         NOSLIT.type = LITTYPEREAL;
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         EMIT2("fadd_i r%d", tos);
         X86_emit_op_Tm(FADD_I, tos);
         pop();
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*add*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            op_sizeof             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_sizeof(void)
 {
     int isize = 0;
	 int scalarlength;
	 int pointersize = X64gen ? sizeof(__int64) : sizeof(__int32);

     ensureoneoperand();

     if (TOSOBJECT.kind == VAROBJECTKIND) {
         //an lvalue
		 scalarlength = abs(TOSVAR.nbytes);
         if (TOSVAR.typeindex == 0)
         {   /* integer var*/
             isize = scalarlength;	//handled by parser for now??
         }
         else if (TOSVAR.typeindex == 126)
         {   /* real var */
             isize = scalarlength;	//handled by parser for now??
         }
         else if (TOSVAR.typeindex == 1234)
         {   /* integer function */
             isize = pointersize;
         }
         else if (TOSVAR.typeindex == 1237)
         {   /* real function */
             isize = pointersize;
         }
         else if (TOSVAR.typeindex == 1238)
         {   /* region function */
             isize = pointersize;
         }
         else if (TOSVAR.typeindex == 4321)
         {   /* array of char */
             isize = 256; //for now, is always 256 bytes!!
         }
         else {
             Mde_abort_command(MD386__BADMIX, 0, 0, 0); //for now
         }
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND) {
         //an rvalue
         if (TOSLIT.type == LITTYPEINTEGER)
         {   /* integer value */
             isize = pointersize;	//32 or 64 bits (calc32 or calc64)
         }
         else if (TOSLIT.type == LITTYPEREAL)
         {   /* real value */
             isize = sizeof(long double);	//?? calc80
         }
         else if (TOSLIT.type == LITTYPECHAR)
         {   /* char value */
             isize = sizeof(char);
         }
         else if (TOSLIT.type == LITTYPELOGICAL)
         {   /* boolean value */
             isize = 1; //for now??? later may change to 4
         }
         else if (TOSLIT.type == LITTYPEREGION)
         {   /* region value */
             isize = pointersize;	//??
         }
         else {
             Mde_abort_command(MD386__BADMIX, 0, 0, 0); //for now
         }
     }
     else {
         Mde_abort_command(MD386__BADMIX, 0, 0, 0); //for now
     }
     //change TOS to an integer value
     TOSOBJECT.kind = LITOBJECTKIND;
     TOSLIT.type = LITTYPEINTEGER;
     TOSLIT.nbytes = 4;	//??
     TOSLIT.ivalue = isize; //size of var or literal value in bytes
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
     EMIT3("mov r%d, %d", tos, TOSLIT.ivalue);
     X86_emit_op_T_imm32(MOVI_D, tos, TOSLIT.ivalue);	//X64 zero-extends
} /*op_sizeof*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            op_postinc            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_postinc(void)
 {
     INT4			scalarlength;

     ensureoneoperand();

     if (TOSOBJECT.kind != VAROBJECTKIND) {
         //must be an lvalue
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }

	 scalarlength = TOSVAR.nbytes;

	 //check the variable's 'const' attribute...
	 if (TOSVAR.flags & fmfCONST) {
		 //can't modify a const var!!
		 //for now, we just issue a warning and
		 //proceed anyway
         warn("Attempt to increment a 'const' variable: %s",
			  (UINT_PTR)TOSVAR.namep, 0, 0);
	 }

     if (TOSVAR.typeindex == 0)
     {   /* int++ */
#if 0	//fm 1.0 code
         //need to check that tos+1 doesn't overflow # of temps...
		 push(LITOBJECTKIND); dec_stack();	//also updates tosmax
         EMIT3("movsxd r%d, [r%d] ;op_load", tos+1, tos);
         X86_emit_op_TTi(MOVSXD_V, tos+1, tos);	//X64 op_load assumes int32 for now
         EMIT2("inc dword ptr[r%d] ;int32++", tos);
         X86_emit_op_Ti(INC_D, tos);			//X64 assumes 32-bit var for now!!!
         EMIT3("mov r%d, r%d", tos, tos+1);
         X86_emit_op_TT(MOV_V, tos, tos+1);
         //change TOS to an integer value
         TOSOBJECT.kind = LITOBJECTKIND;
         TOSLIT.type = LITTYPEINTEGER;
         TOSLIT.nbytes = 4;
         TOSLIT.ivalue = 0;
         TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
#elif 0	//fm 2.0 code
		 // Cheap kludge for now: preinc the operand, then decrement rvalue
		 // (mod 2^n??) to return original value of the operand...
		 // quick & dirty implementation based on preinc:
		 //	op_preinc(tos);		//tos = var+1 mod size
		 //	dec_v tos			//may overflow
		 // mov[s/z][b/w/d/q]x	//narrow/widen to orig value (omit if mov[s/z][v]v!!)
		 op_preinc();
		 EMIT2("dec_v r%d ;restore orig val", tos);
		 X86_emit_op_T(DEC_V, tos);
		 //still need to narrow/widen the original value!!
#elif 1
		 // assumes dereference(tos) doesn't clobber tos+1:
		 duplicate();				//copy var object in tos to tos+1(check ovfl)
		 dec_stack();				//pop tmp stack
		 dereference();				//load original value in tos
		 // inc_/b/w/d/q [tos+1]	//increment the var in mem based on var size
		 switch (abs(scalarlength)) {
		 case 1: //Increment byte/int8
			 EMIT2("inc_b [r%d] ;int8++", tos+1);
			 X86_emit_op_Ti(INC_B, tos+1);
			 break;
		 case 2: //Increment word/int16
			 EMIT2("inc_w [r%d] ;int16++", tos+1);
			 X86_emit_op_Ti(INC_W, tos+1);
			 break;
		 case 4: //Increment dword/int32
			 EMIT2("inc_d [r%d] ;int32++", tos+1);
			 X86_emit_op_Ti(INC_D, tos+1);
			 break;
		 case 8: //Increment qword/int64
			 if (X64gen) {
				 EMIT2("inc_q [r%d] ;int64++", tos+1);
				 X86_emit_op_Ti(INC_Q, tos+1);
				 break;
			 }//else X32gen falls through to default...
		 default: //unknown type size, or not yet implemented
			 Mde_abort_command(MD386__NYI, 30, 0, 0);
			 break;
		 }//switch
#endif //fm 2.0 code
     }
     else if (TOSVAR.typeindex == 126)
     {   /* real++ */
         //need to check that FPU stack doesn't overflow...
		 switch (scalarlength) {
		 case 4: //post-increment float
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT2("fld_f [r%d]", tos);			//load x; st = x
			 X86_emit_op_Ti(FLD_F, tos);		//op_load float
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT("fld1");						//st = 1.0, x
			 X86_emit_op(FLD1);
			 EMIT("fadd st, st(1) ;float++");	//st = x+1.0, x
			 X86_emit_op_R(FADD_STN, 1);
			 EMIT2("fstp_f [r%d]", tos);		//store x=x+1.0; st = x
			 check_ndp_stack_room(-1);			//free one ndp stack entry
			 X86_emit_op_Ti(FSTP_F, tos);		//op_store float
			 break;
		 case 8: //post-increment double
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT2("fld_d [r%d]", tos);			//load x; st = x
			 X86_emit_op_Ti(FLD_D, tos);		//op_load double
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT("fld1");						//st = 1.0, x
			 X86_emit_op(FLD1);
			 EMIT("fadd st, st(1) ;double++");	//st = x+1.0, x
			 X86_emit_op_R(FADD_STN, 1);
			 EMIT2("fstp_d [r%d]", tos);		//store x=x+1.0; st = x
			 check_ndp_stack_room(-1);			//free one ndp stack entry
			 X86_emit_op_Ti(FSTP_D, tos);		//op_store double
			 break;
		 case 10: //post-increment tempreal
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT2("fld_t [r%d]", tos);			//load x; st = x
			 X86_emit_op_Ti(FLD_T, tos);		//op_load tempreal
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT("fld1");						//st = 1.0, x
			 X86_emit_op(FLD1);
			 EMIT("fadd st, st(1) ;treal++");	//st = x+1.0, x
			 X86_emit_op_R(FADD_STN, 1);
			 EMIT2("fstp_t [r%d]", tos);		//store x=x+1.0; st = x
			 check_ndp_stack_room(-1);			//free one ndp stack entry
			 X86_emit_op_Ti(FSTP_T, tos);		//op_store tempreal
			 break;
		 case 2: // post-increment half
			 //fall thru to NYI
		 case -10: // post-increment bcd
			 //fall thru to NYI
		 default: //all other types
			 Mde_abort_command(MD386__NYI, 31, 0, 0);
			 break;
		 }//switch
         //change TOS to a real value
         TOSOBJECT.kind = LITOBJECTKIND;
         TOSLIT.type = LITTYPEREAL;
         TOSLIT.nbytes = 8;
         TOSLIT.rvalue = 0.0;
         TOSLIT.valuep = (UINT1 *)&TOSLIT.rvalue;
     }
#if 0
     //this is not right yet!!!	//X64!!
     else if (TOSVAR.typeindex == ptrtype???)
     {   /* ptr++ */
         UINT4 scalefactor;
         if (Mdg_language == LANG_C)
         {   /* For C-language, need to scale the addition. */
             scalefactor = gettypelength(TOSVAR.typeindex);
             // Should complain if pointer type is void * or a function ptr!!!
             if (scalefactor == 0)
             {   /* Illegal to perform arithmetic on pointer to void. */
                 Mde_warn(MD386__VOIDPTR, 0, 0, 0);
                 scalefactor = 1;
             }
         }
         else
         {
             /* Languages such as PLM don't scale pointer arithmetic. */
             scalefactor = 1;
             warn("Scaled pointer arithmetic is not yet implemented;\n"
                  "unscaled arithmetic will be used instead.", 0, 0, 0);
         }

         //need to check that tos+1 doesn't overflow # of temps...
		 push(); dec_stack();	//also updates tosmax
         EMIT3("mov r%d, [r%d]", tos+1, tos);
         X86_emit_op_TTi(MOV_V, tos+1, tos);	//X64 variable ptr size
         EMIT3("add [r%d], %d ;ptr++", tos, scalefactor);	//or INC_V if scalefactor==1??
         X86_emit_op_Ti_simm32(ADDI_V, tos, scalefactor);	//X64 variable ptr size
         EMIT3("mov r%d, r%d", tos, tos+1);
         X86_emit_op_TT(MOV_V, tos, tos+1);	//X64 variable ptr size
     }
#endif
     else {
         //for now...
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }
 } /*op_postinc*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            op_postdec            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_postdec(void)
 {
     INT4			scalarlength;

     ensureoneoperand();

     if (TOSOBJECT.kind != VAROBJECTKIND) {
         //must be an lvalue
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }

	 scalarlength = TOSVAR.nbytes;

	 //check the variable's 'const' attribute...
	 if (TOSVAR.flags & fmfCONST) {
		 //can't modify a const var!!
		 //for now, we just issue a warning and
		 //proceed anyway
         warn("Attempt to decrement a 'const' variable: %s",
			  (UINT_PTR)TOSVAR.namep, 0, 0);
	 }

     if (TOSVAR.typeindex == 0)
     {   /* int-- */
#if 0	//fm 1.0 code
         //need to check that tos+1 doesn't overflow # of temps...
		 push(LITOBJECTKIND); dec_stack();	//also updates tosmax
         EMIT3("movsxd r%d, [r%d] ;op_load", tos+1, tos);
         X86_emit_op_TTi(MOVSXD_V, tos+1, tos);	//X64 op_load assumes int32 for now
         EMIT2("dec dword ptr[r%d] ;int32--", tos);
         X86_emit_op_Ti(DEC_D, tos);	//X64 assumes 32-bit var for now!!!
         EMIT3("mov r%d, r%d", tos, tos+1);
         X86_emit_op_TT(MOV_V, tos, tos+1);
         //change TOS to an integer value
         TOSOBJECT.kind = LITOBJECTKIND;
         TOSLIT.type = LITTYPEINTEGER;
         TOSLIT.nbytes = 4;
         TOSLIT.ivalue = 0;
         TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
#elif 0	//fm 2.0 code
		 // Cheap kludge for now: predec the operand, then increment rvalue
		 // (mod 2^n??) to return original value of the operand...
		 // quick & dirty implementation based on predec:
		 //	op_predec(tos);		//tos = var-1 mod size
		 //	inc_v tos			//may overflow
		 // mov[s/z][b/w/d/q]x	//narrow/widen to orig value (omit if mov[s/z][v]v!!)
		 op_predec();
		 EMIT2("inc_v r%d ;restore orig val", tos);
		 X86_emit_op_T(INC_V, tos);
		 //still need to narrow/widen the original value!!
#elif 1
		 // assumes dereference(tos) doesn't clobber tos+1:
		 duplicate();				//copy var object in tos to tos+1(check ovfl)
		 dec_stack();				//pop tmp stack
		 dereference();				//load original value in tos
		 // dec_/b/w/d/q [tos+1]	//decrement the var in mem based on var size
		 switch (abs(scalarlength)) {
		 case 1: //Decrement byte/int8
			 EMIT2("dec_b [r%d] ;int8--", tos+1);
			 X86_emit_op_Ti(DEC_B, tos+1);
			 break;
		 case 2: //Decrement word/int16
			 EMIT2("dec_w [r%d] ;int16--", tos+1);
			 X86_emit_op_Ti(DEC_W, tos+1);
			 break;
		 case 4: //Decrement dword/int32
			 EMIT2("dec_d [r%d] ;int32--", tos+1);
			 X86_emit_op_Ti(DEC_D, tos+1);
			 break;
		 case 8: //Decrement qword/int64
			 if (X64gen) {
				 EMIT2("dec_q [r%d] ;int64--", tos+1);
				 X86_emit_op_Ti(DEC_Q, tos+1);
				 break;
			 }//else X32gen falls through to default...
		 default: //unknown type size, or not yet implemented
			 Mde_abort_command(MD386__NYI, 32, 0, 0);
			 break;
		 }//switch
#endif //fm 2.0 code
     }
     else if (TOSVAR.typeindex == 126)
     {   /* real-- */
         //need to check that FPU stack doesn't overflow...
		 switch (scalarlength) {
		 case 4: //post-decrement float
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT2("fld_f [r%d]", tos);			//load x; st = x
			 X86_emit_op_Ti(FLD_F, tos);		//op_load float
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT("fld1");						//st = 1.0, x
			 X86_emit_op(FLD1);
			 EMIT("fsubr st, st(1) ;float--");	//st = x-1.0, x
			 X86_emit_op_R(FSUBR_STN, 1);
			 EMIT2("fstp_f [r%d]", tos);		//store x=x-1.0; st = x
			 check_ndp_stack_room(-1);			//free one ndp stack entry
			 X86_emit_op_Ti(FSTP_F, tos);		//op_store float
			 break;
		 case 8: //post-decrement double
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT2("fld_d [r%d]", tos);			//load x; st = x
			 X86_emit_op_Ti(FLD_D, tos);		//op_load double
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT("fld1");						//st = 1.0, x
			 X86_emit_op(FLD1);
			 EMIT("fsubr st, st(1) ;double--");	//st = x-1.0, x
			 X86_emit_op_R(FSUBR_STN, 1);
			 EMIT2("fstp_d [r%d]", tos);		//store x=x-1.0; st = x
			 check_ndp_stack_room(-1);			//free one ndp stack entry
			 X86_emit_op_Ti(FSTP_D, tos);		//op_store double
			 break;
		 case 10: //post-decrement tempreal
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT2("fld_t [r%d]", tos);			//load x; st = x
			 X86_emit_op_Ti(FLD_T, tos);		//op_load tempreal
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT("fld1");						//st = 1.0, x
			 X86_emit_op(FLD1);
			 EMIT("fsubr st, st(1) ;treal--");	//st = x-1.0, x
			 X86_emit_op_R(FSUBR_STN, 1);
			 EMIT2("fstp_t [r%d]", tos);		//store x=x-1.0; st = x
			 check_ndp_stack_room(-1);			//free one ndp stack entry
			 X86_emit_op_Ti(FSTP_T, tos);		//op_store tempreal
			 break;
		 case 2: // post-decrement half
			 //fall thru to NYI
		 case -10: // post-decrement bcd
			 //fall thru to NYI
		 default: //all other types
			 Mde_abort_command(MD386__NYI, 33, 0, 0);
			 break;
		 }//switch
        //change TOS to a real value
         TOSOBJECT.kind = LITOBJECTKIND;
         TOSLIT.type = LITTYPEREAL;
         TOSLIT.nbytes = 8;
         TOSLIT.rvalue = 0.0;
         TOSLIT.valuep = (UINT1 *)&TOSLIT.rvalue;
     }
#if 0
     //this is not right yet!!!	//X64!!
     else if (TOSVAR.typeindex == ptrtype???)
     {   /* ptr-- */
         UINT4 scalefactor;
         if (Mdg_language == LANG_C)
         {   /* For C-language, need to scale the addition. */
             scalefactor = gettypelength(TOSVAR.typeindex);
             // Should complain if pointer type is void * or a function ptr!!!
             if (scalefactor == 0)
             {   /* Illegal to perform arithmetic on pointer to void. */
                 Mde_warn(MD386__VOIDPTR, 0, 0, 0);
                 scalefactor = 1;
             }
         }
         else
         {
             /* Languages such as PLM don't scale pointer arithmetic. */
             scalefactor = 1;
             warn("Scaled pointer arithmetic is not yet implemented;\n"
                  "unscaled arithmetic will be used instead.", 0, 0, 0);
         }

         //need to check that tos+1 doesn't overflow # of temps...
		 push(); dec_stack();	//also updates tosmax
         X86_emit_op_TTi(MOV_V, tos+1, tos);	//X64 variable ptr size
         EMIT3("sub [r%d], %d ;ptr--", tos, scalefactor);
         X86_emit_op_Ti_simm32(SUBI_V, tos, scalefactor);	//or DEC_V if scalefactor==1??
         EMIT3("mov r%d, r%d", tos, tos+1);
         X86_emit_op_TT(MOV_V, tos, tos+1);	//X64 variable ptr size
     }
#endif
     else {
         //for now...
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }
 } /*op_postdec*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            op_preinc             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_preinc(void)
 {
     INT4			scalarlength;

     ensureoneoperand();

     if (TOSOBJECT.kind != VAROBJECTKIND) {
         //must be an lvalue
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }

	 scalarlength = TOSVAR.nbytes;

	 //check the variable's 'const' attribute...
	 if (TOSVAR.flags & fmfCONST) {
		 //can't modify a const var!!
		 //for now, we just issue a warning and
		 //proceed anyway
         warn("Attempt to increment a 'const' variable: %s",
			  (UINT_PTR)TOSVAR.namep, 0, 0);
	 }

     if (TOSVAR.typeindex == 0)
     {   /* ++int */
#if 0	//fm 1.0 code
         EMIT2("inc dword ptr[r%d] ;++int32", tos);
         X86_emit_op_Ti(INC_D, tos);			//X64 assumes 32-bit var for now!!!
         EMIT3("movsxd r%d, [r%d] ;op_load", tos, tos);
         X86_emit_op_TTi(MOVSXD_V, tos, tos);	//X64 op_load assumes int32 for now
         //change TOS to an integer value
         TOSOBJECT.kind = LITOBJECTKIND;
         TOSLIT.type = LITTYPEINTEGER;
         TOSLIT.nbytes = 4;
         TOSLIT.ivalue = 0;
         TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
#else	//fm 2.0 code
		 // Ignore signed/unsigned by taking absolute value of
		 // scalar length
		 switch (abs(scalarlength)) {
		 case 1:
			 //X64: increment int8/byte
			 EMIT2("inc byte ptr[r%d] ;++int8", tos);
			 X86_emit_op_Ti(INC_B, tos);
			 break;
		 case 2:
			 //X64: increment int16/word
			 EMIT2("inc word ptr[r%d] ;++int16", tos);
			 X86_emit_op_Ti(INC_W, tos);
			 break;
		 case 4:
			 //X64: increment int32/dword
			 EMIT2("inc dword ptr[r%d] ;++int32", tos);
			 X86_emit_op_Ti(INC_D, tos);
			 break;
		 case 8:
			 if (X64gen) {
			 //X64: increment int64/qword
				 EMIT2("inc qword ptr[r%d] ;++int64", tos);
				 X86_emit_op_Ti(INC_Q, tos);
				 break;
			 }//X64gen
			 //else if (X32gen)--falls through to default case for now...
		 default:
			 Mde_abort_command(MD386__NYI, 5, 0, 0);
			 break;
		 }//switch
		 dereference();	//finally, op_load the reference
#endif
     }
     else if (TOSVAR.typeindex == 126)
     {   /* ++real */
         //need to check that FPU stack doesn't overflow...
		 switch (scalarlength) {
		 case 4: // pre-increment float
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT("fld1");						// st = 1.0
			 X86_emit_op(FLD1);

			 EMIT2("fadd_f [r%d] ;++float", tos);// st = x+1.0
			 X86_emit_op_Ti(FADD_F, tos);		// add float

			 EMIT2("fst_f [r%d]", tos);			// store x=x+1.0; st = x+1.0
			 X86_emit_op_Ti(FST_F, tos);		// op_store float
		  	 break;
		 case 8: // pre-increment double
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT("fld1");						// st = 1.0
			 X86_emit_op(FLD1);

			 EMIT2("fadd_d [r%d] ;++double", tos);// st = x+1.0
			 X86_emit_op_Ti(FADD_D, tos);		// add double

			 EMIT2("fst_d [r%d]", tos);			// store x=x+1.0; st = x+1.0
			 X86_emit_op_Ti(FST_D, tos);		// op_store double
			 break;
		 case 10: //pre-increment tempreal
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT2("fld_t [r%d]", tos);			// st = x
			 X86_emit_op_Ti(FLD_T, tos);		// load tempreal

			 EMIT("fadd_f =1.0F ;++treal");		// st = x+1.0
			 X86_emit_op_M(FADD_F, FMC(float1p0));	// float1p0 = 1.0F

			 if (check_ndp_stack_room(0) < MAX_NDP_SLOTS) {
				 //at least one free ndp slot...
				 check_ndp_stack_room(1);			//need one more ndp stack entry
				 EMIT("fdup ;++treal");				// st = x+1.0 x+1.0
				 X86_emit_op(FDUP);					// FDUP = FLD ST(0) (saves 1 memref)

				 EMIT2("fstp_t [r%d]", tos);		// st = x+1.0; store x=x+1.0;
				 X86_emit_op_Ti(FSTP_T, tos);		// op_store treal
				 check_ndp_stack_room(-1);			//FSTP_T pops 1 ndp stack entry
			 }
			 else {
				 //no free ndp slots...
				 EMIT2("fstp_t [r%d]", tos);		// store x=x+1.0
				 X86_emit_op_Ti(FSTP_T, tos);		// op_store treal
				 check_ndp_stack_room(-1);			//FSTP_T pops 1 ndp stack entry

				 check_ndp_stack_room(1);			//FLD_T adds 1 ndp stack entry
				 EMIT2("fld_t [r%d] ;++treal", tos);// st = x
				 X86_emit_op_Ti(FLD_T, tos);		// reload tempreal rvalue
			 }
			 break;
		 case 2: // pre-increment half
			 //fall thru to NYI
		 case -10: // pre-increment bcd
			 //fall thru to NYI
		 default: //all other types
			 Mde_abort_command(MD386__NYI, 6, 0, 0);
			 break;
		 }
         //change TOS to a real value
         TOSOBJECT.kind = LITOBJECTKIND;
         TOSLIT.type = LITTYPEREAL;
         TOSLIT.nbytes = 8;
         TOSLIT.rvalue = 0.0;
         TOSLIT.valuep = (UINT1 *)&TOSLIT.rvalue;
     }
#if 0
     //this is not right yet!!!
     else if (TOSVAR.typeindex == ptrtype???)
     {   /* ++ptr */
         UINT4 scalefactor;
         if (Mdg_language == LANG_C)
         {   /* For C-language, need to scale the addition. */
             scalefactor = gettypelength(TOSVAR.typeindex);
             // Should complain if pointer type is void * or a function ptr!!!
             if (scalefactor == 0)
             {   /* Illegal to perform arithmetic on pointer to void. */
                 Mde_warn(MD386__VOIDPTR, 0, 0, 0);
                 scalefactor = 1;
             }
         }
         else
         {
             /* Languages such as PLM don't scale pointer arithmetic. */
             scalefactor = 1;
             warn("Scaled pointer arithmetic is not yet implemented;\n"
                  "unscaled arithmetic will be used instead.", 0, 0, 0);
         }

         //need to check that tos+1 doesn't overflow # of temps???	//WRONG, DOES p++, not ++p?????
		 push(); dec_stack();	//also updates tosmax
         EMIT3("mov r%d, [r%d]", tos+1, tos);
         X86_emit_op_TTi(MOV_V, tos+1, tos);	//X64 variable ptr size
         EMIT3("add [r%d], %d ;++ptr", tos, scalefactor);
         X86_emit_op_Ti_simm32(ADDI_V, tos, scalefactor);	//or INC_V if scalefactor==1??
         EMIT3("mov r%d, r%d", tos, tos+1);
         X86_emit_op_TT(MOV_V, tos, tos+1);	//X64 variable ptr size
     }
#endif
     else {
         //for now...
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }
 } /*op_preinc*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            op_predec             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_predec(void)
 {
     INT4			scalarlength;

     ensureoneoperand();

     if (TOSOBJECT.kind != VAROBJECTKIND) {
         //must be an lvalue
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }

	 scalarlength = TOSVAR.nbytes;

	 //check the variable's 'const' attribute...
	 if (TOSVAR.flags & fmfCONST) {
		 //can't modify a const var!!
		 //for now, we just issue a warning and
		 //proceed anyway
         warn("Attempt to decrement a 'const' variable: %s",
			  (UINT_PTR)TOSVAR.namep, 0, 0);
	 }

     if (TOSVAR.typeindex == 0)
     {   /* --int */
#if 0	//fm 1.0 code
         EMIT2("dec dword ptr[r%d] ;--int32", tos);
         X86_emit_op_Ti(DEC_D, tos);			//X64 assumes 32-bit var for now!!!
         EMIT3("movsxd r%d, [r%d] ;op_load", tos, tos);
         X86_emit_op_TTi(MOVSXD_V, tos, tos);	//X64 op_load assumes int32 for now
         //change TOS to an integer value
         TOSOBJECT.kind = LITOBJECTKIND;
         TOSLIT.type = LITTYPEINTEGER;
         TOSLIT.nbytes = 4;
         TOSLIT.ivalue = 0;
         TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
#else	//fm 2.0 code
		 // Ignore signed/unsigned by taking absolute value of
		 // scalar length
		 switch (abs(scalarlength)) {
		 case 1:
			 //X64: decrement int8/byte
			 EMIT2("dec byte ptr[r%d] ;--int8", tos);
			 X86_emit_op_Ti(DEC_B, tos);
			 break;
		 case 2:
			 //X64: decrement int16/word
			 EMIT2("dec word ptr[r%d] ;--int16", tos);
			 X86_emit_op_Ti(DEC_W, tos);
			 break;
		 case 4:
			 //X64: decrement int32/dword
			 EMIT2("dec dword ptr[r%d] ;--int32", tos);
			 X86_emit_op_Ti(DEC_D, tos);
			 break;
		 case 8:
			 if (X64gen) {
			 //X64: decrement int64/qword
				 EMIT2("dec qword ptr[r%d] ;--int64", tos);
				 X86_emit_op_Ti(DEC_Q, tos);
				 break;
			 }//X64gen
			 //else if (X32gen)--falls through to default case for now...
		 default:
			 Mde_abort_command(MD386__NYI, 7, 0, 0);
			 break;
		 }//switch
		 dereference();	//finally, op_load the reference
#endif
	 }
     else if (TOSVAR.typeindex == 126)
     {   /* --real */
         //need to check that FPU stack doesn't overflow...
		 switch (scalarlength) {
		 case 4: // pre-decrement float
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT("fld1");						// st = 1.0
			 X86_emit_op(FLD1);

			 EMIT2("fsubr_f [r%d] ;--float", tos);// st = x-1.0
			 X86_emit_op_Ti(FSUBR_F, tos);		// reverse subtract float

			 EMIT2("fst_f [r%d]", tos);			// store x=x-1.0; st = x-1.0
			 X86_emit_op_Ti(FST_F, tos);		// op_store float
		  	 break;
		 case 8: // pre-decrement double
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT("fld1");						// st = 1.0
			 X86_emit_op(FLD1);

			 EMIT2("fsubr_d [r%d] ;--double", tos);// st = x-1.0
			 X86_emit_op_Ti(FSUBR_D, tos);		// reverse subtract double

			 EMIT2("fst_d [r%d]", tos);			// store x=x-1.0; st = x-1.0
			 X86_emit_op_Ti(FST_D, tos);		// op_store double
			 break;
		 case 10: //pre-decrement tempreal
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT2("fld_t [r%d]", tos);			// st = x
			 X86_emit_op_Ti(FLD_T, tos);		// load tempreal

			 EMIT("fsub_f =1.0F ;--treal");		// st = x-1.0
			 X86_emit_op_M(FSUB_F, FMC(float1p0));	// float1p0 = 1.0F

			 if (check_ndp_stack_room(0) < MAX_NDP_SLOTS) {
				 //at least one free ndp slot...
				 check_ndp_stack_room(1);			//need one more ndp stack entry
				 EMIT("fdup ;--treal");				// st = x-1.0 x-1.0
				 X86_emit_op(FDUP);					// FDUP = FLD ST(0) (saves 1 memref later)

				 EMIT2("fstp_t [r%d]", tos);		// st = x-1.0; store x=x-1.0;
				 X86_emit_op_Ti(FSTP_T, tos);		// op_store treal
				 check_ndp_stack_room(-1);			//FSTP_T pops 1 ndp stack entry
			 }
			 else {
				 //no free ndp slots...
				 EMIT2("fstp_t [r%d]", tos);		// store x=x-1.0
				 X86_emit_op_Ti(FSTP_T, tos);		// op_store treal
				 check_ndp_stack_room(-1);			//FSTP_T pops 1 ndp stack entry

				 check_ndp_stack_room(1);			//FLD_T adds 1 ndp stack entry
				 EMIT2("fld_t [r%d] ;--treal", tos);// st = x
				 X86_emit_op_Ti(FLD_T, tos);		// reload tempreal rvalue
			 }
			 break;
		 case 2: // pre-decrement half
			 //fall thru to NYI
		 case -10: // pre-decrement bcd
			 //fall thru to NYI
		 default: //all other types
			 Mde_abort_command(MD386__NYI, 8, 0, 0);
			 break;
		 }
         //change TOS to a real value
         TOSOBJECT.kind = LITOBJECTKIND;
         TOSLIT.type = LITTYPEREAL;
         TOSLIT.nbytes = 8;
         TOSLIT.rvalue = 0.0;
         TOSLIT.valuep = (UINT1 *)&TOSLIT.rvalue;
     }
#if 0
     //this is not right yet!!!
     else if (TOSVAR.typeindex == ptrtype???)
     {   /* --ptr */
         UINT4 scalefactor;
         if (Mdg_language == LANG_C)
         {   /* For C-language, need to scale the addition. */
             scalefactor = gettypelength(TOSVAR.typeindex);
             // Should complain if pointer type is void * or a function ptr!!!
             if (scalefactor == 0)
             {   /* Illegal to perform arithmetic on pointer to void. */
                 Mde_warn(MD386__VOIDPTR, 0, 0, 0);
                 scalefactor = 1;
             }
         }
         else
         {
             /* Languages such as PLM don't scale pointer arithmetic. */
             scalefactor = 1;
             warn("Scaled pointer arithmetic is not yet implemented;\n"
                  "unscaled arithmetic will be used instead.", 0, 0, 0);
         }

         //need to check that tos+1 doesn't overflow # of temps???	//WRONG, DOES p--, not --p?????
		 push(); dec_stack();	//also updates tosmax
         EMIT3("mov r%d, [r%d]", tos+1, tos);
         X86_emit_op_TTi(MOV_V, tos+1, tos);	//X64 variable ptr size
         EMIT3("sub [r%d], %d ;--ptr", tos, scalefactor);
         X86_emit_op_Ti_simm32(SUBI_V, tos, scalefactor);	//or DEC_V if scalefactor==1??
         EMIT3("mov r%d, r%d", tos, tos+1);
         X86_emit_op_TT(MOV_V, tos, tos+1);	//X64 variable ptr size
     }
#endif
     else {
         //for now...
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }
 } /*op_predec*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            subtract              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void subtract(void)
 {
     double         dpresult;
     UINT4          scalefactor;

     ensuretwooperands();

     if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == LITOBJECTKIND && 
         TOSLIT.type == LITTYPEINTEGER)
     {   /* ptr - int --> ptr */
         if (Mdg_language == LANG_C)
         {   /* For C-language, need to scale the subtraction. */
             scalefactor = gettypelength(NOSVAR.typeindex);
             // Should complain if pointer type is void * or a function ptr!!!
             if (scalefactor == 0)
             {   /* Illegal to perform arithmetic on pointer to void. */
                 Mde_warn(MD386__VOIDPTR, 0, 0, 0);
                 scalefactor = 1;
             }
         }
         else
         {
             /* Languages such as PLM don't scale pointer arithmetic. */
             scalefactor = 1;
             warn("Scaled pointer arithmetic is not yet implemented;\n"
                  "unscaled arithmetic will be used instead.", 0, 0, 0);
         }
         if (scalefactor == 2) {
             EMIT2("shl r%d, 1", tos);
             X86_emit_op_T(SHL1_V, tos);	//X64 variable ptr size -- CORRECT???
         }
         else if (scalefactor == 4) {
             EMIT2("shl r%d, 2", tos);
             X86_emit_op_T_imm8(SHLI_V, tos, 2);	//X64 variable ptr size
         }
         else if (scalefactor == 8) {
             EMIT2("shl r%d, 3", tos);
             X86_emit_op_T_imm8(SHLI_V, tos, 3);	//X64 variable ptr size
         }
         else if (scalefactor == 16) {
             EMIT2("shl r%d, 4", tos);
             X86_emit_op_T_imm8(SHLI_V, tos, 4);	//X64 variable ptr size
         }
         else if (scalefactor != 1) {
             EMIT3("imul r%d, %d", tos, scalefactor);
             X86_emit_op_T_simm32(IMULI_V, tos, scalefactor);	//X64 variable ptr size
         }
         EMIT3("sub r%d, r%d", tos-1, tos);
         X86_emit_op_TT(SUB_V, tos-1, tos);	//X64 variable ptr size
         NOSVAR.offset = (UINT4) (NOSVAR.offset - integervalue(tosp)*scalefactor);
         pop();
     }
     else if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == VAROBJECTKIND)
     {   /* ptr - ptr --> int */
         INT4   diff;
         if (Mdg_language == LANG_C)
         {   /* For C-language, need to scale the difference. */
             scalefactor = gettypelength(NOSVAR.typeindex);
             // For ANSI C, should warn if ptrs are not pointing to the same array!
             // Operation is illegal for void * or function ptr types...
             if (scalefactor != gettypelength(TOSVAR.typeindex))
             {   /* ptr types are (way!) incompatible */
                 Mde_warn(MD386__PTRDIFF, 0, 0, 0);
             }
             if (scalefactor == 0)
             {   /* Illegal to take difference of void * ptrs. */
                 Mde_warn(MD386__VOIDPTR, 0, 0, 0);
                 scalefactor = 1;
             }
         }
         else
         {
             /* Languages such as PLM don't scale pointer arithmetic. */
             scalefactor = 1;
             warn("Scaled pointer arithmetic is not yet implemented;\n"
                  "unscaled arithmetic will be used instead.", 0, 0, 0);
         }
         if (NOSVAR.frame != TOSVAR.frame)
         {   /* Pointers with different selectors are incommensurable. */
             Mde_warn(MD386__PTRISEL, 0, 0, 0);
         }
         /* Maybe should compute diff to 8 bytes??? */
         EMIT("*!!!*scaled ptr diff***");
         EMIT3("sub r%d, r%d", tos-1, tos);
         X86_emit_op_TT(SUB_V, tos-1, tos);	//X64 variable ptr size

         /* Are shifts okay for (negative) division here, since the */
         /* diff should always be a multiple of the scale factor?   */
         if (scalefactor == 2) {
             EMIT2("sar r%d, 1", tos);
             X86_emit_op_T(SAR1_V, tos);	//X64 variable ptr size
         }
         else if (scalefactor == 4) {
             EMIT2("sar r%d, 2", tos);
             X86_emit_op_T_imm8(SARI_V, tos, 2);	//X64 variable ptr size
         }
         else if (scalefactor == 8) {
             EMIT2("sar r%d, 3", tos);
             X86_emit_op_T_imm8(SARI_V, tos, 3);	//X64 variable ptr size
         }
         else if (scalefactor == 16) {
             EMIT2("sar r%d, 4", tos);
             X86_emit_op_T_imm8(SARI_V, tos, 4);	//X64 variable ptr size
         }
         else if (scalefactor != 1) {
             // there's no idiv with immed divisor...
             warn("Unrecognized scale factor: %d", scalefactor, 0, 0);
             EMIT3("idiv r%d, %d", tos, scalefactor);
             ////X86_emit_op_T_simm32(IDIVI_D, tos, scalefactor);	//FIX THIS!!!
         }

         diff = (INT4) (NOSVAR.offset - TOSVAR.offset) / (INT4) scalefactor;
         /* Complain if (NOSVAR.offset - TOSVAR.offset) % scalefactor != 0 ??? */
         pop();
         pop();
         push(LITOBJECTKIND);
         TOSLIT.type = LITTYPEINTEGER;
         TOSLIT.nbytes = 4;	//X64
         TOSLIT.ivalue = diff;
         TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {   /* int - int --> int */
         NOSLIT.nbytes = 4;
         NOSLIT.ivalue = integervalue(nosp) - integervalue(tosp);
         NOSLIT.valuep = (UINT1 *)&NOSLIT.ivalue;
         EMIT3("sub r%d, r%d", tos-1, tos);
         X86_emit_op_TT(SUB_V, tos-1, tos);	//X64
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {   /* real - real --> real */
         dpresult = dpvalue(nosp) - dpvalue(tosp);
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         EMIT("fsub");
		 check_ndp_stack_room(-1);		//FSUB pops 1 ndp stack entry
         X86_emit_op(FSUB);	//X64 okay???
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {   /* int - real --> real */
         dpresult = integervalue(nosp) - dpvalue(tosp);
         NOSLIT.type = LITTYPEREAL;
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         EMIT2("fsubr_i r%d", tos-1);
         X86_emit_op_Tm(FSUBR_I, tos-1);
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {   /* real - int --> real */
         dpresult = dpvalue(nosp) - integervalue(tosp);
         NOSLIT.type = LITTYPEREAL;
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         EMIT2("fsub_i r%d", tos);
         X86_emit_op_Tm(FSUB_I, tos);
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREGION && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREGION)
     {
		 op_region(RGN_DIFF);
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*subtract*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            multiply              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void multiply(void)
 {
     double         dpresult;
     INT4           iresult;

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {   /* int * int --> int */
         iresult = integervalue(nosp) * integervalue(tosp);
         NOSLIT.nbytes = 4;
         NOSLIT.ivalue = iresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.ivalue;
         EMIT3("imul r%d, r%d", tos-1, tos);
         X86_emit_op_TT(IMUL32_V, tos-1, tos);
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {   /* real * real --> real */
         dpresult = dpvalue(nosp) * dpvalue(tosp);
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         EMIT("fmul");
		 check_ndp_stack_room(-1);		//FMUL pops 1 ndp stack entry
         X86_emit_op(FMUL);
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {   /* int * real --> real */
         dpresult = integervalue(nosp) * dpvalue(tosp);
         NOSLIT.type = LITTYPEREAL;
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         EMIT2("fmul_i r%d", tos-1);
         X86_emit_op_Tm(FMUL_I, tos-1);
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {   /* real * int --> real */
         dpresult = dpvalue(nosp) * integervalue(tosp);
         NOSLIT.type = LITTYPEREAL;
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         EMIT2("fmul_i r%d", tos);
         X86_emit_op_Tm(FMUL_I, tos);
         pop();
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*multiply*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             divide               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void divide(void)
 {
     double         dpresult;
     INT4           i1, i2, iresult;

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {   /* int / int --> int */
         i1 = integervalue(nosp);
         i2 = integervalue(tosp);
         if (i2 == 0 || (i1 == 0x80000000 && i2 == -1))
         {  /* would result in runtime integer divide exception! */
             iresult = 0x80000000;  /* integer NaN */
         }
         else
             iresult = i1 / i2;
         NOSLIT.nbytes = 4;
         NOSLIT.ivalue = iresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.ivalue;
         EMIT3("DIVCHK r%d, r%d", tos-1, tos);
         EMIT3("idiv r%d, r%d", tos-1, tos);
         X86_emit_DIVREM_TT(0/*div*/, tos-1, tos, TRUE/*check*/);
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {    /* real / real --> real */
         if (dpvalue(tosp) == 0.0) {
             dpresult = 9e99;
         } else {
             dpresult = dpvalue(nosp) / dpvalue(tosp);
         }
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         EMIT("fdiv");
		 check_ndp_stack_room(-1);		//FDIV pops 1 ndp stack entry
         X86_emit_op(FDIV);
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {   /* int / real --> real */
         if (dpvalue(tosp) == 0.0) {
             dpresult = 9e99;
         } else {
             dpresult = integervalue(nosp) / dpvalue(tosp);
         }
         NOSLIT.type = LITTYPEREAL;
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         EMIT2("fdivr_i r%d", tos-1);
         X86_emit_op_Tm(FDIVR_I, tos-1);
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {   /* real / int --> real */
         if (integervalue(tosp) == 0.0) {
             dpresult = 9e99;
         } else {
             dpresult = dpvalue(nosp) / integervalue(tosp);
         }
         NOSLIT.type = LITTYPEREAL;
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = dpresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;
         EMIT2("fdiv_i r%d", tos);
         X86_emit_op_Tm(FDIV_I, tos);
         pop();
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*divide*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            remainder             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void remainder(void)
 {
     INT4           i1, i2, iresult;

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {   /* int % int --> int */
         i1 = integervalue(nosp);
         i2 = integervalue(tosp);
         if (i2 == 0 || (i1 == 0x80000000 && i2 == -1))
         {  /* would result in runtime integer divide exception! */
             iresult = 0x80000000;  /* integer NaN */
         }
         else
             iresult = i1 % i2;
         NOSLIT.nbytes = 4;
         NOSLIT.ivalue = iresult;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.ivalue;
         EMIT3("DIVCHK r%d, r%d", tos-1, tos);
         EMIT3("irem r%d, r%d", tos-1, tos);
         X86_emit_DIVREM_TT(1/*rem*/, tos-1, tos, TRUE/*check*/);
         pop();
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*remainder*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             op_and               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_and(void)
 {   /* Perform bitwise-and.*/

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         NOSLIT.nbytes = 4;
         NOSLIT.ivalue = integervalue(nosp) & integervalue(tosp);
         NOSLIT.valuep = (UINT1 *)&NOSLIT.ivalue;
         EMIT3("and r%d, r%d", tos-1, tos);
         X86_emit_op_TT(AND_V, tos-1, tos);
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREGION && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREGION)
     {
		 op_region(RGN_AND);
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*op_and*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             op_ior               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_ior(void)
 {   /* Perform bitwise-inclusive-or.*/

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         NOSLIT.nbytes = 4;
         NOSLIT.ivalue = integervalue(nosp) | integervalue(tosp);
         NOSLIT.valuep = (UINT1 *)&NOSLIT.ivalue;
         EMIT3("or r%d, r%d", tos-1, tos);
         X86_emit_op_TT(OR_V, tos-1, tos);
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREGION && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREGION)
     {
		 op_region(RGN_OR);
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*op_ior*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             op_xor               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_xor(void)
 {   /* Perform bitwise-exclusive-or.*/

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         NOSLIT.nbytes = 4;
         NOSLIT.ivalue = integervalue(nosp) ^ integervalue(tosp);
         NOSLIT.valuep = (UINT1 *)&NOSLIT.ivalue;
         EMIT3("xor r%d, r%d", tos-1, tos);
         X86_emit_op_TT(XOR_V, tos-1, tos);
         pop();
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREGION && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREGION)
     {
		 op_region(RGN_XOR);
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*op_xor*/


        /************************************/
        /*                                  */
        /*                                  */
        /*           shiftleft              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void shiftleft(void)
 {   /* Perform bitwise left shift.*/

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         NOSLIT.nbytes = 4;
         NOSLIT.ivalue = integervalue(nosp) << integervalue(tosp);
         NOSLIT.valuep = (UINT1 *)&NOSLIT.ivalue;
         EMIT3("shl r%d, r%d", tos-1, tos);
         X86_emit_op_RT(MOV_D, ecx, tos);
         X86_emit_op_T(SHL_V, tos-1);
         pop();
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*shiftleft*/


        /************************************/
        /*                                  */
        /*                                  */
        /*           shiftright             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void shiftright(void)
 {   /* Perform bitwise right shift.*/
	 /* C uses SAR or SHL if opnd1 is signed or unsigned, resp */
	 /* but FF *always* uses SHR.  True???					   */

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         NOSLIT.nbytes = 4;
         NOSLIT.ivalue = integervalue(nosp) >> integervalue(tosp);
         NOSLIT.valuep = (UINT1 *)&NOSLIT.ivalue;
         EMIT3("shr r%d, r%d", tos-1, tos);
         X86_emit_op_RT(MOV_D, ecx, tos);	//count uses only the low 5 or 6 bits of CL
         X86_emit_op_T(SHR_V, tos-1);       //SHR for FF compatibilty!!!!
         pop();
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*shiftright*/


        /************************************/
        /*                                  */
        /*                                  */
        /*           logical_and            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void logical_and(void)
 {   /* Perform logical-and (NOT short-circuited!). */
	 bug(7911);	//not used at present!!

#if 0
     UINT1          result1, result2;

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(nosp))
             result1 = 1;
         else
             result1 = 0;
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(nosp))
             result1 = 1;
         else
             result1 = 0;
     }
     else if (NOSOBJECT.kind == VAROBJECTKIND)
     {   /* pointer */
         if (NOSVAR.offset|NOSVAR.frame)
             result1 = 1;
         else
             result1 = 0;
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(tosp))
             result2 = 1;
         else
             result2 = 0;
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(tosp))
             result2 = 1;
         else
             result2 = 0;
     }
     else if (TOSOBJECT.kind == VAROBJECTKIND)
     {   /* pointer */
         if (TOSVAR.offset|TOSVAR.frame)
             result2 = 1;
         else
             result2 = 0;
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();
     result1 &= result2;
     TOSLIT.kind = LITOBJECTKIND;
     TOSLIT.type = LITTYPEINTEGER;
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = result1;
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
#endif
 } /*logical_and*/


        /************************************/
        /*                                  */
        /*                                  */
        /*           logical_or             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void logical_or(void)
 {   /* Perform logical-or (NOT short-circuited!). */
	 bug(7912);	//not used at present!!

#if 0
     UINT1          result1, result2;

	 ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(nosp))
             result1 = 1;
         else
             result1 = 0;
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(nosp))
             result1 = 1;
         else
             result1 = 0;
     }
     else if (NOSOBJECT.kind == VAROBJECTKIND)
     {   /* pointer */
         if (NOSVAR.offset|NOSVAR.frame)
             result1 = 1;
         else
             result1 = 0;
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(tosp))
             result2 = 1;
         else
             result2 = 0;
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(tosp))
             result2 = 1;
         else
             result2 = 0;
     }
     else if (TOSOBJECT.kind == VAROBJECTKIND)
     {   /* pointer */
         if (TOSVAR.offset|TOSVAR.frame)
             result2 = 1;
         else
             result2 = 0;
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();
     result1 |= result2;
     TOSLIT.kind = LITOBJECTKIND;
     TOSLIT.type = LITTYPEINTEGER;
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = result1;
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
#endif
 } /*logical_or*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_eq               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_eq(void)
 {
     UINT1          result;

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(nosp) == integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT3("cmp r%d, r%d", tos-1, tos);
         //EMIT2("mov r%d, 0", tos-1);
         EMIT2("sete r%d", tos-1);
         X86_emit_op_TT(CMP_V, tos-1, tos);
         //X86_emit_op_T_imm32(MOVI_D, tos-1, 0);
         X86_emit_op_T(SETZ_D, tos-1);	//X64 zero-extends (??)
         //EMIT2("mov r%d, 1", tos-1);
         //EMIT("beq .+2");
         //EMIT2("xor r%d, r%d", tos-1);
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(nosp) == dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT("fcompp");
		 check_ndp_stack_room(-2);		//popped 2 ndp stack entries
         EMIT("fnstsw ax");				//VC does fnstsw ax/test ah,...
         EMIT("sahf");
         EMIT2("sete r%d", tos-1);		//We ignore unordered case for now!
         X86_emit_op(FCOMPP);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETE_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (integervalue(nosp) == dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos-1);
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("sete r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos-1);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETE_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (dpvalue(nosp) == integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos);
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("sete r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETE_D, tos-1);	//X64 zero-extends (??)
     }
#if 0   // PTR COMPARES ARE NOT YET IMPLEMENTED!!!
     else if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == VAROBJECTKIND)
     {   /* compare two pointers */
         if (NOSVAR.frame == TOSVAR.frame && NOSVAR.offset == TOSVAR.offset)
             result = 1;
         else
             result = 0;
     }
     else if (NOSOBJECT.kind == VAROBJECTKIND &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER &&
              integer_is_zero(tosp))
     {   /* compare pointer to integer constant 0 */
         if (NOSVAR.frame == 0 && NOSVAR.offset == 0)
             result = 1;
         else
             result = 0;
     }
     else if (TOSOBJECT.kind == VAROBJECTKIND &&
              NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              integer_is_zero(nosp))
     {   /* compare pointer to integer constant 0 */
         if (TOSVAR.frame == 0 && TOSVAR.offset == 0)
             result = 1;
         else
             result = 0;
     }
#endif //PTR COMPARES
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();
     TOSLIT.kind = LITOBJECTKIND;
     TOSLIT.type = LITTYPEINTEGER;
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = result;
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_eq*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_ne               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_ne(void)
 {
     UINT1          result;

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(nosp) != integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT3("cmp r%d, r%d", tos-1, tos);
         //EMIT2("mov r%d, 0", tos-1);
         EMIT2("setne r%d", tos-1);
         X86_emit_op_TT(CMP_V, tos-1, tos);
         //X86_emit_op_T_imm32(MOVI_D, tos-1, 0);
         X86_emit_op_T(SETNE_D, tos-1);	//X64 zero-extends (??)
         //EMIT2("mov r%d, 1", tos-1);
         //EMIT("bne .+2");
         //EMIT2("xor r%d, r%d", tos-1);
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(nosp) != dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT("fcompp");
		 check_ndp_stack_room(-2);		//popped 2 ndp stack entries
         EMIT("fnstsw ax");  //VC does fnstsw ax/test ah,...
         EMIT("sahf");
         EMIT2("setne r%d", tos-1);  //We ignore unordered case for now!
         X86_emit_op(FCOMPP);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETNE_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (integervalue(nosp) != dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos-1);
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("setne r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos-1);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETNE_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (dpvalue(nosp) != integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos);
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("setne r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETNE_D, tos-1);	//X64 zero-extends (??)
     }
#if 0   // PTR COMPARES ARE NOT YET IMPLEMENTED!!!
     else if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == VAROBJECTKIND)
     {   /* compare two pointers */
         if (NOSVAR.frame == TOSVAR.frame && NOSVAR.offset == TOSVAR.offset)
             result = 0;
         else
             result = 1;
     }
     else if (NOSOBJECT.kind == VAROBJECTKIND &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER &&
              integer_is_zero(tosp))
     {   /* compare pointer to integer constant 0 */
         if (NOSVAR.frame == 0 && NOSVAR.offset == 0)
             result = 0;
         else
             result = 1;
     }
     else if (TOSOBJECT.kind == VAROBJECTKIND &&
              NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              integer_is_zero(nosp))
     {   /* compare pointer to integer constant 0 */
         if (TOSVAR.frame == 0 && TOSVAR.offset == 0)
             result = 0;
         else
             result = 1;
     }
#endif //PTR COMPARES
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();
     TOSLIT.kind = LITOBJECTKIND;
     TOSLIT.type = LITTYPEINTEGER;
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = result;
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_ne*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_lt               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_lt(void)
 {
     UINT1          result;

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(nosp) < integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT3("cmp r%d, r%d", tos-1, tos);
         //EMIT2("mov r%d, 0", tos-1);
         EMIT2("setl r%d", tos-1);
         X86_emit_op_TT(CMP_V, tos-1, tos);
         //X86_emit_op_T_imm32(MOVI_D, tos-1, 0);
         X86_emit_op_T(SETL_D, tos-1);	//X64 zero-extends (??)
         //EMIT2("mov r%d, 1", tos-1);
         //EMIT("jlt .+2");
         //EMIT2("xor r%d, r%d", tos-1);
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(nosp) < dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT("fcompp");    //Sense is backwards, so we flip the test below.
                            //(VC fixes this by swapping st,st(1) with fxch,
                            // but that costs extra!!!
		 check_ndp_stack_room(-2);		//popped 2 ndp stack entries
         EMIT("fnstsw ax");				//VC does fnstsw ax/test ah,...
         EMIT("sahf");					//Is SAHF always available in X64??
         EMIT2("seta r%d", tos-1);		//We ignore unordered case for now!
         X86_emit_op(FCOMPP);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETA_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (integervalue(nosp) < dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos-1);    //Reversed sense.
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("seta r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos-1);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETA_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (dpvalue(nosp) < integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos);		//Correct sense.
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("setb r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETB_D, tos-1);	//X64 zero-extends (??)
     }
#if 0   // PTR COMPARES ARE NOT YET IMPLEMENTED!!!
     else if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == VAROBJECTKIND)
     {   /* compare two pointers */
         if (NOSVAR.frame != TOSVAR.frame)
         {   /* Pointers with different selectors are incommensurable. */
             Mde_warn(MD386__PTRISEL, 0, 0, 0);
             result = 0;    /* incommensurate */
         }
         else if (NOSVAR.offset < TOSVAR.offset)
             result = 1;
         else
             result = 0;
     }
#endif //PTR COMPARES
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();
     TOSLIT.kind = LITOBJECTKIND;
     TOSLIT.type = LITTYPEINTEGER;
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = result;
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_lt*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_le               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_le(void)
 {
     UINT1          result;

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(nosp) <= integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT3("cmp r%d, r%d", tos-1, tos);
         //EMIT2("mov r%d, 0", tos-1);
         EMIT2("setle r%d", tos-1);
         X86_emit_op_TT(CMP_V, tos-1, tos);
         //X86_emit_op_T_imm32(MOVI_D, tos-1, 0);
         X86_emit_op_T(SETLE_D, tos-1);
         //EMIT2("mov r%d, 1", tos-1);
         //EMIT("ble .+2");
         //EMIT2("xor r%d, r%d", tos-1);
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(nosp) <= dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT("fcompp");    //Sense is backwards, so we flip the test below.
		 check_ndp_stack_room(-2);		//popped 2 ndp stack entries
         EMIT("fnstsw ax");				//VC does fnstsw ax/test ah,...
         EMIT("sahf");
         EMIT2("setae r%d", tos-1);		 //We ignore unordered case for now!
         X86_emit_op(FCOMPP);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETAE_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (integervalue(nosp) <= dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos-1);    //Reversed sense.
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("setae r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos-1);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETAE_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (dpvalue(nosp) <= integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos);		//Correct sense.
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("setbe r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETBE_D, tos-1);	//X64 zero-extends (??)
     }
#if 0   // PTR COMPARES ARE NOT YET IMPLEMENTED!!!
     else if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == VAROBJECTKIND)
     {   /* compare two pointers */
         if (NOSVAR.frame != TOSVAR.frame)
         {   /* Pointers with different selectors are incommensurable. */
             Mde_warn(MD386__PTRISEL, 0, 0, 0);
             result = 0;    /* incommensurate */
         }
         else if (NOSVAR.offset <= TOSVAR.offset)
             result = 1;
         else
             result = 0;
     }
#endif //PTR COMPARES
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();
     TOSLIT.kind = LITOBJECTKIND;
     TOSLIT.type = LITTYPEINTEGER;
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = result;
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_le*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_gt               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_gt(void)
 {
     UINT1          result;

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(nosp) > integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT3("cmp r%d, r%d", tos-1, tos);
         //EMIT2("mov r%d, 0", tos-1);
         EMIT2("setg r%d", tos-1);
         X86_emit_op_TT(CMP_V, tos-1, tos);
         //X86_emit_op_T_imm32(MOVI_D, tos-1, 0);
         X86_emit_op_T(SETG_D, tos-1);
         //EMIT2("mov r%d, 1", tos-1);
         //EMIT("bgt .+2");
         //EMIT2("xor r%d, r%d", tos-1);
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(nosp) > dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT("fcompp");    //Sense is backwards, so we flip the test below.
		 check_ndp_stack_room(-2);		//popped 2 ndp stack entries
         EMIT("fnstsw ax");				//VC does fnstsw ax/test ah,...
         EMIT("sahf");
         EMIT2("setb r%d", tos-1);		//We ignore unordered case for now!
         X86_emit_op(FCOMPP);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETB_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (integervalue(nosp) > dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos-1);    //Reversed sense.
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("setb r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos-1);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETB_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (dpvalue(nosp) > integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos);		//Correct sense.
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("seta r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETA_D, tos-1);	//X64 zero-extends (??)
     }
#if 0   // PTR COMPARES ARE NOT YET IMPLEMENTED!!!
     else if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == VAROBJECTKIND)
     {   /* compare two pointers */
         if (NOSVAR.frame != TOSVAR.frame)
         {   /* Pointers with different selectors are incommensurable. */
             Mde_warn(MD386__PTRISEL, 0, 0, 0);
             result = 0;    /* incommensurate */
         }
         else if (NOSVAR.offset > TOSVAR.offset)
             result = 1;
         else
             result = 0;
     }
#endif //PTR COMPARES
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();
     TOSLIT.kind = LITOBJECTKIND;
     TOSLIT.type = LITTYPEINTEGER;
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = result;
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_gt*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_ge               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_ge(void)
 {
     UINT1          result;

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(nosp) >= integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT3("cmp r%d, r%d", tos-1, tos);
         //EMIT2("mov r%d, 0", tos-1);
         EMIT2("setge r%d", tos-1);
         X86_emit_op_TT(CMP_V, tos-1, tos);
         //X86_emit_op_T_imm32(MOVI_D, tos-1, 0);
         X86_emit_op_T(SETGE_D, tos-1);
         //EMIT2("mov r%d, 1", tos-1);
         //EMIT("bge .+2");
         //EMIT2("xor r%d, r%d", tos-1);
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(nosp) >= dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT("fcompp");    //Sense is backwards, so we flip the test below.
		 check_ndp_stack_room(-2);		//popped 2 ndp stack entries
         EMIT("fnstsw ax");  //VC does fnstsw ax/test ah,...
         EMIT("sahf");
         EMIT2("setbe r%d", tos-1);  //We ignore unordered case for now!
         X86_emit_op(FCOMPP);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETBE_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (integervalue(nosp) >= dpvalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos-1);    //Reversed sense.
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("setbe r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos-1);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETBE_D, tos-1);	//X64 zero-extends (??)
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (dpvalue(nosp) >= integervalue(tosp))
             result = 1;
         else
             result = 0;
         EMIT2("ficomp r%d", tos);		//Correct sense.
		 check_ndp_stack_room(-1);		//popped 1 ndp stack entry
         EMIT("fnstsw ax");
         EMIT("sahf");
         EMIT2("setae r%d", tos-1);
         X86_emit_op_Tm(FCOMP_I, tos);
         X86_emit_op(FNSTSW_AX);
         X86_emit_op(SAHF);
         X86_emit_op_T(SETAE_D, tos-1);	//X64 zero-extends (??)
     }
#if 0   // PTR COMPARES ARE NOT YET IMPLEMENTED!!!
     else if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == VAROBJECTKIND)
     {   /* compare two pointers */
         if (NOSVAR.frame != TOSVAR.frame)
         {   /* Pointers with different selectors are incommensurable. */
             Mde_warn(MD386__PTRISEL, 0, 0, 0);
             result = 0;    /* incommensurate */
         }
         else if (NOSVAR.offset >= TOSVAR.offset)
             result = 1;
         else
             result = 0;
     }
#endif //PTR COMPARES
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();
     TOSLIT.kind = LITOBJECTKIND;
     TOSLIT.type = LITTYPEINTEGER;
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = result;
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_ge*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              negate              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void negate(void)
 {
     double         dpresult;

     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         TOSLIT.nbytes = 4;
         TOSLIT.ivalue = -integervalue(tosp);
         TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
         EMIT2("neg r%d", tos);
         X86_emit_op_T(NEG_V, tos);
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         dpresult = -dpvalue(tosp);
         TOSLIT.nbytes = 8;
         TOSLIT.rvalue = dpresult;
         TOSLIT.valuep = (UINT1 *)&TOSLIT.rvalue;
         EMIT("fchs");
         X86_emit_op(FCHS);
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*negate*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_not              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_not(void)
 {   /* Perform bit-wise not. */

     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         TOSLIT.nbytes = 4;
         TOSLIT.ivalue = ~integervalue(tosp);
         TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
         EMIT2("not r%d", tos);
         X86_emit_op_T(NOT_V, tos);
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*op_not*/


        /************************************/
        /*                                  */
        /*                                  */
        /*           logical_not            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void logical_not(void)
 {   /* Returns 1 if value on TOS is 0, else 0. */
     UINT1          result;

     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(tosp))
             result = 0;
         else
             result = 1;
         EMIT2("cmp r%d, 0", tos);	//or test tos, tos if isReg(tos)???
         //EMIT2("mov r%d, 0", tos);
         EMIT2("sete r%d", tos);
         X86_emit_op_T_simm32(CMPI_V, tos, 0);
         //X86_emit_op_T_imm32(MOVI_D, tos, 0);
         X86_emit_op_T(SETZ_D, tos);
         //EMIT3("test r%d, r%d", tos, tos);
         //EMIT2("mov r%d, 1", tos);
         //EMIT("be .+2");
         //EMIT2("xor r%d, r%d", tos);
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(tosp))
             result = 0;
         else
             result = 1;
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);	//OOPS: FIX THIS LATER!!!
     }
     else if (TOSOBJECT.kind == VAROBJECTKIND)
     {   /* pointer */
         if (TOSVAR.offset|TOSVAR.frame)
             result = 0;
         else
             result = 1;
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);	//OOPS: FIX THIS LATER!!!
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.kind = LITOBJECTKIND;
     TOSLIT.type = LITTYPEINTEGER;
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = result;
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*logical_not*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             op_then              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_then(void)
 {   /* Process true part of then-else-if. */
     UINT1          result;
     int            current_if_depth;

     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         if (integervalue(tosp))
             result = 1;
         else
             result = 0;
         assert(if_depth >= 1);
         if_info[if_depth-1].else_lab = X86_genlab();
         if_info[if_depth-1].endif_lab = X86_genlab();
         EMIT2("cmp r%d, 0", tos);
         EMIT2("je cgl%d  ;else", if_info[if_depth-1].else_lab);
         X86_emit_op_T_simm32(CMPI_V, tos, 0);
         X86_emit_op_lab(JE, if_info[if_depth-1].else_lab);
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         if (dpvalue(tosp))
             result = 1;
         else
             result = 0;
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);	//OOPS: FIX THIS LATER!!!
     }
     else if (TOSOBJECT.kind == VAROBJECTKIND)
     {   /* pointer */
         if (TOSVAR.offset|TOSVAR.frame)
             result = 1;
         else
             result = 0;
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);	//OOPS: FIX THIS LATER!!!
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();

     if (!compile_flag) {								//UNUSED??
         if (!result) {
             /* skip to matching OP_ELSE or OP_IF */
             current_if_depth = if_depth;
             do advance_sdlp();
                while (!eosdl && 
                       !(sdlop == SDTYPE(OP_ELSE) && if_depth == current_if_depth) &&
                       !(sdlop == SDTYPE(OP_IF)   && if_depth == current_if_depth - 1));
             if (eosdl) bug(15);
             /* Resume execution after OP_ELSE or OP_IF... */
         }
     }
 } /*op_then*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             op_else              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_else(void)
 {   /* Process false part of then-else-if. */
     int            current_if_depth;
     
     if (!compile_flag) {								//UNUSED??
         /* Skip to matching OP_IF... */
         current_if_depth = if_depth;
         do advance_sdlp();
            while (!eosdl && !(sdlop == SDTYPE(OP_IF) && if_depth == current_if_depth - 1));
         if (eosdl) bug(16);
         /* Resume execution after OP_IF... */
     }
     else {
         //pop();
         assert(if_depth >= 1);
         EMIT2("jmp cgl%d", if_info[if_depth-1].endif_lab);
         EMIT2L("cgl%d:    ;else", if_info[if_depth-1].else_lab);
         X86_emit_op_lab(JMP, if_info[if_depth-1].endif_lab);
         X86_deflab(if_info[if_depth-1].else_lab);
         if_info[if_depth-1].else_lab = -1; //mark as defined
     }
 } /*op_else*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_if               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_if(void)
 {   /* Process end of then-else-if. */
     /* Nothing to do... advance_sdlp() has already done it. */

	 assert(if_depth >= 0);
     if (if_info[if_depth].else_lab != -1) {
         //else_lab not yet defined (because there was no OP_ELSE)...
         EMIT2L("cgl%d:    ;else", if_info[if_depth].else_lab);
         X86_deflab(if_info[if_depth].else_lab);
     }
     EMIT2L("cgl%d:    ;endif", if_info[if_depth].endif_lab);
     X86_deflab(if_info[if_depth].endif_lab);
 } /*op_if*/


 /********************** arglist stack ************************/

#define MAX_ARGLIST_STACK_DEPTH 1024

 static int call_depth = 0;
 static int narg = 0;		//must push/pop across CALLs
 static int argoff = 0;		//must push/pop across CALLs
 static CGL ref_nargsize;	//must push/pop across CALLs
 static int argtypemask = 0;//must push/pop across CALLs
 static struct {
	 int narg;
	 int argoff;
	 CGL ref_nargsize;
	 int argtypemask;
 } arglist_stack[MAX_ARGLIST_STACK_DEPTH];

        /************************************/
        /*                                  */
        /*                                  */
        /*          init_arglist            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void init_arglist(void)
 {
	 call_depth = 0;
	 narg = 0;
	 argoff = 0;
	 ref_nargsize = -1;
	 argtypemask = 0;
 } /*init_arglist*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          push_arglist            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void push_arglist(void)
 {
#ifdef DEBUG_OP_CALL
     PRINTF4("++ push_arglist, call_depth=%d, narg=%d, argoff=%d\r\n",
		     call_depth, narg, argoff);
#endif

	 if (call_depth >= MAX_ARGLIST_STACK_DEPTH) {
		 // function calls too deeply nested
		 X86_abort2("Function calls nested too deeply (%d.)", call_depth);
	 }
	 arglist_stack[call_depth].narg = narg;
	 arglist_stack[call_depth].argoff = argoff;
	 arglist_stack[call_depth].ref_nargsize = ref_nargsize;
	 arglist_stack[call_depth].argtypemask = argtypemask;
	 ++call_depth;
 } /*push_arglist*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          pop_arglist             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pop_arglist(void)
 {
#ifdef DEBUG_OP_CALL
     PRINTF4("++ pop_arglist, call_depth=%d, narg=%d, argoff=%d\r\n",
		     call_depth, narg, argoff);
#endif

	 if (call_depth <= 0) bug(223344);
	 --call_depth;
	 narg = arglist_stack[call_depth].narg;
	 argoff = arglist_stack[call_depth].argoff;
	 ref_nargsize = arglist_stack[call_depth].ref_nargsize;
	 argtypemask = arglist_stack[call_depth].argtypemask;

#ifdef DEBUG_OP_CALL
     PRINTF4("-- pop_arglist, call_depth=%d, narg=%d, argoff=%d\r\n",
		     call_depth, narg, argoff);
#endif
 } /*pop_arglist*/


        /************************************/
        /*                                  */
        /*                                  */
        /*           op_arglist             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_arglist(void)
 {
#ifdef DEBUG_OP_CALL
     PRINTF("++ op_arglist\r\n");
#endif

	 push_arglist();
	 narg = 0;
	 argoff = 0;
	 argtypemask = 0;	//first 4 args assumed to be int
#ifdef X64
	 if (X64gen) {
#if 0 //not necessary if prolog always assures stack is aligned mod 16
		 //ensure stack is aligned mod 16
		 //(optional if assume ALWAYS aligned mod 16)
		 if (call_depth <= 1) {
			 // only ever necessary at outermost call level!!
			 EMIT("and RSP, -16 ;align 16");
			 X86_emit_op_R_simm32(ANDI_V, esp, -16);
		 }
#endif //not necessary
		 // X64gen *always* allocates at least a 32-byte shadow arglist frame for
		 // every call, even if the call has 0 actual args.
		 // QUESTION: If nargsize > 4KB do we need to do stack probes, for
		 // example by calling _alloca()?  Or will the stores into the arglist
		 // suffice for probes?
		 // Q2: Should we use LEA rather than SUB to allocate the arglist, per
		 //		Jonathan Pollard?  Per JP:
		 //     To increment or decrement ESP, it is preferred to use LEA ESP,[] rather than
		 //     add/sub esp,const to avoid mixing ALU with AGU (address generation unit)
		 //     accesses (???).
		 ref_nargsize = X86_genlab();
		 if (call_depth > 1) {
			 //allocate arglist dynamically
#ifndef PERJP
			 //forward reference to nargsize from OP_CALL
			 EMIT2("sub RSP, nargsize%d ;allocate arglist", ref_nargsize);
			 X86_emit_op_R_ref_simm32(SUBI_V, esp, ref_nargsize);
#else //per JP
			 //forward reference to -nargsize!!! from OP_CALL
			 EMIT2("lea RSP, [RSP-nargsize] ;allocate arglist", ref_nargsize);
			 X86_emit_op_RM_ref_simm32(LEA_V, esp, esp, no_reg, 0, ref_nargsize); //NYI!!
#endif //per JP
		 }//call_depth > 1
	 } //X64gen
#endif //X64

#ifdef DEBUG_OP_CALL
     PRINTF4("-- op_arglist, call_depth=%d, narg=%d, argoff=%d\r\n",
		     call_depth, narg, argoff);
#endif
} /*op_arglist*/


//// TODO -check all *movd, *movq;
////	  -implement ARGLIST;
////      -allocate a "static" stack frame for level-0 CALLs
//// XXX Bug?  Will calls by subsequent args destroy value of outer arg1 cached in RCX/XMM0?
////	 Need to stash such args in the arglist, then load the regs into their dedicated regs
////	 just before the actual call ???
////	 But home the args only if necessary.  Not necessary at level 0 (which covers the 
////	 majority of cases)!!!  
////	 Before starting work on a new arg list, cache any existing register args in their
////	 home locations, and mark them as "cached" at that level, so they can be recalled
////	 into the proper regs immediately before their matching call.
////	 Even better: Only when preparing to load an arg into its register, check to see if
////	 that register is already occupied by an arg at a lower call level; if so, home this
////	 arg and mark it as such.

////  How to implement a "static" arglist for level-0 calls:
////	For all level-0 calls, do NOT allocate a new arglist on the stack -- assume an
////	arglist of appropriate size has already been allocated, and just keep track of
////	the max size arglist needed at level-0 (call it MAX_ARGL and set the minimum
////	value to 4 slots == 32 bytes if any calls were made, else set it to -1).
////	In the prolog, allocate a presumptive level-0 arglist:
////		and rsp, -16	;align stack if not already aligned
////		sub rsp, REF_ARGL0 ;allocate 8*MAX_ARGL bytes on stack.
////	In the epilog fix up the foward reference to REF_ARGL0:
////		if (MAX_ARGL == -1) then no calls were made; set MAX_ARGL=0;
////		else set MAX_ARGL = max(MAX_ARGL, 4); //arglist must have at least 4 slots
////		Def_lab_ref(REF_ARGL0, 8*MAX_ARGL);
////	Note that 0 bytes will be allocated for the level-0 arglist if no calls
////	were made; otherwise at least 32 bytes will be allocated.  Also note
////	that MAX_ARGL only takes into account level-0 calls; a level-1 call may
////	use 99 args for example, but the level-0 arglist may only accommodate 4
////	args if no level-0 call used 5 or more args.
////	Later the FM compiler make break up nested calls into a series of level-0
////	calls, thus obviating the need to dynamically allocate arglists for
////	inner-level (level > 0) calls, and making it unecessary to cache register args
////	on the stack when processing nested calls. But not at present!


        /************************************/
        /*                                  */
        /*                                  */
        /*             op_arg               */
        /*                                  */
        /*                                  */
        /************************************/

/**procedure**/
 static void op_arg(void)
 {
#ifdef DEBUG_OP_CALL
     PRINTF4("++ op_arg, call_depth=%d, narg=%d, argoff=%d\r\n",
		     call_depth, narg, argoff);
#endif

	 narg++;	//count args

	 if (X86gen && argoff == 0) {
		 //First arg, start the arglist for X86.
		 //For X86gen, we wait until now, so that we don't
		 //needlessly emit SUB ESP,0 for a call with no args...
		 // QUESTION: If nargsize > 4KB do we need to do stack probes, for
		 // example by calling _alloca()?  Or will the stores into the arglist
		 // suffice for probes?
		 // Q2: Should we use LEA rather than SUB to allocate the arglist, per
		 //		Jonathan Pollard?
		 ref_nargsize = X86_genlab();
		 if (call_depth > 1) {
			 //allocate arglist dynamically
#ifndef PERJP
			 //forward reference to nargsize from OP_CALL
			 EMIT2("sub ESP, nargsize%d ;alloc arglist", ref_nargsize);
			 X86_emit_op_R_ref_simm32(SUBI_V, esp, ref_nargsize);
#else //per JP
			 //forward reference to -nargsize!!! from OP_CALL
			 EMIT2("lea ESP, [ESP-nargsize] ;alloc arglist", ref_nargsize);
			 X86_emit_op_RM_ref_simm32(LEA_V, esp, esp, no_reg, 0, ref_nargsize); //NYI!!
#endif //per JP
		 }//call_depth > 1
	 } //X86gen && argoff == 0

     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type != LITTYPEREAL ||
		 TOSOBJECT.kind == VAROBJECTKIND)
     {
#ifdef X64
		 if (X64gen) {
			 switch (narg) {
#if 0 //bug fix
			 //move first 4 integer args to their fastcall registers
			 case 1:
				 EMIT2("mov RCX, r%d", tos);
				 X86_emit_op_RT(MOV_Q, rcx, tos);
				 break;
			 case 2:
				 EMIT2("mov RDX, r%d", tos);
				 X86_emit_op_RT(MOV_Q, rdx, tos);
				 break;
			 case 3:
				 EMIT2("mov R8, r%d", tos);
				 X86_emit_op_RT(MOV_Q, r8, tos);
				 break;
			 case 4:
				 EMIT2("mov R9, r%d", tos);
				 X86_emit_op_RT(MOV_Q, r9, tos);
				 break;
#endif //bug fix
			 default:
				 //move remaining args to their stack slots
				 EMIT3("mov qword ptr[RSP%+d], r%d", argoff, tos);
				 X86_emit_op_MT(MOV_Q, rsp, no_reg, 0, argoff, tos);
				 break;
			 }//switch
			 argoff += 8;
		 } else
#endif //X64
		 {	//Could collapse with above code???
			 EMIT3("mov dword ptr[ESP%+d], r%d", argoff, tos);
			 X86_emit_op_MT(MOV_V, esp, no_reg, 0, argoff, tos);
			 argoff += 4;
		 }
		 pop();
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
#ifdef X64
		 if (X64gen) {
			 //move f.p. args to their stack slots
			 EMIT2("fstp qword ptr[RSP%+d]", argoff);
			 check_ndp_stack_room(-1);		//popped 1 arg from ndp stack
			 X86_emit_op_M(FSTP_D, rsp, no_reg, 0, argoff);
		 } else
#endif //X64
		 {	//Could collapse with above code???
			 EMIT2("fstp qword ptr[ESP%+d]", argoff);
			 check_ndp_stack_room(-1);		//popped 1 arg from ndp stack
			 X86_emit_op_M(FSTP_D, esp, no_reg, 0, argoff);
		 }
		 if (narg <= 32) {
			 //beware that << takes the shift count mod 32 on x86 (mod 64 on x64)!
			 //set REAL flag for this arg (only 1st 4 matter)
			 argtypemask |= 1 << (narg - 1);
		 }
		 argoff += 8;
         TOSLIT.type = LITTYPEINTEGER;  //so pop won't pop ndp stack (since fstp already did!)
         pop();
     }
#if 0 //these can be omitted...
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPELOGICAL)
     {
         //EMIT2("push r%d", tos);
         //X86_emit_op_T(PUSH_D, tos);
         //pop();
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPECHAR)
     {
         //EMIT2("push r%d", tos);
         //X86_emit_op_T(PUSH_D, tos);
         //pop();
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREGION)
     {
         //assert(TOSLIT.nbytes == 4);	//What is X64 REGION size ?????
         //EMIT2("push r%d", tos);
         //X86_emit_op_T(PUSH_D, tos);
         //pop();
     }
#endif //omit these
     else {
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }

 #ifdef DEBUG_OP_CALL
     PRINTF4("-- op_arg, call_depth=%d, narg=%d, argoff=%d\r\n",
		     call_depth, narg, argoff);
#endif
} /*op_arg*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             op_call              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_call(void)
 {
     int nargs;
     int nargsize = 0;
     int i;
     INT4 iresult = 0;
     UINT4 resultTypeIndex;

#ifdef DEBUG_OP_CALL
     PRINTF4("++ op_call, call_depth=%d, narg=%d, argoff=%d\r\n",
		     call_depth, narg, argoff);
#endif

     ensuretwooperands();	//ensure #args and fn address

     // TOS = actual # of args (n)
#if 0
     // TOS-1 = param n
     // TOS-2 = param n-1
     // ...
     // TOS-n = param 1
#endif
     // TOS-n-1 = address of function

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER) {
         nargs = integervalue(tosp);
#ifdef DEBUG_OP_CALL
     PRINTF2("++ nargs=%d\r\n", nargs);
#endif
         if (nargs < 0) bug(20);
     }
     else bug(21);

     pop();	//pop nargs

#if 1	//push args left-to-right
	 if (X64gen || argoff) {	//either X64 code, or more than 0 args in X32 code
		 nargsize = argoff;
#ifdef X64
		 if (X64gen) {
			 //ensure nargsize rounded up to multiple of 16 (to maintain stack alignment)
			 nargsize = (nargsize + 15) & -16;
			 //nargsize must be a minimum of 32 to shadow first 4 args (even if 0 actual args)
			 nargsize = max(nargsize, 32);
		 }
#endif //X64
		 // Update max size of a level-1 arglist in this function (for now --
		 // later we will use the "static arglist busy" scheme).
		 if (call_depth == 1) {
			 cbp->func.cbStaticArglistSize =	/// N.B. X64 nargsize rounded up to 16 here!!!
				 max(cbp->func.cbStaticArglistSize, nargsize);
		 }
		 // Fixup forward refs to nargsize, now that it's known...
		 assert(ref_nargsize != -1);
		 EMIT3L("nargsize%d = %d ;fixup", ref_nargsize, nargsize);
#ifndef PERJP
		 X86_def_abs32(ref_nargsize, nargsize);
#else //per JP
		 //fixing up LEA rSP,[rSP-nargsize} requires NEGATIVE fixup value!!
		 X86_def_abs32(ref_nargsize, -nargsize);
#endif //per JP
	 } //if X64gen||argoff
#endif	//push args left-to-right


#if 0	//push args right-to-left
	 //no longer used!!
#endif //push args right-to-left

#if 1 //push args left-to-right
#ifdef X64
	 if (X64gen) {
		 int imax = min(nargs, 4);
		 for (i = 1; i <= imax; i++) {
			 if ( argtypemask & (1 << (i - 1)) ) {
				 //REAL type bit set in arg type mask...
				 //move first 4 f.p. args to their FASTCALL registers
				 //also to int reg if vararg [or unprototyped??]
				 switch (i) {
				 case 1:
#ifndef USE_MOVSDX
					 // Use movd/movq to copy mem to xmm reg.
					 EMIT("movq xmm0, qword ptr[RSP]");
					 X86_emit_op_RM(XMOVQ, xmm0, rsp, no_reg, 0, 0);
#else
					 // VC++ uses MOVSDX. Why? (fewer bytes??)
					 EMIT("movsdx xmm0, qword ptr[RSP]");
					 X86_emit_op_RM(MOVSDX, xmm0, rsp, no_reg, 0, 0);
#endif //USE_MOVSDX
#ifndef NOTVARARG
					 // If it's a vararg [or unprototyped arg??], we also need
					 // to replicate the double value into the corresponding 
					 // integer fastcall reg... (see todo.txt).
					 // We also unconditionally replicate the arg if the FML
					 // version is <= 1.5, for backwards compatibility...
					 if (nProgRevMajor < 1 ||
						 nProgRevMajor == 1 && nProgRevMinor <= 5 ||
						 TOSVAR.extradata == -12345 ||	//flag meaning all varargs (unprototyped)
						 TOSVAR.extradata < 0 && TOSVAR.extradata > -i) //arg within ellipses
					 {
						 EMIT("movq RCX, xmm0 ;vararg copy");
						 X86_emit_op_RR(XMOVQR, xmm0, rcx);	//note reversed move op
					 }
#endif //NOTVARARG
					 break;
				 case 2:
#ifndef USE_MOVSDX
					 // Use movd/movq to copy mem to xmm reg.
					 EMIT("movq xmm1, qword ptr[RSP+8]");
					 X86_emit_op_RM(XMOVQ, xmm1, rsp, no_reg, 0, 8);
#else
					 // VC++ uses MOVSDX. Why?
					 EMIT("movsdx xmm1, qword ptr[RSP+8]");
					 X86_emit_op_RM(MOVSDX, xmm1, rsp, no_reg, 0, 8);
#endif //USE_MOVSDX
#ifndef NOTVARARG
					 // If it's a vararg [or unprototyped arg??], we also need
					 // to replicate the double value into the corresponding 
					 // integer fastcall reg... (see todo.txt).
					 // We also unconditionally replicate the arg if the FML
					 // version is <= 1.5, for backwards compatibility...
					 if (nProgRevMajor < 1 ||
						 nProgRevMajor == 1 && nProgRevMinor <= 5 ||
						 TOSVAR.extradata == -12345 ||	//flag meaning all varargs
						 TOSVAR.extradata < 0 && TOSVAR.extradata > -i) //arg within ellipses
					 {
						 EMIT("movq RDX, xmm1 ;vararg copy");
						 X86_emit_op_RR(XMOVQR, xmm1, rdx);	//note reversed move op
					 }
#endif //NOTVARARG
					 break;
				 case 3:
#ifndef USE_MOVSDX
					 // Use movd/movq to copy mem to xmm reg.
					 EMIT("movq xmm2, qword ptr[RSP+16]");
					 X86_emit_op_RM(XMOVQ, xmm2, rsp, no_reg, 0, 16);
#else
					 // VC++ uses MOVSDX. Why?
					 EMIT("movsdx xmm2, qword ptr[RSP+16]");
					 X86_emit_op_RM(MOVSDX, xmm2, rsp, no_reg, 0, 16);
#endif //USE_MOVSDX
#ifndef NOTVARARG
					 // If it's a vararg [or unprototyped arg??], we also need
					 // to replicate the double value into the corresponding 
					 // integer fastcall reg... (see todo.txt).
					 // We also unconditionally replicate the arg if the FML
					 // version is <= 1.5, for backwards compatibility...
					 if (nProgRevMajor < 1 ||
						 nProgRevMajor == 1 && nProgRevMinor <= 5 ||
						 TOSVAR.extradata == -12345 ||	//flag meaning all varargs
						 TOSVAR.extradata < 0 && TOSVAR.extradata > -i) //arg within ellipses
					 {
						 EMIT("movq R8, xmm2 ;vararg copy");
						 X86_emit_op_RR(XMOVQR, xmm2, r8);	//note reversed move op
					 }
#endif //NOTVARARG
					 break;
				 case 4:
#ifndef USE_MOVSDX
					 // Use movd/movq to copy mem to xmm reg.
					 EMIT("movq xmm3, qword ptr[RSP+24]");
					 X86_emit_op_RM(XMOVQ, xmm3, rsp, no_reg, 0, 24);
#else
					 // VC++ uses MOVSDX. Why?
					 EMIT("movsdx xmm3, qword ptr[RSP+24]");
					 X86_emit_op_RM(MOVSDX, xmm3, rsp, no_reg, 0, 24);
#endif //USE_MOVSDX
#ifndef NOTVARARG
					 // If it's a vararg [or unprototyped arg??], we also need
					 // to replicate the double value into the corresponding 
					 // integer fastcall reg... (see todo.txt).
					 // We also unconditionally replicate the arg if the FML
					 // version is <= 1.5, for backwards compatibility...
					 if (nProgRevMajor < 1 ||
						 nProgRevMajor == 1 && nProgRevMinor <= 5 ||
						 TOSVAR.extradata == -12345 ||	//flag meaning all varargs
						 TOSVAR.extradata < 0 && TOSVAR.extradata > -i) //arg within ellipses
					 {
						 EMIT("movq R9, xmm3 ;vararg copy");
						 X86_emit_op_RR(XMOVQR, xmm3, r9);	//note reversed move op
					 }
#endif //NOTVARARG
					 break;
				 }//switch
			 } else { //not REAL
				 //any kind of arg except real value...
				 //move first 4 integral args to their FASTCALL registers
				 switch (i) {
				 case 1:
					 EMIT("mov RCX, qword ptr[RSP]");
					 X86_emit_op_RM(MOV_Q, rcx, rsp, no_reg, 0, 0);
					 break;
				 case 2:
					 EMIT("mov RDX, qword ptr[RSP+8]");
					 X86_emit_op_RM(MOV_Q, rdx, rsp, no_reg, 0, 8);
					 break;
				 case 3:
					 EMIT("mov R8, qword ptr[RSP+16]");
					 X86_emit_op_RM(MOV_Q, r8, rsp, no_reg, 0, 16);
					 break;
				 case 4:
					 EMIT("mov R9, qword ptr[RSP+24]");
					 X86_emit_op_RM(MOV_Q, r9, rsp, no_reg, 0, 24);
					 break;
				 }//switch
			 }//if not REAL
		 } /*for i*/	//SHOULD consolidate all cases and eliminate both switches ABOVE!!
	 } //if X64gen
#endif //X64
#endif //push args left-to-right

     ensureoneoperand();	//the function address
     if (TOSOBJECT.kind == VAROBJECTKIND && TOSVAR.typeindex == 1234) {
         // TOS is an integer function address
         resultTypeIndex = LITTYPEINTEGER;
     }
     else if (TOSOBJECT.kind == VAROBJECTKIND && TOSVAR.typeindex == 1237) {
         // TOS is a real function address
         resultTypeIndex = LITTYPEREAL;
     }
     else if (TOSOBJECT.kind == VAROBJECTKIND && TOSVAR.typeindex == 1238) {
         // TOS is a region function address
         resultTypeIndex = LITTYPEREGION;
     }
#if 1	//kludge to allow primitive form of indirect function call
	 else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type != LITTYPEREAL) {
		 //assume it's an indirect call to an integer function...
		 ////Warn("Assuming indirect call to integer function");
		 resultTypeIndex = LITTYPEINTEGER;
		 TOSVAR.extradata = -12345;	// magic number meaning any number of args is okay 
		 //OOPS, it's a LIT not a VAR!!! Probably works anyway, since extradata is high up,
		 //but probably should move extradata to common head of object structure,
		 //or change current object to VAROBJECTKIND
	 }
#endif //kludge
     else {
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
         return;
     }

	 // Check for correct number of actual arguments

     if (TOSVAR.extradata >= 0 && TOSVAR.extradata != nargs) {
         // wrong number of arguments to FPL function
		 // "incorrect number of arguments in function call (!SL instead of !SL)"
         Mde_abort_command(MD386__NARGS, nargs, TOSVAR.extradata, 0);
         //return;
     }
#if 1	//kludge for indirect and unprototyped calls
	 else if (TOSVAR.extradata == -12345) {
		 // -12345 is magic number meaning VPL with nargs >= 0 (i.e., unprototyped)
	 }
#endif //kludge
     else if (nargs < -TOSVAR.extradata) {
         // for VPL function, -TOSVAR.extradata is minimum number
         // of required arguments
		 // Should also allow a VPL function with >= 0 args (not just >= 1) -- see -12345 above
         Mde_abort_command(MD386__NARGS, nargs, -TOSVAR.extradata, 0);
		 //Add new msg, should say e.g. "too few args (1 instead of 2 or more)"
     }

	 //Pass nargs to callee per FM32/FM64 convention...
	 //Note to self:  Could we remember #args from earlier
	 //and just emit mov ecx, #nargs here???? -- but nargs is already in a temp reg!!
#ifdef X64
#if 1	//push args left-to-right
	 nargs = 0;	//already popped off stack
#endif 	//push args left-to-right
	 if (X64gen) {
		 //for X64, pass nargs in rax...
	     EMIT2("mov RAX, r%d ;nargs", tos+nargs+1);	//CAUTION--nargs tmp is no longer allocated!
		 X86_emit_op_RT(MOV_Q, rax, tos+nargs+1);
		//X86_emit_op_R_imm32(MOVI_Q, rax, nargs); //would this be better if T is not an R???
		//or X86_emit_op_RR(XOR_V, rax, rax);	//if nargs==0??
	 } else
#endif
	 {	//for X86, pass nargs in ecx...
	     EMIT2("mov ECX, r%d ;nargs", tos+nargs+1);	//CAUTION--nargs tmp is no longer allocated!
		 X86_emit_op_RT(MOV_D, ecx, tos+nargs+1);
		//X86_emit_op_R_imm32(MOVI_D, ecx, nargs); //would this be better if T is not an R???
		//or X86_emit_op_RR(XOR_, ecx, ecx);	//if nargs==0??
	 }
#if 0
     //following would be useful to optimize thunks in universal function calls...
	 //at least for x86, where nargsize is not a fixed multiple of nargs...
#ifdef X64
	 if (X64gen) {
		 //for X64, pass NARGSIZE in R10
		 //don't really need NARGSIZE for FM64; it's always 8*NARGS!!!
		 EMIT2("mov r10, %d ;nargsize", nargsize);
		 X86_emit_op_R_imm32(Q(MOVI_D), r10, nargsize);	//x64???
	 } else
#endif
	 {	//for x86, pass NARGSIZE in EDX
		 EMIT2("mov edx, %d ;nargsize", nargsize);
		 X86_emit_op_R_imm32(MOVI_D, edx, nargsize);
	 }
#endif

	 // Now make the CALL, and restore the stack after the CALL.
	 //
	 // X64 always uses FASTCALL/CDECL in which the callee returns with rsp set to its
	 // value just before the CALL.  The caller must pop any args off the stack, plus
	 // remove any extra padding on the stack if the number of args was odd.  We don't
	 // necessarily pop the stack all the way back to its original value (unless we are
	 // relying on [rsp] indexing of local temps and vars).  Otherwise, we can just pop
	 // back to where it was 16-byte aligned; then no further alignment is required on
	 // subsequent calls within this call frame.  In fact, if we can assure that the
	 // stack is always 16-byte aligned BEFORE args are pushed, then we could eliminate
	 // the alignment operation altogether (e.g., if we use a fixed 16-byte aligned
	 // argument stack frame).
#ifdef X64
	 if (X64gen) {
		 EMIT2("call r%d", tos);
		 X86_emit_op_T(CALLI_N, tos);
		 if (call_depth > 1 && nargsize > 0) {
			 //using a dynamically allocated arglist..
			 //deallocate it
			 EMIT2("add RSP, %d ;deallocate arglist", nargsize);
			 X86_emit_op_R_simm32(ADDI_V, rsp, nargsize);
		 }
	 } else
#endif
	 // For x86, for CDECL calls we only need to restore the stack to its value just
	 // before the topmost arg was pushed (assuming stack is always 4- (or 8-) byte
	 // aligned). For x86 STDCALL calls, we assume the callee has restored the stack
	 // to that same point.  If we don't know whether the callee uses the CDECL or
	 // STDCALL convention, we can make a UNIVERSAL call by saving and restoring the
	 // value of esp across the entire call (including arg pushes) in a nonvolatile
	 // register or in a temp on the caller's stack frame.
	 {	 //must be X86gen
#if 0
		 //following works only for x86 CDECL calls (or STDCALL calls with no args)
		 //and is same as X64 code above!!
		 EMIT2("call r%d", tos);
		 X86_emit_op_T(CALLI_N, tos);
		 if (call_depth > 1 && nargsize > 0) {
			 //using a dynamically allocated arglist..
			 //deallocate it
			 EMIT2("add ESP, %d ;restore esp", nargsize);
			 X86_emit_op_R_simm32(ADDI_V, esp, nargsize);
		 }
#else
		 //following works for both CDECL and STDCALL calls (i.e. it's a UNIVERSAL call)
		 //(Assumes the STDCALL callee does not pop more args off the stack than were pushed
		 //by the caller!)
		 if (nargsize == 0) {
			 // no args to pop ==> stack is already correct
			 EMIT2("call r%d", tos);	//proc address is in tos
			 X86_emit_op_T(CALLI_N, tos);
		 }
		 else {
			 //if nargsize > 0, need to save/restore stack ptr in case callee is STDCALL
			 //in which case the callee pops args off the stack
	 //CAN WE OMIT this if we KNOW the callee is cdecl??
			 push(VAROBJECTKIND);	//allocate a tmp -- hope it's a (nonvolatile) register!!
			 EMIT2("mov r%d, esp ;save stack ptr", tos);
			 X86_emit_op_TR(MOV_V, tos, esp);
			 EMIT2("call r%d", tos - 1);	//proc address is in nos
			 X86_emit_op_T(CALLI_N, tos - 1);
			 if (call_depth == 1) {
				 //level-1 call, using static arglist
				 //just restore the stack ptr
				 EMIT2("mov esp, r%d ;restore stack ptr", tos);
				 X86_emit_op_RT(MOV_V, esp, tos);
			 }
			 else if (isReg(tos)) {
				 // using dynamic arglist, and tos is a register...
				 EMIT3("lea esp, [r%d%+d] ;restore stack ptr", tos, nargsize);
				 X86_emit_op_RM(LEA_V, esp, tmpReg(tos), no_reg, 0, nargsize);
			 }
			 else {
				 // using dynamic arglist, and tos is in local temps stack frame
				 EMIT2("mov esp, r%d ;restore stack ptr", tos);
				 EMIT2("add esp, %d ;deallocate arglist", nargsize);
				 X86_emit_op_RT(MOV_V, esp, tos);
				 X86_emit_op_R_simm32(ADDI_V, esp, nargsize);
			 }
			 pop();		//deallocate the tmp
		 }
#endif
	 }//x86

     if (resultTypeIndex != LITTYPEREAL) {
         // move result from rAX to reg we're expecting...
         EMIT2("mov r%d, rAX ;result", tos);
         X86_emit_op_TR(MOV_V, tos, eax);
     }
#ifdef X64
	 else if (X64gen) {
		 //for X64, real result is returned in xmm0.
		 //push it onto the ndp stack for now...
		 //(Need a temp in memory, so don't allocate a tmp,
		 //which might be a register -- although we *could*
		 //use its home in memory.  Anyway, a fixed "temp"
		 //in our stack frame or FMC/tls should suffice.)
#ifndef USE_MOVSDX
		 // Use movd/movq to copy mem to xmm reg.
		 EMIT("movq temp, xmm0");
		 X86_emit_op_MR(XMOVQ, FMC(temp), xmm0);		//translates to op_RM(XMOVQR)
#else
		 // VC++ uses MOVSDX...
		 EMIT("movsdx temp, xmm0");
		 X86_emit_op_MR(MOVSDX, FMC(temp), xmm0);		//translates to op_RM(MOVSDXR)
#endif //USE_MOVSDX
		 check_ndp_stack_room(1);						//allocate 1 ndp entry for result
		 EMIT("fld_d temp");
		 X86_emit_op_M(FLD_D, FMC(temp));
	 } else
#endif //X64
	 {
		//X86: real result is returned in st(0)
		 check_ndp_stack_room(1);						//allocate 1 ndp entry for result
	 }//X86

     //pop();
	 assert(tos >= 0);
     TOSOBJECT.kind = LITOBJECTKIND;	// change tos to LIT containing result
     TOSLIT.type = resultTypeIndex;
#ifdef X64
	 if (X64gen) {
		 TOSLIT.nbytes = 8;				// sizeof(double), sizeof(int64)
	 } else
#endif //X64
     if (resultTypeIndex == LITTYPEREAL) {
		 TOSLIT.nbytes = 8;				// sizeof(double)
	 } else {
		 TOSLIT.nbytes = 4;				// sizeof(int)
	 }
     TOSLIT.ivalue = iresult;
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;

	 pop_arglist();	//restore previous arglist parameters

#ifdef DEBUG_OP_CALL
     PRINTF4("-- op_call, call_depth=%d, narg=%d, argoff=%d\r\n",
		     call_depth, narg, argoff);
#endif
 } /*op_call*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_min              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_min(void)
 {
     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         EMIT3("cmp r%d, r%d", tos-1, tos);
         EMIT3("cmovg r%d, r%d", tos-1, tos);
         if (isReg(tos-1)) {
			 X86_emit_op_RT(CMP_V, tmpReg(tos-1), tos);
             X86_emit_op_RT(V(CMOVG), tmpReg(tos-1), tos);
         }
         else {
             X86_emit_op_RT(MOV_V, eax, tos-1);
             X86_emit_op_RT(CMP_V, eax, tos);
             X86_emit_op_RT(V(CMOVG), eax, tos);
             X86_emit_op_TR(MOV_V, tos-1, eax);
         }
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
		 // Note: Could use direct integer64 conparison (with proper adjustment
		 // of signed magnitude to two's complement representation) if willing
		 // to accept whatever behavior results for NaNs (behavior for infinities
		 // should be correct). But would it gain any useful speedup?
         if (gFmcp->use_CMOV) {
			 //FCOMI and FCMOVcc are available...
			 EMIT("fcomi");				//compare st to st(1), results to flags
			 EMIT("fcmovnbe st(1) ;op_min");	//load st from st(1) if st > st(1)
			 EMIT("fstp st(1)");		//move result to st(1) and pop
			 check_ndp_stack_room(-1);	//popped off 1 ndp stack entry

			 X86_emit_op(FCOMI);
			 X86_emit_opreg(FCMOVNBE_STN, 1);
			 X86_emit_op_R(FSTP_STN, 1);

			 TOSLIT.type = LITTYPEINTEGER;  //so pop() won't pop ndp stack below...
		 }
		 else {
			 //No FCOMI or FCMOVcc available...
			 CGL cgl1 = X86_genlab();

			 EMIT("fcom");              //compare st to st(1)
			 EMIT("fnstsw ax");
			 EMIT("test ah, 41h ;test C3,C0");
			 EMIT2("je cgl%d", cgl1);   //C3C0=00 => st > st(1)
			 EMIT("fst st(1)");         //C3=1 or C0=1 => st <= st(1)
			 EMIT2L("cgl%d:", cgl1);
			 ///EMIT("fpop ;st=min");   //done by pop() below...

			 X86_emit_op(FCOM);
			 X86_emit_op(FNSTSW_AX);
			 X86_emit_op_R_imm8(TESTI_B, ah, 0x41);
			 X86_emit_op_lab_disp8(JE, cgl1);
			 X86_emit_op_R(FST_STN, 1);
			 X86_deflab(cgl1);
			 ///X86_emit_op(FPOP);  //done by pop() below...
		 }//!use_CMOV
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
		 //Should convert int arg to real and return real?
         Mde_abort_command(MD386__NYI, 9, 0, 0); // for now
         NOSLIT.type = LITTYPEREAL;
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
		 //Should convert int arg to real and return real?
         Mde_abort_command(MD386__NYI, 10, 0, 0); // for now
         TOSLIT.type = LITTYPEREAL;
     }
     else if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == VAROBJECTKIND)
     {   /* compare two pointers */
         if (NOSVAR.frame != TOSVAR.frame)
         {   /* Pointers with different selectors are incommensurable. */
             Mde_warn(MD386__PTRISEL, 0, 0, 0);
         }
         Mde_abort_command(MD386__NYI, 11, 0, 0); // for now
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_min*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_max              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_max(void)
 {
     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         EMIT3("cmp r%d, r%d", tos-1, tos);
         EMIT3("cmovl r%d, r%d", tos-1, tos);
         if (isReg(tos-1)) {
             X86_emit_op_RT(CMP_V, tmpReg(tos-1), tos);
             X86_emit_op_RT(V(CMOVL), tmpReg(tos-1), tos);
         }
         else {
             X86_emit_op_RT(MOV_V, eax, tos-1);
             X86_emit_op_RT(CMP_V, eax, tos);
             X86_emit_op_RT(V(CMOVL), eax, tos);
             X86_emit_op_TR(MOV_V, tos-1, eax);
         }
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
		 // Note: Could use direct integer64 conparison (with proper adjustment
		 // of signed magnitude to two's complement representation) if willing
		 // to accept whatever behavior results for NaNs (behavior for infinities
		 // should be correct). But would it gain any useful speedup?
         if (gFmcp->use_CMOV) {
			 //FCOMI and FCMOVcc are available...
			 EMIT("fcomi");				//compare st to st(1), results to flags
			 EMIT("fcmovb st(1) ;op_max");	//load st from st(1) if st < st(1)
			 EMIT("fstp st(1)");		//move result to st(1) and pop
			 check_ndp_stack_room(-1);	//popped off 1 ndp stack entry

			 X86_emit_op(FCOMI);
			 X86_emit_opreg(FCMOVB_STN, 1);
			 X86_emit_op_R(FSTP_STN, 1);

			 TOSLIT.type = LITTYPEINTEGER;  //so pop() won't pop ndp stack below...
		 }
		 else {
			 //No FCOMI or FCMOVcc available...
			 CGL cgl1 = X86_genlab();

			 EMIT("fcom");              //compare st to st(1)
			 EMIT("fnstsw ax");
			 EMIT("test ah, 1 ;test C0");
			 EMIT2("jne cgl%d", cgl1);  //C0=1 => st < st(1)
			 EMIT("fst st(1)");         //C0=0 => st >= st(1)
			 EMIT2L("cgl%d:", cgl1);
			 ///EMIT("fpop ;st=max");   //done by pop() below...

			 X86_emit_op(FCOM);
			 X86_emit_op(FNSTSW_AX);
			 X86_emit_op_R_imm8(TESTI_B, ah, 1);
			 X86_emit_op_lab_disp8(JNE, cgl1);
			 X86_emit_op_R(FST_STN, 1);
			 X86_deflab(cgl1);
			 ///X86_emit_op(FPOP);  //done by pop() below...
		 }//!use_CMOV
	 }
	 else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
		 //Should convert int arg to real and return real?
         Mde_abort_command(MD386__NYI, 12, 0, 0); // for now
         NOSLIT.type = LITTYPEREAL;
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
		 //Should convert int arg to real and return real?
         Mde_abort_command(MD386__NYI, 13, 0, 0); // for now
         TOSLIT.type = LITTYPEREAL;
     }
     else if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == VAROBJECTKIND)
     {   /* compare two pointers */
         if (NOSVAR.frame != TOSVAR.frame)
         {   /* Pointers with different selectors are incommensurable. */
             Mde_warn(MD386__PTRISEL, 0, 0, 0);
         }
         Mde_abort_command(MD386__NYI, 14, 0, 0); // for now
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_max*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_dif              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_dif(void)
 {
     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         EMIT2("mov rAX, r%d", tos-1);
         EMIT2("sub rAX, r%d", tos);
		 if (X64gen)
	         EMIT("cqo ;dif(a,b)=abs(a-b)");
		 else
	         EMIT("cdq ;dif(a,b)=abs(a-b)");
         EMIT("xor rAX, rDX");
         EMIT("sub rAX, rDX");
         EMIT2("mov r%d, rAX", tos-1);

		 X86_emit_op_RT(MOV_V, eax, tos-1);
         X86_emit_op_RT(SUB_V, eax, tos);
         X86_emit_op(CDQ_V);
         X86_emit_op_RR(XOR_V, eax, edx);
         X86_emit_op_RR(SUB_V, eax, edx);
         X86_emit_op_TR(MOV_V, tos-1, eax);
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         Mde_abort_command(MD386__NYI, 15, 0, 0); // for now
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         Mde_abort_command(MD386__NYI, 16, 0, 0); // for now
         NOSLIT.type = LITTYPEREAL;
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         Mde_abort_command(MD386__NYI, 17, 0, 0); // for now
         TOSLIT.type = LITTYPEREAL;
     }
     else if (NOSOBJECT.kind == VAROBJECTKIND && TOSOBJECT.kind == VAROBJECTKIND)
     {   /* subtract two pointers */
         if (NOSVAR.frame != TOSVAR.frame)
         {   /* Pointers with different selectors are incommensurable. */
             Mde_warn(MD386__PTRISEL, 0, 0, 0);
         }
         Mde_abort_command(MD386__NYI, 18, 0, 0); // for now
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_dif*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_abs              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_abs(void)
 {
     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         EMIT2("mov rAX, r%d ;op_abs", tos);
		 if (X64gen)
	         EMIT("cqo");
		 else
	         EMIT("cdq");
         EMIT("xor rAX, rDX");
         EMIT("sub rAX, rDX");
         EMIT2("mov r%d, rAX", tos);

		 X86_emit_op_RT(MOV_V, eax, tos);
         X86_emit_op(CDQ_V);
         X86_emit_op_RR(XOR_V, eax, edx);
         X86_emit_op_RR(SUB_V, eax, edx);
         X86_emit_op_TR(MOV_V, tos, eax);
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         EMIT("fabs");
         X86_emit_op(FABS);
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_abs*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_floor            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_floor(void)
 {
     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
#if 0
         EMIT("fnstcw [rBX].current_cw ;save cw");    //for now; later set globally
#endif
         EMIT("fldcw [rBX].CW_FLOOR ;=173FH");
         EMIT("frndint ;floor");
         EMIT("fldcw [rBX].current_cw ;restore cw");

#if 0
         X86_emit_op_M(FNSTCW, FMC(current_cw));    //for now; later set globally
#endif
         X86_emit_op_M(FLDCW, FMC(CW_FLOOR));
         X86_emit_op(FRNDINT);
         X86_emit_op_M(FLDCW, FMC(current_cw));
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_floor*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_ceil             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_ceil(void)
 {
     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
#if 0
         EMIT("fnstcw [rBX].current_cw ;save cw");    //for now; later set globally
#endif
         EMIT("fldcw [rBX].CW_CEIL ;=1B3FH");
         EMIT("frndint ;ceil");
         EMIT("fldcw [rBX].current_cw ;restore cw");

#if 0
         X86_emit_op_M(FNSTCW, FMC(current_cw));    //for now; later set globally
#endif
         X86_emit_op_M(FLDCW, FMC(CW_CEIL));
         X86_emit_op(FRNDINT);
         X86_emit_op_M(FLDCW, FMC(current_cw));
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_ceil*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_round            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_round(void)
 {
     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
#if 0
         EMIT("fnstcw [rBX].current_cw ;save cw");    //for now; later set globally
#endif
         ///EMIT("fldcw [rBX].CW_ROUND ;=133FH");
         EMIT("frndint ;round");
         ///EMIT("fldcw [rBX].current_cw ;restore cw");

#if 0
         X86_emit_op_M(FNSTCW, FMC(current_cw));    //for now; later set globally
#endif
         ///X86_emit_op_M(FLDCW, FMC(CW_ROUND));
         X86_emit_op(FRNDINT);
         ///X86_emit_op_M(FLDCW, FMC(current_cw));
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_round*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_chop             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_chop(void)
 {
     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
#if 0
         EMIT("fnstcw [rBX].current_cw ;save cw");    //for now; later set globally
#endif
         EMIT("fldcw [rBX].CW_CHOP ;=1F3FH");
         EMIT("frndint ;chop");
         EMIT("fldcw [rBX].current_cw ;restore cw");

#if 0
         X86_emit_op_M(FNSTCW, FMC(current_cw));    //for now; later set globally
#endif
         X86_emit_op_M(FLDCW, FMC(CW_CHOP));
         X86_emit_op(FRNDINT);
         X86_emit_op_M(FLDCW, FMC(current_cw));
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_chop*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_sqr              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_sqr(void)
 {
     ensureoneoperand();

     //should check for NDP stack overflow????

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
#if 0
         //convert to double, take fsqrt, convert back to integer.
         //note that this uses the current rounding mode of the NDP,
         //which is probably round-to-even, not truncate!!!
		 check_ndp_stack_room(1);	//need 1 stack slot
         EMIT2("fild r%d", tos);
         X86_emit_op_Tm(FILD_V, tos);	//FILD_V = FILD_I or FILD_L
         EMIT("fsqrt");
         X86_emit_op(FSQRT);
#if 1
         //simulate chop instead of round for FF compatibility?
         EMIT("fsub_d [rBX].magic5 ;approx. floor");
         X86_emit_op_M(FSUB_D, FMC(magic5));
#endif
         EMIT2("fistp r%d ;integer sqrt", tos);
         X86_emit_op_Tm(FISTP_V, tos);	//FISTP_V = FISTP_I or FISTP_L
		 check_ndp_stack_room(-1);	//stack slot was popped
#elif 0
         //out-of-line isqrt() is faster than in-line fsqrt!
#ifdef X64
#error Not yet implemented for X64
#endif
         EMIT2("push r%d", tos);
         X86_emit_op_T(PUSH_D, tos);
         //EMIT("mov ECX, 1 ;nargs");			//nargs is unneeded
         //X86_emit_op_R_imm32(MOVI_D, ecx, 1); //nargs is unneeded
         EMIT("call [rBX].isqrt");
         X86_emit_op_M(CALLI_N, FMC(isqrt));
         EMIT("add ESP, 4 ;pop arg");
         X86_emit_op_R_simm32(ADDI_D, esp, 4);
         EMIT2("mov r%d, EAX", tos);
         X86_emit_op_TR(MOV_D, tos, eax);
#else
         //use FASTCALL instead of CDECL...
#ifdef X64
#ifdef _DEBUG
		 //Debug x64 version of isqrt_fast homes its arg,
		 //so requires shadow stack space; the release
		 //version doesn't use the shadow space, but we
		 //should always check whenever the VC compiler
		 //version changes, or if compile options change!
		 if (X64gen) {
			 //create the arglist
 			 ///EMIT("and RSP, -16 ;align 16");			//not needed if stack always aligned
			 ///X86_emit_op_R_simm32(ANDI_V, rsp, -16);	//not needed if stack always aligned
			 EMIT("sub RSP, 32 ;alloc arglist");		//32 = 4 * 8 shadow space
			 X86_emit_op_R_simm32(SUBI_V, rsp, 32);		//(used by isqrt_fast only under _DEBUG)
		 } //X64gen
#endif //_DEBUG
#endif //X64
         EMIT2("mov rCX, r%d", tos);
         X86_emit_op_RT(MOV_V, ecx, tos);
         EMIT("call [rBX].isqrt_fast");					//nargs is not needed
         X86_emit_op_M(CALLI_N, FMC(isqrt_fast));
#ifdef X64
#ifdef _DEBUG
		 //See note above re shadow stack!!
		 if (X64gen) {
			 //delete the arglist
			 EMIT("add RSP, 32 ;delete arglist");		//32 = 4 * 8 shadow space
			 X86_emit_op_R_simm32(ADDI_V, rsp, 32);		//(Probably not needed -- see above!!)
		 } //X64gen
#endif //_DEBUG
#endif //X64
         EMIT2("mov r%d, rAX", tos);					//move result from rAX to TOS
         X86_emit_op_TR(MOV_V, tos, eax);
#endif
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         EMIT("fsqrt");
         X86_emit_op(FSQRT);
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_sqr*/


 //Should break out OP_POW into its own module!!!

        /************************************/
        /*                                  */
        /*                                  */
        /*              op_pow              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_pow(void)
 {
     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {   
		 /*------------------------------------------*/
		 /* pow(int, int) --> ipow(int, int) --> int */
		 /*------------------------------------------*/
         NOSLIT.nbytes = 4;
         NOSLIT.ivalue = 0;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.ivalue;

////#define INLINE_IPOW 1
#ifdef INLINE_IPOW			//only works if TOS is a known integer constant...
#if 1
         if (1) { //force it for testing...
#else
         if (TOSLIT.tag == LITKNOWNVAL) {
#endif
             // in-line ipow(i, n) if n <= +18
             int n = TOSLIT.ivalue;
             // mustn't pop() yet if we might fall thru to next case...
             if (n < 0) {
                 // handle ipow(i, -n) special cases...
                 pop();     // i is now tos
                #error
                 return;    // return i for now... WRONG!!!!
             }
             else if (n <= 18) {
                 // do it in-line
                 pop();     // i is now tos (okay to pop now, since we won't fall thru)
                 switch (n) {
                 case 0:     // 0 mulops: 1.0
                     // ipow(i, 0) = 1, but for i=0 we should 
                     //  set errno = EDOM???
                     EMIT2("mov r%d, 1", tos);
                     X86_emit_op_T_imm32(MOVI_D, tos, 1);	//X64 zero-extends
                     break;
                 case 1:     // 0 mulops: i
                     // no op
                     break;									// tos = i
                 case 2:     // 1 mulop: i * i
                     if (isReg(tos)) {
                         EMIT3("imul r%d, r%d ;sqr", tos, tos);
                         X86_emit_op_TT(IMUL32_V, tos, tos);
                     }
                     else { //can gen better code than op_MM by ourselves:
                         EMIT2("mov rAX, r%d", tos);
                         EMIT("imul rAX, rAX ;sqr");
                         EMIT2("mov r%d, rAX", tos);
                         X86_emit_op_RM(MOV_V, eax, TMP(tos));
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_MR(MOV_V, TMP(tos), eax);
                     }
                     break;
                 case 3:     // 2 mulops: i * i * i
                     // can't use eax as buffer, since op_TR 
                     // uses eax if !isReg(tos); so use edx
                     EMIT2("mov rDX, r%d", tos);
                     EMIT("imul rDX, rDX ;i^2");
                     EMIT2("imul r%d, rDX ;i^3", tos);
                     X86_emit_op_RT(MOV_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_TR(IMUL32_V, tos, edx);
                     break;
                 case 4:     // 2 mulops: (i * i)^2
                     if (isReg(tos)) {
                         EMIT3("imul r%d, r%d ;i^2", tos, tos);
                         EMIT3("imul r%d, r%d ;i^4", tos, tos);
                         X86_emit_op_TT(IMUL32_V, tos, tos);
                         X86_emit_op_TT(IMUL32_V, tos, tos);
                     }
                     else { //can gen better code than op_TT/op_TT:
                         EMIT2("mov rAX, r%d", tos);
                         EMIT("imul rAX, rAX ;i^2");
                         EMIT("imul rAX, rAX ;i^4");
                         EMIT2("mov r%d, rAX", tos);
                         X86_emit_op_RT(MOV_V, eax, tos);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_TR(MOV_V, tos, eax);
                     }
                     break;
                 case 5:     // 3 mulops: (i^2)^2 * i
                     // can't use eax as buffer, since op_TR 
                     // uses eax if !isReg(tos); so use edx
                     EMIT2("mov rDX, r%d", tos);
                     EMIT("imul rDX, rDX ;i^2");
                     EMIT("imul rDX, rDX ;i^4");
                     EMIT2("imul r%d, rDX ;i^5", tos);
                     X86_emit_op_RT(MOV_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_TR(IMUL32_V, tos, edx);
                     break;
                 case 6:     // 3 mulops: (i * i * i)^2
                     // can't use eax as buffer, since op_TR 
                     // uses eax if !isReg(tos); so use edx
                     EMIT2("mov rDX, r%d", tos);
                     EMIT("imul rDX, rDX ;i^2");
                     EMIT2("imul r%d, rDX ;i^3", tos);
                     EMIT3("imul r%d, r%d ;i^6", tos, tos);
                     X86_emit_op_RT(MOV_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_TR(IMUL32_V, tos, edx);           //can gen better code!!!
                     X86_emit_op_TT(IMUL32_V, tos, tos);
                     break;
                 case 7:     // 4 mulops: (i * i * i)^2 * i
                     // can't use eax as buffer, since op_TR 
                     // uses eax if !isReg(tos); so use edx
                     EMIT2("mov rDX, r%d", tos);
                     EMIT("imul rDX, rDX ;i^2");
                     EMIT2("imul rDX, r%d ;i^3", tos);
                     EMIT("imul rDX, rDX ;i^6");
                     EMIT2("imul r%d, rDX ;i^7", tos);
                     X86_emit_op_RT(MOV_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RT(IMUL32_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_TR(IMUL32_V, tos, edx);           //can gen better code???
                    break;
                 case 8:     // 3 mulops: ((i * i)^2)^2
                     if (isReg(tos)) {
                         EMIT3("imul r%d, r%d ;i^2", tos, tos);
                         EMIT3("imul r%d, r%d ;i^4", tos, tos);
                         EMIT3("imul r%d, r%d ;i^8", tos, tos);
                         X86_emit_op_TT(IMUL32_V, tos, tos);
                         X86_emit_op_TT(IMUL32_V, tos, tos);
                         X86_emit_op_TT(IMUL32_V, tos, tos);
                     }
                     else { //can gen better code than op_TT/op_TT/op_TT:
                         EMIT2("mov rAX, r%d", tos);
                         EMIT("imul rAX, rAX ;i^2");
                         EMIT("imul rAX, rAX ;i^4");
                         EMIT("imul rAX, rAX ;i^8");
                         EMIT2("mov r%d, rAX", tos);
                         X86_emit_op_RT(MOV_V, eax, tos);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_TR(MOV_V, tos, eax);
                     }
                     break;
                 case 9:     // 4 mulops ((i * i)^2)^2 * i
                     if (isReg(tos)) {
                         EMIT2("mov rAX, r%d", tos);
                         EMIT("imul rAX, rAX ;i^2");
                         EMIT("imul rAX, rAX ;i^4");
                         EMIT("imul rAX, rAX ;i^8");
                         EMIT2("imul r%d, rAX ;i^9", tos);
                         X86_emit_op_RT(MOV_V, eax, tos);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_TR(IMUL32_V, tos, eax);
                     }
                     else { //can gen better code than op_TR:
                         EMIT2("mov rAX, r%d", tos);
                         EMIT("imul rAX, rAX ;i^2");
                         EMIT("imul rAX, rAX ;i^4");
                         EMIT("imul rAX, rAX ;i^8");
                         EMIT2("imul rAX, r%d ;i^9", tos);
                         EMIT2("mov r%d, rAX", tos);
                         X86_emit_op_RT(MOV_V, eax, tos);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RT(IMUL32_V, eax, tos);
                         X86_emit_op_TR(MOV_V, tos, eax);
                     }
                     break;
                 case 10:    // 4 mulops: ((i * i)^2 * i)^2
                     // can't use eax as buffer, since op_TR 
                     // uses eax if !isReg(tos); so use edx
                     EMIT2("mov rDX, r%d", tos);
                     EMIT("imul rDX, rDX ;i^2");
                     EMIT("imul rDX, rDX ;i^4");
                     EMIT2("imul r%d, rDX ;i^5", tos);
                     EMIT3("imul r%d, r%d ;i^10", tos, tos);
                     X86_emit_op_RT(MOV_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_TR(IMUL32_V, tos, edx);
                     X86_emit_op_TT(IMUL32_V, tos, tos);  //can do better if TMP???
                     break;
                 case 11:    // 5 mulops: ((i * i)^2 * i)^2 * i
                     // can't use eax as buffer, since op_TR 
                     // uses eax if !isReg(tos); so use edx
                     EMIT2("mov rDX, r%d", tos);
                     EMIT("imul rDX, rDX ;i^2");
                     EMIT("imul rDX, rDX ;i^4");
                     EMIT2("imul rDX, r%d ;i^5", tos);
                     EMIT("imul rDX, rDX ;i^10");
                     EMIT2("imul r%d, rDX ;i^11", tos);
                     X86_emit_op_RT(MOV_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RT(IMUL32_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_TR(IMUL32_V, tos, edx);  //can do better if TMP???
                     break;
                 case 12:    // 4 mulops: ((i * i * i)^2)^2
                     // can't use eax as buffer, since op_TR 
                     // uses eax if !isReg(tos); so use edx === not true here??
                     EMIT2("mov rDX, r%d", tos);
                     EMIT("imul rDX, rDX ;i^2");
                     EMIT2("imul rDX, r%d ;i^3", tos);
                     EMIT("imul rDX, rDX ;i^6");
                     EMIT("imul rDX, rDX ;i^12");
                     EMIT2("mov r%d, rDX", tos);
                     X86_emit_op_RT(MOV_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RT(IMUL32_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_TR(MOV_V, tos, edx);   //can eliminate if isReg(tos)???
                     break;
                 case 13:    // 5 mulops: ((i * i * i)^2)^2 * i
                     EMIT2("mov rDX, r%d", tos);
                     EMIT("imul rDX, rDX ;i^2");
                     EMIT2("imul rDX, r%d ;i^3", tos);
                     EMIT("imul rDX, rDX ;i^6");
                     EMIT("imul rDX, rDX ;i^12");
                     EMIT2("imul rDX, r%d ;i^13", tos);
                     EMIT2("mov r%d, rDX", tos);
                     X86_emit_op_RT(MOV_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RT(IMUL32_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RT(IMUL32_V, edx, tos);
                     X86_emit_op_TR(MOV_V, tos, edx);   //can eliminate if isReg(tos)???
                     break;
                 case 14:    // 5 mulops: ((i * i * i)^2 * i)^2
                     EMIT2("mov rDX, r%d", tos);
                     EMIT("imul rDX, rDX ;i^2");
                     EMIT2("imul rDX, r%d ;i^3", tos);
                     EMIT("imul rDX, rDX ;i^6");
                     EMIT2("imul rDX, r%d ;i^7", tos);
                     EMIT2("imul rDX, rDX ;i^14", tos);
                     EMIT2("mov r%d, rDX", tos);
                     X86_emit_op_RT(MOV_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RT(IMUL32_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);
                     X86_emit_op_RT(IMUL32_V, edx, tos);
                     X86_emit_op_RR(IMUL32_V, edx, edx);  //can gen better code???
                     X86_emit_op_TR(MOV_V, tos, edx);
                     break;
                 case 15:    // 5 mulops (1st time add-chain beats binary exponentiation (6 mulops) per [Knuth]!!)
                     // temp = (i * i)^2 * i = i^5; (temp * temp * temp) = (i^5)^3 = i^15
                     EMIT2("mov rAX, r%d", tos);
                     EMIT("imul rAX, rAX ;i^2");
                     EMIT("imul rAX, rAX ;i^4");
                     EMIT2("imul rAX, r%d ;i^5", tos);
                     EMIT("mov rDX, rAX ;i^5");
                     EMIT("imul rAX, rAX ;i^10");
                     EMIT("imul rAX, rDX ;i^15");
                     EMIT2("mov r%d, rAX", tos);
                     X86_emit_op_RT(MOV_V, eax, tos);
                     X86_emit_op_RR(IMUL32_V, eax, eax);  //i^2
                     X86_emit_op_RR(IMUL32_V, eax, eax);  //i^4
                     X86_emit_op_RT(IMUL32_V, eax, tos);  //i^5
                     X86_emit_op_RR(MOV_V, edx, eax);	  //i^5
                     X86_emit_op_RR(IMUL32_V, eax, eax);  //i^10
                     X86_emit_op_RR(IMUL32_V, eax, edx);  //i^15
                     X86_emit_op_TR(MOV_V, tos, eax);
                     break;
                 case 16:    // 4 mulops: (((i * i)^2)^2)^2
                     if (isReg(tos)) {
                         EMIT3("imul r%d, r%d ;i^2", tos, tos);
                         EMIT3("imul r%d, r%d ;i^4", tos, tos);
                         EMIT3("imul r%d, r%d ;i^8", tos, tos);
                         EMIT3("imul r%d, r%d ;i^16", tos, tos);
                         X86_emit_op_TT(IMUL32_V, tos, tos);  //i^2
                         X86_emit_op_TT(IMUL32_V, tos, tos);  //i^4
                         X86_emit_op_TT(IMUL32_V, tos, tos);  //i^8
                         X86_emit_op_TT(IMUL32_V, tos, tos);  //i^16
                     }
                     else { //can gen better code when !isReg(tos):
                         EMIT2("mov rAX, r%d", tos);
                         EMIT("imul rAX, rAX ;i^2");
                         EMIT("imul rAX, rAX ;i^4");
                         EMIT("imul rAX, rAX ;i^8");
                         EMIT("imul rAX, rAX ;i^16");
                         EMIT2("mov r%d, rAX", tos);
                         X86_emit_op_RT(MOV_V, eax, tos);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_TR(MOV_V, tos, eax);
                     }
                     break;
                 case 17:    // 5 mulops: (((i * i)^2)^2)^2 * i
                     EMIT2("mov rAX, r%d", tos);
                     EMIT("imul rAX, rAX ;i^2");
                     EMIT("imul rAX, rAX ;i^4");
                     EMIT("imul rAX, rAX ;i^8");
                     EMIT("imul rAX, rAX ;i^16");
                     EMIT2("imul rAX, r%d ;i^17", tos);
                     EMIT2("mov r%d, rAX", tos);
                     X86_emit_op_RT(MOV_V, eax, tos);
                     X86_emit_op_RR(IMUL32_V, eax, eax);
                     X86_emit_op_RR(IMUL32_V, eax, eax);
                     X86_emit_op_RR(IMUL32_V, eax, eax);
                     X86_emit_op_RR(IMUL32_V, eax, eax);
                     X86_emit_op_RT(IMUL32_V, eax, tos);
                     X86_emit_op_TR(MOV_V, tos, eax);
                     break;
                 case 18:    // 5 mulops: (((i * i)^2)^2 * i)^2
                     if (isReg(tos)) {
                         EMIT2("mov rAX, r%d", tos);
                         EMIT1("imul rAX, rAX ;i^2");
                         EMIT1("imul rAX, rAX ;i^4");
                         EMIT1("imul rAX, rAX ;i^8");
                         EMIT2("imul r%d, rAX ;i^9", tos);
                         EMIT3("imul r%d, r%d ;i^18", tos, tos);
                         X86_emit_op_RT(MOV_V, eax, tos);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_TR(IMUL32_V, tos, eax);
                         X86_emit_op_TT(IMUL32_V, tos, tos);
                     }
                     else { //can gen better code than op_TR:
                         EMIT2("mov rAX, r%d", tos);
                         EMIT1("imul rAX, rAX ;i^2");
                         EMIT1("imul rAX, rAX ;i^4");
                         EMIT1("imul rAX, rAX ;i^8");
                         EMIT2("imul rAX, r%d ;i^9", tos);
                         EMIT1("imul rAX, rAX ;i^18");
                         EMIT2("mov r%d, rAX", tos);
                         X86_emit_op_RT(MOV_V, eax, tos);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_RT(IMUL32_V, eax, tos);
                         X86_emit_op_RR(IMUL32_V, eax, eax);
                         X86_emit_op_TR(MOV_V, tos, eax);
                     }
                     break;
                 }//switch n
                 return; //done!
             }// n <= 18
         }//if LITKNOWNVAL

         //fall through to general case... (n is unknown or > 18)

#endif //INLINE_IPOW

         // cdecl call ipow(arg1, arg2)
#ifdef X64
		 if (X64gen) {
			 //create the arglist						//CAN WE USE THE STATIC ARGLIST IF FREE??
 			 ///EMIT("and RSP, -16 ;align 16");			//not needed if stack always aligned
			 ///X86_emit_op_R_simm32(ANDI_V, rsp, -16);	//not needed if stack always aligned
			 EMIT("sub RSP, 32 ;alloc arglist");		//32 = 4 * 8 min stack shadow space
			 X86_emit_op_R_simm32(SUBI_V, rsp, 32);		//(used by ipow only under _DEBUG???)
			 EMIT2("mov RCX, r%d", tos-1);				//arg1
			 X86_emit_op_RT(MOV_V, rcx, tos-1);
			 EMIT2("mov RDX, r%d", tos);				//arg2
			 X86_emit_op_RT(MOV_V, rdx, tos);
		 } //X64gen
		 else 
#endif //X64
		 {//X32
			 EMIT2("push r%d", tos);					//push arg2
			 X86_emit_op_T(PUSH_D, tos);
			 EMIT2("push r%d", tos-1);					//push arg1
			 X86_emit_op_T(PUSH_D, tos-1);
		 }
		 //no need to empty the NDP stack--ipow is leaf and doesn't use ndp???	//WRONG!!
		 //Use FASTCALL for x86??
		 //nargs not needed
         EMIT("call [rBX].ipow");
         X86_emit_op_M(CALLI_N, FMC(math_ipow));
#ifdef X64
		 if (X64gen) {
			 //delete the arglist
			 EMIT("add RSP, 32 ;delete arglist");		//32 = 4 * 8 min stack shadow space
			 X86_emit_op_R_simm32(ADDI_V, rsp, 32);		//(maybe not needed -- see above!!)
		 } //X64gen
		 else
#endif //X64
		 {
			 //X86 -- pop the args
			 EMIT("add esp, 8 ;pop args");
			 X86_emit_op_R_simm32(ADDI_D, esp, 8);
		 }
         EMIT2("mov r%d, rAX", tos-1);					//move result from rAX to NOS
         X86_emit_op_TR(MOV_V, tos-1, eax);
         pop();											//NOS becomes TOS
     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL && 
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {   
		 /*--------------------------*/
		 /* pow(real, real) --> real */
		 /*--------------------------*/
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = 0.0;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;

         // cdecl call pow(arg1, arg2)
#ifdef X64
		 if (X64gen) {
			 //create the arglist						//CAN WE USE THE STATIC ARGLIST IF FREE??
 			 ///EMIT("and RSP, -16 ;align 16");			//not needed if stack always aligned
			 ///X86_emit_op_R_simm32(ANDI_V, rsp, -16);	//not needed if stack always aligned
			 EMIT("sub RSP, 32 ;alloc arglist");		//x64: alloc 32 bytes min arglist
			 X86_emit_op_R_simm32(SUBI_V, rsp, 32);
		 } //X64gen
		 else
#endif //X64
		 {	 //x86 allocate arglist
			 EMIT("sub ESP, 16");						//allocate arglist
			 X86_emit_op_R_simm32(SUBI_V, esp, 16);
		 }//x86
		 EMIT("fstp [rSP+8]");							//store arg2
		 check_ndp_stack_room(-1);						//popped one ndp entry
         X86_emit_op_M(FSTP_D, esp, no_reg, 0, 8);
         TOSLIT.type = LITTYPEINTEGER;					//so pop won't pop ndp stack
         pop();
         EMIT("fstp [rSP]");							//store arg1
		 check_ndp_stack_room(-1);						//popped another ndp entry
         X86_emit_op_M(FSTP_D, esp, no_reg, 0, 0);
#ifdef X64
		 if (X64gen) { //move args to f.p. fastcall registers
#ifndef USE_MOVSDX
			 // Use movd/movq to copy mem to xmm reg.
			 EMIT("movd xmm1, qword ptr[RSP+8]");
			 X86_emit_op_RM(XMOVQ, xmm1, rsp, no_reg, 0, 8);
			 EMIT("movd xmm0, qword ptr[RSP]");
			 X86_emit_op_RM(XMOVQ, xmm0, rsp, no_reg, 0, 0);
#else
			 // VC++ uses movsdx instead of movd/q. Why?
			 EMIT("movsdx xmm1, qword ptr[RSP+8]");
			 X86_emit_op_RM(MOVSDX, xmm1, rsp, no_reg, 0, 8);
			 EMIT("movsdx xmm0, qword ptr[RSP]");
			 X86_emit_op_RM(MOVSDX, xmm0, rsp, no_reg, 0, 0);
#endif //USE_MOVSDX
		 } //X64gen
#endif //X64
		 //empty the NDP stack if necessary??? !!!!
		 //Use FASTCALL for x86??
		 //nargs is unneeded by pow()
         EMIT("call [rBX].pow");
         X86_emit_op_M(CALLI_N, FMC(math_pow));
		 //restore the NDP stack if necessary!!!
#ifdef X64
		 if (X64gen) {
			 //delete the arglist
			 //(maybe not needed if static arglist??)
			 EMIT("add RSP, 32 ;delete arglist");		//32 = min stack shadow space
			 X86_emit_op_R_simm32(ADDI_V, rsp, 32);

			 //for X64, real result is returned in xmm0.
			 //push it onto the ndp stack for now...
#ifndef USE_MOVSDX
			 // Use movq to copy xmm reg to m64.
			 EMIT("movq temp, xmm0");
			 X86_emit_op_MR(XMOVQ, FMC(temp), xmm0);
#else
			 EMIT("movsdx temp, xmm0");
			 X86_emit_op_MR(MOVSDX, FMC(temp), xmm0);
#endif //USE_MOVSDX
			 check_ndp_stack_room(1);					//push 1 ndp entry
			 EMIT("fld_d temp");
			 X86_emit_op_M(FLD_D, FMC(temp));
		 } //X64gen
		 else
#endif //X64
		 {
			 //X86 -- call to f.p. function pushed return
			 //       value onto the ndp stack
			 check_ndp_stack_room(1);					//real call pushed 1 ndp entry
			 //X86 -- pop the args
			 EMIT("add ESP, 16 ;pop args");				//deallocate arglist
			 X86_emit_op_R_simm32(ADDI_V, esp, 16);		//bug fix (was 8 not 16)
		 }
         //real result is now in ndp st(0), at TOS

     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {  
		 /*--------------------------------------------*/
		 /* pow(int, real) --> pow(real, real) -> real */
		 /*--------------------------------------------*/

         // Optimize to sqrt() when exp = 0.5, etc. ????? Or does CRT pow() do that anyway???

         NOSLIT.type = LITTYPEREAL;
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = 0.0;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;

         // cdecl call pow( (double)arg1, arg2 )
#ifdef X64
		 if (X64gen) {
			 //create the arglist						//CAN WE USE THE STATIC ARGLIST IF FREE??
 			 ///EMIT("and RSP, -16 ;align 16");			//not needed if stack always aligned
			 ///X86_emit_op_R_simm32(ANDI_V, rsp, -16);	//not needed if stack always aligned
			 EMIT("sub RSP, 32");						//x64: alloc 32 bytes min arglist
			 X86_emit_op_R_simm32(SUBI_V, rsp, 32);
		 } //X64gen
		 else
#endif //X64
		 {	 //x86 allocate arglist
			 EMIT("sub ESP, 16");						//allocate arglist
			 X86_emit_op_R_simm32(SUBI_V, esp, 16);
		 }//x86
         EMIT("fstp [rSP+8]");							//store arg2 in arglist
         X86_emit_op_M(FSTP_D, esp, no_reg, 0, 8);
		 check_ndp_stack_room(-1);						//popped 1 ndp entry
         TOSLIT.type = LITTYPEINTEGER;					//so next pop won't pop ndp stack
         pop();
         // convert arg1 to double
		 check_ndp_stack_room(1);						//need room for 1 ndp entry
         EMIT2("fild r%d ;op_double", tos);				//load double from int32/64
         X86_emit_op_Tm(FILD_V, tos);
         EMIT("fstp [rSP]");							//store double arg1
         X86_emit_op_M(FSTP_D, esp, no_reg, 0, 0);
		 check_ndp_stack_room(-1);						//entry has been popped
#ifdef X64
		 if (X64gen) { //move args to f.p. fastcall registers
#ifndef USE_MOVSDX
			 // Use movd/movq to copy mem to xmm reg.
			 EMIT("movd xmm1, qword ptr[RSP+8]");
			 X86_emit_op_RM(XMOVQ, xmm1, rsp, no_reg, 0, 8);
			 EMIT("movd xmm0, qword ptr[RSP]");
			 X86_emit_op_RM(XMOVQ, xmm0, rsp, no_reg, 0, 0);
#else
			 // VC++ uses movsdx instead of movd/q. Why?
			 EMIT("movsdx xmm1, qword ptr[RSP+8]");
			 X86_emit_op_RM(MOVSDX, xmm1, rsp, no_reg, 0, 8);
			 EMIT("movsdx xmm0, qword ptr[RSP]");
			 X86_emit_op_RM(MOVSDX, xmm0, rsp, no_reg, 0, 0);
#endif //USE_MOVSDX
		 } //X64gen
#endif //X64
		 //empty the NDP stack if necessary???
		 //Use FASTCALL for x86??
		 //nargs is unneeded by pow()
         EMIT("call [rBX].pow");
         X86_emit_op_M(CALLI_N, FMC(math_pow));
		 //restore the NDP stack if necessary!!!
#ifdef X64
		 if (X64gen) {
			 //delete the arglist
			 //(maybe not needed if static arglist??)
			 EMIT("add RSP, 32 ;delete arglist");		//32 = min stack shadow space
			 X86_emit_op_R_simm32(ADDI_V, rsp, 32);

			 //for X64, real result is returned in xmm0.
			 //push it onto the ndp stack for now...
#ifndef USE_MOVSDX
			 // Use movq to copy xmm reg to m64.
			 EMIT("movq temp, xmm0");
			 X86_emit_op_MR(XMOVQ, FMC(temp), xmm0);
#else
			 EMIT("movsdx temp, xmm0");
			 X86_emit_op_MR(MOVSDX, FMC(temp), xmm0);
#endif //USE_MOVSDX
			 check_ndp_stack_room(1);					//push 1 ndp entry
			 EMIT("fld_d temp");
			 X86_emit_op_M(FLD_D, FMC(temp));
		 } //X64gen
		 else
#endif //X64
		 {
			 //X86 -- call to f.p. function pushed return
			 //       value onto the ndp stack
			 check_ndp_stack_room(1);					//real call pushed 1 ndp entry
			 //X86 -- pop the args
			 EMIT("add ESP, 16 ;pop args");				//deallocate arglist
			 X86_emit_op_R_simm32(ADDI_V, esp, 16);		//bug fix (was 8 not 16)
		 }
         //real result is now in ndp st(0), at TOS

     }
     else if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEREAL &&
              TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {   
		 /*--------------------------------------------*/
		 /* pow(real, int) --> powi(real, int) -> real */
		 /*--------------------------------------------*/
         NOSLIT.nbytes = 8;
         NOSLIT.rvalue = 0.0;
         NOSLIT.valuep = (UINT1 *)&NOSLIT.rvalue;

///#define INLINE_POWI 1
#ifdef INLINE_POWI
#if 1
         if (1) { //force it for testing...
#else
         if (TOSLIT.tag == LITKNOWNVAL) {
#endif
			 // in-line powi(x, n) if -18 <= n <= +18
             int n = abs(TOSLIT.ivalue);
             if (n <= 18) {
                 // do it in-line
                 switch (n) {
                 case 0:     // 0 mulops: 1.0
                     // powi(x, 0) = 1 
                     // but set EDOM if x == 0.0 ??
                     EMIT("fpop");					// throw away x
					 check_ndp_stack_room(-1);		//free the top ndp stack entry
					 check_ndp_stack_room(1);		//reallocate the ndp stack entry
                     EMIT("fld1");					// return 1.0
                     X86_emit_op(FPOP);
                     X86_emit_op(FLD1);
                     break;
                 case 1:     // 0 mulops: x
					 // no op
                     break;							// = x
                 case 2:     // 1 mulop: x * x
                     EMIT("fmul st, st(0)");		// =fsqr x^2
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     break;
                 case 3:     // 2 mulops: x * x * x
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmulp st(1), st");		// =fmul x^2 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     break;
                 case 4:     // 2 mulops: (x * x)^2
                     EMIT("fmul st, st(0)");		// =fsqr x^2
                     EMIT("fmul st, st(0)");		// =fsqr x^4
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     break;
                 case 5:     // 3 mulops: (x^2)^2 * x
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmul st, st(0)");		// =fsqr x^4, x
                     EMIT("fmulp st(1), st");		// =fmul x^4 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     break;
                 case 6:     // 3 mulops: (x * x * x)^2
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmulp st(1), st");		// =fmul x^2 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     EMIT("fmul st, st(0)");		// =fsqr (x^2 * x)^2
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     break;
                 case 7:     // 4 mulops: (x * x * x)^2 * x			<<<=== RESUME HERE
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmul st, st(1)");		// =fmul x^2 * x, x
                     EMIT("fmul st, st(0)");		// =fsqr (x^2 * x)^2, x
                     EMIT("fmulp st(1), st");		// =fmul (x^2 * x)^2 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 1);    // FMUL ST, ST(1)
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     break;
                 case 8:     // 3 mulops: ((x * x)^2)^2
                     EMIT("fmul st, st(0)");		// =fsqr x^2
                     EMIT("fmul st, st(0)");		// =fsqr x^4
                     EMIT("fmul st, st(0)");		// =fsqr x^8
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     break;
                 case 9:     // 4 mulops ((x * x)^2)^2 * x
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmul st, st(0)");		// =fsqr x^4, x
                     EMIT("fmul st, st(0)");		// =fsqr x^8, x
                     EMIT("fmulp st(1), st");		// =fmul x^8 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     break;
                 case 10:    // 4 mulops: ((x * x)^2 * x)^2
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmul st, st(0)");		// =fsqr x^4, x
                     EMIT("fmulp st(1), st");		// =fmul x^4 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     EMIT("fmul st, st(0)");		// =fsqr (x^4 * x)^2
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     break;
                 case 11:    // 5 mulops: ((x * x)^2 * x)^2 * x
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmul st, st(0)");		// =fsqr x^4, x
                     EMIT("fmul st, st(1)");		// =fmul x^4 * x, x
                     EMIT("fmul st, st(0)");		// =fsqr (x^4 * x)^2, x
                     EMIT("fmulp st(1), st");		// =fmul (x^4 * x)^2 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 1);    // FMUL ST, ST(1)
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     break;
                 case 12:    // 4 mulops: ((x * x * x)^2)^2
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmulp st(1), st");		// =fmul x^2 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     EMIT("fmul st, st(0)");		// =fsqr (x^2 * x)^2
                     EMIT("fmul st, st(0)");		// =fsqr ((x^2 * x)^2)^2
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     break;
                 case 13:    // 5 mulops: ((x * x * x)^2)^2 * x
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmul st, st(1)");		// =fmul x^2 * x, x
                     EMIT("fmul st, st(0)");		// =fsqr (x^2 * x)^2, x
                     EMIT("fmul st, st(0)");		// =fsqr ((x^2 * x)^2)^2, x
                     EMIT("fmulp st(1), st");		// =fmul ((x^2 * x)^2)^2 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 1);    // FMUL ST, ST(1)
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     break;
                 case 14:    // 5 mulops: ((x * x * x)^2 * x)^2
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmul st, st(1)");		// =fmul x^2 * x, x
                     EMIT("fmul st, st(0)");		// =fsqr (x^2 * x)^2, x
                     EMIT("fmulp st(1), st");		// =fmul (x^2 * x)^2 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     EMIT("fmul st, st(0)");		// =fsqr ((x^2 * x)^2 * x)^2
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 1);    // FMUL ST, ST(1)
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     break;
                 case 15:    // 5 mulops (1st time add-chain beats binary exponentiation (6 mulops) per [Knuth]!!)
                     // temp = (x * x)^2 * x = x^5; (temp * temp * temp) = (x^5)^3 = x^15
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmul st, st(0)");		// =fsqr x^4, x
                     EMIT("fmulp st(1), st");		// =fmul x^4 * x = x^5
					 //check_ndp_stack_room(-1);	//free the extra ndp stack entry -- these 2 cancel out
					 //check_ndp_stack_room(1);		//need one more ndp stack entry  -- these 2 cancel out
                     EMIT("fld st(0)");				// =fdup x^5, x^5
                     EMIT("fmul st, st(0)");		// =fsqr x^10, x^5
                     EMIT("fmulp st(1), st");		// =fmul x^15
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     break;
                 case 16:    // 4 mulops: (((x * x)^2)^2)^2
                     EMIT("fmul st, st(0)");		// =fsqr x^2
                     EMIT("fmul st, st(0)");		// =fsqr x^4
                     EMIT("fmul st, st(0)");		// =fsqr x^8
                     EMIT("fmul st, st(0)");		// =fsqr x^16
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     break;
                 case 17:    // 5 mulops: (((x * x)^2)^2)^2 * x
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmul st, st(0)");		// =fsqr x^4, x
                     EMIT("fmul st, st(0)");		// =fsqr x^8, x
                     EMIT("fmul st, st(0)");		// =fsqr x^16, x
                     EMIT("fmulp st(1), st");		// =fmul x^16 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     break;
                 case 18:    // 5 mulops: (((x * x)^2)^2 * x)^2
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld st(0)");				// =fdup x, x
                     EMIT("fmul st, st(0)");		// =fsqr x^2, x
                     EMIT("fmul st, st(0)");		// =fsqr x^4, x
                     EMIT("fmul st, st(0)");		// =fsqr x^8, x
                     EMIT("fmulp st(1), st");		// =fmul x^8 * x
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     EMIT("fmul st, st(0)");		// =fsqr (x^9)^2
                     X86_emit_op_R(FLD_STN, 0);     // FDUP
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     X86_emit_op(FMUL);             // FMUL
                     X86_emit_op_R(FMUL_STN, 0);    // FSQR
                     break;
                 }//switch n
                 if (TOSLIT.ivalue < 0){
                     // return reciprocal if exponent was negative
#if 0
					 check_ndp_stack_room(1);		//need one more ndp stack entry
                     EMIT("fld1");
                     EMIT("fdivp st(1),st ;reciprocal");
					 check_ndp_stack_room(-1);		//free the extra ndp stack entry
                     X86_emit_op(FLD1);
                     X86_emit_op(FDIV);
#else
                     // following is faster on 387:
                     EMIT("fdivr =1.0F ;reciprocal");   // 1.0F (float) is faster than 1.0 (double)
                     X86_emit_op_M(FDIVR_F, FMC(float1p0));
#endif
                 }
                 pop(); //real result is in ndp st(0)
                 return; //done!
             }// n <= 18
         }//if LITKNOWNVAL

         //fall through to other cases...

#endif //INLINE_POWI

         // cdecl call powi(arg1, arg2)
#ifdef X64
		 if (X64gen) {
 			 ///EMIT("and RSP, -16 ;align 16");			//not needed if stack always aligned
			 ///X86_emit_op_R_simm32(ANDI_V, rsp, -16);	//not needed if stack always aligned
			 EMIT("sub RSP, 32");						//x64: alloc 32 bytes min arglist
			 X86_emit_op_R_simm32(SUBI_V, esp, 32);
		 } //X64gen
		 else
#endif //X64
		 {	 //x86 allocate arglist (8+4=12 bytes)
			 EMIT("sub ESP, 12");						//allocate arglist
			 X86_emit_op_R_simm32(SUBI_V, esp, 12);
		 }//x86
         EMIT2("mov [rSP+8], r%d", tos);				//move arg2 to arglist
         X86_emit_op_MT(MOV_V, esp, no_reg, 0, 8, tos);
         pop();
         EMIT("fstp [rSP]");							//move arg1 to arglist
		 check_ndp_stack_room(-1);						//popped 1 ndp entry
         X86_emit_op_M(FSTP_D, esp, no_reg, 0, 0);
#ifdef X64
		 if (X64gen) { //move args to f.p. fastcall registers
			 EMIT("mov RDX, qword ptr[RSP+8]");				//move arg2 to RDX
			 X86_emit_op_RM(MOV_V, rdx, rsp, no_reg, 0, 8);
#ifndef USE_MOVSDX
			 // Use movd/movq to copy mem to xmm reg.	//movq == movd qword ptr???
			 EMIT("movd xmm0, qword ptr[RSP]");				//move arg1 to xmm0	//uses movq???
			 X86_emit_op_RM(XMOVQ, xmm0, rsp, no_reg, 0, 0);
#else
			 // VC++ uses movsdx instead of movd/q. Why?
			 EMIT("movsdx xmm0, qword ptr[RSP]");			//move arg1 to xmm0
			 X86_emit_op_RM(MOVSDX, xmm0, rsp, no_reg, 0, 0);
#endif //USE_MOVSDX
		 } //X64gen
#endif //X64
		 //empty NDP stack if necessary???
		 //nargs is unneeded by math_powi
         EMIT("call [rBX].powi");
         X86_emit_op_M(CALLI_N, FMC(math_powi));
#ifdef X64
		 if (X64gen) {
			 //delete the arglist
			 //(maybe not needed if static arglist??)
			 EMIT("add RSP, 32 ;delete arglist");		//32 = min stack shadow space
			 X86_emit_op_R_simm32(ADDI_V, rsp, 32);

			 //for X64, real result is returned in xmm0.
			 //push it onto the ndp stack for now...
#ifndef USE_MOVSDX
			 // Use movq to copy xmm reg to m64.
			 EMIT("movq temp, xmm0");
			 X86_emit_op_MR(XMOVQ, FMC(temp), xmm0);
#else
			 EMIT("movsdx temp, xmm0");
			 X86_emit_op_MR(MOVSDX, FMC(temp), xmm0);
#endif //USE_MOVSDX
			 check_ndp_stack_room(1);					//push 1 ndp entry
			 EMIT("fld_d temp");
			 X86_emit_op_M(FLD_D, FMC(temp));
		 } //X64gen
		 else
#endif //X64
		 {
			 //X86 -- call to f.p. function pushed return
			 //       value onto the ndp stack
			 check_ndp_stack_room(1);					//real call pushed 1 ndp entry
			 //X86 -- pop the args
			 EMIT("add ESP, 12 ;pop args");				//deallocate arglist (8+4=12 bytes)
			 X86_emit_op_R_simm32(ADDI_V, esp, 12);		//bug fix (was 8)
		 }
         //real result is now in ndp st(0), at TOS

     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*op_pow*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_ctl              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_ctl(void)
 {
     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         int reg;
         if (isReg(tos)) {
             reg = tmpReg(tos);
         }
         else {
             reg = eax;
			 EMIT3("mov %s, r%d ;index reg", GPR_V(reg), tos);
             X86_emit_op_RT(MOV_V, reg, tos);
         }
         // N_CTLS must be a power of 2!!!!
         assert(IS_POWER_OF_TWO(N_CTLS));
         assert(sizeof(gFmcp->pre_ctl[0]) == 4);
         EMIT3("and %s, %d ;safe index", GPR_D(reg), N_CTLS-1);	//should use tmpReg pseudo-name, not actual GPR??
         X86_emit_op_R_simm32(ANDI_D, reg, N_CTLS-1);	//X64 zero-extends to 64-bits
#ifdef X64
		 if (X64gen) {
			 //X64: sign-extend 32-bit value of pre_ctl[reg] to 64-bits
			 EMIT4("movsxd r%d, [%s+4*%s].pre_ctl", tos, GPR_V(ebx), GPR_V(reg));
			 X86_emit_op_TM(MOVSXD_Q, tos, FMCX(pre_ctl,reg,4));
		 } else
#endif
		 {
			 //X86: load 32-bit value of pre_ctl[reg]
			 EMIT4("mov r%d, [%s+4*%s].pre_ctl", tos, GPR_V(ebx), GPR_V(reg));
			 X86_emit_op_TM(MOV_D, tos, FMCX(pre_ctl,reg,4));
		 }
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

#ifdef FML
     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
#else
     TOSLIT.nbytes = 1;
     TOSLIT.valuep = storeliteral(1, &result);
#endif
 } /*op_ctl*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             op_cnvX_Y            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_cnvX_Y(int x_or_y)					//X64 NEEDS WORK!!??
 {
     // val = cnv*(k, offset, d, pfGet, x, y, z)
     // [temp:     r1   r2    r3  r4    r5 r6 r7

     CGL loop = X86_genlab();
     int reg;

     ensurenoperands(7);

     // tos = z
     if (TOSOBJECT.kind != LITOBJECTKIND || TOSLIT.type != LITTYPEINTEGER)
     {
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
         return;
     }
     EMIT2("push r%d ;z", tos);
     X86_emit_op_T(PUSH_D, tos);
     pop();

     // tos = y
     if (TOSOBJECT.kind != LITOBJECTKIND || TOSLIT.type != LITTYPEINTEGER)
     {
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
         return;
     }
     if (x_or_y == 1) {
         // cnvY
         EMIT3("sub r%d, r%d ;y -= k", tos, tos-5);
         X86_emit_op_TT(SUB_V, tos, tos-5);
     }
     EMIT2("push r%d ;y", tos);
     X86_emit_op_T(PUSH_D, tos);
     pop();

     // tos = x
     if (TOSOBJECT.kind != LITOBJECTKIND || TOSLIT.type != LITTYPEINTEGER)
     {
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
         return;
     }
     if (x_or_y == 0) {
         // cnvX
         EMIT3("sub r%d, r%d ;x -= k", tos, tos-4);
         X86_emit_op_TT(SUB_V, tos, tos-4);			//X64??
     }
     EMIT2("push r%d ;x", tos);
     X86_emit_op_T(PUSH_D, tos);

     //pop();  don't pop r5 yet -- we're going to use it for sum.
     EMIT2("mov r%d, 0 ;sum = 0", tos);		// or xor if isreg??
     X86_emit_op_T_imm32(MOVI_D, tos, 0);

     if (isReg(tos-4)) {
         EMIT3("lea r%d, [2*r%d+1] ;k=2*k+1", tos-4, tos-4);
         X86_emit_op_RM(LEA, tmpReg(tos-4), no_reg, tmpReg(tos-4), 2, 1);	//LEA, not LEA_V!!
     }
     else {
         EMIT2("mov eax, r%d ;k", tos-4);
         X86_emit_op_RM(MOV_D, eax, TMP(tos-4));
         EMIT2("lea r%d, [2*eax+1] ;k = 2*k + 1", tos-4);
         X86_emit_op_MM(LEA, TMP(tos-4), no_reg, eax, 2, 1);	//LEA, not LEA_V!!
     }

   EMIT2L("cgl%d:    ;top", loop);
   X86_deflab(loop);

     EMIT("mov ecx, 3 ;nargs");
     X86_emit_op_R_imm32(MOVI_D, ecx, 3);
     EMIT2("call r%d ;pfGet()", tos-1);
     X86_emit_op_T(CALLI_N, tos-1);		//X64??	NFG	//RESUME X64 CALLI_N fixups here...
     //pop args later...
     if (isReg(tos-3)) {
         reg = tmpReg(tos-3);
     }
     else {
         reg = edx;
         EMIT3("mov %s, r%d ;offset", X86_reg32_name(reg), tos-3);
         X86_emit_op_RT(MOV_D, reg, tos-3);
     }
     assert(sizeof(gFmcp->cell[0]) == 4);
#if 1   //wrap offset for safety
     EMIT3("and %s, %d ;wrap offset", X86_reg32_name(reg), (N_CELLS-1));
     X86_emit_op_R_simm32(ANDI_D, reg, (N_CELLS-1));
#endif
#if 1
     EMIT2("imul [rBX+4*%s].cell ;get*cell[offset]", X86_reg32_name(reg));
     X86_emit_op_M(IMUL_D, FMCX(cell,reg,4));
#else
     //or could use imul32, which doesn't clobber edx but is 1 byte longer???
     EMIT2("imul32 eax, [rBX+4*%s].cell ;get*cell[offset]", X86_reg32_name(reg));
     X86_emit_op_RM(IMUL32, eax, FMCX(cell,reg,4));
#endif
     EMIT2("add r%d, eax ;sum += get*cell", tos);
     X86_emit_op_TR(ADD_D, tos, eax);
     EMIT2("inc r%d ;offset++", tos-3);
     X86_emit_op_T(INC_D, tos-3);
     if (x_or_y == 0) {
         //cnvX
         EMIT("inc [esp] ;x++");
         X86_emit_op_M(INC_D, esp, no_reg, 0, 0);
     }
     else {
         //cnvY
         EMIT("inc [esp+4] ;y++");
         X86_emit_op_M(INC_D, esp, no_reg, 0, 4);
     }
     EMIT2("dec r%d ;k--", tos-4);
     X86_emit_op_T(DEC_D, tos-4);
     EMIT2("jnz cgl%d   ;loop", loop);
     X86_emit_op_lab(JNZ, loop);

     //end of loop; clean up stack, divide sum by d...
     EMIT("add esp, 12 ;pop 3 args");
     X86_emit_op_R_simm32(ADDI_D, esp, 12);
     EMIT2("mov eax, r%d ;sum", tos);
     X86_emit_op_RT(MOV_D, eax, tos);
     EMIT("cdq");
     X86_emit_op(CDQ);
     EMIT2("idiv eax, r%d ;sum /= d", tos-2);
     X86_emit_op_T(IDIV_D, tos-2);
     EMIT2("mov r%d, eax ;final result", tos-4);
     X86_emit_op_TR(MOV_D, tos-4, eax);

     //pop all but last evaluation stack entry, performing
     //final type checks along the way...
     pop();
     // tos = pfGet
     if (TOSOBJECT.kind != VAROBJECTKIND || TOSVAR.typeindex != 1234)
     {
         //must be an integer function.
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }
     pop();
     // tos = d
     if (TOSOBJECT.kind != LITOBJECTKIND || TOSLIT.type != LITTYPEINTEGER)
     {
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }
     pop();
     // tos = offset
     if (TOSOBJECT.kind != LITOBJECTKIND || TOSLIT.type != LITTYPEINTEGER)
     {
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }
     pop();
     // tos = k
     if (TOSOBJECT.kind != LITOBJECTKIND || TOSLIT.type != LITTYPEINTEGER)
     {
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_cnvX_Y*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            op_gamma              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_gamma(void)
 {
     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         int N = sizeof(gFmcp->pre_gamma)/sizeof(gFmcp->pre_gamma[0]);
         int reg;
         if (isReg(tos)) {
             reg = tmpReg(tos);
         }
         else {
             reg = eax;
             EMIT3("mov %s, r%d ;index reg", X86_reg32_name(reg), tos);
             X86_emit_op_RT(MOV_D, reg, tos);
         }
         assert(N == 256);
         assert(sizeof(gFmcp->pre_gamma[0]) == 4);
         EMIT3("and %s, %d ;safe index", X86_reg32_name(reg), N-1);
         EMIT3("mov r%d, [rBX+4*%s].pre_gamma", tos, X86_reg32_name(reg));
         X86_emit_op_R_simm32(ANDI_D, reg, N-1);
         X86_emit_op_TM(MOV_D, tos, FMCX(pre_gamma,reg,4));
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_gamma*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_val              */
        /*                                  */
        /*                                  */
        /************************************/

#if 0 //temporarily no in-line for "val" -- reinstate later??
 /**procedure**/
 static void op_val(void)
 {
#ifndef FML
     UINT1          result;
     UINT2          value1[4];
     UINT2          value2[4];
     int            i;
#endif

     ensurethreeoperands();

     if (NOS2OBJECT.kind == LITOBJECTKIND && NOS2LIT.type == LITTYPEINTEGER && 
         NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         int reg;
         // val(i,a,b) = ctl(i)*(b-a)/255 + a
         //NB:  should really be:
         //     (ctl(i) - ctl[i].lo)*(b-a)/(ctl[i].hi - ctl[i].lo) + a
         int i = tos-2;
         int a = tos-1;
         int b = tos;
         EMIT2("mov eax, r%d ;b", b);
         EMIT2("sub eax, r%d ;(b-a)", a);
         // inc eax
         X86_emit_op_RT(MOV_D, eax, b);
         X86_emit_op_RT(SUB_D, eax, a);
         if (isReg(i)) {
             reg = tmpReg(i);
         }
         else {
             reg = edx;
             EMIT3("mov %s, r%d ;i", X86_reg32_name(reg), i);
             X86_emit_op_RT(MOV_D, reg, i);
         }
         // N_CTLS must be a power of 2!!!!
         assert(IS_POWER_OF_TWO(N_CTLS));
         assert(sizeof(gFmcp->pre_ctl[0]) == 4);
         EMIT3("and %s, %d ;safe index i", X86_reg32_name(reg), N_CTLS-1);
         EMIT2("imul [rBX+4*%s].pre_ctl ;ctl(i)*(b-a)", X86_reg32_name(reg));
         EMIT("mov ecx, 255");
         EMIT("idiv eax, ecx ;ctl(i)*(b-a)/255");
         EMIT2("add eax, r%d ;ctl(i)*(b-a)/255+a", a);
         EMIT2("mov r%d, eax ;=val(i,a,b)", tos-2);
         X86_emit_op_R_simm32(ANDI_D, reg, N_CTLS-1);
         X86_emit_op_M(IMUL_D, FMCX(pre_ctl,reg,4));
         X86_emit_op_R_imm32(MOVI_D, ecx, 255);
         X86_emit_op_R(IDIV_D, ecx);
         X86_emit_op_RT(ADD_D, eax, a);
         X86_emit_op_TR(MOV_D, tos-2, eax);
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     // pop two args, leaving result on stack
     pop();
     pop();

	 TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_val*/
#endif

        /************************************/
        /*                                  */
        /*                                  */
        /*              op_sin              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_sin(void)
 {

     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         int reg;
         if (isReg(tos)) {
             reg = tmpReg(tos);
         }
         else {
             reg = eax;
             EMIT3("mov %s, r%d ;index reg", X86_reg32_name(reg), tos);
             X86_emit_op_RT(MOV_D, reg, tos);	//X64 okay, only need low bits of index
         }
         assert(sizeof(gFmcp->pre_sin)/sizeof(gFmcp->pre_sin[0]) == 1024);	//needed??
         assert(sizeof(gFmcp->pre_sin[0]) == 4);	//assumes 32-bit entries
         EMIT3("and %s, %d ;safe index",
             X86_reg32_name(reg),
             sizeof(gFmcp->pre_sin)/sizeof(gFmcp->pre_sin[0])-1);
         EMIT3("movsxd r%d, [rBX+4*%s].pre_sin", tos, GPR_V(reg));
         X86_emit_op_R_simm32(ANDI_D, reg, sizeof(gFmcp->pre_sin)/sizeof(gFmcp->pre_sin[0])-1);
         X86_emit_op_TM(MOVSXD_V, tos, FMCX(pre_sin,reg,4));	//sign-extend for X64
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         EMIT("fsin");
         X86_emit_op(FSIN);
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_sin*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_cos              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_cos(void)
 {

     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         // cos(i) = sin(i+256)
         int reg;
         assert(sizeof(gFmcp->pre_sin)/sizeof(gFmcp->pre_sin[0]) == 1024);	//needed??
         assert(sizeof(gFmcp->pre_sin[0]) == 4);	//assumes 32-bit entries
         if (isReg(tos)) {
             reg = tmpReg(tos);
         }
         else {
             reg = eax;
             EMIT3("mov %s, r%d ;index reg", X86_reg32_name(reg), tos);
             X86_emit_op_RT(MOV_D, reg, tos);	//X64 okay, only need low bits of index
         }
         EMIT2("add %s, 256 ;cos(i)=sin(i+256)", X86_reg32_name(reg));
         EMIT3("and %s, %d ;safe index",
             X86_reg32_name(reg),
             sizeof(gFmcp->pre_sin)/sizeof(gFmcp->pre_sin[0])-1);
         EMIT3("movsxd r%d, [rBX+4*%s].pre_sin", tos, GPR_V(reg));
         X86_emit_op_R_simm32(ADDI_D, reg, 256);
         X86_emit_op_R_simm32(ANDI_D, reg, sizeof(gFmcp->pre_sin)/sizeof(gFmcp->pre_sin[0])-1);
         X86_emit_op_TM(MOVSXD_V, tos, FMCX(pre_sin,reg,4));	//sign-extend for X64
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
         EMIT("fcos");
         X86_emit_op(FCOS);
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_cos*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_tan              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_tan(void)
 {

     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         int reg;
         if (isReg(tos)) {
             reg = tmpReg(tos);
         }
         else {
             reg = eax;
             EMIT3("mov %s, r%d ;index reg", X86_reg32_name(reg), tos);
             X86_emit_op_RT(MOV_D, reg, tos);	//X64 okay, only need low bits of index
         }
         assert(sizeof(gFmcp->pre_tan)/sizeof(gFmcp->pre_tan[0]) == 512);
         assert(sizeof(gFmcp->pre_tan[0]) == 4);	//assumes 32-bit entries
         EMIT3("and %s, %d ;safe index",
             X86_reg32_name(reg),
             sizeof(gFmcp->pre_tan)/sizeof(gFmcp->pre_tan[0])-1);
         EMIT3("movsxd r%d, [rBX+4*%s].pre_tan", tos, GPR_V(reg));
         X86_emit_op_R_simm32(ANDI_D, reg, sizeof(gFmcp->pre_tan)/sizeof(gFmcp->pre_tan[0])-1);
         X86_emit_op_TM(MOVSXD_V, tos, FMCX(pre_tan,reg,4));	//sign-extend for X64
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL)
     {
		 check_ndp_stack_room(1);			//fptan pushed an extra +1.0 onto the ndp stack
         EMIT("fptan");
         X86_emit_op(FPTAN);
         EMIT("fpop   ;pop extra 1.0");
         X86_emit_op(FPOP);
		 check_ndp_stack_room(-1);			//popped 1 ndp stack slot
	 }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_tan*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_get              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_get(void)
 {

     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         int reg;
         if (isReg(tos)) {
             reg = tmpReg(tos);
         }
         else {
             reg = eax;
             EMIT3("mov %s, r%d ;index reg", X86_reg32_name(reg), tos);
             X86_emit_op_RT(MOV_D, reg, tos);	//X64 okay, only need low bits of index
         }
         assert(sizeof(gFmcp->cell[0]) == 4);	//assuming cells are 32 bits
         EMIT3("and %s, %d ;safe index", X86_reg32_name(reg), (N_CELLS-1));
         EMIT3("movsxd r%d, [rBX+4*%s].cell", tos, GPR_V(reg));
         X86_emit_op_R_simm32(ANDI_D, reg, (N_CELLS-1));
         X86_emit_op_TM(MOVSXD_V, tos, FMCX(cell,reg,4));	//sign-extends for X64
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_get*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              op_put              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void op_put(void)
 {

     ensuretwooperands();

     if (NOSOBJECT.kind == LITOBJECTKIND && NOSLIT.type == LITTYPEINTEGER && 
         TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         int reg;
         if (isReg(tos)) {
             reg = tmpReg(tos);
         }
         else {
             reg = edx;
             EMIT3("mov %s, r%d ;index reg", X86_reg32_name(reg), tos);
             X86_emit_op_RT(MOV_D, reg, tos);	//X64 okay, only need low bits of index
         }
         assert(sizeof(gFmcp->cell[0]) == 4);	//assuming cells are 32 bits
         EMIT3("and %s, %d ;safe index", X86_reg32_name(reg), (N_CELLS-1));
         EMIT3("movd [rBX+4*%s].cell, r%d", GPR_V(reg), tos-1);
         X86_emit_op_R_simm32(ANDI_D, reg, (N_CELLS-1));
         //CAUTION!!! MT(MOV_D) might use eax for temp!!!
         X86_emit_op_MT(MOV_D, FMCX(cell,reg,4), tos-1);	//X64 int32 store okay
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);

     pop();     //value is now in TOS

	 TOSLIT.nbytes = 4;
     TOSLIT.ivalue = 0; //for now
     TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
 } /*op_put*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            dereference           */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void dereference(void)
 {
     UINT4          typeindex;
     INT4			scalarlength;
	 UINT1          kind;
	 UINT4          offset;

     /* main body of 'dereference' */

     ensureoneoperand();

	 //// What if tos is a pointer value???  Can it even happen??

     kind = TOSOBJECT.kind;
     if (kind == VAROBJECTKIND)
     {   /* dereferencing a variable */
         offset = TOSVAR.offset;	//unused?
         typeindex = TOSVAR.typeindex;
		 scalarlength = TOSVAR.nbytes;
     }
     else 
         Mde_abort_command(MD386__BADREF, 0, 0, 0);

     if (typeindex >= 1234 && typeindex <= 1238) {
         // dereferencing the address of a function is a no-op
         return;
     }
     else if (typeindex == 4321) {
         // dereferencing an array of char yields a pointer to int for now...
         typeindex = 0;		// EH?? Was this meant to set TOSVAR.typeindex ???
         return;
     }
     else if (typeindex == 0) {
         // int
         TOSOBJECT.kind = LITOBJECTKIND;
         TOSLIT.type = LITTYPEINTEGER;
         TOSLIT.nbytes = scalarlength;
         TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue;
		 //need to switch on signed vs. unsigned types...
		 switch (scalarlength) {
		 //////// signed types /////////
		 case 1:
			 //sign-extend int8 to int32 or int64
			 EMIT3("movsxb r%d, [r%d] ;op_load int8", tos, tos);
			 //special logic because there is no MR form of MOVSXB_V
			 if (isReg(tos)) {
				 X86_emit_op_TTi(MOVSXB_V, tos, tos);
			 }
			 else {	//could later move this logic into X86_emit_op_TTi??
				 //tos is not a reg
				 X86_emit_op_RM(MOV_V, eax, TMP(tos));
				 X86_emit_op_RM(MOVSXB_V, eax, eax, no_reg, 0, 0);
				 X86_emit_op_MR(MOV_V, TMP(tos), eax);
			 }
			 break;
		 case 2:
			 //sign-extend int16 to int32 or int64
			 EMIT3("movsxw r%d, [r%d] ;op_load int16", tos, tos);
			 //special logic because there is no MR form of MOVSXW_V
			 if (isReg(tos)) {
				 X86_emit_op_TTi(MOVSXW_V, tos, tos);
			 }
			 else {	//could later move this logic into X86_emit_op_TTi??
				//tos is not a reg
				X86_emit_op_RM(MOV_V, eax, TMP(tos));
				X86_emit_op_RM(MOVSXW_V, eax, eax, no_reg, 0, 0);
				X86_emit_op_MR(MOV_V, TMP(tos), eax);
			 }
			 break;
		 case 4:
			 //sign-extend int32 to int32 or int64
			 EMIT3("movsxd r%d, [r%d] ;op_load int32", tos, tos);
			 //special logic because there is no MR form of MOVSXD_Q
			 if (isReg(tos)) {
				 X86_emit_op_TTi(MOVSXD_V, tos, tos);
			 }
			 else {	//could later move this logic into X86_emit_op_TTi??
				//tos is not a reg
				X86_emit_op_RM(MOV_V, eax, TMP(tos));
				X86_emit_op_RM(MOVSXD_V, eax, eax, no_reg, 0, 0);
				X86_emit_op_MR(MOV_V, TMP(tos), eax);
			 }
			 break;
		 case 8:
			 if (X64gen) {
				 //X64: load directly into int64 tos
				 EMIT3("mov_q r%d, [r%d] ;op_load int64", tos, tos);
				 X86_emit_op_TTi(MOV_Q, tos, tos);
			 }
			 else { //x32: 64-bit math not yet implemented
				 Mde_abort_command(MD386__NYI, 19, 0, 0);
			 }
			 break;
		 ///////// unsigned types /////////
		 case -1:
			 //zero-extend byte to dword (and hence to qword for X64)
			 EMIT3("movzxb r%d, [r%d] ;op_load byte", tos, tos);
			 //special logic because there is no MR form of MOVZXB_D
			 if (isReg(tos)) {
				 X86_emit_op_TTi(MOVZXB_D, tos, tos);
			 }
			 else {	//could later move this logic into X86_emit_op_TTi??
				//tos is not a reg
				X86_emit_op_RM(MOV_V, eax, TMP(tos));
				X86_emit_op_RM(MOVZXB_D, eax, eax, no_reg, 0, 0);
				X86_emit_op_MR(MOV_V, TMP(tos), eax);
			 }
			 break;
		 case -2:
			 //zero-extend word to dword (and hence to qword for X64)
			 EMIT3("movzxw r%d, [r%d] ;op_load word", tos, tos);
			 //special logic because there is no MR form of MOVZXW_D
			 if (isReg(tos)) {
				 X86_emit_op_TTi(MOVZXW_D, tos, tos);
			 }
			 else {	//could later move this logic into X86_emit_op_TTi??
				//tos is not a reg
				X86_emit_op_RM(MOV_V, eax, TMP(tos));
				X86_emit_op_RM(MOVZXW_D, eax, eax, no_reg, 0, 0);
				X86_emit_op_MR(MOV_V, TMP(tos), eax);
			 }
			 break;
		 case -4:
			 //load dword, with implicit zero-extension to qword for X64
			 EMIT3("movzxd r%d, [r%d] ;op_load dword", tos, tos);
			 //special logic because there is no MR form of MOV_D (WRONG???)
			 if (isReg(tos)) {
				 X86_emit_op_TTi(MOV_D, tos, tos);
			 }
			 else {	//could later move this logic into X86_emit_op_TTi??
				//tos is not a reg
				X86_emit_op_RM(MOV_V, eax, TMP(tos));
				X86_emit_op_RM(MOV_D, eax, eax, no_reg, 0, 0);
				X86_emit_op_MR(MOV_V, TMP(tos), eax);
			 }
			 break;
		 case -8:
			 if (X64gen) {
				 //X64: load qword
				 EMIT3("mov_q r%d, [r%d] ;op_load qword", tos, tos);
				 X86_emit_op_TTi(MOV_Q, tos, tos);
			 }
			 else { //x32: 64-bit math not yet implemented
				 Mde_abort_command(MD386__NYI, 20, 0, 0);
			 }
			 break;
		 default:
			 Mde_abort_command(MD386__NYI, 21, 0, 0);
			 break;
		 }//switch
     }
     else if (typeindex == 126) {
         //double
         TOSOBJECT.kind = LITOBJECTKIND;
         TOSLIT.type = LITTYPEREAL;
         TOSLIT.nbytes = scalarlength;
		 switch (scalarlength) {
		 case 2:
			 //load 16-bit (half) floating point value
			 //(needs to call halfToFloat() )
             Mde_abort_command(MD386__NYI, 22, 0, 0);
			 break;
		 case 4:
			 //load 32-bit (float) floating point value
			 check_ndp_stack_room(1);				//need 1 ndp stack entry
			 EMIT2("fld_f [r%d] ;op_load float", tos);
			 X86_emit_op_Ti(FLD_F, tos);
			 break;
		 case 8:
			 //load 64-bit (double) floating point value
			 check_ndp_stack_room(1);				//need 1 ndp stack entry
			 EMIT2("fld_d [r%d] ;op_load double", tos);
			 X86_emit_op_Ti(FLD_D, tos);
			 break;
		 case 10:
			 //load 80-bit (tempreal) floating point value
			 check_ndp_stack_room(1);				//need 1 ndp stack entry
			 EMIT2("fld_t [r%d] ;op_load tempreal", tos);
			 X86_emit_op_Ti(FLD_T, tos);
			 break;
		 default:
             Mde_abort_command(MD386__NYI, 23, 0, 0);
			 break;
		 }//switch
         //pop();
         //ndp_st++;
     }
     else {
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }
 } /*dereference*/


        /************************************/
        /*                                  */
        /*                                  */
        /*              store               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void store(void)
 {
     UINT4          typeindex;
     INT4			scalarlength;
	 UINT1          kind;
	 UINT4          offset;

     /* main body of 'store' */

     ensuretwooperands();

	 //// What if nos is a pointer value???  Can it even happen??

	 kind = NOSOBJECT.kind;
     if (kind == VAROBJECTKIND)
     {   /* dereferencing a variable */
         offset = NOSVAR.offset;	//unused??
         typeindex = NOSVAR.typeindex;
		 scalarlength = NOSVAR.nbytes;
		 //check the variable's 'const' attribute...
		 if (NOSVAR.flags & fmfCONST) {
			 //can't store into a const var!!
			 //for now, we just issue a warning and
			 //proceed anyway
             warn("Attempt to modify a 'const' variable: %s",
				  (UINT_PTR)NOSVAR.namep, 0, 0);
		 }
#if 0	//could use another flag to mark 'predef' vars??
		 else if (NOSVAR.pType && (NOSVAR.pType->flags & fmfCONST)) {
			 //can't store into a const type!!
			 //for now, we just issue a warning and
			 //proceed anyway
             warn("Attempt to modify a 'const' (predefined) variable",
				  0, 0, 0);
		 }
#endif
     }
#ifndef FML
     else if (kind == REGOBJECTKIND)
     {   /* dereferencing a register */
         regindex = NOSREG.regindex;
         typeindex = NOSREG.typeindex;
     }
#endif
     else 
         Mde_abort_command(MD386__BADREF, 0, 0, 0);

     if (TOSOBJECT.kind != LITOBJECTKIND) {
         //for now...
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }
     if (typeindex == 0 && TOSLIT.type == LITTYPEINTEGER)
     {   /* int <- int */
		 // Ignore signed/unsigned by taking absolute value of
		 // scalar length
		 switch (abs(scalarlength)) {
		 case 1:
			 //store int8/byte
			 if (X64gen) { //x64 can use SIL and DIL temp regs
				 EMIT3("mov_b [r%d], r%d ;op_store int8", tos-1, tos);
				 X86_emit_op_TiT(XL(MOV_B), tos-1, tos);	//XL to use SIL, DIL
			 }
			 else { //x32 can't use SIL and DIL!!!
				 EMIT2("mov_d EAX, r%d", tos);
				 X86_emit_op_RT(MOV_D, eax, tos);
				 EMIT2("mov_b [r%d], AL ;op_store int8", tos-1);
				 X86_emit_op_TiR(MOV_B, tos-1, al);
			 }
			 break;
		 case 2:
			 //store int16/word
			 EMIT3("mov_w [r%d], r%d ;op_store int16", tos-1, tos);
			 X86_emit_op_TiT(MOV_W, tos-1, tos);
			 break;
		 case 4:
			 //store int32/dword
			 EMIT3("mov_d [r%d], r%d ;op_store int32", tos-1, tos);
			 X86_emit_op_TiT(MOV_D, tos-1, tos);
			 break;
		 case 8:
			 //store int64/qword
			 if (X64gen) {
				 EMIT3("mov_q [r%d], r%d ;op_store int64", tos-1, tos);
				 X86_emit_op_TiT(MOV_Q, tos-1, tos);
			 }
			 else { //X32: 64-bit integers not yet implemented in X32
				 Mde_abort_command(MD386__NYI, 24, scalarlength, tos); // for now
			 }
			 break;
		 default:
			 //indicate -9999 if scalarlength is 0, else it doesn't show up!
			 Mde_abort_command(MD386__NYI, 25, scalarlength?scalarlength:-9999, tos);
			 break;
		 }//switch
		 /* Successfully stored the value!      */
         /* Pop the rhs and reload the lhs to get */
         /* the actual value of the store op.     */
         pop();
         dereference();
     }
     else if (typeindex == 0 && TOSLIT.type == LITTYPEREAL) {
         /* int <- real */           //// SHOULDN'T THIS BE int <- ichop(real) AKA _ftol(real)????
#if 0
         EMIT2("fstp_i [r%d] ;op_store", tos-1);
		 check_ndp_stack_room(-1);		//popped 1 entry from ndp stack
         X86_emit_op_Ti(FSTP_I, tos-1);
         TOSLIT.type = LITTYPEINTEGER;  // so pop() won't pop ndp stack
         pop();
         dereference();                 // reload the integer value
#else
		 // Ignore signed/unsigned by taking absolute value of
		 // scalar length
		 switch (abs(scalarlength)) {
		 case 1:
			 //store int8/byte
			 EMIT("fstp_l temp ;op_store int64");
			 check_ndp_stack_room(-1);		//popped 1 entry from ndp stack
			 X86_emit_op_M(FSTP_L, FMC(temp));
			 EMIT("mov_v rCX, temp");		//Uses rCX temporarily!
			 X86_emit_op_RM(MOV_V, rcx, FMC(temp));
			 EMIT2("mov_b [r%d], CL ;op_store byte", tos-1);
			 X86_emit_op_TiR(MOV_B, tos-1, cl);
			 break;
		 case 2:
			 //store int16/word
			 EMIT2("fstp_s [r%d] ;op_store int16", tos-1);
			 check_ndp_stack_room(-1);		//popped 1 entry from ndp stack
			 X86_emit_op_Ti(FSTP_S, tos-1);
			 break;
		 case 4:
			 //store int32/dword
			 EMIT2("fstp_i [r%d] ;op_store int32", tos-1);
			 check_ndp_stack_room(-1);		//popped 1 entry from ndp stack
			 X86_emit_op_Ti(FSTP_I, tos-1);
			 break;
		 case 8:
			 //store int64/qword
			 EMIT2("fstp_l [r%d] ;op_store int64", tos-1);
			 check_ndp_stack_room(-1);		//popped 1 entry from ndp stack
			 X86_emit_op_Ti(FSTP_L, tos-1);
			 break;
		 default:
			 Mde_abort_command(MD386__NYI, 26, scalarlength, tos);
			 break;
		 }//switch
		 /* Successfully stored the value!      */
         /* Pop the rhs and reload the lhs to get */
         /* the actual value of the store op.     */
         TOSLIT.type = LITTYPEINTEGER;  // so pop() won't pop ndp stack
         pop();
         dereference();                 // reload the integer value
#endif
     }
     else if (typeindex == 126) {
         if (TOSLIT.type == LITTYPEINTEGER) {
             // real <- int
             // convert the int to real, then treat as real <- real
#if 0
			 // Would using a temp in FMC be faster than push/add? (FMC is thread-local)
			 EMIT2("push r%d", tos);
			 check_ndp_stack_room(1);		//need room for 1 ndp stack entry
             EMIT("fld_i [esp] ;op_load");	//X64: op_load use FLD_I or FLD_L???
			 EMIT("add esp, 4");			//X64: add rsp,8
             X86_emit_op_T(PUSH_D, tos);
             X86_emit_op_M(FLD_I, esp, no_reg, 0, 0);
             X86_emit_op_R_simm32(ADDI_D, esp, 4);
#else
			 if (X64gen) {
				 check_ndp_stack_room(1);		//need room for 1 ndp stack entry
				 EMIT2("fld_l r%d ;op_load int64", tos);	//X64: op_load FLD_L
				 X86_emit_op_Tm(FLD_L, tos);
			 }
			 else { //X32
				 check_ndp_stack_room(1);		//need room for 1 ndp stack entry
				 EMIT2("fld_i r%d ;op_load int32", tos);	//X32: op_load FLD_I
				 X86_emit_op_Tm(FLD_I, tos);
			 }
#endif
             TOSLIT.type = LITTYPEREAL;
         } // then fall through to real <- real!!
         else if (TOSLIT.type != LITTYPEREAL) {
             Mde_abort_command(MD386__BADMIX, 0, 0, 0);
         }
         // real <- real
#if 0
         //should really do this in case there is a precision change,
         //but the resulting code is far too ugly...
         EMIT2("fstp_d [r%d] ;op_store", tos-1);
		 check_ndp_stack_room(-1);		//popped 1 arg from ndp stack
         X86_emit_op_Ti(FSTP_D, tos-1);
         /* Successfully stored the value!      */
         /* Pop the rhs and reload the lhs to get */
         /* the actual value of the store op.     */
         TOSLIT.type = LITTYPEINTEGER;  // so pop() won't pop ndp stack
         pop();
         dereference();
#else
		 switch (scalarlength) {
		 case 2:
			 //store 16-bit (half) floating point value
			 //(needs to call floatToHalf() )
             Mde_abort_command(MD386__NYI, 27, 0, 0);
			 break;
		 case 4:
			 //store 32-bit (float) floating point value
			 EMIT2("fst_f [r%d] ;op_store float", tos-1); //later do FPOP...
			 X86_emit_op_Ti(FST_F, tos-1);
			 break;
		 case 8:
			 //store 64-bit (double) floating point value
			 EMIT2("fst_d [r%d] ;op_store double", tos-1); //later do FPOP...
			 X86_emit_op_Ti(FST_D, tos-1);
			 break;
		 case 10:
			 //Store 80-bit (tempreal) floating point value.
#if 1	//// use this if low on ndp slots!!!
			 //C semantics prefers this implementation: (where
			 //value of an assignment is the value of the lhs,
			 //not the value of the rhs). This only makes a
			 //difference if the type of the lhs differs
			 //from the type of the rhs.
			 //There is no FST_T opcode, so synthesize one
			 //as FSTP_T followed by FLD_T
			 EMIT2("fstp_t [r%d] ;op_store tempreal", tos-1);
			 check_ndp_stack_room(-1);			//popped 1 entry from ndp stack
			 X86_emit_op_Ti(FSTP_T, tos-1);
			 check_ndp_stack_room(1);			//reloading 1 ndp stack entry
			 EMIT2("fld_t [r%d] ;op_load", tos-1);	//reload lhs in case value still needed
			 X86_emit_op_Ti(FLD_T, tos-1);		//later do FPOP...
#else	//// use this if not low on ndp slots!!!
			 //There is no FST_T opcode, so synthesize one as an
			 //FDUP followed by an FSTP_T (requires 1 less memref
			 //than above, but 1 more ndp slot).
			 //N.B. This does not strictly follow C semantics if the
			 //type of the lhs differs from the type of the rhs.
			 check_ndp_stack_room(1);			//need one more ndp stack entry
			 EMIT("fdup");						// st = rhs rhs
			 X86_emit_op(FDUP);					// FDUP = FLD ST(0)
			 EMIT2("fstp_t [r%d] ;op_store tempreal", tos-1);	// st = rhs
			 check_ndp_stack_room(-1);			//popped 1 entry from ndp stack
			 X86_emit_op_Ti(FSTP_T, tos-1);
#endif
			 break;
		 default:
             Mde_abort_command(MD386__NYI, 28, 0, 0);
			 break;
		 }//switch
         *nosp = *tosp;
         TOSLIT.type = LITTYPEINTEGER;  // so pop() won't pop ndp stack
         pop();
         //ndp_st--;
#endif
     }
     else
         Mde_abort_command(MD386__BADREF, 0, 0, 0);

     /*** But C should allow structure/union assignments???  ***/
     
 } /*store*/


#if 0
        /************************************/
        /*                                  */
        /*                                  */
        /*             qualify              */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void qualify(void)
 {
     void           *qualifierp;
     #define QUALIFIER (*(QUALIFIERSTRUCTURE *) qualifierp)

     ensuretwooperands();

     if (NOSOBJECT.kind == IDOBJECTKIND && (TOSOBJECT.kind == IDOBJECTKIND || 
         TOSOBJECT.kind == LINOBJECTKIND || TOSOBJECT.kind == SEGOBJECTKIND || 
         TOSOBJECT.kind == GRPOBJECTKIND))
     {
         if (TOSID.qualp == NULL)
             TOSID.qualp = installqualifier(NOSID.lnamep);
         else 
         {
             qualifierp = TOSID.qualp;
             while (QUALIFIER.link != NULL)
             {
                 qualifierp = QUALIFIER.link;
             }
             QUALIFIER.link = installqualifier(NOSID.lnamep);
         }
         NOSID = TOSID;
         pop();
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     
     #undef QUALIFIER
 } /*qualify*/
#endif


        /************************************/
        /*                                  */
        /*                                  */
        /*            match$ID              */
        /*                                  */
        /*                                  */
        /************************************/

 // Could use strncmp () here ??? 
 // Make this a macro???

 /**procedure**/
 static int matchID(STRINGDESCRIPTOR *lnamep, char* sz)
{
     // lnamep points to the 1st dword of a string descriptor; i.e.,
     // to the count field, which is followed by the address
     // of the id chars themselves.
     // lnamep->ptr is address of an ASCIZ string -- Always true???  Does parser guarantee this?? YES
     return strcmp(lnamep->ptr, sz) == 0;
 } /*matchID*/

        /************************************/
        /*                                  */
        /*                                  */
        /*            lookup$ID             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void lookupid(void)
 {								//should check tos is ID???
	 STRINGDESCRIPTOR stringdesc;
	 STRINGDESCRIPTOR *currentlnamep;
     char			  *idp;
     int              nChars;
     pfm_OBJ          pSym, pType;

     //make a local copy of the string descriptor before it
     //gets overwritten... (obsolescent code!!! still needed???)
     stringdesc.len = nChars = TOSID.nbytes;	// number of chars in name
     stringdesc.ptr = idp = TOSID.charp;		// pointer to 1st char of name
     currentlnamep = &stringdesc;

     //now switch object to a VAR...
     TOSOBJECT.kind = VAROBJECTKIND;
     TOSVAR.typeindex = Mdg_current_default_type_index; // always int for now.
     TOSVAR.frame = ebx;    //assume static
	 TOSVAR.offset = PREDEFSYMOFF(undefined);		// in case undefined
     TOSVAR.extradata = 0;  // holds nargs for funcs
	 TOSVAR.nbytes = 0;	//unknown at this point
	 TOSVAR.pSym = pSym = NULL;
	 TOSVAR.pType = pType = NULL;
	 TOSVAR.flags = 0;
	 TOSVAR.namep = NULL;

#if 0
     PRINTF4("lookupid: nChars=%d, idp->%.4s\r\n", nChars, idp);
#endif

     // First, search for the symbol in the user-defined symbol table...
     if ((pSym = lookup_symbol(idp))) {
         // It's a user-defined symbol.
         if (pSym->what != fmVAR) {
             // Not a VAR.  Give up for now.
             Mde_abort_command(MD386__BADMIX, 0, 0, 0);
         }
         pType = pSym->pType;
         if (pSym->am == amSTATIC) {
            TOSVAR.frame = ebx;
         }
         else {
            TOSVAR.frame = ebp;  //assume amSTACK for now...
         }
         TOSVAR.offset = pSym->offset;
		 TOSVAR.nbytes = pSym->size;
		 TOSVAR.pSym = pSym;
		 TOSVAR.pType = pType;
		 TOSVAR.flags = pSym->flags;
		 TOSVAR.namep = pSym->name;
#if 0		//doesn't work for now!!
		 if (pType->flags & fmfUNSIGNED) {
			 //set nbytes negative to flag an unsigned type
			 TOSVAR.nbytes = -TOSVAR.nbytes;
		 }
#endif
         TOSVAR.typeindex = Mdg_current_default_type_index; // always int for now. //already set above???
         if (pType->what == fmSCALAR) {
            if (pType->basetype == btINTEGRAL) {
                TOSVAR.typeindex = Mdg_current_default_type_index;	//already set???
				if (pSym->flags & fmfNONNEG) {
					//set nbytes negative to flag an unsigned type
					TOSVAR.nbytes = -TOSVAR.nbytes;
				}
            }//btINTEGRAL
            else if (pType->basetype == btREAL) {
                TOSVAR.typeindex = 126;
            }
            else bug(8866); //for now
         }//fmSCALAR
         else if (pType->what == fmARRAY) {
            TOSVAR.typeindex = 4321; //for now
         }
         else if (pType->what == fmPOINTER) {
            TOSVAR.typeindex = 0; //for now
			TOSVAR.nbytes = -TOSVAR.nbytes;	//pointers are UNSIGNED
         }
         else bug(8877); //for now
         EMIT5("lea r%d, [%s%+d] ;&%s",
              tos,
			  X64gen?X86_reg64_name(TOSVAR.frame):X86_reg32_name(TOSVAR.frame),
			  TOSVAR.offset, pSym->name);
         X86_emit_op_TM(LEA_V, tos, TOSVAR.frame, no_reg, 0, TOSVAR.offset);
         return;
     }

     // Not a user-defined symbol.  Lookup in predefined symbol table...
     if (nChars > 0)
     switch (*idp) {
         // A little speedup by branching on the first char of the name...
     
    case '_':
         if (matchID(currentlnamep, "_strerror")) {
             TOSVAR.offset = PREDEFSYMOFF(_strerror);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		
		 else if (matchID(currentlnamep, "__fgetArray")) {
             TOSVAR.offset = PREDEFSYMOFF(UNSAFE_fgetArray);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
		 else if (matchID(currentlnamep, "__fputArray")) {
             TOSVAR.offset = PREDEFSYMOFF(UNSAFE_fputArray);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }		 
         else if (matchID(currentlnamep, "__getArray")) {
             TOSVAR.offset = PREDEFSYMOFF(UNSAFE_getArray);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "__putArray")) {
             TOSVAR.offset = PREDEFSYMOFF(UNSAFE_putArray);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         break;
     
    case 'a':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(a);
         }
         else if (matchID(currentlnamep, "a0")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(a0);
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "a1")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(a1);
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "abort")) {
             TOSVAR.offset = PREDEFSYMOFF(abort);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "abort_mode")) {
             TOSVAR.offset = PREDEFSYMOFF(abort_mode);
         }
         else if (matchID(currentlnamep, "abs")) {
             TOSVAR.offset = PREDEFSYMOFF(abs);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "acos")) {
             TOSVAR.offset = PREDEFSYMOFF(math_acos);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "add")) {
             TOSVAR.offset = PREDEFSYMOFF(add);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#if 0
         else if (matchID(currentlnamep, "alloca")) {
             TOSVAR.offset = PREDEFSYMOFF(alloca);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif

         //Added by Harald Heim, Apr 9, 2002
         else if (matchID(currentlnamep, "allocArray")) {
             TOSVAR.offset = PREDEFSYMOFF(allocArray);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         //Added by Harald Heim, Oct 12, 2003
         else if (matchID(currentlnamep, "allocArrayPad")) {
             TOSVAR.offset = PREDEFSYMOFF(allocArrayPad);
             TOSVAR.extradata = 6;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "allocHost")) {
             TOSVAR.offset = PREDEFSYMOFF(allocHost);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

#ifdef IMGSOURCE
         //Added by Harald Heim, Oct 25, 2002
         else if (matchID(currentlnamep, "applyFX")) {
             TOSVAR.offset = PREDEFSYMOFF(applyFX);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif

         else if (matchID(currentlnamep, "amax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(amax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "amin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(amin);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "appendEllipsis")) {
             TOSVAR.offset = PREDEFSYMOFF(appendEllipsis);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function for now (should be char*)
         }
         else if (matchID(currentlnamep, "asin")) {
             TOSVAR.offset = PREDEFSYMOFF(math_asin);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "atan")) {
             TOSVAR.offset = PREDEFSYMOFF(math_atan);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "atan2")) {
             TOSVAR.offset = PREDEFSYMOFF(math_atan2);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }

         //Added by Harald Heim, Jul 5, 2002
         else if (matchID(currentlnamep, "autoMask")) {
             TOSVAR.offset = PREDEFSYMOFF(autoMask);
         }
         break;

     case 'b':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(b);
         }
         else if (matchID(currentlnamep, "b0")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(b0);
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "b1")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(b1);
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "backColor")) {
             TOSVAR.offset = PREDEFSYMOFF(backColor);
			 TOSVAR.flags |= fmfCONST;
         }
         
         else if (matchID(currentlnamep, "bandWidth")) {
             TOSVAR.offset = PREDEFSYMOFF(bandWidth);
         }
         
         else if (matchID(currentlnamep, "bgColor")) {
             TOSVAR.offset = PREDEFSYMOFF(bgColor);
			 TOSVAR.flags |= fmfCONST;
         }

        //added by Harald Heim, Feb 04, 2002
         else if (matchID(currentlnamep, "blend")) {
             TOSVAR.offset = PREDEFSYMOFF(blend);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		 else if (matchID(currentlnamep, "blendRGBA")) {
             TOSVAR.offset = PREDEFSYMOFF(blendRGBA);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Feb 13, 2002
         else if (matchID(currentlnamep, "bCircle")) {
             TOSVAR.offset = PREDEFSYMOFF(bCircle);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Feb 13, 2002
         else if (matchID(currentlnamep, "bRect")) {
             TOSVAR.offset = PREDEFSYMOFF(bRect);
             TOSVAR.extradata = 6;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Feb 13, 2002
         else if (matchID(currentlnamep, "bRect2")) {
             TOSVAR.offset = PREDEFSYMOFF(bRect2);
             TOSVAR.extradata = 6;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Feb 13, 2002
         else if (matchID(currentlnamep, "bTriangle")) {
             TOSVAR.offset = PREDEFSYMOFF(bTriangle);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "bmax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(bmax);
 			 TOSVAR.flags |= fmfCONST;
        }
         else if (matchID(currentlnamep, "bmin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(bmin);
			 TOSVAR.flags |= fmfCONST;
         }
		 else if (matchID(currentlnamep, "bufferSpace")) {
             TOSVAR.offset = PREDEFSYMOFF(bufferSpace);
         }
         break;

     case 'c':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(c);
         }
         else if (matchID(currentlnamep, "c0")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(c0);
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "c1")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(c1);
             gFmcp->needPremiere = TRUE;
         }

#define USE_OLD_DLL_NAMES 1
		 else if (matchID(currentlnamep, "callLib")) {
             TOSVAR.offset = PREDEFSYMOFF(callLib);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //int function
         }

		 else if (matchID(currentlnamep, "callLibFmc")) {
             TOSVAR.offset = PREDEFSYMOFF(callLibFmc);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //int function
         }
#ifdef USE_OLD_DLL_NAMES
         else if (matchID(currentlnamep, "callLibraryDouble")) {
             TOSVAR.offset = PREDEFSYMOFF(fcallLib);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1237; //double function
         }

         else if (matchID(currentlnamep, "callLibraryDoubleFast")) {
             TOSVAR.offset = PREDEFSYMOFF(fcallLibFmc);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1237; //double function
         }

         else if (matchID(currentlnamep, "callLibraryInt")) {
             TOSVAR.offset = PREDEFSYMOFF(callLib);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "callLibraryIntFast")) {
             TOSVAR.offset = PREDEFSYMOFF(callLibFmc);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //int function
         }
#endif //USE_OLD_DLL_NAMES
         //added by Harald Heim, Jun 4, 2002
         else if (matchID(currentlnamep, "calcSBD")) {
             TOSVAR.offset = PREDEFSYMOFF(calcSBD);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         else if (matchID(currentlnamep, "calloc")) {
             TOSVAR.offset = PREDEFSYMOFF(calloc);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "ceil")) {
             TOSVAR.offset = PREDEFSYMOFF(math_ceil);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
        

         else if (matchID(currentlnamep, "checkArrayAlloc")) {
             TOSVAR.offset = PREDEFSYMOFF(checkArrayAlloc);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "checkCtlFocus")) {
             TOSVAR.offset = PREDEFSYMOFF(checkCtlFocus);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		
		else if (matchID(currentlnamep, "checkDialogFocus")) {
             TOSVAR.offset = PREDEFSYMOFF(checkDialogFocus);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
        }

        else if (matchID(currentlnamep, "checkScriptVal")) {
             TOSVAR.offset = PREDEFSYMOFF(checkScriptVal);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }


         else if (matchID(currentlnamep, "chooseColor")) {
             TOSVAR.offset = PREDEFSYMOFF(chooseColor);
             TOSVAR.extradata = -2;  //nargs >= 2
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "chop")) {
             TOSVAR.offset = PREDEFSYMOFF(math_chop);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "clearerr")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_clearerr);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "cmax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(cmax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "cmin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(cmin);
			 TOSVAR.flags |= fmfCONST;
         }
         
         //added by Harald Heim, Feb 05, 2002
         else if (matchID(currentlnamep, "contrast")) {
             TOSVAR.offset = PREDEFSYMOFF(contrast);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Feb 10, 2002
         else if (matchID(currentlnamep, "cell_initialize")) {
             TOSVAR.offset = PREDEFSYMOFF(cell_initialize);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "cell_preserve")) {
             TOSVAR.offset = PREDEFSYMOFF(cell_initialize);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "ctl")) {
             TOSVAR.offset = PREDEFSYMOFF(ctl);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
#ifdef DEBUG_OP_CALL
             PRINTF2("++ lookupid: ctl, offset=%x\r\n", TOSVAR.offset);
#endif
         }
         else if (matchID(currentlnamep, "chk")) {
             // 'chk' is a deprecated synonym for 'ctl'
             TOSVAR.offset = PREDEFSYMOFF(ctl);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "c2d")) {
             TOSVAR.offset = PREDEFSYMOFF(c2d);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "c2m")) {
             TOSVAR.offset = PREDEFSYMOFF(c2m);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Feb 08, 2002
         else if (matchID(currentlnamep, "cmyk2rgb")) {
             TOSVAR.offset = PREDEFSYMOFF(cmyk2rgb);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }   

         else if (matchID(currentlnamep, "cnv")||
                  matchID(currentlnamep, "cnv0")||       //Premiere compatibility
                  matchID(currentlnamep, "cnv1")) {      //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(cnv);
             TOSVAR.extradata = 10;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->isTileable = FALSE; //for now...
             gParams->flags &= ~AFH_IS_TILEABLE;
             //1-pixel padding needed for 3x3 convolution kernel (use max(needPadding,1)???)
             gFmcp->needPadding = 1;
         }
         else if (matchID(currentlnamep, "cnvX")) {
             TOSVAR.offset = PREDEFSYMOFF(cnvX);
             TOSVAR.extradata = 7;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gFmcp->isTileable = FALSE; //for now...
             gParams->flags &= ~AFH_IS_TILEABLE;
             //k-pixel padding needed for radius k convolution kernel
             gFmcp->needPadding = 99999;    //for now
         }
         else if (matchID(currentlnamep, "cnvY")) {
             TOSVAR.offset = PREDEFSYMOFF(cnvY);
             TOSVAR.extradata = 7;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gFmcp->isTileable = FALSE; //for now...
             gParams->flags &= ~AFH_IS_TILEABLE;
             //k-pixel padding needed for radius k convolution kernel
             gFmcp->needPadding = 99999;    //for now
         }

#ifdef THEPLUGINSITE
         //Added by Harald Heim, Jun 05, 2002
         else if (matchID(currentlnamep, "cnvX_edge")) {
             TOSVAR.offset = PREDEFSYMOFF(cnvX_edge);
             TOSVAR.extradata = 9;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gFmcp->isTileable = FALSE; //for now...
             gParams->flags &= ~AFH_IS_TILEABLE;
             //k-pixel padding needed for radius k convolution kernel
             gFmcp->needPadding = 99999;    //for now
         }

         //Added by Harald Heim, Jun 05, 2002
         else if (matchID(currentlnamep, "cnvY_edge")) {
             TOSVAR.offset = PREDEFSYMOFF(cnvY_edge);
             TOSVAR.extradata = 11;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gFmcp->isTileable = FALSE; //for now...
             gParams->flags &= ~AFH_IS_TILEABLE;
             //k-pixel padding needed for radius k convolution kernel
             gFmcp->needPadding = 99999;    //for now
         }

         
         else if (matchID(currentlnamep, "copyImage2Array")) {
             TOSVAR.offset = PREDEFSYMOFF(copyImage2Array);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }  
#endif


         //added by Harald Heim, Oct 25, 2002
         else if (matchID(currentlnamep, "copyArray")) {
             TOSVAR.offset = PREDEFSYMOFF(copyArray);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		 else if (matchID(currentlnamep, "copyResToArray")) {
             TOSVAR.offset = PREDEFSYMOFF(copyResToArray);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "cos")) {
             TOSVAR.offset = PREDEFSYMOFF(cos);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "cosh")) {
             TOSVAR.offset = PREDEFSYMOFF(math_cosh);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "col")) {
             TOSVAR.offset = PREDEFSYMOFF(col);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "columns")) {
             TOSVAR.offset = PREDEFSYMOFF(columns);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "clearCtlProperties")) {
             TOSVAR.offset = PREDEFSYMOFF(clearCtlProperties);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "clearDialogEvent")) {
             TOSVAR.offset = PREDEFSYMOFF(clearDialogEvent);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "clearDialogStyle")) {
             TOSVAR.offset = PREDEFSYMOFF(clearDialogStyle);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "clearDialogStyleEx")) {
             TOSVAR.offset = PREDEFSYMOFF(clearDialogStyleEx);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "clearCtlStyle")) {
             TOSVAR.offset = PREDEFSYMOFF(clearCtlStyle);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "clearCtlStyleEx")) {
             TOSVAR.offset = PREDEFSYMOFF(clearCtlStyleEx);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "clearCtlBuddyStyle")) {
             TOSVAR.offset = PREDEFSYMOFF(clearCtlBuddyStyle);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "clearCtlBuddyStyleEx")) {
             TOSVAR.offset = PREDEFSYMOFF(clearCtlBuddyStyleEx);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "clock")) {
             TOSVAR.offset = PREDEFSYMOFF(clock);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "createCtl")) {
             TOSVAR.offset = PREDEFSYMOFF(createCtl);
             TOSVAR.extradata = 11;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "createFont")) {
             TOSVAR.offset = PREDEFSYMOFF(createFont);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "createRectRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(createRectRgn);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1238; //region function
         }
         else if (matchID(currentlnamep, "createRoundRectRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(createRoundRectRgn);
             TOSVAR.extradata = 6;  //nargs
             TOSVAR.typeindex = 1238; //region function
         }
         else if (matchID(currentlnamep, "createCircularRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(createCircularRgn);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1238; //region function
         }
         else if (matchID(currentlnamep, "createEllipticRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(createEllipticRgn);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1238; //region function
         }
         else if (matchID(currentlnamep, "createPolyRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(createPolyRgn);
             TOSVAR.extradata = -8;  //nargs >= 8
             TOSVAR.typeindex = 1238; //region function
         }
         else if (matchID(currentlnamep, "createDialogRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(createDialogRgn);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1238; //region function
         }
         else if (matchID(currentlnamep, "createTitlebarRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(createTitlebarRgn);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1238; //region function
         }
         else if (matchID(currentlnamep, "createMenubarRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(createMenubarRgn);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1238; //region function
         }
         else if (matchID(currentlnamep, "createClientRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(createClientRgn);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1238; //region function
         }
         else if (matchID(currentlnamep, "createStatusbarRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(createStatusbarRgn);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1238; //region function
         }


         else if (matchID(currentlnamep, "createMenu")) {
             TOSVAR.offset = PREDEFSYMOFF(createMenu);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "createPopupMenu")) {
             TOSVAR.offset = PREDEFSYMOFF(createPopupMenu);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "createSync")) {
             TOSVAR.offset = PREDEFSYMOFF(createSync);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "createCriticalSection")) {
             TOSVAR.offset = PREDEFSYMOFF(createCriticalSection);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int (actually handle) function
         }
#ifdef HAVE_XP
         else if (matchID(currentlnamep, "createCriticalSectionAndSpinCount")) {
             TOSVAR.offset = PREDEFSYMOFF(createCriticalSectionAndSpinCount);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int (actually handle) function
         }
#endif
#ifdef HAVE_VISTA
         else if (matchID(currentlnamep, "createCriticalSectionEx")) {
             TOSVAR.offset = PREDEFSYMOFF(createCriticalSectionEx);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int (actually handle) function
         }
#endif
         
         else if (matchID(currentlnamep, "ctlClass")) {
             TOSVAR.offset = PREDEFSYMOFF(ctlClass);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "ctlCode")) {
             TOSVAR.offset = PREDEFSYMOFF(ctlCode);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "ctlMsg")) {
             TOSVAR.offset = PREDEFSYMOFF(ctlMsg);
			 TOSVAR.flags |= fmfCONST;
         }
         
         //added by Harald Heim, Oct 25, 2002
         else if (matchID(currentlnamep, "ctlEnabled")) {
             TOSVAR.offset = PREDEFSYMOFF(ctlEnabled);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //MWVDL 2008-08-19
         else if (matchID(currentlnamep, "ctlEnabledAs")) {
             TOSVAR.offset = PREDEFSYMOFF(ctlEnabledAs);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

#ifdef THEPLUGINSITE
         //added by Harald Heim, Nov 16, 2002
         else if (matchID(currentlnamep, "ct2R")) {
             TOSVAR.offset = PREDEFSYMOFF(ct2R);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Nov 16, 2002
         else if (matchID(currentlnamep, "ct2G")) {
             TOSVAR.offset = PREDEFSYMOFF(ct2G);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Nov 16, 2002
         else if (matchID(currentlnamep, "ct2B")) {
             TOSVAR.offset = PREDEFSYMOFF(ct2B);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		 else if (matchID(currentlnamep, "calcEdgeMask")) {
             TOSVAR.offset = PREDEFSYMOFF(calcEdgeMask);
             TOSVAR.extradata = 7;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "calcYCC")) {
             TOSVAR.offset = PREDEFSYMOFF(calcYCC);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "calcSpecialMinMax")) {
             TOSVAR.offset = PREDEFSYMOFF(calcSpecialMinMax);
             TOSVAR.extradata = 10;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "calcSpecialMinMaxFast")) {
             TOSVAR.offset = PREDEFSYMOFF(calcSpecialMinMaxFast);
             TOSVAR.extradata = 8;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "calcTheMix")) {
             TOSVAR.offset = PREDEFSYMOFF(calcTheMix);
             TOSVAR.extradata = 10;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "calcCMFinal")) {
             TOSVAR.offset = PREDEFSYMOFF(calcCMFinal);
             TOSVAR.extradata = 8;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

#endif

		 else if (matchID(currentlnamep, "countProcessors")) {
             TOSVAR.offset = PREDEFSYMOFF(countProcessors);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "cubicInterpolate")) {
             TOSVAR.offset = PREDEFSYMOFF(cubicInterpolate);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "cosineInterpolate")) {
             TOSVAR.offset = PREDEFSYMOFF(cosineInterpolate);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "chdir")) {
             TOSVAR.offset = PREDEFSYMOFF(direct_chdir);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "current_cw")) {
             TOSVAR.offset = PREDEFSYMOFF(current_cw);
         }
         break;

     case 'd':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(d);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_d = TRUE;
         }
         else if (matchID(currentlnamep, "d0")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(d0);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_d = TRUE;
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "d1")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(d1);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_d = TRUE;
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "deleteCtl")) {
             TOSVAR.offset = PREDEFSYMOFF(deleteCtl);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "deleteCtlItem")) {
             TOSVAR.offset = PREDEFSYMOFF(deleteCtlItem);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "deleteCtlItems")) {
             TOSVAR.offset = PREDEFSYMOFF(deleteCtlItems);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "deleteRegSubKey")) {
             TOSVAR.offset = PREDEFSYMOFF(deleteRegSubKey);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "deleteRegValue")) {
             TOSVAR.offset = PREDEFSYMOFF(deleteRegValue);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "deleteRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(deleteRgn);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "deleteFont")) {
             TOSVAR.offset = PREDEFSYMOFF(deleteFont);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "deleteCriticalSection")) {
             TOSVAR.offset = PREDEFSYMOFF(deleteCriticalSection);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int (actually bool) function
         }
         else if (matchID(currentlnamep, "deleteSync")) {
             TOSVAR.offset = PREDEFSYMOFF(deleteSync);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "destroyMenu")) {
             TOSVAR.offset = PREDEFSYMOFF(destroyMenu);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "dif")) {
             TOSVAR.offset = PREDEFSYMOFF(dif);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "displayDialog")) {
             TOSVAR.offset = PREDEFSYMOFF(displayDialog);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "dividend")) {
             TOSVAR.offset = PREDEFSYMOFF(dividend);
         }
         else if (matchID(currentlnamep, "divisor")) {
             TOSVAR.offset = PREDEFSYMOFF(divisor);
         }
         else if (matchID(currentlnamep, "dmax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(dmax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "dmin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(dmin);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "doAction")) {
             TOSVAR.offset = PREDEFSYMOFF(doAction);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "doEvents")) {
             TOSVAR.offset = PREDEFSYMOFF(doEvents);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "doForEveryRow")) {
             TOSVAR.offset = PREDEFSYMOFF(doForEveryRow);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "doingProxy")) {
             TOSVAR.offset = PREDEFSYMOFF(doingProxy);
			 //pType = pfmCONST_INT;
			 TOSVAR.flags |= fmfCONST;
         }

#ifdef SCRIPTABLE
         else if (matchID(currentlnamep, "doingScripting")) {
             TOSVAR.offset = PREDEFSYMOFF(doingScripting);
			 TOSVAR.flags |= fmfCONST;
         }
#endif


#ifdef THEPLUGINSITE         
         else if (matchID(currentlnamep, "drawTextDB")) {
             TOSVAR.offset = PREDEFSYMOFF(drawTextDB);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "dynaContra")) {
             TOSVAR.offset = PREDEFSYMOFF(dynaContra);
             TOSVAR.extradata = 6;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

#endif

         
#ifdef IMGSOURCE
         //added by Harald Heim, Oct 25, 2002
         else if (matchID(currentlnamep, "drawArray")) {
             TOSVAR.offset = PREDEFSYMOFF(drawArray);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif
         
         else if (matchID(currentlnamep, "dstColDelta")) {
             TOSVAR.offset = PREDEFSYMOFF(dstColDelta);
 			 TOSVAR.flags |= fmfCONST;
        }
         else if (matchID(currentlnamep, "dstRowDelta")) {
             TOSVAR.offset = PREDEFSYMOFF(dstRowDelta);
			 TOSVAR.flags |= fmfCONST;
         }
         break;

     case 'e':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(e);
         }
         
         //added by Harald Heim, Feb 06, 2002
         else if (matchID(currentlnamep, "egw")) {
             TOSVAR.offset = PREDEFSYMOFF(egw);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         else if (matchID(currentlnamep, "enableCtl")) {
             TOSVAR.offset = PREDEFSYMOFF(enableCtl);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "enableScriptVal")) {
             TOSVAR.offset = PREDEFSYMOFF(enableScriptVal);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }


		 //added by Martijn van der Lee, Oct 28, 2008
         else if (matchID(currentlnamep, "enableToolTipBalloon")) {
             TOSVAR.offset = PREDEFSYMOFF(enableToolTipBalloon);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "endSetPixel")) {
             TOSVAR.offset = PREDEFSYMOFF(endSetPixel);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		 else if (matchID(currentlnamep, "endSetPixelSS")) {
             TOSVAR.offset = PREDEFSYMOFF(endSetPixel);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "endSetPixelBuffer")) {
             TOSVAR.offset = PREDEFSYMOFF(endSetPixelBuffer);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

#ifdef THEPLUGINSITE
         else if (matchID(currentlnamep, "endSetPixelDB")) {
             TOSVAR.offset = PREDEFSYMOFF(endSetPixelDB);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif         

         else if (matchID(currentlnamep, "enlargeFactor")) {
             TOSVAR.offset = PREDEFSYMOFF(enlargeFactor);

         }else if (matchID(currentlnamep, "enumRegSubKey")) {
             TOSVAR.offset = PREDEFSYMOFF(enumRegSubKey);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "enumRegValue")) {
             TOSVAR.offset = PREDEFSYMOFF(enumRegValue);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "errno")) {
             TOSVAR.offset = PREDEFSYMOFF(fm_errno);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 2468; //function returning address of int
         }
         else if (matchID(currentlnamep, "exp")) {
             TOSVAR.offset = PREDEFSYMOFF(math_exp);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "expand")) {
             TOSVAR.offset = PREDEFSYMOFF(expand);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "enterCriticalSection")) {
             TOSVAR.offset = PREDEFSYMOFF(enterCriticalSection);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int (actually bool) function
         }
         break;

     case 'f':
         if (matchID(currentlnamep, "fabs")) {
             TOSVAR.offset = PREDEFSYMOFF(math_fabs);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }


#ifdef THEPLUGINSITE

         else if (matchID(currentlnamep, "fastMotion")) {
             TOSVAR.offset = PREDEFSYMOFF(fastMotion);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         
         else if (matchID(currentlnamep, "fastBox")) {
             TOSVAR.offset = PREDEFSYMOFF(fastBox);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "fastGauss")) {
             TOSVAR.offset = PREDEFSYMOFF(fastGauss);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 
		 else if (matchID(currentlnamep, "fastPreciseGauss")) {
             TOSVAR.offset = PREDEFSYMOFF(fastPreciseGauss);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "fastMotionGauss")) {
             TOSVAR.offset = PREDEFSYMOFF(fastMotionGauss);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

#endif
         
         
         else if (matchID(currentlnamep, "fclose")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fclose);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fcloseall")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fcloseall);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fcos")) {
             TOSVAR.offset = PREDEFSYMOFF(math_cos);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "feof")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_feof);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "ferror")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_ferror);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fflush")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fflush);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fgColor")) {
             TOSVAR.offset = PREDEFSYMOFF(fgColor);
 			 TOSVAR.flags |= fmfCONST;
        }
         else if (matchID(currentlnamep, "fgetc")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fgetc);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fgetpos")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fgetpos);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fgets")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fgets);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "fmax")) {
             TOSVAR.offset = PREDEFSYMOFF(fmax);
             TOSVAR.extradata = -2;  //nargs >= 2
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "fmin")) {
             TOSVAR.offset = PREDEFSYMOFF(fmin);
             TOSVAR.extradata = -2;  //nargs >= 2
             TOSVAR.typeindex = 1237; //real function
         }

         else if (matchID(currentlnamep, "fillArray")) {
             TOSVAR.offset = PREDEFSYMOFF(fillArray);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "ffillArray")) {
             TOSVAR.offset = PREDEFSYMOFF(ffillArray);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "fillDir")) {
             TOSVAR.offset = PREDEFSYMOFF(fillDir);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

		 else if (matchID(currentlnamep, "fgetArray")) {
             TOSVAR.offset = PREDEFSYMOFF(fgetArray);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
		 else if (matchID(currentlnamep, "fputArray")) {
             TOSVAR.offset = PREDEFSYMOFF(fputArray);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 		 
         else if (matchID(currentlnamep, "fcallLib")) {
             TOSVAR.offset = PREDEFSYMOFF(fcallLib);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1237; //double function
         }

         else if (matchID(currentlnamep, "fcallLibFmc")) {
             TOSVAR.offset = PREDEFSYMOFF(fcallLibFmc);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1237; //double function
         }

         else if (matchID(currentlnamep, "filterAuthorText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterAuthorText);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterCase")) {
             TOSVAR.offset = PREDEFSYMOFF(filterCase);
 			 TOSVAR.flags |= fmfCONST;
        }
         else if (matchID(currentlnamep, "filterCaseText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterCaseText);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterCategoryText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterCategoryText);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterCopyrightText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterCopyrightText);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterDescriptionText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterDescriptionText);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterFilenameText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterFilenameText);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterHostText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterHostText);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterImageModeText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterImageModeText);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterInstallDir")) {
             TOSVAR.offset = PREDEFSYMOFF(filterInstallDir);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterOrganizationText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterOrganizationText);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterRectBottom")) {
             TOSVAR.offset = PREDEFSYMOFF(filterRect32.bottom);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "filterRectHeight")) {
             TOSVAR.offset = PREDEFSYMOFF(filterRectHeight32);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "filterRectLeft")) {
             TOSVAR.offset = PREDEFSYMOFF(filterRect32.left);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "filterRectRight")) {
             TOSVAR.offset = PREDEFSYMOFF(filterRect32.right);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "filterRectTop")) {
             TOSVAR.offset = PREDEFSYMOFF(filterRect32.top);
 			 TOSVAR.flags |= fmfCONST;
        }
         else if (matchID(currentlnamep, "filterRectWidth")) {
             TOSVAR.offset = PREDEFSYMOFF(filterRectWidth32);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "filterTitleText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterTitleText);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterURLText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterURLText);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "filterVersionText")) {
             TOSVAR.offset = PREDEFSYMOFF(filterVersionText);
             TOSVAR.typeindex = 4321; //array of char
         }

         else if (matchID(currentlnamep, "filterUniqueID")) {
             TOSVAR.offset = PREDEFSYMOFF(filterUniqueID);
             TOSVAR.typeindex = 4321; //array of char
         }

         else if (matchID(currentlnamep, "findFirstFile")) {
             TOSVAR.offset = PREDEFSYMOFF(findFirstFile);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "findNextFile")) {
             TOSVAR.offset = PREDEFSYMOFF(findNextFile);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "findClose")) {
             TOSVAR.offset = PREDEFSYMOFF(findClose);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "finit")) {
             TOSVAR.offset = PREDEFSYMOFF(finit);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int (actually void) function
         }
         else if (matchID(currentlnamep, "fldcw")) {
             TOSVAR.offset = PREDEFSYMOFF(fldcw);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int (actually void) function
         }


#ifdef THEPLUGINSITE

         else if (matchID(currentlnamep, "floodFill")) {
             TOSVAR.offset = PREDEFSYMOFF(floodFill);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif
         
         else if (matchID(currentlnamep, "floatCoordHeight")) {
             TOSVAR.offset = PREDEFSYMOFF(floatCoord32.v);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "floatCoordWidth")) {
             TOSVAR.offset = PREDEFSYMOFF(floatCoord32.h);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "floor")) {
             TOSVAR.offset = PREDEFSYMOFF(math_floor);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "flushall")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_flushall);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fmod")) {
             TOSVAR.offset = PREDEFSYMOFF(math_fmod);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "fopen")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fopen);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "foreColor")) {
             TOSVAR.offset = PREDEFSYMOFF(foreColor);
 			 TOSVAR.flags |= fmfCONST;
        }
         else if (matchID(currentlnamep, "formatString")) {
             TOSVAR.offset = PREDEFSYMOFF(formatString);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function for now (should be char*)
         }
         else if (matchID(currentlnamep, "fprintf")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fprintf);
             TOSVAR.extradata = -2;  //nargs >= 2
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fputc")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fputc);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fputs")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fputs);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fread")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fread);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "free")) {
             TOSVAR.offset = PREDEFSYMOFF(free);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         //Added by Harald Heim, Apr 9, 2002
         else if (matchID(currentlnamep, "freeArray")) {
             TOSVAR.offset = PREDEFSYMOFF(freeArray);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         else if (matchID(currentlnamep, "freeHost")) {
             TOSVAR.offset = PREDEFSYMOFF(freeHost);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "freeLib")) {
             TOSVAR.offset = PREDEFSYMOFF(freeLib);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#ifdef USE_OLD_DLL_NAMES
         else if (matchID(currentlnamep, "freeLibrary")) {
             TOSVAR.offset = PREDEFSYMOFF(freeLib);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif //USE_OLD_DLL_NAMES
         else if (matchID(currentlnamep, "freeArray")) {
             TOSVAR.offset = PREDEFSYMOFF(freeArray);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }


         //added by Harald Heim, Jun 4, 2002
         else if (matchID(currentlnamep, "freeSBD")) {
             TOSVAR.offset = PREDEFSYMOFF(freeSBD);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         else if (matchID(currentlnamep, "freopen")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_freopen);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fscanf")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fscanf);
             TOSVAR.extradata = -2;  //nargs >= 2
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fseek")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fseek);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fsetpos")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fsetpos);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "fr2x")) {
             TOSVAR.offset = PREDEFSYMOFF(fr2x);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "fr2y")) {
             TOSVAR.offset = PREDEFSYMOFF(fr2y);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "fc2m")) {
             TOSVAR.offset = PREDEFSYMOFF(fc2m);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "fc2d")) {
             TOSVAR.offset = PREDEFSYMOFF(fc2d);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }

         else if (matchID(currentlnamep, "fsin")) {
             TOSVAR.offset = PREDEFSYMOFF(math_sin);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "fstcw")) {
             TOSVAR.offset = PREDEFSYMOFF(fstcw);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fstsw")) {
             TOSVAR.offset = PREDEFSYMOFF(fstsw);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "ftan")) {
             TOSVAR.offset = PREDEFSYMOFF(math_tan);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "ftell")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_ftell);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "fwrite")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_fwrite);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         break;

     case 'g':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(g);
         }
         else if (matchID(currentlnamep, "g0")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(g0);
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "g1")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(g1);
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "gamma")) {
             TOSVAR.offset = PREDEFSYMOFF(gamma);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "get")) {
             TOSVAR.offset = PREDEFSYMOFF(get);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "getc")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_getc);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "getCpuReg")) {
             TOSVAR.offset = PREDEFSYMOFF(getCpuReg);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "getSystemDefaultLCID")) {
             TOSVAR.offset = PREDEFSYMOFF(getSystemDefaultLCID);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "getSysDefaultLCID")) {	//alternate spelling??
             TOSVAR.offset = PREDEFSYMOFF(getSystemDefaultLCID);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "getUserDefaultLCID")) {
             TOSVAR.offset = PREDEFSYMOFF(getUserDefaultLCID);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

		 else if (matchID(currentlnamep, "getCtlImage")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlImage);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "getCtlColor")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlColor);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "getCtlVal")) {
             TOSVAR.offset = PREDEFSYMOFF(ctl);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "getFilterInstallDir")) {
             TOSVAR.offset = PREDEFSYMOFF(getFilterInstallDir);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function for now (should be char*)
         }

          else if (matchID(currentlnamep, "getImageTitle")) {
             TOSVAR.offset = PREDEFSYMOFF(getImageTitle);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function for now (should be char*)
         }
         else if (matchID(currentlnamep, "getLibFn")) {
             TOSVAR.offset = PREDEFSYMOFF(getLibFn);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
#ifdef USE_OLD_DLL_NAMES
         else if (matchID(currentlnamep, "getLibraryFunction")) {
             TOSVAR.offset = PREDEFSYMOFF(getLibFn);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
#endif //USE_OLD_DLL_NAMES

         else if (matchID(currentlnamep, "getLocaleInfo")) {
             TOSVAR.offset = PREDEFSYMOFF(getLocaleInfo);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "getOpenFileName")) {
             TOSVAR.offset = PREDEFSYMOFF(getOpenFileName);
             TOSVAR.extradata = 15;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "getRegData")) {
             TOSVAR.offset = PREDEFSYMOFF(getRegData);
             TOSVAR.extradata = -3;  //nargs >= 3
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "getRegInt")) {
             TOSVAR.offset = PREDEFSYMOFF(getRegInt);
             TOSVAR.extradata = -2;  //nargs >= 2
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "getRegPath")) {
             TOSVAR.offset = PREDEFSYMOFF(getRegPath);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "getRegRoot")) {
             TOSVAR.offset = PREDEFSYMOFF(getRegRoot);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "getRegString")) {
             TOSVAR.offset = PREDEFSYMOFF(getRegString);
             TOSVAR.extradata = -3;  //nargs >= 3
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "getSaveFileName")) {
             TOSVAR.offset = PREDEFSYMOFF(getSaveFileName);
             TOSVAR.extradata = 15;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "getSpecialFolder")) {
             TOSVAR.offset = PREDEFSYMOFF(getSpecialFolder);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "getSysColor")) {
             TOSVAR.offset = PREDEFSYMOFF(getSysColor);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		else if (matchID(currentlnamep, "getSysMem")) {
             TOSVAR.offset = PREDEFSYMOFF(getSysMem);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         //Added by Harald Heim, Apr 9, 2002
         else if (matchID(currentlnamep, "getArray")) {
             TOSVAR.offset = PREDEFSYMOFF(getArray);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "getArrayString")) {
             TOSVAR.offset = PREDEFSYMOFF(getArrayString);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //Added by Harald Heim, Apr 9, 2002
         else if (matchID(currentlnamep, "getDisplaySettings")) {
             TOSVAR.offset = PREDEFSYMOFF(getDisplaySettings);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         //Added by Harald Heim, Oct 14, 2003
         else if (matchID(currentlnamep, "getDialogWidth")) {
             TOSVAR.offset = PREDEFSYMOFF(getDialogWidth);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "getDialogHeight")) {
             TOSVAR.offset = PREDEFSYMOFF(getDialogHeight);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "getDialogPos")) {
             TOSVAR.offset = PREDEFSYMOFF(getDialogPos);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "getDialogHandle")) {
             TOSVAR.offset = PREDEFSYMOFF(getDialogHandle);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "getCtlHandle")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlHandle);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		 else if (matchID(currentlnamep, "getResAddress")) {
             TOSVAR.offset = PREDEFSYMOFF(getResAddress);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "getResSize")) {
             TOSVAR.offset = PREDEFSYMOFF(getResSize);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 
		 
		 //Added by Harald Heim, Jun 4, 2002
         else if (matchID(currentlnamep, "getSBD")) {
             TOSVAR.offset = PREDEFSYMOFF(getSBD);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         //Added by Harald Heim, Jun 4, 2002
         else if (matchID(currentlnamep, "getSBDX")) {
             TOSVAR.offset = PREDEFSYMOFF(getSBDX);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
                 //Added by Harald Heim, Jun 4, 2002
         else if (matchID(currentlnamep, "getSBDY")) {
             TOSVAR.offset = PREDEFSYMOFF(getSBDY);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         //Added by Harald Heim, Jun 4, 2002
         else if (matchID(currentlnamep, "getSBDangle")) {
             TOSVAR.offset = PREDEFSYMOFF(getSBDangle);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         //Added by Harald Heim, Aug 9, 2002
         else if (matchID(currentlnamep, "getPreviewCoordX")) {
             TOSVAR.offset = PREDEFSYMOFF(getPreviewCoordX);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         //Added by Harald Heim, Aug 9, 2002
         else if (matchID(currentlnamep, "getPreviewCoordY")) {
             TOSVAR.offset = PREDEFSYMOFF(getPreviewCoordY);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         else if (matchID(currentlnamep, "getCtlCoord")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlCoord);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //Added by Martijn W. van der Lee, June 29, 2008
         else if (matchID(currentlnamep, "getCtlDivisor")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlDivisor);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 
         else if (matchID(currentlnamep, "getCtlPos")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlPos);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "getCtlRange")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlRange);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "getCtlText")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlText);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "getCtlClass")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlClass);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "getCtlItemCount")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlItemCount);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "getCtlItemTop")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlItemTop);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "getCtlItemText")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlItemText);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "getCtlTab")) {
             TOSVAR.offset = PREDEFSYMOFF(getCtlTab);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		else if (matchID(currentlnamep, "getAppTheme")) {
             TOSVAR.offset = PREDEFSYMOFF(getAppTheme);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "getPreviewCursor")) {
             TOSVAR.offset = PREDEFSYMOFF(getPreviewCursor);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }


         else if (matchID(currentlnamep, "getScriptVal")) {
             TOSVAR.offset = PREDEFSYMOFF(getScriptVal);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 
         //Added by Harald Heim, Oct 25, 2002
         else if (matchID(currentlnamep, "getArrayDim")) {
             TOSVAR.offset = PREDEFSYMOFF(getArrayDim);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         //Added by Harald Heim, Aug 12, 2002
         else if (matchID(currentlnamep, "getAsyncKeyState")) {
             TOSVAR.offset = PREDEFSYMOFF(getAsyncKeyState);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         //Added by Harald Heim, Aug 12, 2002
         else if (matchID(currentlnamep, "getAsyncKeyStateF")) {
             TOSVAR.offset = PREDEFSYMOFF(getAsyncKeyStateF);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		else if (matchID(currentlnamep, "getArrayAddress")) {
             TOSVAR.offset = PREDEFSYMOFF(getArrayAddress);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		else if (matchID(currentlnamep, "getBufferAddress")) {
             TOSVAR.offset = PREDEFSYMOFF(getBufferAddress);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		 else if (matchID(currentlnamep, "getcwd")) {
             TOSVAR.offset = PREDEFSYMOFF(direct_getcwd);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

        //Added by Harald Heim, Oct 7, 2002
         /*else if (matchID(currentlnamep, "getKey")) {
             TOSVAR.offset = PREDEFSYMOFF(getKey);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }*/
         //Added by Harald Heim, January, 2002
         else if (matchID(currentlnamep, "getWindowsVersion")) {
             TOSVAR.offset = PREDEFSYMOFF(getWindowsVersion);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         else if (matchID(currentlnamep, "getPixel")) {
             TOSVAR.offset = PREDEFSYMOFF(getPixel);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "getSetPixelWidth")) {
             TOSVAR.offset = PREDEFSYMOFF(getSetPixelWidth);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "getSetPixelHeight")) {
             TOSVAR.offset = PREDEFSYMOFF(getSetPixelHeight);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "gmax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(gmax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "gmin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(gmin);
			 TOSVAR.flags |= fmfCONST;
         }
         
         //added by Harald Heim, Feb 07, 2002
         else if (matchID(currentlnamep, "grad2D")) {
             TOSVAR.offset = PREDEFSYMOFF(grad2D);
             TOSVAR.extradata = 7;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Feb 06, 2002
         else if (matchID(currentlnamep, "gray")) {
             TOSVAR.offset = PREDEFSYMOFF(gray);
             TOSVAR.extradata = 6;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		 else if (matchID(currentlnamep, "getThreadRetVal")) {
             TOSVAR.offset = PREDEFSYMOFF(getThreadRetVal);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         break;

     case 'h':
         if (matchID(currentlnamep, "haveMask")) {
             TOSVAR.offset = PREDEFSYMOFF(haveMask);
			 TOSVAR.flags |= fmfCONST;
         }
         
         else if (matchID(currentlnamep, "hermiteInterpolate")) {
             TOSVAR.offset = PREDEFSYMOFF(hermiteInterpolate);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         else if (matchID(currentlnamep, "hostSerialNumber")) {
			 //deprecated
             TOSVAR.offset = PREDEFSYMOFF(hostSerialNumber);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "hostSig")) {
             TOSVAR.offset = PREDEFSYMOFF(hostSig);
 			 TOSVAR.flags |= fmfCONST;
        }
         else if (matchID(currentlnamep, "hypot")) {
             TOSVAR.offset = PREDEFSYMOFF(math_hypot);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         //Added by Harald Heim, Dec 23, 2002
         else if (matchID(currentlnamep, "hsl2rgb")) {
             TOSVAR.offset = PREDEFSYMOFF(hsl2rgb);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //real function
         }
         break;

     case 'i':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(i);
             gFmcp->need_iuv = TRUE;
         }
         else if (matchID(currentlnamep, "imageHeight")) {
             TOSVAR.offset = PREDEFSYMOFF(imageSize32.v);
 			 TOSVAR.flags |= fmfCONST;
        }
         else if (matchID(currentlnamep, "imageMode")) {
             TOSVAR.offset = PREDEFSYMOFF(imageMode);
 			 TOSVAR.flags |= fmfCONST;
        }
         else if (matchID(currentlnamep, "imageSizeH")) {
             //deprecated
             TOSVAR.offset = PREDEFSYMOFF(imageSize32.h);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "imageSizeV")) {
             //deprecated
             TOSVAR.offset = PREDEFSYMOFF(imageSize32.v);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "imageWidth")) {
             TOSVAR.offset = PREDEFSYMOFF(imageSize32.h);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "imax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(imax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "imin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(imin);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "inPlanes")) {
             TOSVAR.offset = PREDEFSYMOFF(inPlanes);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "isFloating")) {
             TOSVAR.offset = PREDEFSYMOFF(isFloating);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "isqrt")) {
             TOSVAR.offset = PREDEFSYMOFF(isqrt);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "isTileable")) {
             TOSVAR.offset = PREDEFSYMOFF(isTileable);  //need to synch  gParams->flags & AFH_IS_TILEABLE if store to isTileable???
			 //TOSVAR.flags |= fmfCONST;	//some users set it!!??
         }
         else if (matchID(currentlnamep, "initCtl")) {
             TOSVAR.offset = PREDEFSYMOFF(initCtl);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Feb 08, 2002
         else if (matchID(currentlnamep, "iuv2rgb")) {
             TOSVAR.offset = PREDEFSYMOFF(iuv2rgb);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         //added by Harald Heim, Feb 10, 2002
         else if (matchID(currentlnamep, "iget")) {
             TOSVAR.offset = PREDEFSYMOFF(iget);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

#ifdef THEPLUGINSITE
         //added by Harald Heim, July 25, 2002
         else if (matchID(currentlnamep, "igetArray")) {
             TOSVAR.offset = PREDEFSYMOFF(igetArray);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif

         else if (matchID(currentlnamep, "imageHRes")) {
             TOSVAR.offset = PREDEFSYMOFF(imageHRes);
             TOSVAR.typeindex = 126;    //double
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "imageVRes")) {
             TOSVAR.offset = PREDEFSYMOFF(imageVRes);
             TOSVAR.typeindex = 126;    //double
			 TOSVAR.flags |= fmfCONST;
         }

         else if (matchID(currentlnamep, "insertMenuItem")) {
             TOSVAR.offset = PREDEFSYMOFF(insertMenuItem);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "iceil")) {
             TOSVAR.offset = PREDEFSYMOFF(math_iceil);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "ichop")) {
             TOSVAR.offset = PREDEFSYMOFF(math_ichop);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "ifloor")) {
             TOSVAR.offset = PREDEFSYMOFF(math_ifloor);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "ipow")) {
             TOSVAR.offset = PREDEFSYMOFF(math_ipow);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "iround")) {
             TOSVAR.offset = PREDEFSYMOFF(math_iround);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		
		 else if (matchID(currentlnamep, "isThreadActive")) {
             TOSVAR.offset = PREDEFSYMOFF(isThreadActive);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

		 else if (matchID(currentlnamep, "interlockedIncrement")) {
             TOSVAR.offset = PREDEFSYMOFF(interlockedIncrement);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "interlockedDecrement")) {
             TOSVAR.offset = PREDEFSYMOFF(interlockedDecrement);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "interlockedExchange")) {
             TOSVAR.offset = PREDEFSYMOFF(interlockedExchange);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "interlockedExchangeAdd")) {
             TOSVAR.offset = PREDEFSYMOFF(interlockedExchangeAdd);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "interlockedCompareExchange")) {
             TOSVAR.offset = PREDEFSYMOFF(interlockedCompareExchange);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

#if 0       //Too awful!!!
         else if (matchID(currentlnamep, "i0")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(i0);
             gFmcp->need_iuv = TRUE;
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "i1")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(i1);
             gFmcp->need_iuv = TRUE;
             gFmcp->needPremiere = TRUE;
         }
#else
         else if (matchID(currentlnamep, "i0")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.i0);
         }
         else if (matchID(currentlnamep, "i1")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.i1);
         }
#endif
         else if (matchID(currentlnamep, "i2")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.i2);
         }
         else if (matchID(currentlnamep, "i3")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.i3);
         }
         else if (matchID(currentlnamep, "i4")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.i4);
         }
         else if (matchID(currentlnamep, "i5")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.i5);
         }
         else if (matchID(currentlnamep, "i6")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.i6);
         }
         else if (matchID(currentlnamep, "i7")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.i7);
         }
         else if (matchID(currentlnamep, "i8")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.i8);
         }
         else if (matchID(currentlnamep, "i9")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.i9);
         }
         break;

     case 'j':
         if (matchID(currentlnamep, "j0")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.j0);
         }
         else if (matchID(currentlnamep, "j1")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.j1);
         }
         else if (matchID(currentlnamep, "j2")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.j2);
         }
         else if (matchID(currentlnamep, "j3")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.j3);
         }
         else if (matchID(currentlnamep, "j4")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.j4);
         }
         else if (matchID(currentlnamep, "j5")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.j5);
         }
         else if (matchID(currentlnamep, "j6")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.j6);
         }
         else if (matchID(currentlnamep, "j7")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.j7);
         }
         else if (matchID(currentlnamep, "j8")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.j8);
         }
         else if (matchID(currentlnamep, "j9")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.j9);
         }
         break;

     case 'k':
         if (matchID(currentlnamep, "k0")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.k0);
         }
         else if (matchID(currentlnamep, "k1")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.k1);
         }
         else if (matchID(currentlnamep, "k2")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.k2);
         }
         else if (matchID(currentlnamep, "k3")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.k3);
         }
         else if (matchID(currentlnamep, "k4")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.k4);
         }
         else if (matchID(currentlnamep, "k5")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.k5);
         }
         else if (matchID(currentlnamep, "k6")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.k6);
         }
         else if (matchID(currentlnamep, "k7")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.k7);
         }
         else if (matchID(currentlnamep, "k8")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.k8);
         }
         else if (matchID(currentlnamep, "k9")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.k9);
         }
         break;

     case 'l':
         if (matchID(currentlnamep, "ldexp")) {
             TOSVAR.offset = PREDEFSYMOFF(math_ldexp);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }

         //Added by Harald Heim, Dec 7, 2002
         else if (matchID(currentlnamep, "lab2rgb")) {
             TOSVAR.offset = PREDEFSYMOFF(lab2rgb);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //int function
         } 
         else if (matchID(currentlnamep, "leaveCriticalSection")) {
             TOSVAR.offset = PREDEFSYMOFF(leaveCriticalSection);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int (actually bool) function
         }

	      else if (matchID(currentlnamep, "linearInterpolate")) {
             TOSVAR.offset = PREDEFSYMOFF(linearInterpolate);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "loadLib")) {
             TOSVAR.offset = PREDEFSYMOFF(loadLib);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
#ifdef USE_OLD_DLL_NAMES
         else if (matchID(currentlnamep, "loadLibrary")) {
             TOSVAR.offset = PREDEFSYMOFF(loadLib);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         } 
#endif //USE_OLD_DLL_NAMES

         else if (matchID(currentlnamep, "loadResource")) {
             TOSVAR.offset = PREDEFSYMOFF(copyResToArray);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         

         //added by Harald Heim, Apr 6, 2002
         else if (matchID(currentlnamep, "lockWindow")) {
             TOSVAR.offset = PREDEFSYMOFF(lockWindow);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         } 
         
         else if (matchID(currentlnamep, "lockCtlScaling")) {
             TOSVAR.offset = PREDEFSYMOFF(lockCtlScaling);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "lockHost")) {
             TOSVAR.offset = PREDEFSYMOFF(lockHost);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         } 

         /*
         //Added by Harald Heim, Apr 6, 2002
         else if (matchID(currentlnamep, "lockCtl")) {
             TOSVAR.offset = PREDEFSYMOFF(lockCtl);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         */
        
         else if (matchID(currentlnamep, "log")) {
             TOSVAR.offset = PREDEFSYMOFF(math_log);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "log10")) {
             TOSVAR.offset = PREDEFSYMOFF(math_log10);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         break;

     case 'm':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(m);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_m = TRUE;
         }
         else if (matchID(currentlnamep, "m0")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(m0);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_m = TRUE;
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "m1")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(m1);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_m = TRUE;
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "malloc")) {
             TOSVAR.offset = PREDEFSYMOFF(malloc);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "map")) {
             TOSVAR.offset = PREDEFSYMOFF(map);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "max")) {
             TOSVAR.offset = PREDEFSYMOFF(max);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "maxSpace")) {
             TOSVAR.offset = PREDEFSYMOFF(maxSpace);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "memccpy")) {
             TOSVAR.offset = PREDEFSYMOFF(memccpy);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "memchr")) {
             TOSVAR.offset = PREDEFSYMOFF(memchr);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "memcmp")) {
             TOSVAR.offset = PREDEFSYMOFF(memcmp);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "memcpy")) {
             TOSVAR.offset = PREDEFSYMOFF(memcpy);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "memicmp")) {
             TOSVAR.offset = PREDEFSYMOFF(memicmp);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "memmove")) {
             TOSVAR.offset = PREDEFSYMOFF(memmove);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "memset")) {
             TOSVAR.offset = PREDEFSYMOFF(memset);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "min")) {
             TOSVAR.offset = PREDEFSYMOFF(min);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "mix")) {
             TOSVAR.offset = PREDEFSYMOFF(mix);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "mix1")) {
             TOSVAR.offset = PREDEFSYMOFF(mix1);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "mix2")) {
             TOSVAR.offset = PREDEFSYMOFF(mix2);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "mmax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(mmax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "mmin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(mmin);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "msize")) {
             TOSVAR.offset = PREDEFSYMOFF(msize);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         
         //added by Harald Heim, Jun 06, 2002
         else if (matchID(currentlnamep, "mouseOverWhenInvisible")) {
             TOSVAR.offset = PREDEFSYMOFF(mouseOverWhenInvisible);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

		 else if (matchID(currentlnamep, "mkdir")) {
             TOSVAR.offset = PREDEFSYMOFF(direct_mkdir);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         //added by Harald Heim, Feb 06, 2002
         else if (matchID(currentlnamep, "msk")) {
             TOSVAR.offset = PREDEFSYMOFF(msk);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "mskColDelta")) {
             TOSVAR.offset = PREDEFSYMOFF(mskColDelta);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "mskRowDelta")) {
             TOSVAR.offset = PREDEFSYMOFF(mskRowDelta);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "msgBox")) {
             TOSVAR.offset = PREDEFSYMOFF(msgBox);
             TOSVAR.extradata = -3;  //nargs >= 3
             TOSVAR.typeindex = 1234; //function
         }
         break;

     case 'n':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(n);
         }
         else if (matchID(currentlnamep, "need_iuv")) {
             TOSVAR.offset = PREDEFSYMOFF(need_iuv);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "need_d")) {
             TOSVAR.offset = PREDEFSYMOFF(need_d);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "need_m")) {
             TOSVAR.offset = PREDEFSYMOFF(need_m);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "need_tbuf")) {
             TOSVAR.offset = PREDEFSYMOFF(need_tbuf);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "need_t2buf")) {
             TOSVAR.offset = PREDEFSYMOFF(need_t2buf);
			 TOSVAR.flags |= fmfCONST;
		 }
		 else if (matchID(currentlnamep, "need_t3buf")) {
             TOSVAR.offset = PREDEFSYMOFF(need_t3buf);
			 TOSVAR.flags |= fmfCONST;
         }
		 else if (matchID(currentlnamep, "need_t4buf")) {
             TOSVAR.offset = PREDEFSYMOFF(need_t4buf);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "needPadding")) {
             TOSVAR.offset = PREDEFSYMOFF(needPadding);
         }
         else if (matchID(currentlnamep, "needPremiere")) {
             TOSVAR.offset = PREDEFSYMOFF(needPremiere);
			 TOSVAR.flags |= fmfCONST;
         }
#ifdef THEPLUGINSITE
         else if (matchID(currentlnamep, "noise")) {
             TOSVAR.offset = PREDEFSYMOFF(noise);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
#endif
         break;
     
     case 'o':

#ifdef IMGSOURCE
         //added by Harald Heim, Oct 25, 2002
         if (matchID(currentlnamep, "openImage")) {
             TOSVAR.offset = PREDEFSYMOFF(openImage);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif
         break;
         
     case 'p':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(p);   //Premiere compatibility
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "pget")) {
             TOSVAR.offset = PREDEFSYMOFF(pget);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gFmcp->isTileable = FALSE;
             gParams->flags &= ~AFH_IS_TILEABLE;
         }
         else if (matchID(currentlnamep, "pgetr")) {
             TOSVAR.offset = PREDEFSYMOFF(pgetr);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gFmcp->isTileable = FALSE;
             gParams->flags &= ~AFH_IS_TILEABLE;
         }
         
         //added by Harald Heim, Feb 05, 2002
         else if (matchID(currentlnamep, "phaseshift")) {
             TOSVAR.offset = PREDEFSYMOFF(phaseshift);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, January 2003
         else if (matchID(currentlnamep, "platformData")) {
             TOSVAR.offset = PREDEFSYMOFF(platformData);	//X64 void * ???
			 TOSVAR.flags |= fmfCONST;
         }



         else if (matchID(currentlnamep, "plane")) {
             TOSVAR.offset = PREDEFSYMOFF(plane);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "planes")) {
             TOSVAR.offset = PREDEFSYMOFF(planes);
			 TOSVAR.flags |= fmfCONST;
         }

         else if (matchID(currentlnamep, "planesWithoutAlpha")) {
             TOSVAR.offset = PREDEFSYMOFF(planesWithoutAlpha);
			 TOSVAR.flags |= fmfCONST;
         }

         else if (matchID(currentlnamep, "playSoundWave")) {
             TOSVAR.offset = PREDEFSYMOFF(playSoundWave);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gParams->gWavesSpecified = TRUE;
         }
         else if (matchID(currentlnamep, "playSoundWaveLoop")) {
             TOSVAR.offset = PREDEFSYMOFF(playSoundWaveLoop);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gParams->gWavesSpecified = TRUE;
         }
         else if (matchID(currentlnamep, "playSoundWaveSync")) {
             TOSVAR.offset = PREDEFSYMOFF(playSoundWaveSync);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gParams->gWavesSpecified = TRUE;
         }
         else if (matchID(currentlnamep, "pmax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(pmax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "pmin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(pmin);
			 TOSVAR.flags |= fmfCONST;
         }

         
         //added by Harald Heim, Feb 04, 2002
         else if (matchID(currentlnamep, "pointer_to_buffer")) {
             TOSVAR.offset = PREDEFSYMOFF(pointer_to_buffer);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Feb 05, 2002
         else if (matchID(currentlnamep, "posterize")) {
             TOSVAR.offset = PREDEFSYMOFF(posterize);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "pow")) {
             TOSVAR.offset = PREDEFSYMOFF(math_pow);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "powi")) {
             TOSVAR.offset = PREDEFSYMOFF(math_powi);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }

         //Added by Harald Heim, May 22, 2003
         else if (matchID(currentlnamep, "previous")) {
            TOSVAR.offset = PREDEFSYMOFF(previous);
         }
         
         else if (matchID(currentlnamep, "printf")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_printf);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "pset")) {
             TOSVAR.offset = PREDEFSYMOFF(pset);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->isTileable = FALSE;
             gParams->flags &= ~AFH_IS_TILEABLE;
         }
         else if (matchID(currentlnamep, "psetr")) {
             TOSVAR.offset = PREDEFSYMOFF(psetr);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->isTileable = FALSE;
             gParams->flags &= ~AFH_IS_TILEABLE;
         }
         else if (matchID(currentlnamep, "put")) {
             TOSVAR.offset = PREDEFSYMOFF(put);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "putRegData")) {
             TOSVAR.offset = PREDEFSYMOFF(putRegData);
             TOSVAR.extradata = -3;  //nargs >= 3
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "putRegInt")) {
             TOSVAR.offset = PREDEFSYMOFF(putRegInt);
             TOSVAR.extradata = -2;  //nargs >= 2
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "putRegString")) {
             TOSVAR.offset = PREDEFSYMOFF(putRegString);
             TOSVAR.extradata = -2;  //nargs >= 2
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "putc")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_putc);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         //Added by Harald Heim, Apr 9, 2002
         else if (matchID(currentlnamep, "putArray")) {
             TOSVAR.offset = PREDEFSYMOFF(putArray);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "putArrayString")) {
             TOSVAR.offset = PREDEFSYMOFF(putArrayString);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#ifdef HDRpro
		 else if (matchID(currentlnamep, "processHDR")) {
             TOSVAR.offset = PREDEFSYMOFF(processHDR);
             TOSVAR.extradata = 6;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "processHDR2")) {
             TOSVAR.offset = PREDEFSYMOFF(processHDR2);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif

#ifdef THEPLUGINSITE
		 else if (matchID(currentlnamep, "processACE")) {
             TOSVAR.offset = PREDEFSYMOFF(processACE);
             TOSVAR.extradata = 11;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif


          //Added by Harald Heim, January 2003
         else if (matchID(currentlnamep, "pgetp")) {
             TOSVAR.offset = PREDEFSYMOFF(pgetp);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gFmcp->isTileable = FALSE;
             gParams->flags &= ~AFH_IS_TILEABLE;
         }
         //Added by Harald Heim, January 2003
         else if (matchID(currentlnamep, "psetp")) {
             TOSVAR.offset = PREDEFSYMOFF(psetp);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gFmcp->isTileable = FALSE;
             gParams->flags &= ~AFH_IS_TILEABLE;
         }

         break;

    case 'q':
        //Added by Harald Heim. March 17, 2003
         if (matchID(currentlnamep, "quickSort")) {
             TOSVAR.offset = PREDEFSYMOFF(quickSort);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "quickMedian")) {
             TOSVAR.offset = PREDEFSYMOFF(quickMedian);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "quickFill")) {
             TOSVAR.offset = PREDEFSYMOFF(quickFill);
             TOSVAR.extradata = 9;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "quickFillArray")) {
             TOSVAR.offset = PREDEFSYMOFF(quickFillArray);
             TOSVAR.extradata = 9;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

#ifdef THEPLUGINSITE
         /*else if (matchID(currentlnamep, "quickGauss")) {
             TOSVAR.offset = PREDEFSYMOFF(quickGauss);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }*/
#endif

         break;

     case 'r':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(r);
         }
         else if (matchID(currentlnamep, "r0")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(r0);
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "r1")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(r1);
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "r2x")) {
             TOSVAR.offset = PREDEFSYMOFF(r2x);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "r2y")) {
             TOSVAR.offset = PREDEFSYMOFF(r2y);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "rad")||
                  matchID(currentlnamep, "rad0")||       //Premiere compatibility
                  matchID(currentlnamep, "rad1")) {      //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(rad);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->isTileable = FALSE;
             gParams->flags &= ~AFH_IS_TILEABLE;
         }
         else if (matchID(currentlnamep, "rand")) {
             TOSVAR.offset = PREDEFSYMOFF(stdlib_rand);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "realloc")) {
             TOSVAR.offset = PREDEFSYMOFF(realloc);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         
         //Added by Harald Heim, Apr 6, 2002
         else if (matchID(currentlnamep, "refreshCtl")) {
             TOSVAR.offset = PREDEFSYMOFF(refreshCtl);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //Added by Harald Heim, Apr 6, 2002
         else if (matchID(currentlnamep, "refreshWindow")) {
             TOSVAR.offset = PREDEFSYMOFF(refreshWindow);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         //Added by Harald Heim, Apr 6, 2002
         else if (matchID(currentlnamep, "refreshRgn")) {
             TOSVAR.offset = PREDEFSYMOFF(refreshRgn);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "requestRect")) {
             TOSVAR.offset = PREDEFSYMOFF(requestRect);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "restoreRect")) {
             TOSVAR.offset = PREDEFSYMOFF(restoreRect);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		 else if (matchID(currentlnamep, "resetAllCtls")) {
             TOSVAR.offset = PREDEFSYMOFF(resetAllCtls);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         
         
         else if (matchID(currentlnamep, "remove")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_remove);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "rename")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_rename);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "rewind")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_rewind);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

		 else if (matchID(currentlnamep, "rmdir")) {
             TOSVAR.offset = PREDEFSYMOFF(direct_rmdir);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         //added by Harald Heim, Feb 06, 2002
         else if (matchID(currentlnamep, "rgb2iuv")) {
             TOSVAR.offset = PREDEFSYMOFF(rgb2iuv);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         //added by Harald Heim, Feb 08, 2002
         else if (matchID(currentlnamep, "rgb2cmyk")) {
             TOSVAR.offset = PREDEFSYMOFF(rgb2cmyk);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

#ifdef THEPLUGINSITE

         else if (matchID(currentlnamep, "rgb2ct")) {
             TOSVAR.offset = PREDEFSYMOFF(rgb2ct);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
#endif

         //Added by Harald Heim, Dec 7, 2002
         else if (matchID(currentlnamep, "rgb2lab")) {
             TOSVAR.offset = PREDEFSYMOFF(rgb2lab);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         //Added by Harald Heim, Dec 23, 2002
         else if (matchID(currentlnamep, "rgb2hsl")) {
             TOSVAR.offset = PREDEFSYMOFF(rgb2hsl);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         //Added by Harald Heim, Dec 23, 2002
         else if (matchID(currentlnamep, "rgb2ycbcr")) {
             TOSVAR.offset = PREDEFSYMOFF(rgb2ycbcr);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         

         else if (matchID(currentlnamep, "rmax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(rmax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "rmin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(rmin);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "rnd")) {
             TOSVAR.offset = PREDEFSYMOFF(rnd);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "round")) {
             TOSVAR.offset = PREDEFSYMOFF(math_round);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "row")) {
             TOSVAR.offset = PREDEFSYMOFF(row);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "rows")) {
             TOSVAR.offset = PREDEFSYMOFF(rows);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "rst")) {
             TOSVAR.offset = PREDEFSYMOFF(rst);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         break;

     case 's':
         if (matchID(currentlnamep, "src")||
             matchID(currentlnamep, "src0")||       //Premiere compatibility
             matchID(currentlnamep, "src1")) {      //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(src);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->isTileable = FALSE;
             gParams->flags &= ~AFH_IS_TILEABLE;
         }
         else if (matchID(currentlnamep, "samplingSupport")) {
             TOSVAR.offset = PREDEFSYMOFF(samplingSupport);
			 TOSVAR.flags |= fmfCONST;
         }

        //added by Harald Heim, Feb 05, 2002
        else if (matchID(currentlnamep, "saturation")) {
             TOSVAR.offset = PREDEFSYMOFF(saturation);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "scaleFactor")) {
             TOSVAR.offset = PREDEFSYMOFF(scaleFactor);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "scl")) {
             TOSVAR.offset = PREDEFSYMOFF(scl);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

        //added by Harald Heim, Dec 14, 2002
         else if (matchID(currentlnamep, "scrollPreview")) {
             TOSVAR.offset = PREDEFSYMOFF(scrollPreview);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
          
         //added by Harald Heim, Dec 13, 2002
         else if (matchID(currentlnamep, "setClickDrag")) {
             TOSVAR.offset = PREDEFSYMOFF(setClickDrag);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
          else if (matchID(currentlnamep, "setTimerEvent")) {
             TOSVAR.offset = PREDEFSYMOFF(setTimerEvent);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "setCtlAction")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlAction);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlColor")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlColor);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "setCtlSysColor")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlSysColor);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "setCtlDivisor")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlDivisor);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlEditSize")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlEditSize);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
            
         else if (matchID(currentlnamep, "setEditBoxUpdate")) {
             TOSVAR.offset = PREDEFSYMOFF(setEditBoxUpdate);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "setCtlFocus")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlFocus);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "setCtlStyle")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlStyle);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "setCtlStyleEx")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlStyleEx);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "setCtlBuddyStyle")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlBuddyStyle);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "setCtlBuddyStyleEx")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlBuddyStyleEx);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "setCtlGamma")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlGamma);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "setCtlFontColor")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlFontColor);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlFontSysColor")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlFontSysColor);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlFont")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlFont);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "setCtlImage")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlImage);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gParams->gCtlImagesSet = TRUE;
         }
         else if (matchID(currentlnamep, "setCtlLineSize")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlLineSize);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlPageSize")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlPageSize);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
	     else if (matchID(currentlnamep, "setCtlThumbSize")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlThumbSize);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlOrder")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlOrder);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlPos")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlPos);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlPixelPos")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlPixelPos);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlProperties")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlProperties);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlRange")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlRange);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlText")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlText);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlTextColor")) {  //alternate name
             TOSVAR.offset = PREDEFSYMOFF(setCtlFontColor);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlTextv")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlTextv);
             TOSVAR.extradata = -2;  //nargs >= 2
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlTicFreq")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlTicFreq);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlToolTip")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlToolTip);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "setToolTipDelay")) {
             TOSVAR.offset = PREDEFSYMOFF(setToolTipDelay);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlVal")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlVal);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlDefVal")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlDefVal);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setDialogColor")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogColor);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "setDialogSysColor")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogSysColor);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setDialogDragMode")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogDragMode);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setDialogGradient")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogGradient);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setDialogImage")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogImage);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setDialogImageMode")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogImageMode);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         else if (matchID(currentlnamep, "setDialogMinMax")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogMinMax);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setDialogSizeGrip")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogSizeGrip);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setCtlItemTop")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlItemTop);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "setCtlItemText")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlItemText);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "setDialogPos")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogPos);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setDialogRegion")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogRegion);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "setDialogShowState")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogShowState);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         else if (matchID(currentlnamep, "setDialogStyle")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogStyle);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setDialogStyleEx")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogStyleEx);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setDialogText")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogText);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setDialogTextv")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogTextv);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //function
         }

		 else if (matchID(currentlnamep, "setDialogTheme")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogTheme);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "setCtlTheme")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlTheme);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		 else if (matchID(currentlnamep, "setDialogEvent")) {
             TOSVAR.offset = PREDEFSYMOFF(setDialogEvent);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

		 else if (matchID(currentlnamep, "setCtlTab")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlTab);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "setCtlAnchor")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlAnchor);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "scaleCtls")) {
             TOSVAR.offset = PREDEFSYMOFF(scaleCtls);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "setCtlScripting")) {
             TOSVAR.offset = PREDEFSYMOFF(setCtlScripting);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setScriptVal")) {
             TOSVAR.offset = PREDEFSYMOFF(setScriptVal);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }


         else if (matchID(currentlnamep, "setGamma")) {
             TOSVAR.offset = PREDEFSYMOFF(setGamma);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "setRegRoot")) {
             TOSVAR.offset = PREDEFSYMOFF(setRegRoot);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "setRegPath")) {
             TOSVAR.offset = PREDEFSYMOFF(setRegPath);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //int function
         }
#ifdef THEPLUGINSITE
         else if (matchID(currentlnamep, "setFontDB")) {
             TOSVAR.offset = PREDEFSYMOFF(setFontDB);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
#endif

         else if (matchID(currentlnamep, "setMenu")) {
             TOSVAR.offset = PREDEFSYMOFF(setMenu);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         //Added by Harald Heim, April 02, 2002
         else if (matchID(currentlnamep, "setZoom")) {
             TOSVAR.offset = PREDEFSYMOFF(setZoom);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		else if (matchID(currentlnamep, "setZoomEx")) {
             TOSVAR.offset = PREDEFSYMOFF(setZoomEx);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "setPixel")) {
             TOSVAR.offset = PREDEFSYMOFF(setPixel);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "setBitmap")) {
             TOSVAR.offset = PREDEFSYMOFF(setBitmap);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setBitmapStretch")) {
             TOSVAR.offset = PREDEFSYMOFF(setBitmapStretch);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setBitmapTile")) {
             TOSVAR.offset = PREDEFSYMOFF(setBitmapTile);
             TOSVAR.extradata = 6;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setBitmapTransparent")) {
             TOSVAR.offset = PREDEFSYMOFF(setBitmapTransparent);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "setBitmapStretchTransparent")) {
             TOSVAR.offset = PREDEFSYMOFF(setBitmapStretchTransparent);
             TOSVAR.extradata = 6;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#ifdef HAVE_XP
         else if (matchID(currentlnamep, "setCriticalSectionSpinCount")) {
             TOSVAR.offset = PREDEFSYMOFF(setCriticalSectionSpinCount);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int (actually DWORD) function
         }
#endif
         else if (matchID(currentlnamep, "startSetPixel")) {
             TOSVAR.offset = PREDEFSYMOFF(startSetPixel);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "startSetPixelSS")) {
             TOSVAR.offset = PREDEFSYMOFF(startSetPixelSS);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "startSetPixelBuffer")) {
             TOSVAR.offset = PREDEFSYMOFF(startSetPixelBuffer);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

#ifdef THEPLUGINSITE
		 else if (matchID(currentlnamep, "stretchContrast")) {
             TOSVAR.offset = PREDEFSYMOFF(stretchContrast);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
#endif

		 else if (matchID(currentlnamep, "setFont")) {
             TOSVAR.offset = PREDEFSYMOFF(setFont);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setText")) {
             TOSVAR.offset = PREDEFSYMOFF(setText);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setTextv")) {
             TOSVAR.offset = PREDEFSYMOFF(setTextv);
             TOSVAR.extradata = -5;  //nargs >= 5
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setLine")) {
             TOSVAR.offset = PREDEFSYMOFF(setLine);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setEllipse")) {
             TOSVAR.offset = PREDEFSYMOFF(setEllipse);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setEllipseFill")) {
             TOSVAR.offset = PREDEFSYMOFF(setEllipseFill);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setRectFrame")) {
             TOSVAR.offset = PREDEFSYMOFF(setRectFrame);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setRectangle")) {
             TOSVAR.offset = PREDEFSYMOFF(setRectangle);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setFill")) {
             TOSVAR.offset = PREDEFSYMOFF(setFill);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setRectFill")) {
             TOSVAR.offset = PREDEFSYMOFF(setRectFill);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setRectGradient")) {
             TOSVAR.offset = PREDEFSYMOFF(setRectGradient);
             TOSVAR.extradata = 7;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setPenWidth")) {
             TOSVAR.offset = PREDEFSYMOFF(setPenWidth);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setThinLineAA")) {
             TOSVAR.offset = PREDEFSYMOFF(setThinLineAA);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setThinEllipseAA")) {
             TOSVAR.offset = PREDEFSYMOFF(setThinEllipseAA);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setEndCaps")) {
             TOSVAR.offset = PREDEFSYMOFF(setEndCaps);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setJoin")) {
             TOSVAR.offset = PREDEFSYMOFF(setJoin);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setInsideFrame")) {
             TOSVAR.offset = PREDEFSYMOFF(setInsideFrame);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setAngleArcFill")) {
             TOSVAR.offset = PREDEFSYMOFF(setAngleArcFill);
             TOSVAR.extradata = 6;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
		 else if (matchID(currentlnamep, "setAngleArc")) {
             TOSVAR.offset = PREDEFSYMOFF(setAngleArc);
             TOSVAR.extradata = 6;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

#ifdef THEPLUGINSITE
         else if (matchID(currentlnamep, "setPixelDB")) {
             TOSVAR.offset = PREDEFSYMOFF(setPixelDB);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "startSetPixelDB")) {
             TOSVAR.offset = PREDEFSYMOFF(startSetPixelDB);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif
         //Added by Harald Heim, January, 2003
         else if (matchID(currentlnamep, "setPreviewCursor")) {
             TOSVAR.offset = PREDEFSYMOFF(setPreviewCursor);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "setPreviewDragMode")) {
             TOSVAR.offset = PREDEFSYMOFF(setPreviewDragMode);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "set_edge_mode")) {
             TOSVAR.offset = PREDEFSYMOFF(set_edge_mode);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "set_bitdepth_mode")) {
             TOSVAR.offset = PREDEFSYMOFF(set_bitdepth_mode);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "set_psetp_mode")) {
             TOSVAR.offset = PREDEFSYMOFF(set_psetp_mode);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

		else if (matchID(currentlnamep, "set_array_mode")) {
             TOSVAR.offset = PREDEFSYMOFF(set_array_mode);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "shellExec")) {
             TOSVAR.offset = PREDEFSYMOFF(shellExec);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "sin")) {
             TOSVAR.offset = PREDEFSYMOFF(sin);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         //added by Harald Heim, Feb 08, 2002
          else if (matchID(currentlnamep, "sinbell")) {
             TOSVAR.offset = PREDEFSYMOFF(sinbell);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         else if (matchID(currentlnamep, "sinh")) {
             TOSVAR.offset = PREDEFSYMOFF(math_sinh);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "sleep")) {
             TOSVAR.offset = PREDEFSYMOFF(sleep);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

		 else if (matchID(currentlnamep, "scrollFactor")) {
             TOSVAR.offset = PREDEFSYMOFF(scrollFactor);
         }
         //Added by Harald Heim, January 11, 2003
         else if (matchID(currentlnamep, "srcp")) {
             TOSVAR.offset = PREDEFSYMOFF(srcp);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
             gFmcp->isTileable = FALSE;
             gParams->flags &= ~AFH_IS_TILEABLE;
         }

         //added by Harald Heim, Feb 06, 2002
         else if (matchID(currentlnamep, "solarize")) {
             TOSVAR.offset = PREDEFSYMOFF(solarize);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "snprintf")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_snprintf);
             TOSVAR.extradata = -3;  //nargs >= 3
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "sprintf")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_sprintf);
             TOSVAR.extradata = -2;  //nargs >= 2
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "sqr")) {
             TOSVAR.offset = PREDEFSYMOFF(sqr);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "sqrt")) {
             TOSVAR.offset = PREDEFSYMOFF(math_sqrt);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "srand")) {
             TOSVAR.offset = PREDEFSYMOFF(stdlib_srand);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int (actually void) function
         }
         else if (matchID(currentlnamep, "srcColDelta")) {
             TOSVAR.offset = PREDEFSYMOFF(srcColDelta);
 			 TOSVAR.flags |= fmfCONST;
        }
         else if (matchID(currentlnamep, "srcRowDelta")) {
             TOSVAR.offset = PREDEFSYMOFF(srcRowDelta);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "sscanf")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_sscanf);
             TOSVAR.extradata = -2;  //nargs >= 2
             TOSVAR.typeindex = 1234; //int function
         }
#ifdef _MSC_VER
         // WORKAROUND FOR MSVC FATAL ERROR:
         //
         //     fatal error C1026: parser stack overflow, 
         //     program too complex
         //
         // MSVC 5.0 (and others?) gets a parser stack overflow
         // error when if-statements are nested more deeply than
         // 120 or so levels.  Insert this workaround anywhere
         // as needed to break up the nesting into no more than
         // 100 or so levels.
         if (TOSVAR.offset != PREDEFSYMOFF(undefined)) {
             /* symbol or label already found */
             break;
         }
#endif
         else if (matchID(currentlnamep, "str0")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str0);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str1")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str1);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str2")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str2);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str3")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str3);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str4")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str4);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str5")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str5);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str6")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str6);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str7")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str7);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str8")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str8);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str9")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str9);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str10")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str10);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str11")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str11);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str12")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str12);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str13")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str13);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str14")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str14);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str15")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str15);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str16")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str16);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str17")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str17);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str18")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str18);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "str19")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.str19);
             TOSVAR.typeindex = 4321; //array of char
         }
         else if (matchID(currentlnamep, "strcat")) {
             TOSVAR.offset = PREDEFSYMOFF(strcat);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strchr")) {
             TOSVAR.offset = PREDEFSYMOFF(strchr);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strcmp")) {
             TOSVAR.offset = PREDEFSYMOFF(strcmp);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strcoll")) {
             TOSVAR.offset = PREDEFSYMOFF(strcoll);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strcpy")) {
             TOSVAR.offset = PREDEFSYMOFF(strcpy);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strcspn")) {
             TOSVAR.offset = PREDEFSYMOFF(strcspn);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strdate")) {
             TOSVAR.offset = PREDEFSYMOFF(strdate);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strdup")) {
             TOSVAR.offset = PREDEFSYMOFF(strdup);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strerror")) {
             TOSVAR.offset = PREDEFSYMOFF(strerror);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "stricmp")) {
             TOSVAR.offset = PREDEFSYMOFF(stricmp);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "stricoll")) {
             TOSVAR.offset = PREDEFSYMOFF(stricoll);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "stripEllipsis")) {
             TOSVAR.offset = PREDEFSYMOFF(stripEllipsis);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function for now (should be char*)
         }
         else if (matchID(currentlnamep, "strlen")) {
             TOSVAR.offset = PREDEFSYMOFF(strlen);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strlwr")) {
             TOSVAR.offset = PREDEFSYMOFF(strlwr);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strncat")) {
             TOSVAR.offset = PREDEFSYMOFF(strncat);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strncmp")) {
             TOSVAR.offset = PREDEFSYMOFF(strncmp);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strncoll")) {
             TOSVAR.offset = PREDEFSYMOFF(strncoll);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strncpy")) {
             TOSVAR.offset = PREDEFSYMOFF(strncpy);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strnicmp")) {
             TOSVAR.offset = PREDEFSYMOFF(strnicmp);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strnicoll")) {
             TOSVAR.offset = PREDEFSYMOFF(strnicoll);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strnset")) {
             TOSVAR.offset = PREDEFSYMOFF(strnset);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strpbrk")) {
             TOSVAR.offset = PREDEFSYMOFF(strpbrk);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strrchr")) {
             TOSVAR.offset = PREDEFSYMOFF(strrchr);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strrev")) {
             TOSVAR.offset = PREDEFSYMOFF(strrev);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strset")) {
             TOSVAR.offset = PREDEFSYMOFF(strset);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strspn")) {
             TOSVAR.offset = PREDEFSYMOFF(strspn);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strstr")) {
             TOSVAR.offset = PREDEFSYMOFF(strstr);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strtime")) {
             TOSVAR.offset = PREDEFSYMOFF(strtime);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strtod")) {
             TOSVAR.offset = PREDEFSYMOFF(strtod);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "strtok")) {
             TOSVAR.offset = PREDEFSYMOFF(strtok);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strtol")) {
             TOSVAR.offset = PREDEFSYMOFF(strtol);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strtoul")) {
             TOSVAR.offset = PREDEFSYMOFF(strtoul);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strupr")) {
             TOSVAR.offset = PREDEFSYMOFF(strupr);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "strxfrm")) {
             TOSVAR.offset = PREDEFSYMOFF(strxfrm);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "sub")) {
             TOSVAR.offset = PREDEFSYMOFF(sub);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         break;

     case 't':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(t);   //Premiere compatibility
         }
         else if (matchID(currentlnamep, "tan")) {
             TOSVAR.offset = PREDEFSYMOFF(tan);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "tanh")) {
             TOSVAR.offset = PREDEFSYMOFF(math_tanh);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1237; //real function
         }
         else if (matchID(currentlnamep, "tget")) {
             TOSVAR.offset = PREDEFSYMOFF(tget);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_tbuf = TRUE;
         }
		 else if (matchID(currentlnamep, "terminateThread")) {
             TOSVAR.offset = PREDEFSYMOFF(terminateThread);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }


        //Added by Harald Heim, January 2003
        else if (matchID(currentlnamep, "tgetp")) {
             TOSVAR.offset = PREDEFSYMOFF(tgetp);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_tbuf = TRUE;
         }
        //Added by Harald Heim, March 9, 2003
        else if (matchID(currentlnamep, "t2getp")) {
             TOSVAR.offset = PREDEFSYMOFF(t2getp);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t2buf = TRUE;
         }
		 else if (matchID(currentlnamep, "t3getp")) {
             TOSVAR.offset = PREDEFSYMOFF(t3getp);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t3buf = TRUE;
         }
		 else if (matchID(currentlnamep, "t4getp")) {
             TOSVAR.offset = PREDEFSYMOFF(t4getp);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t4buf = TRUE;
         }

         else if (matchID(currentlnamep, "tgetr")) {
             TOSVAR.offset = PREDEFSYMOFF(tgetr);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_tbuf = TRUE;
         }
         else if (matchID(currentlnamep, "t2get")) {
             TOSVAR.offset = PREDEFSYMOFF(t2get);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t2buf = TRUE;
         }
         else if (matchID(currentlnamep, "t2getr")) {
             TOSVAR.offset = PREDEFSYMOFF(t2getr);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t2buf = TRUE;
         }
		 else if (matchID(currentlnamep, "t3get")) {
             TOSVAR.offset = PREDEFSYMOFF(t3get);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t3buf = TRUE;
         }
		 else if (matchID(currentlnamep, "t3getr")) {
             TOSVAR.offset = PREDEFSYMOFF(t3getr);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t3buf = TRUE;
         }
		  else if (matchID(currentlnamep, "t4get")) {
             TOSVAR.offset = PREDEFSYMOFF(t4get);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t4buf = TRUE;
         }
		 else if (matchID(currentlnamep, "t4getr")) {
             TOSVAR.offset = PREDEFSYMOFF(t4getr);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t4buf = TRUE;
         }
         else if (matchID(currentlnamep, "testAbort")) {
             TOSVAR.offset = PREDEFSYMOFF(testAbort);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "time")) {
             TOSVAR.offset = PREDEFSYMOFF(time);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "tmax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(tmax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "tmin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(tmin);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "tmpfile")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_tmpfile);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "tmpnam")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_tmpnam);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }

         //added by Harald Heim, Feb 08, 2002
         else if (matchID(currentlnamep, "tone")) {
             TOSVAR.offset = PREDEFSYMOFF(tone);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "total")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(total);
			 TOSVAR.flags |= fmfCONST;
         }

         
         else if (matchID(currentlnamep, "trackPopupMenu")) {
             TOSVAR.offset = PREDEFSYMOFF(trackPopupMenu);
             TOSVAR.extradata = 5;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         //added by Harald Heim, Feb 07, 2002
         else if (matchID(currentlnamep, "tri")) {
             TOSVAR.offset = PREDEFSYMOFF(tri);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         
         //added by Harald Heim, Feb 07, 2002
         else if (matchID(currentlnamep, "tricos")) {
             TOSVAR.offset = PREDEFSYMOFF(tricos);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "triggerEvent")) {
             TOSVAR.offset = PREDEFSYMOFF(triggerEvent);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "triggerThread")) {
             TOSVAR.offset = PREDEFSYMOFF(triggerThread);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }


         //Added by Harald Heim, Aug 14, 2002
         /*else if (matchID(currentlnamep, "threeDto2D_X")) {
             TOSVAR.offset = PREDEFSYMOFF(threeDto2D_X);
             TOSVAR.extradata = 10;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "threeDto2D_Y")) {
             TOSVAR.offset = PREDEFSYMOFF(threeDto2D_Y);
             TOSVAR.extradata = 10;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }*/
        
         //Added by Harald Heim, January 2003
         else if (matchID(currentlnamep, "tsetp")) {
             TOSVAR.offset = PREDEFSYMOFF(tsetp);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_tbuf = TRUE;
         }
        //Added by Harald Heim, March 9, 2003
         else if (matchID(currentlnamep, "t2setp")) {
             TOSVAR.offset = PREDEFSYMOFF(t2setp);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t2buf = TRUE;
         }
		 else if (matchID(currentlnamep, "t3setp")) {
             TOSVAR.offset = PREDEFSYMOFF(t3setp);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t3buf = TRUE;
         }
		 else if (matchID(currentlnamep, "t4setp")) {
             TOSVAR.offset = PREDEFSYMOFF(t4setp);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t4buf = TRUE;
         }

         else if (matchID(currentlnamep, "tset")) {
             TOSVAR.offset = PREDEFSYMOFF(tset);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_tbuf = TRUE;
         }
         else if (matchID(currentlnamep, "tsetr")) {
             TOSVAR.offset = PREDEFSYMOFF(tsetr);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_tbuf = TRUE;
         }
         else if (matchID(currentlnamep, "t2set")) {
             TOSVAR.offset = PREDEFSYMOFF(t2set);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t2buf = TRUE;
         }
         else if (matchID(currentlnamep, "t2setr")) {
             TOSVAR.offset = PREDEFSYMOFF(t2setr);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t2buf = TRUE;
         }
		 else if (matchID(currentlnamep, "t3set")) {
             TOSVAR.offset = PREDEFSYMOFF(t3set);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t3buf = TRUE;
         }
		 else if (matchID(currentlnamep, "t3setr")) {
             TOSVAR.offset = PREDEFSYMOFF(t3setr);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t3buf = TRUE;
         }
		 else if (matchID(currentlnamep, "t4set")) {
             TOSVAR.offset = PREDEFSYMOFF(t4set);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t4buf = TRUE;
         }
		 else if (matchID(currentlnamep, "t4setr")) {
             TOSVAR.offset = PREDEFSYMOFF(t4setr);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
             gFmcp->need_t4buf = TRUE;
         }
#if _WIN32_WINNT >= 0x0400 
         else if (matchID(currentlnamep, "tryEnterCriticalSection")) {
             TOSVAR.offset = PREDEFSYMOFF(tryEnterCriticalSection);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int (actually bool) function
         }
#endif
         break;

     case 'u':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(u);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_iuv = TRUE;
         }
         else if (matchID(currentlnamep, "u0")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(u0);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_iuv = TRUE;
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "u1")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(u1);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_iuv = TRUE;
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "umax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(umax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "umin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(umin);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "ungetc")) {
             TOSVAR.offset = PREDEFSYMOFF(stdio_ungetc);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "updatePreview")) {
             TOSVAR.offset = PREDEFSYMOFF(updatePreview);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "updateProgress")) {
             TOSVAR.offset = PREDEFSYMOFF(updateProgress);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "updateAnchors")) {
             TOSVAR.offset = PREDEFSYMOFF(updateAnchors);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "use_CMOV")) {
             TOSVAR.offset = PREDEFSYMOFF(use_CMOV);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "use_MMX")) {
             TOSVAR.offset = PREDEFSYMOFF(use_MMX);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "use_SSE")) {
             TOSVAR.offset = PREDEFSYMOFF(use_SSE);
			 TOSVAR.flags |= fmfCONST;
         }
         break;

     case 'v':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(v);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_iuv = TRUE;
         }
         else if (matchID(currentlnamep, "v0")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(v0);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_iuv = TRUE;
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "v1")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(v1);
			 TOSVAR.flags |= fmfCONST;
             gFmcp->need_iuv = TRUE;
             gFmcp->needPremiere = TRUE;
         }
         else if (matchID(currentlnamep, "val")) {
             TOSVAR.offset = PREDEFSYMOFF(val);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "vmax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(vmax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "vmin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(vmin);
			 TOSVAR.flags |= fmfCONST;
         }
         break;

     case 'w':
         if (matchID(currentlnamep, "wholeHeight")) {
             TOSVAR.offset = PREDEFSYMOFF(wholeSize32.v);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "wholeWidth")) {
             TOSVAR.offset = PREDEFSYMOFF(wholeSize32.h);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "writeAlpha")) {
             TOSVAR.offset = PREDEFSYMOFF(writeAlpha);
			 TOSVAR.flags |= fmfCONST;
         }
 		 else if (matchID(currentlnamep, "waitForThread")) {
             TOSVAR.offset = PREDEFSYMOFF(waitForThread);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "waitForSync")) {
             TOSVAR.offset = PREDEFSYMOFF(waitForSync);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

#ifdef WATERMARK        
         else if (matchID(currentlnamep, "watermark")) {
             TOSVAR.offset = PREDEFSYMOFF(watermark);
             TOSVAR.extradata = 2;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
#endif
         break;

     case 'x':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(x);
         }
         else if (matchID(currentlnamep, "x_end")) {
             TOSVAR.offset = PREDEFSYMOFF(x_end);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "x_start")) {
             TOSVAR.offset = PREDEFSYMOFF(x_start);
			 TOSVAR.flags |= fmfCONST;
         }

        //Added by Harald Heim, Feb 10, 2002
         else if (matchID(currentlnamep, "xyzcnv")) {
             TOSVAR.offset = PREDEFSYMOFF(xyzcnv);
             TOSVAR.extradata = 13;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         else if (matchID(currentlnamep, "x0")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.x0);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "x1")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.x1);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "x2")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.x2);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "x3")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.x3);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "x4")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.x4);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "x5")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.x5);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "x6")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.x6);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "x7")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.x7);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "x8")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.x8);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "x9")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.x9);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "xmax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(xmax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "xmin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(xmin);
			 TOSVAR.flags |= fmfCONST;
         }
         break;

     case 'y':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(y);
         }
         else if (matchID(currentlnamep, "y_end")) {
             TOSVAR.offset = PREDEFSYMOFF(y_end);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "y_start")) {
             TOSVAR.offset = PREDEFSYMOFF(y_start);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "y0")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.y0);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "y1")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.y1);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "y2")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.y2);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "y3")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.y3);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "y4")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.y4);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "y5")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.y5);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "y6")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.y6);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "y7")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.y7);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "y8")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.y8);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "y9")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.y9);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "ymax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(ymax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "ymin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(ymin);
			 TOSVAR.flags |= fmfCONST;
         }

          //Added by Harald Heim, Dec 23, 2002
         else if (matchID(currentlnamep, "ycbcr2rgb")) {
             TOSVAR.offset = PREDEFSYMOFF(ycbcr2rgb);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //function
         }

         break;

     case 'z':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(z);
         }
         else if (matchID(currentlnamep, "z0")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.z0);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "z1")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.z1);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "z2")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.z2);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "z3")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.z3);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "z4")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.z4);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "z5")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.z5);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "z6")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.z6);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "z7")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.z7);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "z8")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.z8);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "z9")) {
             TOSVAR.offset = PREDEFSYMOFF(globvars.z9);
             TOSVAR.typeindex = 126;    //double
         }
         else if (matchID(currentlnamep, "zmax")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(zmax);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "zmin")) {   //Premiere compatibility
             TOSVAR.offset = PREDEFSYMOFF(zmin);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "zoomFactor")) {
             TOSVAR.offset = PREDEFSYMOFF(zoomFactor);
			 TOSVAR.flags |= fmfCONST;
         }
         break;

     case 'A':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(A);
         }
         else if (matchID(currentlnamep, "Aval")) {
             TOSVAR.offset = PREDEFSYMOFF(Aval);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         break;

     case 'B':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(B);
         }
         else if (matchID(currentlnamep, "Bval")) {
             TOSVAR.offset = PREDEFSYMOFF(Bval);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         break;

     case 'C':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(C);
         }
         break;

     case 'D':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(D);
			 TOSVAR.flags |= fmfCONST;
         } 
		 else if (matchID(currentlnamep, "DESIGNTIME")) {
             TOSVAR.offset = PREDEFSYMOFF(DESIGNTIME);
			 TOSVAR.flags |= fmfCONST;
         }
         break;

     case 'E':
         if (matchID(currentlnamep, "Error")) {
             TOSVAR.offset = PREDEFSYMOFF(Error);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "ErrorOk")) {
             TOSVAR.offset = PREDEFSYMOFF(ErrorOk);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //function
         }
		 else if (matchID(currentlnamep, "ENTRYPOINT")) {
             TOSVAR.offset = PREDEFSYMOFF(entryPoint);
         }
         break;

	 case 'F':
         if (matchID(currentlnamep, "FMC_COMPILE_PLATFORM")) {		//caps for future macro name
             TOSVAR.offset = PREDEFSYMOFF(FMC_compile_platform);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "FMC_MAJOR_VERSION")) {	//caps for future macro name
             TOSVAR.offset = PREDEFSYMOFF(FMC_major_version);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "FMC_MINOR_VERSION")) {	//caps for future macro name
             TOSVAR.offset = PREDEFSYMOFF(FMC_minor_version);
			 TOSVAR.flags |= fmfCONST;
         }
         else if (matchID(currentlnamep, "FMC_TARGET")) {			//caps for future macro name
             TOSVAR.offset = PREDEFSYMOFF(FMC_target);
			 TOSVAR.flags |= fmfCONST;
         }
		 break;

     case 'G':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(G);
         }
         else if (matchID(currentlnamep, "Gval")) {
             TOSVAR.offset = PREDEFSYMOFF(Gval);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         break;

     case 'H':
         if (matchID(currentlnamep, "HDBUsToPixels")) {
             TOSVAR.offset = PREDEFSYMOFF(HDBUsToPixels);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         break;

     case 'I':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(I);
         }
         else if (matchID(currentlnamep, "Info")) {
             TOSVAR.offset = PREDEFSYMOFF(Info);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //function
         }
         break;

     case 'R':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(R);
         }
         else if (matchID(currentlnamep, "RGB")) {
             TOSVAR.offset = PREDEFSYMOFF(RGB);
             TOSVAR.extradata = 3;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "RGBA")) {
             TOSVAR.offset = PREDEFSYMOFF(RGBA);
             TOSVAR.extradata = 4;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         else if (matchID(currentlnamep, "Rval")) {
             TOSVAR.offset = PREDEFSYMOFF(Rval);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         break;

     case 'M':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(M);
			 TOSVAR.flags |= fmfCONST;
         }
         break;

     case 'P':
         if (matchID(currentlnamep, "PixelsToHDBUs")) {
             TOSVAR.offset = PREDEFSYMOFF(PixelsToHDBUs);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "PixelsToVDBUs")) {
             TOSVAR.offset = PREDEFSYMOFF(PixelsToVDBUs);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //function
         }
         break;

     case 'U':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(U);
         }
         else if (matchID(currentlnamep, "UserFunction0")) {    //for debugging only!!!
             TOSVAR.offset = PREDEFSYMOFF(UserFunctions);
             TOSVAR.extradata = 0;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         break;

     case 'V':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(V);
         }
         else if (matchID(currentlnamep, "VDBUsToPixels")) {
             TOSVAR.offset = PREDEFSYMOFF(VDBUsToPixels);
             TOSVAR.extradata = 1;  //nargs
             TOSVAR.typeindex = 1234; //int function
         }
         break;

     case 'W':
         if (matchID(currentlnamep, "Warn")) {
             TOSVAR.offset = PREDEFSYMOFF(Warn);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //function
         }
         break;

     case 'X':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(X);
         }
         break;

     case 'Y':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(Y);
         }
         else if (matchID(currentlnamep, "YesNo")) {
             TOSVAR.offset = PREDEFSYMOFF(YesNo);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //function
         }
         else if (matchID(currentlnamep, "YesNoCancel")) {
             TOSVAR.offset = PREDEFSYMOFF(YesNoCancel);
             TOSVAR.extradata = -1;  //nargs >= 1
             TOSVAR.typeindex = 1234; //function
         }
         break;

     case 'Z':
         if (nChars == 1) {
             TOSVAR.offset = PREDEFSYMOFF(Z);
         }
         break;
     default:
         break;
     } /*switch*/

	 TOSVAR.pSym = pSym;
	 if (pSym) {
		 TOSVAR.flags |= pSym->flags;
	 }
	 TOSVAR.pType = pType;
	 if (pType) {
		 TOSVAR.flags |= pType->flags;
	 }
	 TOSVAR.namep = idp;

     if (TOSVAR.offset == PREDEFSYMOFF(undefined)) {
         /* symbol or label not found */
         Mde_error(MD386__SYMNOTFND, (UINT_PTR) currentlnamep, 0, 0);
     }

     if (TOSVAR.typeindex >= 1234 && TOSVAR.typeindex <= 1238) {
         //(address of)function -- should really be a LITOBECTKIND??
         EMIT5("mov r%d, [rBX%+d] ;&%.*s", tos, TOSVAR.offset, nChars, idp);
         X86_emit_op_TM(MOV_V, tos, ebx, no_reg, 0, TOSVAR.offset);
         TOSVAR.offset = *(UINT4 *)((UINT1 *)gFmcp + TOSVAR.offset);   //for &func -- needed??
     }
     else if (TOSVAR.typeindex == 2468) {
         // magic number for int variable based on function call
         // I.e., errno = (*_errno()) is handled via this kludge for now...
         TOSVAR.typeindex = Mdg_current_default_type_index; // always int for now.
		 TOSVAR.nbytes = 4;	//default int size
		 if (X64gen) {
			 EMIT("sub RSP, 32 ;alloc arglist");		//X64: alloc 32 bytes min arglist
			 X86_emit_op_R_simm32(SUBI_V, rsp, 32);
		 }
         EMIT4("call [rBX%+d] ;&_%.*s", TOSVAR.offset, nChars, idp);
         X86_emit_op_M(CALLI_N, ebx, no_reg, 0, TOSVAR.offset);
         EMIT4("mov r%d, rAX ;&%.*s", tos, nChars, idp);
         X86_emit_op_TR(MOV_V, tos, eax);	//X64??
		 if (X64gen) {
			 //X64: delete the arglist
			 //(maybe not needed if static arglist??)
			 EMIT("add RSP, 32 ;delete arglist");		//32 = min stack shadow space
			 X86_emit_op_R_simm32(ADDI_V, rsp, 32);
		 }
     }
     else {
         //variable
		 if (TOSVAR.nbytes == 0) {
			 //size not explicitly set;
			 //set default size from variable type
			 if (TOSVAR.typeindex == 0) {
				 //int: default size is 4 bytes
				 TOSVAR.nbytes = 4;
			 }
			 else if (TOSVAR.typeindex == 126) {
				 //real: default size is 8 bytes
				 TOSVAR.nbytes = 8;
			 }
		 }
         EMIT5("lea r%d, [rBX%+d] ;&%.*s", tos, TOSVAR.offset, nChars, idp);
         X86_emit_op_TM(LEA_V, tos, ebx, no_reg, 0, TOSVAR.offset);	//X64
     }
 } /*lookupid*/


        /************************************/
        /*                                  */
        /*                                  */
        /*             lookup               */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void lookup(void)
 {
     ensureoneoperand();

     if (TOSOBJECT.kind == IDOBJECTKIND)
         lookupid();
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*lookup*/


        /************************************/
        /*                                  */
        /*                                  */
        /*         get$type$length          */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static UINT4 gettypelength(UINT4 typeindex)
 {
     // Warn about dubious type (e.g., LABEL) ???
     if (typeindex == 126)
         return 8;      /* real */
     else
         return 4;  /* default type is int4 */	//X64!!
 } /*gettypelength*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            duplicate             */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void duplicate(void)
 {
     ensureoneoperand();
     push(0);
     *tosp = *nosp;
     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL) {
		 check_ndp_stack_room(1);		//need room for 1 more ndp stack entry
         EMIT("fld st(0) ;op_dupl");
         X86_emit_op_R(FLD_STN, 0);
     }
     else if (TOSOBJECT.kind == LITOBJECTKIND || TOSOBJECT.kind == VAROBJECTKIND) {
         EMIT3("mov_v r%d, r%d ;op_dupl", tos, tos-1);
         X86_emit_op_TT(MOV_V, tos, tos-1);
     }
 } /*duplicate*/


        /************************************/
        /*                                  */
        /*                                  */
        /*           physical$address       */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void physicaladdress(void)
 {
     ensureoneoperand();

     if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
     {
         TOSVAR.offset = integervalue(tosp);       /* for now */
         TOSVAR.frame = 0;                         /* for now */
         TOSOBJECT.kind = VAROBJECTKIND;
         TOSVAR.typeindex = Mdg_current_default_type_index;
     }
     else 
         Mde_abort_command(MD386__BADMIX, 0, 0, 0);
 } /*physicaladdress*/


        /************************************/
        /*                                  */
        /*                                  */
        /*            mark$stack            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void markstack(void)
 {
     push(MARKEROBJECTKIND);
 } /*markstack*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          advance_sdlp            */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void advance_sdlp(void)
 {
     M3__sdlp = (void *) ((UINT1 *) M3__sdlp + 4 + (SDLITEM.nbytes + 3)/4*4);

     sdlop = 0;

     if (SDLITEM.kind <= 0 || SDLITEM.kind > SDKIND(LAST_VALID_SD)) {
         /* End of Semantic Descriptor List */
         eosdl = TRUE;
     }
     else if (SDLITEM.kind == 3) {
         /* operator */
         sdlop = SDLITEM.type;
         if (sdlop == SDTYPE(OP_THEN)) {
             /* OP_THEN */
             if_depth++;
             if (if_depth > sizeof(if_info)/sizeof(if_info[0])) bug(24);
         }
         else if (sdlop == SDTYPE(OP_IF)) {
             /* OP_IF */
             if (if_depth <= 0) bug(13);
             if_depth--;
         }
     }
 } /*advance_sdlp*/

 
        /************************************/
        /*                                  */
        /*    Iteration statement stack.    */
        /*                                  */
        /************************************/

 typedef struct {
     int    iVal;
     CGL    cgl;
 } CaseInfo_ts;

 typedef struct Loop_tag
 {
     struct Loop_tag *prev_loop;
     struct Loop_tag *curr_switch;
     CGL    cglTop;
     CGL    cglContinue;
     CGL    cglBreak;
     CGL    cglDefault;
     CGL    cglJumper;
     CGL    cglJtable;
     CB*    prev_cb;
     CB     cbE3;
     union {            //Caution!!!
         BYTE   E3_Code[2048];              //used by for-loops
         CaseInfo_ts Case[MAX_CASE_LABELS]; //used by switch statements
     };
     int    nCases;
     int    minCase;
     int    maxCase;
 } Loop_t;


 static Loop_t Loop = {NULL, NULL, -1, -1, -1, -1, -1, -1};
 static CGL cglReturn = -1;
 static CB *cbpPrevious = NULL;
 static CB *cbpOriginal = NULL;
 static BOOL bReturnEmitted = FALSE;

 /* Case-comparison routine for use by qsort() below */

 static int compareCase(const void *case1, const void *case2) {
#if 1
     return (((CaseInfo_ts *)case1)->iVal < ((CaseInfo_ts *)case2)->iVal) ? -1 : (((CaseInfo_ts *)case1)->iVal > ((CaseInfo_ts *)case2)->iVal) ? 1 : 0;
#else //would the following work??  Maybe not if tne subtraction overflows!!
	 return ((CaseInfo_ts *)case1)->iVal - ((CaseInfo_ts *)case2)->iVal;
#endif
 }

 
static void push_loop_context(void)
 {
     Loop_t *newLoop = calloc(1, sizeof(Loop_t));
     if (newLoop == NULL) {
         X86_abort2("Not enough memory (%d. bytes) for iteration statement nesting",
					sizeof(Loop_t));
     }
     else {
         *newLoop = Loop;
         Loop.prev_loop = newLoop;
         Loop.cglTop = -1;  //for debugging
         //Loop.cglContinue = -1;   //SWITCH inherits outer CONTINUE
         Loop.cglBreak = -1;    //for debugging
         Loop.cglDefault = -1;    //for debugging
         Loop.nCases = 0;
         Loop.minCase = MAXLONG;
         Loop.maxCase = MINLONG;
         if (Loop.curr_switch == &Loop) {
             //old context was a switch
             Loop.curr_switch = newLoop;
         }
         Loop.cbE3.cgls = NULL;     //BUG FIX: don't inherit CGLs!
     }
 } /*push_loop_context*/

 static void pop_loop_context(void)
 {
     Loop_t *prev_loop = Loop.prev_loop;
     if (prev_loop == NULL) bug(24357);
     Loop = *prev_loop;
     free(prev_loop);
 } /*pop_loop_context*/

 static void init_loop_context(void)
 {
     //release any previously allocated loop contexts
     //in case we aborted without popping back to initial level...
     while (Loop.prev_loop)
         pop_loop_context();
     Loop.curr_switch = NULL;
     Loop.cglTop = -1;
     Loop.cglContinue = -1;
     Loop.cglBreak = -1;
     Loop.cglDefault = -1;

     cglReturn = -1;
     cbpPrevious = NULL;
 } /*init_loop_context*/

 static void do_binary_search(int lo, int hi)	//x64??
 {
     // Do a binary search on the cases between case lo and case hi.
     int ncases = hi - lo + 1;
     assert(ncases >= 1);
     if (ncases > 2) {
         //round pivot up, so # lower cases >= # higher cases
         int pivot = (lo + hi + 1)/2;
         CGL cglHigh = X86_genlab();

         EMIT2("cmp r0, %d", Loop.Case[pivot].iVal);
         X86_emit_op_T_simm32(CMPI_D, 0, Loop.Case[pivot].iVal);	//X64??
         EMIT2("jg cgl%d", cglHigh);
         X86_emit_op_lab(JG, cglHigh);
         EMIT2("je cgl%d", Loop.Case[pivot].cgl);
         X86_emit_op_lab(JE, Loop.Case[pivot].cgl);
         do_binary_search(lo, pivot-1);
         EMIT2L("cgl%d:", cglHigh);
         X86_deflab(cglHigh);
         do_binary_search(pivot+1, hi);
     }
     else {
         // ncases = 1 or 2
         if (ncases == 2) {
             EMIT2("cmp r0, %d", Loop.Case[hi].iVal);
             X86_emit_op_T_simm32(CMPI_D, 0, Loop.Case[hi].iVal);	//X64??
             EMIT2("je cgl%d", Loop.Case[hi].cgl);
             X86_emit_op_lab(JE, Loop.Case[hi].cgl);
         }
         EMIT2("cmp r0, %d", Loop.Case[lo].iVal);
         X86_emit_op_T_simm32(CMPI_D, 0, Loop.Case[lo].iVal);	//X64??
         EMIT2("je cgl%d", Loop.Case[lo].cgl);
         X86_emit_op_lab(JE, Loop.Case[lo].cgl);
         EMIT2("jmp cgl%d ;default case", Loop.cglDefault);
         X86_emit_op_lab(JMP, Loop.cglDefault);
     }
 } /*do_binary_search*/


        /************************************/
        /*                                  */
        /*                                  */
        /*         perform$operation        */
        /*                                  */
        /*                                  */
        /************************************/

 static int nTemps = MAX_TEMPS;		 ///NO LONGER NEEDED!!!
 static int cbLocals = MAX_LOCALS*4; ///NO LONGER NEEDED!!!

 /**procedure**/
 static void performoperation(void)
 {
     UINT1  op;
     int i;
     static CGL cglSkip = -1;
     static pfm_OBJ pVar;

     op = SDLITEM.type;

     switch (op)
     {
       case SDTYPE(OP_ADD):
           add();
           break;
       case SDTYPE(OP_SUB):
           subtract();
           break;
       case SDTYPE(OP_NEG):
           negate();
           break;
       case SDTYPE(OP_LOAD):
           dereference();
           break;
       case SDTYPE(OP_STORE):
           store();
           break;
#if 0
       case SDTYPE(OP_QUAL):
           qualify();
           break;
#endif
       case SDTYPE(OP_LOOKUP):
           lookup();
           break;
       case SDTYPE(OP_DUPL):
           duplicate();
           break;
       case SDTYPE(OP_PHYS):
           physicaladdress();
           break;
       case SDTYPE(OP_MARK):
           markstack();
           break;
       case SDTYPE(OP_VAR2TYPE):
           settypefromvar();
           break;
       case SDTYPE(OP_LSH):
           shiftleft();
           break;
       case SDTYPE(OP_RSH):
           shiftright();
           break;
       case SDTYPE(OP_MUL):
           multiply();
           break;
       case SDTYPE(OP_DIV):
           divide();
           break;
       case SDTYPE(OP_REM):
           remainder();
           break;
#if 0
       case SDTYPE(OP_ASS_ADD):
           assign_add();
           break;
       case SDTYPE(OP_ASS_SUB):
           assign_sub();
           break;
       case SDTYPE(OP_ASS_MUL):
           assign_mul();
           break;
       case SDTYPE(OP_ASS_DIV):
           assign_div();
           break;
       case SDTYPE(OP_ASS_REM):
           assign_rem();
           break;
       case SDTYPE(OP_ASS_AND):
           assign_and();
           break;
       case SDTYPE(OP_ASS_XOR):
           assign_xor();
           break;
       case SDTYPE(OP_ASS_IOR):
           assign_ior();
           break;
       case SDTYPE(OP_ASS_LSH):
           assign_lsh();
           break;
       case SDTYPE(OP_ASS_RSH):
           assign_rsh();
           break;
       case SDTYPE(OP_ASS_POW):
           assign_pow();
           break;
#endif
       case SDTYPE(OP_LOG_AND):
           logical_and();
           break;
       case SDTYPE(OP_LOG_OR):
           logical_or();
           break;
       case SDTYPE(OP_LOG_NOT):
           logical_not();
           break;
       case SDTYPE(OP_IF):
           op_if();
           break;
       case SDTYPE(OP_AND):
           op_and();
           break;
       case SDTYPE(OP_IOR):
           op_ior();
           break;
       case SDTYPE(OP_XOR):
           op_xor();
           break;
       case SDTYPE(OP_NOT):
           op_not();
           break;
       case SDTYPE(OP_POSTINC):
           op_postinc();
           break;
       case SDTYPE(OP_POSTDEC):
           op_postdec();
           break;
       case SDTYPE(OP_PREINC):
           op_preinc();
           break;
       case SDTYPE(OP_PREDEC):
           op_predec();
           break;
       case SDTYPE(OP_ARGLIST):
           op_arglist();
           break;
       case SDTYPE(OP_ARG):
           op_arg();
           break;
       case SDTYPE(OP_CALL):
           op_call();
           break;
       case SDTYPE(OP_SIZEOF):
           op_sizeof();
           break;
#if 0
       case SDTYPE(OP_ADDR):
           op_addr();
           break;
#endif
       case SDTYPE(OP_NOP):
           /* op_nop(); */ /*highly optimized!*/
           break;
       case SDTYPE(OP_EQ):
           op_eq();
           break;
       case SDTYPE(OP_NE):
           op_ne();
           break;
       case SDTYPE(OP_LT):
           op_lt();
           break;
       case SDTYPE(OP_GT):
           op_gt();
           break;
       case SDTYPE(OP_LE):
           op_le();
           break;
       case SDTYPE(OP_GE):
           op_ge();
           break;
#if 0
       case SDTYPE(OP_CAST):
           op_cast();
           break;
#endif
       case SDTYPE(OP_THEN):
           op_then();
           break;
       case SDTYPE(OP_ELSE):
           op_else();
           break;
       case SDTYPE(OP_POP):
           pop();
           break;
       case SDTYPE(OP_DEC_STACK):
           dec_stack();
           break;
       case SDTYPE(OP_XCHG):
           op_xchg();
           break;
       case SDTYPE(OP_SETCHAN_R):
           gFmcp->z = 0;                    // eh?  Of any use at compile time??
           gFmcp->c = gFmcp->r;             // eh?  Of any use at compile time??
           EMIT("<Start Channel R>");
           cbpPrevious = X86_switch_to(&cbChannelCode);
           if (cbpPrevious != &cbInitCode &&
               cbpPrevious != &cbChannelCode) {
               Mde_error(MD386__ILLRGBA, 0, 0, 0);
           }
           EMIT("mov z, 0");
           X86_emit_op_M_imm32(MOVI_D, FMC(z), 0);
           if (gFmcp->needPremiere) {
               EMIT("mov p, 0");
               X86_emit_op_M_imm32(MOVI_D, FMC(p), 0);
           }
           EMIT("mov c, r");
           X86_emit_op_RM(MOV_D, eax, FMC(r));
           X86_emit_op_MR(MOV_D, FMC(c), eax);
           if (gFmcp->needPremiere) {
               EMIT("mov c0, r");
               X86_emit_op_MR(MOV_D, FMC(c0), eax);
               EMIT("mov c1, r");
               X86_emit_op_MR(MOV_D, FMC(c1), eax);
           }
#ifndef NO_DEBUG_LABELS
           X86_deflab(X86_genlab());    //superfluous label for debugging
#endif
           break;
       case SDTYPE(OP_SETCHAN_G):       //fold other cases with above, using macro or common subroutine??
           gFmcp->z = 1;
           gFmcp->c = gFmcp->g;
           EMIT("<Start Channel G>");
           cbpPrevious = X86_switch_to(&cbChannelCode);
           if (cbpPrevious != &cbInitCode &&
               cbpPrevious != &cbChannelCode) {
               Mde_error(MD386__ILLRGBA, 0, 0, 0);
           }
           EMIT("mov z, 1");
           X86_emit_op_M_imm32(MOVI_D, FMC(z), 1);
           if (gFmcp->needPremiere) {
               EMIT("mov p, 1");
               X86_emit_op_M_imm32(MOVI_D, FMC(p), 1);
           }
           EMIT("mov c, g");
           X86_emit_op_RM(MOV_D, eax, FMC(g));
           X86_emit_op_MR(MOV_D, FMC(c), eax);
           if (gFmcp->needPremiere) {
               EMIT("mov c0, g");
               X86_emit_op_MR(MOV_D, FMC(c0), eax);
               EMIT("mov c1, g");
               X86_emit_op_MR(MOV_D, FMC(c1), eax);
           }
#ifndef NO_DEBUG_LABELS
           X86_deflab(X86_genlab());    //superfluous label for debugging
#endif
           break;
       case SDTYPE(OP_SETCHAN_B):
           gFmcp->z = 2;
           gFmcp->c = gFmcp->b;
           EMIT("<Start Channel B>");
           cbpPrevious = X86_switch_to(&cbChannelCode);
           if (cbpPrevious != &cbInitCode &&
               cbpPrevious != &cbChannelCode) {
               Mde_error(MD386__ILLRGBA, 0, 0, 0);
           }
           EMIT("mov z, 2");
           X86_emit_op_M_imm32(MOVI_D, FMC(z), 2);
           if (gFmcp->needPremiere) {
               EMIT("mov p, 2");
               X86_emit_op_M_imm32(MOVI_D, FMC(p), 2);
           }
           EMIT("mov c, b");
           X86_emit_op_RM(MOV_D, eax, FMC(b));
           X86_emit_op_MR(MOV_D, FMC(c), eax);
           if (gFmcp->needPremiere) {
               EMIT("mov c0, b");
               X86_emit_op_MR(MOV_D, FMC(c0), eax);
               EMIT("mov c1, b");
               X86_emit_op_MR(MOV_D, FMC(c1), eax);
           }
#ifndef NO_DEBUG_LABELS
           X86_deflab(X86_genlab());    //superfluous label for debugging
#endif
           break;
       case SDTYPE(OP_SETCHAN_A):
           gFmcp->z = 3;
           gFmcp->c = gFmcp->a;
           EMIT("<Start Channel A>");
           cbpPrevious = X86_switch_to(&cbChannelCode);
           if (cbpPrevious != &cbInitCode &&
               cbpPrevious != &cbChannelCode) {
               Mde_error(MD386__ILLRGBA, 0, 0, 0);
           }
           EMIT("mov z, 3");
           X86_emit_op_M_imm32(MOVI_D, FMC(z), 3);
           if (gFmcp->needPremiere) {
               EMIT("mov p, 3");
               X86_emit_op_M_imm32(MOVI_D, FMC(p), 3);
           }
           EMIT("mov c, a");
           X86_emit_op_RM(MOV_D, eax, FMC(a));
           X86_emit_op_MR(MOV_D, FMC(c), eax);
           if (gFmcp->needPremiere) {
               EMIT("mov c0, a");
               X86_emit_op_MR(MOV_D, FMC(c0), eax);
               EMIT("mov c1, a");
               X86_emit_op_MR(MOV_D, FMC(c1), eax);
           }
#ifndef NO_DEBUG_LABELS
           X86_deflab(X86_genlab());    //superfluous label for debugging
#endif
           break;
       case SDTYPE(OP_SETVAL_R):
           //assert(tos == 0);
           //assert(tos == lit_integer)
#ifndef NO_DEBUG_LABELS
           X86_deflab(X86_genlab());    //superfluous label for debugging
#endif
           EMIT2("mov R, r%d", tos);
           EMIT("<End Channel R>");
           X86_emit_op_MT(MOV_D, FMC(R), tos);
           X86_switch_to(cbpPrevious);
           pop();
           break;
       case SDTYPE(OP_SETVAL_G):
           //assert(tos == 0);
           //assert(tos == lit_integer)
#ifndef NO_DEBUG_LABELS
           X86_deflab(X86_genlab());    //superfluous label for debugging
#endif
           EMIT2("mov G, r%d", tos);
           EMIT("<End Channel G>");
           X86_emit_op_MT(MOV_D, FMC(G), tos);
           X86_switch_to(cbpPrevious);
           pop();
           break;
       case SDTYPE(OP_SETVAL_B):
           //assert(tos == 0);
           //assert(tos == lit_integer)
#ifndef NO_DEBUG_LABELS
           X86_deflab(X86_genlab());    //superfluous label for debugging
#endif
           EMIT2("mov B, r%d", tos);
           EMIT("<End Channel B>");
           X86_emit_op_MT(MOV_D, FMC(B), tos);
           X86_switch_to(cbpPrevious);
           pop();
           break;
       case SDTYPE(OP_SETVAL_A):
           //assert(tos == 0);
           //assert(tos == lit_integer)
#ifndef NO_DEBUG_LABELS
           X86_deflab(X86_genlab());    //superfluous label for debugging
#endif
           EMIT2("mov A, r%d", tos);
           EMIT("<End Channel A>");
           X86_emit_op_MT(MOV_D, FMC(A), tos);
           X86_switch_to(cbpPrevious);
           pop();
           break;
       case SDTYPE(OP_SAVE_CHANNEL_CODE):
           EMIT("<Save Channel Code>");
           cbpPrevious = X86_emit_init(&cbSaveChannelCode, CB_saveChannelCode,
                                       sizeof(CB_saveChannelCode), "Saved channel");
           tos = -1;  //force empty stack in case previous code gen didn't clean up properly
           break;
       case SDTYPE(OP_END_CHANNEL_CODE):
           EMIT0("<End Channel Code>");
           //assert(tos == 0);
           //assert(tos == lit_integer)
           //if (tos != 0) bug(2647);
           if (tos > 0) bug(2647);  //for now, until get all the right pop's
           X86_emit_finish();   //Finish off the SaveChannelCode buffer
           X86_switch_to(cbpPrevious);
           break;
       case SDTYPE(OP_DUMP_CHANNEL_CODE):
           EMIT("<Dump Channel Code>");
           X86_append_cb(&cbSaveChannelCode);
           tos = 0;     //force 1 entry, which is how channel code ended
           break;
       case SDTYPE(OP_GET):
           op_get();
           break;
       case SDTYPE(OP_PUT):
           op_put();
           break;
       case SDTYPE(OP_MIN):
           op_min();
           break;
       case SDTYPE(OP_MAX):
           op_max();
           break;
       case SDTYPE(OP_CTL):
           op_ctl();
           break;
#if 0
       case SDTYPE(OP_VAL):
           op_val();
           break;
#endif
       case SDTYPE(OP_SQR):
           op_sqr();
           break;
       case SDTYPE(OP_ABS):
           op_abs();
           break;
       case SDTYPE(OP_DIF):
           op_dif();
           break;
       case SDTYPE(OP_SIN):
           op_sin();
           break;
       case SDTYPE(OP_COS):
           op_cos();
           break;
       case SDTYPE(OP_TAN):
           op_tan();
           break;
       case SDTYPE(OP_CNVX):
           op_cnvX_Y(0);
           break;
       case SDTYPE(OP_CNVY):
           op_cnvX_Y(1);
           break;
       case SDTYPE(OP_GAMMA):
           op_gamma();
           break;
       case SDTYPE(OP_FLOOR):
           op_floor();
           break;
       case SDTYPE(OP_CEIL):
           op_ceil();
           break;
       case SDTYPE(OP_ROUND):
           op_round();
           break;
       case SDTYPE(OP_CHOP):
           op_chop();
           break;
       case SDTYPE(OP_POW):
           op_pow();
           break;
       case SDTYPE(OP_FOR_EVERY_PIXEL):
           assert(tos == -1);
           EMIT0("<ForEveryPixel:>");
           X86_switch_to(&cbChannelCode);
           break;
       case SDTYPE(OP_FOR_EVERY_PIXEL_END):
           assert(tos == 0 || tos == -1);
           EMIT2("<End ForEveryPixel> (tos=%d)", tos);
           X86_switch_to(&cbInitCode);
           //now force tos to be empty...
           tos = -1;
           break;
       case SDTYPE(OP_FOR_EVERY_TILE):
           EMIT0("<ForEveryTile:>");
           X86_emit_init(&cbForEveryTile, gParams->CB_forEveryTile,
                         sizeof(gParams->CB_forEveryTile), "ForEveryTile");
           EMIT("<prolog to load rBX>");
           X86_emit_prolog(nTemps, cbLocals);
           cglReturn = X86_genlab();
           bReturnEmitted = FALSE;
           break;
       case SDTYPE(OP_FOR_EVERY_END):
       case SDTYPE(OP_ON_END):  //should be renamed OP_ON_HANDLER_END ???
           assert(tos == 0 || tos == -1);
           EMIT2("<End external handler> (tos=%d)", tos);
           //force tos to have one element
           tos = 0;
           if (cglReturn == -1) bug(4683);
           if (!bReturnEmitted) {
               if (op == OP_FOR_EVERY_END) {
                   Mde_warn(MD386__NEEDRETURN, (UINT_PTR)"the ForEveryTile handler", 0, 0);
               }
               else {
                   Mde_warn(MD386__NEEDRETURN, (UINT_PTR)"an On-handler", 0, 0);
               }
           }
           EMIT2L("cgl%d:    ;epilog", cglReturn);
           X86_deflab(cglReturn);
           cglReturn = -1;
           EMIT2("mov eax, r%d ;return result in eax", tos);
           EMIT("<External handler epilog>");
           X86_emit_op_RT(MOV_D, eax, tos);
           X86_emit_epilog(nTemps, cbLocals, -1);       //nArgs = 1, vpl
           X86_emit_finish();
           X86_switch_to(&cbInitCode);
           //now force tos to be empty...
           tos = -1;
           //pop(); //ok for int result, but what if FP result???
           break;
       case SDTYPE(OP_ON_HANDLER_END):  //should be renamed OP_ON_FUNCTION_END ???
           //Almost identical to above??? (nargs is different)
           //Could combine????
           assert(tos == 0 || tos == -1);
           EMIT2("<End internal function handler> (tos=%d)", tos);
           //force tos to have one element
           tos = 0;
           if (cglReturn == -1) bug(4683);
           if (!bReturnEmitted) {
               Mde_warn(MD386__NEEDRETURN, (UINT_PTR)"a function body", 0, 0);
           }
           EMIT2L("cgl%d:    ;epilog", cglReturn);
           X86_deflab(cglReturn);
           cglReturn = -1;
           EMIT2("mov eax, r%d ;return result in eax", tos);
           EMIT("<Internal function epilog>");
           X86_emit_op_RT(MOV_D, eax, tos);
           X86_emit_handler_epilog(nTemps, cbLocals, 0);	//nArgs = 0
           X86_emit_finish();                       //resolve labels -- do only once -- clobbers next user func labels???
           X86_switch_to(&cbInitCode);
           //now force tos to be empty...
           tos = -1;
           //pop(); //ok for int result, but what if FP result???
           break;
       case SDTYPE(OP_ON_INIT_DIALOG):
           //op_on_init_dialog();
           bug(67);
           break;
       case SDTYPE(OP_ON_CTL):
           //op_on_ctl();
           EMIT0("<OnCtl:>");
           X86_emit_init(&cbOnCtl, gParams->CB_onCtl, sizeof(gParams->CB_onCtl), "OnCtl");
           X86_emit_prolog(nTemps, cbLocals);
           cglReturn = X86_genlab();
           bReturnEmitted = FALSE;
           break;
       case SDTYPE(OP_ON_FILTER_START):
           //op_on_filter_start();
           EMIT0("<OnFilterStart:>");
           X86_emit_init(&cbOnFilterStart, gParams->CB_onFilterStart,
                         sizeof(gParams->CB_onFilterStart), "OnFilterStart");
           X86_emit_prolog(nTemps, cbLocals);
           cglReturn = X86_genlab();
           bReturnEmitted = FALSE;
           break;
       case SDTYPE(OP_ON_FILTER_END):
           //op_on_filter_end();
           EMIT0("<OnFilterEnd:>");
           X86_emit_init(&cbOnFilterEnd, gParams->CB_onFilterEnd,
                         sizeof(gParams->CB_onFilterEnd), "OnFilterEnd");
           X86_emit_prolog(nTemps, cbLocals);
           cglReturn = X86_genlab();
           bReturnEmitted = FALSE;
           break;
       case SDTYPE(OP_USER_FUNCTION):  //OP_USER_FUNCTION, should switch, not init?
           //op_user_function();
           EMIT2("<UserFunction %d:>", ++(gParams->gUserFunctions));
           X86_switch_to(&cbUserFunctions);
           X86_emit_handler_prolog(nTemps, cbLocals); // correct prolog?  doesn't need to set rBX??? !!!CLOBBERS ecx!!
           cglReturn = X86_genlab();
           bReturnEmitted = FALSE;
           break;
       case SDTYPE(OP_ON_ZERO_DIVIDE):
           //op_on_zero_divide();
           EMIT0("<OnZeroDivide:>");
           X86_emit_init(&cbZeroDivide, gParams->CB_onZeroDivide,
                         sizeof(gParams->CB_onZeroDivide), "OnZeroDivide");
           X86_emit_handler_prolog(nTemps, cbLocals);
           cglReturn = X86_genlab();
           bReturnEmitted = FALSE;
           break;
       case SDTYPE(OP_ON_DIVIDE_OVERFLOW):
           //op_on_divide_overflow();
           EMIT0("<OnDivideOverflow:>");
           X86_emit_init(&cbDivideOverflow, gParams->CB_onDivideOverflow,
                         sizeof(gParams->CB_onDivideOverflow), "OnDivideOverflow");
           X86_emit_handler_prolog(nTemps, cbLocals);
           cglReturn = X86_genlab();
           bReturnEmitted = FALSE;
           break;
       case SDTYPE(OP_FOR1):
           // for ( optional e1;
           push_loop_context();
           Loop.cglTop = X86_genlab();
           Loop.cglContinue = X86_genlab();
           Loop.cglBreak = X86_genlab();
           EMIT2L("cgl%d:    ;top", Loop.cglTop);
           X86_deflab(Loop.cglTop);
           break;
       case SDTYPE(OP_BREAKF):
           // optional e2;
           ensureoneoperand();
           if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
           {
                EMIT2("cmp r%d, 0", tos);
                EMIT2("je cgl%d  ;break", Loop.cglBreak);
                X86_emit_op_T_simm32(CMPI_D, tos, 0);
                X86_emit_op_lab(JE, Loop.cglBreak);
           }
           else
           {
               //for now...
               Mde_abort_command(MD386__BADMIX, 0, 0, 0);
           }
           pop();
           break;
       case SDTYPE(OP_FOR2):
           //save code for e3 in a separate buffer...
           EMIT("<Begin diversion to E3>");
           Loop.prev_cb = X86_emit_init(&Loop.cbE3, Loop.E3_Code, sizeof(Loop.E3_Code), "Loop");
           break;
       case SDTYPE(OP_FOR3):
           // optional e3)
           EMIT("<End diversion to E3>");
           X86_emit_finish();           //finish off the e3 buffer
           X86_switch_to(Loop.prev_cb); // and revert to previous code buffer
           break;
       case SDTYPE(OP_FOR4):
           // s1;
           EMIT2L("cgl%d:    ;continue", Loop.cglContinue);
           X86_deflab(Loop.cglContinue);
           EMIT("<Insert E3 buffer>");
           X86_append_cb(&Loop.cbE3);   // now emit the e3 iterator expression
           // Free up any resources still held by the E3 code buffer...
           X86_emit_free_resources(&Loop.cbE3);
           EMIT2("jmp cgl%d  ;repeat", Loop.cglTop);
           X86_emit_op_lab(JMP, Loop.cglTop);
           EMIT2L("cgl%d:    ;break", Loop.cglBreak);
           X86_deflab(Loop.cglBreak);
           pop_loop_context();
           break;

       case SDTYPE(OP_DO1):     //same as OP_FOR1 ==> OP_LOOP???
           // do
           push_loop_context();
           Loop.cglTop = X86_genlab();
           Loop.cglContinue = X86_genlab();
           Loop.cglBreak = X86_genlab();
           EMIT2L("cgl%d:    ;top", Loop.cglTop);
           X86_deflab(Loop.cglTop);
           break;
       case SDTYPE(OP_DO2):
           // s1
           EMIT2L("cgl%d:    ;continue", Loop.cglContinue);
           X86_deflab(Loop.cglContinue);
           break;
       case SDTYPE(OP_DO3):
           // while (e1);
           ensureoneoperand();
           if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
           {
                EMIT2("cmp r%d, 0", tos);
                EMIT2("jne cgl%d  ;repeat", Loop.cglTop);
                X86_emit_op_T_simm32(CMPI_D, tos, 0);
                X86_emit_op_lab(JNE, Loop.cglTop);
           }
           else
           {
               //for now...
               Mde_abort_command(MD386__BADMIX, 0, 0, 0);
           }
           pop();
           EMIT2L("cgl%d:    ;break", Loop.cglBreak);
           X86_deflab(Loop.cglBreak);
           pop_loop_context();
           break;

       case SDTYPE(OP_SWITCH):
           // switch (e)
           //EMIT("<begin switch>");
           push_loop_context();
           Loop.curr_switch = &Loop;    //new switch context
           Loop.cglJumper = X86_genlab();
           Loop.cglDefault = Loop.cglBreak = X86_genlab();
           //Loop.cglContinue is inherited from outer loop, if any.
           ensureoneoperand();
           if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
           {
               assert(tos == 0);	//must be in r0
           }
           else
           {
               Mde_error(MD386__BADSWITCHVAL, 0, 0, 0);
           }
           EMIT2("jmp cgl%d  ;jumper", Loop.cglJumper);
           X86_emit_op_lab(JMP, Loop.cglJumper);
           pop();
           break;
       case SDTYPE(OP_CASE):
           //nothing to do!!!
           //EMIT("<case>");
           break;
       case SDTYPE(OP_CASETEST):
           // case n:
           //EMIT("<case test>");
           ensureoneoperand();
           if (Loop.curr_switch == NULL) {
               //ErrorOk("case label not within a switch statement"); //for now
               Mde_error(MD386__ILLCASELAB, 0, 0, 0);
           }
           else {
               if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
               {
                    int iCase = integervalue(tosp);
                    CGL cgl;

                    // Check for duplicate case value...
                    for (i = 0; i < Loop.curr_switch->nCases; i++) {
                        if (Loop.curr_switch->Case[i].iVal == iCase) {
                            Mde_error(MD386__DUPCASE, iCase, 0, 0);
                            break;
                        }
                    }

                    if (Loop.curr_switch->nCases < MAX_CASE_LABELS) {
                        if (iCase < Loop.curr_switch->minCase) {
                            Loop.curr_switch->minCase = iCase;
                        }
                        if (iCase > Loop.curr_switch->maxCase) {
                            Loop.curr_switch->maxCase = iCase;
                        }
                        cgl = X86_genlab();
                        EMIT3L("cgl%d:    ;case %d", cgl, iCase);
                        X86_deflab(cgl);
                        Loop.curr_switch->Case[Loop.curr_switch->nCases].cgl = cgl;
                        Loop.curr_switch->Case[Loop.curr_switch->nCases].iVal = iCase;
                        Loop.curr_switch->nCases++;
                    }
                    else {
                        //too many case labels
                        Mde_error(MD386__TOOMANYCASES, MAX_CASE_LABELS, 0, 0);
                    }
               }
               else
               {
                   Mde_error(MD386__BADCASEVAL, 0, 0, 0);
               }
           }
           pop();
           break;
       case SDTYPE(OP_DEFAULT):
           EMIT("<default case>");
           if (Loop.curr_switch == NULL) {
               //ErrorOk("default case label not within a switch statement"); //for now
               Mde_error(MD386__ILLDEFLAB, 0, 0, 0);
           }
           else {
               // check for more than one default: case...
               if (Loop.curr_switch->cglDefault != Loop.curr_switch->cglBreak) {
                   Mde_error(MD386__DUPDEFAULT, 0, 0, 0);
               }
               else {
                   Loop.curr_switch->cglDefault = X86_genlab();
                   EMIT2L("cgl%d:    ;default case", Loop.curr_switch->cglDefault);
                   X86_deflab(Loop.curr_switch->cglDefault);
               }
           }
           break;
       case SDTYPE(OP_ENDSW):
           EMIT("<end of switch>");
           EMIT2("jmp cgl%d  ;break", Loop.cglBreak);
           X86_emit_op_lab(JMP, Loop.cglBreak);
           EMIT2L("cgl%d:    ;jumper", Loop.cglJumper);
           X86_deflab(Loop.cglJumper);
           //EMIT3("<min case = %d, max case = %d>", Loop.minCase, Loop.maxCase);
           // emit the jump table...
           if (Loop.nCases < 5 || Loop.nCases < (Loop.maxCase - Loop.minCase)/3) {//<=make this a compile-time param??
               // For small or sparse cases, use if-then-else logic...
               if (Loop.nCases < 4) {
                   // Don't even bother with a binary search...
                   for (i = 0; i < Loop.nCases; i++) {
                       EMIT2("cmp r0, %d", Loop.Case[i].iVal);
                       X86_emit_op_T_simm32(CMPI_D, 0, Loop.Case[i].iVal);	//X64
                       EMIT2("je cgl%d", Loop.Case[i].cgl);
                       X86_emit_op_lab(JE, Loop.Case[i].cgl);
                   }//for i
                   // jump to default case unless none, in which case
                   // just fall out...
                   if (Loop.cglDefault != Loop.cglBreak) {
                       EMIT2("jmp cgl%d  ;default case", Loop.cglDefault);
                       X86_emit_op_lab(JMP, Loop.cglDefault);
                   }
               }
               else {
                   // Sort the cases and do a binary search...
///#pragma warning( disable : 4090 4028 4024) //dont know why I get these!!!	//Fixed in VS2005???
                   qsort(Loop.Case, Loop.nCases, sizeof(Loop.Case[0]), compareCase);
///#pragma warning( default : 4090 4028 4024)
                   do_binary_search(0, Loop.nCases - 1);
               }
               EMIT2L("cgl%d:    ;break", Loop.cglBreak);
               X86_deflab(Loop.cglBreak);
           }
           else {
               // For dense cases, use a jump table...	//or jmp rel into a table of rel jmps?? -- inherently PIC!!!
               // Sort the cases by case value.
               //IGNORE 5 COMPILER WARNINGS FROM MSVC!!!! -- Fixed in VS2500!!
               int i, j;
               int magic_offset;   //terrible kludge!	//X64!!
///#pragma warning( disable : 4090 4028 4024) //dont know why I get these!!!	//Fixed in VS2005???
               qsort(Loop.Case, Loop.nCases, sizeof(Loop.Case[0]), compareCase);
///#pragma warning( default : 4090 4028 4024)
               EMIT2("cmp r0, %d ;max case", Loop.maxCase);
               X86_emit_op_T_simm32(CMPI_D, 0, Loop.maxCase);	//X64
               EMIT2("jg cgl%d ;default case", Loop.cglDefault);
               X86_emit_op_lab(JG, Loop.cglDefault);
               EMIT2("sub r0, %d ;min case", Loop.minCase);
               X86_emit_op_T_simm32(SUBI_D, 0, Loop.minCase);	//X64
               EMIT2("jl cgl%d ;default case", Loop.cglDefault);	//was overflow possible??
               X86_emit_op_lab(JL, Loop.cglDefault);
               Loop.cglTop = X86_genlab();
               Loop.cglJtable = X86_genlab();
#ifdef X64
			   if (X64gen) {
				   // X64 can use rip-relative addressing, e.g.
				   //    LEA_Q rax,[00000000+RIP] for &cglTop
				   // cglTop:
				   //    MOVSD_Q rdx,(cglJtable-cglTop)[rax+r0*4] for Jtable[r0]
				   // then relocate rdx by &cglTop and jmpi rax:
				   //    ADD_Q rax, rdx
				   //	 JMPI NEAR rax
				   magic_offset = 14;   //terrible kludge!	//X64-specific
				   EMIT1("lea rax, [rip] ;&top");
#if 1	//temporary kludge until X86_emit_op_RM() handles RIP-relative addressing correctly
				   X86_emit_op_RM(LEA_Q, rax, no_reg, no_reg, 0, 0); //use X86 [disp32] for now!!
#else
				   X86_emit_op_RM(LEA_Q, rax, rip, no_reg, 0, 0);
#endif
				   EMIT2L("cgl%d:    ;top", Loop.cglTop);
				   X86_deflab(Loop.cglTop);
				   EMIT3("movsxd rdx, cgl%d-cgl%d[rax+r0*4] ;jtable[r0]", Loop.cglJtable, Loop.cglTop);
				   X86_emit_op_RM(MOVSD_Q, rdx, rax, tmpReg(0), 4, magic_offset);	//X64
				   EMIT1("add rax, rdx");
				   X86_emit_op_RR(ADD_Q, rax, rdx);
				   EMIT("jmp rax");
				   X86_emit_op_R(JMPI_N, rax);
			   } else
#endif //X64
			   {   //X86 must use the call $+n/pop eax trick
				   //to produce PIC code:
				   //    CALL cglTop
				   // cglTop:
				   //    POP eax for &cglTop
				   // fetch and relocate the case table offset
				   //    ADD eax,(cglJtable-cglTop)[eax+r0*4] for Jtable[r0]
				   // then jmpi eax:
				   //	 JMPI NEAR eax
				   magic_offset = 11;   //x32-specific
				   EMIT2("call cgl%d ;top", Loop.cglTop);
				   X86_emit_op_lab(CALL, Loop.cglTop);
				   EMIT2L("cgl%d:    ;top", Loop.cglTop);
				   X86_deflab(Loop.cglTop);
				   EMIT("pop eax ;&top");
				   X86_emit_opreg(POP_REG, eax);
				   EMIT3("add eax, cgl%d-cgl%d[eax+r0*4] ;jtable[r0]", Loop.cglJtable, Loop.cglTop);
				   X86_emit_op_RM(ADD_D, eax, eax, tmpReg(0), 4, magic_offset);	//X64
				   EMIT("jmp eax");
				   X86_emit_op_R(JMPI_N, eax);
			   } //X86
        //align the jump table??? -- but appended buffers might not be aligned???
               EMIT("ud2 ;<jump table>");
               X86_emit_op(UD2);
               EMIT2("dw %d ;<number of dwords to follow>", Loop.maxCase - Loop.minCase + 1);
               X86_emit_word((WORD)(Loop.maxCase - Loop.minCase + 1));
               EMIT2L("cgl%d:    ;jtable", Loop.cglJtable);
               X86_deflab(Loop.cglJtable);
               //assert(X86_lab_offset(Loop.cglJtable) - X86_lab_offset(Loop.cglTop) == magic_offset);
#if 1
               //ISSUE AN ERROR IF ABOVE ASSERT WOULD FAIL!!!
               EMIT3("<magic_offset=%d; should be %d>", magic_offset,
                     X86_lab_offset(Loop.cglJtable) - X86_lab_offset(Loop.cglTop));
               if (X86_lab_offset(Loop.cglJtable) - X86_lab_offset(Loop.cglTop) != magic_offset) {
                   bug(8936);
               }
#endif
               //HMMMM ... in most (if not all) cases, could use a word table
               //instead of a dword table....should be able to tell if word
			   //table is usable by finding max case offset from cglTop
               j = 0;
               for (i = Loop.minCase; i <= Loop.maxCase; i++) {
                   CGL cgl;
                   if (Loop.Case[j].iVal == i) {
                       cgl = Loop.Case[j].cgl;
                       j++;
                       EMIT4("dd cgl%d-cgl%d ;case %d", cgl, Loop.cglTop, i);
                       X86_emit_dword(X86_lab_offset(cgl) - X86_lab_offset(Loop.cglTop));
                   }
                   else {
                       //OOPS! cglDefault is not defined yet if it defaulted to cglBreak!!!
                       //This will cause X86_lab_offset() to fail!!!
                       //We can jump BACKWARDS to the jump-cglBreak at the end of last case,
                       //or we could compute the absolute offset to cglBreak at the end
                       //of the dword jump table...
                       cgl = Loop.cglDefault;
                       EMIT3("dd cgl%d-cgl%d ;default", cgl, Loop.cglTop);
                       if (Loop.cglDefault == Loop.cglBreak) {
                           //special case: cglBreak is not defined yet, so we
                           //compute the offset the hard way...
                           X86_emit_dword(magic_offset + 4*(Loop.maxCase - Loop.minCase + 1));
                       }
                       else {
                           X86_emit_dword(X86_lab_offset(cgl) - X86_lab_offset(Loop.cglTop));
                       }
                   }
               }//for i
               EMIT2L("cgl%d:    ;break", Loop.cglBreak);
               X86_deflab(Loop.cglBreak);
#if 1
               //make sure we correctly computed the absolute offset to cglBreak above...
               if (X86_lab_offset(Loop.cglBreak) - X86_lab_offset(Loop.cglTop) !=
                   magic_offset + 4*(Loop.maxCase - Loop.minCase + 1)) {
                   ErrorOk("Incorrectly computed magic offset to cglBreak for default case.\n"
                           "Should have been %d, but was computed as %d.",
                           X86_lab_offset(Loop.cglBreak) - X86_lab_offset(Loop.cglTop),
                           magic_offset + 4*(Loop.maxCase - Loop.minCase + 1));
                   bug(9836);
               }
#endif
           }//if using jump table
           pop_loop_context();
           break;

       case SDTYPE(OP_BREAK):
           if (Loop.cglBreak == -1) {
               //Not within a breakable statement!
               Mde_error(MD386__ILLBREAK, 0, 0, 0);
           }
           else {
               EMIT2("jmp cgl%d  ;break", Loop.cglBreak);
               X86_emit_op_lab(JMP, Loop.cglBreak);
           }
           break;

       case SDTYPE(OP_CONTINUE):
           if (Loop.cglContinue == -1) {
               //Not within a continuable statement!
               Mde_error(MD386__ILLCONTINUE, 0, 0, 0);
           }
           else {
               EMIT2("jmp cgl%d  ;continue", Loop.cglContinue);
               X86_emit_op_lab(JMP, Loop.cglContinue);
           }
           break;

       case SDTYPE(OP_RETURN):
           if (tos > 0) bug(9467);
           if (cglReturn == -1) {
               //Not within a function body
               Mde_error(MD386__ILLRETURN, 0, 0, 0);
           }
           else {
               EMIT2("jmp cgl%d  ;return", cglReturn);
               X86_emit_op_lab(JMP, cglReturn);
               bReturnEmitted = TRUE;
           }
           if (tos == 0) {
               // return <expression> ;
               ensureoneoperand();
               if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER)
               {
                   //For now, we require the return type to be int
               }
               else
               {
                   Mde_error(MD386__BADMIX, 0, 0, 0);
               }
               pop();   //clear the stack (after the jmp, in case result is on FP stack!)
           }
           else {
               // return ;
               assert(tos == -1);
               // For now, we always require a return value...
               Mde_error(MD386__NEEDRETVAL, 0, 0, 0);
           }
           break;

       case SDTYPE(OP_LABEL):
           //statement labels are simply ignored
           Mde_warn(MD386__IGNLAB, 0, 0, 0);
           break;

       case SDTYPE(OP_SHORT):
       case SDTYPE(OP_INT):
       case SDTYPE(OP_LONG):
       case SDTYPE(OP_BOOL):
#if X64
	   // Not really right yet for X64 !!!!
	   case SDTYPE(OP_INT64):
	   case SDTYPE(OP_QWORD):
#endif //X64
           // all of these are treated as OP_INT for now...
           ensureoneoperand();
           if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL) {
			   //should use FSTP_L for 64-bits !!!
               EMIT2("fstp_i r%d  ;op_int", tos);
			   check_ndp_stack_room(-1);		//popped 1 entry from ndp stack
               X86_emit_op_Tm(FSTP_I, tos);
               TOSLIT.type = LITTYPEINTEGER;
               TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue; //to prevent GPF later
           }
           else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER) {
               //no conversion necessary //WRONG for X64 !!!
           }
#if 1   //beta testing in 1.0 9c...
           else if (TOSOBJECT.kind == VAROBJECTKIND) {
               //convert a pointer into an int...
               //no conversion of representation necessary
               TOSOBJECT.kind = LITOBJECTKIND;
               TOSLIT.type = LITTYPEINTEGER;
               TOSLIT.nbytes = 4;
               TOSLIT.ivalue = 0;
               TOSLIT.valuep = (UINT1 *)&TOSLIT.ivalue; //to prevent GPF later
           }
#endif
           else
               Mde_abort_command(MD386__BADMIX, 0, 0, 0);
           break;

       case SDTYPE(OP_FLOAT):
       case SDTYPE(OP_DOUBLE):
           // all of these are treated as OP_DOUBLE for now...
           ensureoneoperand();
           if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER) {
			   //Cast integer to real
			   //For now, we ignore precision issues...
			   if (X64gen) {
					check_ndp_stack_room(1);					//need room for 1 ndp stack entry
					EMIT2("fld_l r%d  ;op_double", tos);		//X64: opload FLD_L (64-bit temp)
					X86_emit_op_Tm(FLD_L, tos);
			   }
			   else { //X32
					check_ndp_stack_room(1);					//need room for 1 ndp stack entry
					EMIT2("fld_i r%d  ;op_double", tos);		//X32: opload FLD_I (32-bit temp)
					X86_emit_op_Tm(FLD_I, tos);
			   }
			   TOSLIT.type = LITTYPEREAL;
			   TOSLIT.valuep = (UINT1 *)&TOSLIT.rvalue;	//to prevent GPF later
           }
           else if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEREAL) {
			   //Cast real to real
			   //For now, we ignore precision issues,
               //so no conversion is necessary
           }
           else
               Mde_abort_command(MD386__BADMIX, 0, 0, 0);
           break;

       case SDTYPE(OP_ABOUT):
           //temporary until full About: key is implemented...
           ensureoneoperand();
           if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPECHAR) {
               strncpy(gParams->filterAboutText, TOSLIT.valuep,
                       sizeof(gParams->filterAboutText)-1);
               //ensure string is null-terminated
               gParams->filterAboutText[sizeof(gParams->filterAboutText)-1] = '\0';
           }
           else
               bug(1736);
           pop();
           break;

       case SDTYPE(OP_EMBED_BITMAP):
       case SDTYPE(OP_EMBED_ICON):
       case SDTYPE(OP_EMBED_METAFILE):
       case SDTYPE(OP_EMBED_WAVE):
	   case SDTYPE(OP_EMBED_OTHER):
       case SDTYPE(OP_EMBED_CURSOR):
           //temporary until full Embed: key is implemented...
           ensureoneoperand();
           if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPECHAR) {
               int cbNeeded = (int)strlen(TOSLIT.valuep) + 2;
               if (gParams->cbEmbeds + cbNeeded > sizeof(gParams->embeds)) {
                   ErrorOk("No room in string space for embedded resource:\n%s",
                           TOSLIT.valuep);
               }
               else {
                   //Append resource type indicator and filename string to embedded
                   //resource info...
                   gParams->embeds[gParams->cbEmbeds-1] =
                       op == SDTYPE(OP_EMBED_BITMAP)    ? 'B' :   //type=Bitmap
                       op == SDTYPE(OP_EMBED_ICON)      ? 'I' :   //type=Icon
                       op == SDTYPE(OP_EMBED_METAFILE)  ? 'W' :   //type=Wmf
                       op == SDTYPE(OP_EMBED_WAVE)      ? 'A' :   //type=Audio
				       op == SDTYPE(OP_EMBED_OTHER)     ? 'O' :   //type=Other
                       op == SDTYPE(OP_EMBED_CURSOR)    ? 'C' :   //type=Cursor
                       '?';
                   strcpy(&gParams->embeds[gParams->cbEmbeds], TOSLIT.valuep);
                   gParams->cbEmbeds += cbNeeded;
                   gParams->embeds[gParams->cbEmbeds-1] = '\0';  //end of embeds
                   if (op == SDTYPE(OP_EMBED_WAVE)) {
                       gParams->gWavesEmbedded++;
                   }
                   else {
                       gParams->gImagesEmbedded++;
                   }
               }
           }
           else
               bug(1737);
           pop();
           break;

       /********** Declarations... ************/

       case SDTYPE(OP_DECLARE):
           push(TYPOBJECTKIND);					//ties up r0 -- any way around that???
           TOSTYP.pType = newUNDEFtype();
           // remember which CB we were orginally emitting to...
           cbpOriginal = X86_switch_to(NULL);
           break;

       case SDTYPE(OP_END_DECL):
           assert(tos >= 0);
           pop();   //pop the base type
           if (X86_switch_to(NULL) != cbpOriginal) {
               // return to emitting code in original CB...
               EMIT0("<Revert to original CB>");
               X86_switch_to(cbpOriginal);
           }
           break;

       case SDTYPE(OP_COMPOSE_TYPE):
       case SDTYPE(OP_COMPOSE_DECL):
           {
               int vos, dos;
               TYPOBJECT *vosp;
               DECLOBJECT *dosp;

               ensureoneoperand();

               /* skip down over DECL objects to */
               /* find last TYP object on stack */

               dos = tos;
               dosp = (DECLOBJECT *) tosp;
               while (dos >= 1 && dosp->kind == DECLOBJECTKIND)
               {
                   dos--;
                   dosp = (DECLOBJECT *) &Estack[dos];
               }

               /* previous object must be a TYP */
               vos = dos;
               vosp = (TYPOBJECT *) dosp;
               if (vosp->kind != TYPOBJECTKIND) bug(125);

#if 0
               /* now work upward through all DECL objects, */
               /* composing them with the TYP object...     */
               while (dos < tos)
               {
                   dos++;
                   dosp = (DECLOBJECT *) &Estack[dos];
                   vosp->pType = compose_type(vosp->pType, dosp->type,
                                              dosp->dimension);
               }

               // finalize the type (only need to do for OP_COMPOSE_DECL???)
               vosp->pType = finalize_type(vosp->pType);

               //finally, pop the DECL objects off the stack
               //leaving just the final composed TYP object.
               while (tos > vos) pop();
#else
               /* now work backward through all DECL objects, */
               /* composing them with the TYP object...     */
               while (tos > vos)
               {
                   vosp->pType = compose_type(vosp->pType,
                                              ((DECLOBJECT *)tosp)->type,
                                              ((DECLOBJECT *)tosp)->dimension);
                   pop();
               }

               // finalize the type (only need to do for OP_COMPOSE_DECL???)
               //if (op == SDTYPE(OP_COMPOSE_DECL)) {
                   vosp->pType = finalize_type(vosp->pType);
               //}
#endif
           }
           break;

       case SDTYPE(OP_ENTER):
           {
               pfm_OBJ pType;
               //tos is TYP, nos is ID
               ensuretwooperands();
               if (TOSOBJECT.kind != TYPOBJECTKIND ||
                   NOSOBJECT.kind != IDOBJECTKIND) {
                   bug(9988);
               }
               pType = TOSTYP.pType;
               pVar = enter_symbol(NOSID.charp, fmVAR, pType);

               //pop the TYP and change the ID to a VAR object
               pop();
			   //Much of the following code is replicated from the code
			   //following the call to lookup_symbol() in lookupid above.
			   //It MUST be kept in sync with that code! Better yet, they
			   //both should call a function which contains this common
			   //code.  Or maybe the code here can be replaced by a call
			   //to lookupid???
               TOSOBJECT.kind = VAROBJECTKIND;
               if (pVar->am == amSTATIC) {
                   TOSVAR.frame = ebx;
#if 0
                   // The ability to redirect init code generation to the
                   // Init Code buffer is currently unused.  Might be used
                   // again someday to direct init code to a Super Init Code
                   // buffer, which is executed *prior* to the Init Code
                   // buffer to ensure that vars are always inited before
                   // use (which requires limiting inits in the Super Init
                   // buffer to simple inits such as constant assignments,
                   // in order to assure there are no forward-referencing
                   // conflicts, etc.).
                   if (cbpOriginal != &cbInitCode) {
                       //ensure code gets generated in the Init Code buffer...
                       EMIT0("<Switch to Init Code>");
                       X86_switch_to(&cbInitCode);
                   }
#endif
               }
               else {
                   TOSVAR.frame = ebp;  //assume amSTACK for now...
               }
               TOSVAR.offset = pVar->offset;
               TOSVAR.extradata = 0;  // holds nargs for funcs
			   TOSVAR.nbytes = pVar->size;
			   TOSVAR.namep = pVar->name;
			   TOSVAR.pSym = pVar;
			   TOSVAR.pType = pType;
			   //Turn off any 'const' attribute in this newly declared VAR,
			   //so it can be initialized...
			   TOSVAR.flags = (pVar->flags & ~fmfCONST);
               TOSVAR.typeindex = Mdg_current_default_type_index; // always int for now. XXX
               if (pType->what == fmSCALAR) {
                   if (pType->basetype == btINTEGRAL) {
                       TOSVAR.typeindex = Mdg_current_default_type_index;	//already set??
					   //use the new fmfNONNEG flag...
					   if (pVar->flags & fmfNONNEG) {
						   //set nbytes negative to flag an unsigned type
						   TOSVAR.nbytes = -TOSVAR.nbytes;		//<==TIDY UP LATER, just use fmfNONNEG
					   }
				   }//btINTEGRAL
                   else if (pType->basetype == btREAL) {
                       TOSVAR.typeindex = 126;
                   }
                   else bug(9966); //for now
               }
               else if (pType->what == fmARRAY) {
                   TOSVAR.typeindex = 4321; //for now
               }
               else if (pType->what == fmPOINTER) {
                   TOSVAR.typeindex = 0; //for now
				   TOSVAR.nbytes = -TOSVAR.nbytes;	//pointers are UNSIGNED
               }
               else bug(9977); //for now
			   //assume we'll need the var address for an init...
			   //could avoid via lazy generation??
               EMIT5("lea r%d, [%s%+d] ;&%s",
                     tos,
					 (X64gen?X86_reg64_name:X86_reg32_name)(TOSVAR.frame), //neat trick!!
					 TOSVAR.offset, pVar->name);
               X86_emit_op_TM(LEA_V, tos, TOSVAR.frame, no_reg, 0, TOSVAR.offset);
           }
           break;

       case SDTYPE(OP_BEGIN_SCOPE):
           enter_scope();
           break;

       case SDTYPE(OP_END_SCOPE):
			cbLocals = leave_scope();
			cbp->func.cbLocals = max(cbp->func.cbLocals, cbLocals);
           break;

       case SDTYPE(OP_BEGIN_INIT):
           {    // Beginning an <initializer>:
                // If this is a static non-global (level 0) variable, then
                // we need to emit 1st-time code to skip the init
                // code except the very first time
                ensureoneoperand();
                assert(TOSOBJECT.kind == VAROBJECTKIND);
                if (TOSVAR.frame == ebx && !(pVar->flags & fmfGLOBAL)) {
                    //static non-Global var
#ifdef X64
					//For X64, smallest allocation unit is one byte; associated bit offset is
					//offset of the first byte of each global/static var within user_vars
                    int bit_offset = (TOSVAR.offset - offsetof(FMcontext, globvars.user_vars));
#else
					//Prior to X64, smallest allocation unit was one word (4 bytes) per var
                    int bit_offset = (TOSVAR.offset - offsetof(FMcontext, globvars.user_vars))/4;
#endif
                    //compute offset and bit mask into user_vars_init_flags
                    int byte_offset = offsetof(FMcontext, globvars.user_vars_init_flags) + (bit_offset >> 3);
                    int bit_mask = 1 << (bit_offset & 7);
                    cglSkip = X86_genlab();
                    EMIT3("test [rBX%+d], %d ;user_vars_init_flags", byte_offset, bit_mask);
                    EMIT2("jnz cgl%d ;skip init", cglSkip);
                    EMIT3("or [rBX%+d], %d ;set flag bit", byte_offset, bit_mask);
                    X86_emit_op_M_imm8(TESTI_B, ebx, no_reg, 0, byte_offset, (BYTE)bit_mask);
                    X86_emit_op_lab(JNZ, cglSkip); //could probably use X86_emit_op_lab_disp8 ???
                    X86_emit_op_M_imm8(ORI_B, ebx, no_reg, 0, byte_offset, (BYTE)bit_mask);
                }
                else {
                   cglSkip = -1;
                }
           }
           break;

       case SDTYPE(OP_END_INIT):
           {   // End of the <initializer>:
               // Branch here if skipping around 1st-time init code.
               if (cglSkip != -1) {
                   EMIT2L("cgl%d:    ;skip init", cglSkip);
                   X86_deflab(cglSkip);
               }
           }
           break;

       default:
           bug(4);
           break;
     } /*switch*/
 } /*performoperation*/


        /************************************/
        /*                                  */
        /*                                  */
        /*          push$decl$item          */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void pushdeclitem(UINT1 type)
 {
     if (type == SDTYPE(DCL_ARRAYDIM)) {
         // DCL_ARRAYDIM takes one integer arg (the array dimension)
         if (TOSOBJECT.kind == LITOBJECTKIND && TOSLIT.type == LITTYPEINTEGER) {
             INT4 dim;
             ensureoneoperand();
             dim = integervalue(tosp);
             // Change tos to a DECL object.
             TOSOBJECT.kind = DECLOBJECTKIND;
             TOSDCL.type = type;
             TOSDCL.dimension = dim;
         }
         else
             Mde_abort_command(MD386__BADMIX, 0, 0, 0);
     }
     else {
         push(DECLOBJECTKIND);
         TOSDCL.type = type;
         TOSDCL.dimension = 0;
     }
 } /*pushdeclitem*/

        /************************************/
        /*                                  */
        /*                                  */
        /*       release_CB_resources       */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 static void release_CB_resources(void)
 {
     // Call X86_emit_free_resources() for each code buffer to release
     // the CGL's and any other resources that were acquired.
     X86_emit_free_resources(&cbDivCheck);
     X86_emit_free_resources(&cbZeroDivide);
     X86_emit_free_resources(&cbDivideOverflow);
     X86_emit_free_resources(&cbOnCtl);
     X86_emit_free_resources(&cbOnFilterStart);
     X86_emit_free_resources(&cbOnFilterEnd);
     X86_emit_free_resources(&cbInitCode);
     X86_emit_free_resources(&cbForEveryTile);
     X86_emit_free_resources(&cbUserFunctions);
     X86_emit_free_resources(&cbChannelCode);
     X86_emit_free_resources(&cbSaveChannelCode);
     X86_emit_free_resources(&cbCodeBuffer);
 } /*release_CB_resources*/

        /************************************/
        /*                                  */
        /*                                  */
        /*         MDC_EVALUATE_SDL         */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 void *Mdc_evaluate_sdl(void *sdlp, int compile_flag_local)
 {
      int nBytes;

     /*---------------------------------------------*/
     /* PLMC32-GENERATED CODE:                      */
     /*  Make module-level copies of any formal     */
     /*  parameters referenced by inner procedures. */
     /*---------------------------------------------*/
     M3__sdlp = sdlp;
     compile_flag = compile_flag_local;

     /************* Main-line code for MDC_EVALUATE_SDL ******************/

     if (!compile_flag) {
#ifdef X64
#ifdef _WIN64
		 if (X86gen) {
			 ErrorOk("Cannot execute X32 code on Win64.");
			 return NULL;
		 }
#else //_WIN32
		 if (X64gen) {
			 ErrorOk("Cannot execute X64 code on Win32.");
			 return NULL;
		 }
#endif //_WIN32
#endif //X64
         //execute the compiled code...
         if (setjmp(Mde_error_exit_jb)) {
             //someone longjmp'ed back to us...
             return NULL;
         }
         X86_call_fmf1(gParams->codeBuffer, (UINT_PTR)gFmcp);
         return NULL;
     }

     if (setjmp(Mde_error_exit_jb)) {
         //someone longjmp'ed back to us...
         gParams->fValidCompile = FALSE;    //redundant???
         reset_symbol_table();
         //Clean up any resources acquired by the code buffers...
         release_CB_resources();
         return NULL;
     }

	 // Initialize prior to compilation...

	 prog_type = PROG_UNKNOWN;	//program type
	 nProgRevMajor = INT_MAX;	//program major revision level
	 nProgRevMinor = INT_MAX;	//program minor revision level


	 X64gen = X64GEN;
	 X86gen = X86GEN;

	 // following must be (re)set at start of each compile
	 gFmcp->FMC_target = X64GEN ? 64 : 32;
	 gFmcp->abort_mode = -1;	//use default for platform

#ifdef X64
//Should pre-compute isReg at start of compilation???
//E.g. precompute int X64gen = X64GEN at start of compilation, and
//int nTmps = X64gen?(sizeof(tmpRegID)/sizeof(tmpRegID[0])):2, then define isReg as
///#define isReg(tmp) ((tmp) < nTmps)
	 if (X64gen) {
		 // Compiling for X64...
#ifndef _WIN64
		 //Host must be Win32...
		 Warn("Cannot execute 64-bit code in a 32-bit host app.");
#endif
		 nTmps = (sizeof(tmpRegID)/sizeof(tmpRegID[0]));
		 nVolatileTmps = 2; //Could use R10 and R11 as volatile temps,
							//which must be "shrink-wrap" saved across
							//function calls.
		 // Precomputed mnemonics for variable word size opcodes,
		 // defined here as 64-bit.
		 LEA_V	 = LEA_Q;		//64-bit Load Effective Address
		 ADDR_V  = ADDR_Q;
		 ADD_V	 = ADD_Q;
		 ADDI_V	 = Q(ADDI_D);
		 ORR_V	 = ORR_Q;
		 OR_V	 = OR_Q;
		 ADCR_V	 = ADCR_Q;
		 ADC_V	 = ADC_Q;
		 SBBR_V	 = SBBR_Q;
		 SBB_V	 = SBB_Q;
		 ANDR_V  = ANDR_Q;
		 AND_V	 = AND_Q;
		 ANDI_V	 = Q(ANDI_D);
		 SUBR_V  = SUBR_Q;
		 SUB_V	 = SUB_Q;
		 SUBI_V	 = Q(SUBI_D);
		 XORR_V  = XORR_Q;
		 XOR_V	 = XOR_Q;
		 CMPR_V  = CMPR_Q;
		 CMP_V	 = CMP_Q;
		 MOVR_V  = MOVR_Q;
		 MOV_V	 = MOV_Q;
		 XCHG_V  = XCHG_Q;
		 TEST_V  = TEST_Q;
		 INC_V	 = INC_Q;
		 DEC_V	 = DEC_Q;
		 IMUL32_V= IMUL32_Q;
		 IMULI_V = Q(IMULI_D);
		 IDIV_V	 = Q(IDIV_D);
		 CDQ_V	 = Q(CDQ);
		 ROL1_V	 = Q(ROL1_D);
		 ROR1_V  = Q(ROR1_D);
		 SHL1_V	 = Q(SHL1_D);
		 SHLI_V	 = Q(SHLI_D);
		 SAR1_V	 = Q(SAR1_D);
		 SARI_V	 = Q(SARI_D);
		 FILD_V  = FILD_L;
		 FISTP_V = FISTP_L;
		 NEG_V	 = Q(NEG_D);
		 NOT_V	 = Q(NOT_D);
		 MOVSXD_V= MOVSXD_Q;
		 MOVSXB_V= MOVSXB_Q;
		 MOVSXW_V= MOVSXW_Q;
		 SHL_V	 = Q(SHL_D);
		 SHR_V	 = Q(SHR_D);
		 CMPI_V	 = Q(CMPI_D);
	 }
	 else
#endif //X64
	 {	 // Compiling for X86...
#ifdef _WIN64
		 //Host is Win64...
		 Warn("Cannot execute 32-bit code in a 64-bit host app.");
#endif
		 nTmps = 2;  //esi and edi are the only two temps... (or edi and esi???)
		 nVolatileTmps = 0;
		 // Precomputed mnemonics for variable word size opcodes,
		 // defined here as 32-bit.
		 LEA_V	 = LEA;		//32-bit Load Effective Address
		 ADDR_V	 = ADDR_D;
		 ADD_V	 = ADD_D;
		 ADDI_V	 = ADDI_D;
		 ORR_V	 = ORR_D;
		 OR_V	 = OR_D;
		 ADCR_V	 = ADCR_D;
		 ADC_V	 = ADC_D;
		 SBBR_V  = SBBR_D;
		 SBB_V	 = SBB_D;
		 ANDR_V  = ANDR_D;
		 AND_V	 = AND_D;
		 ANDI_V	 = ANDI_D;
		 SUBR_V  = SUBR_D;
		 SUB_V	 = SUB_D;
		 SUBI_V  = SUBI_D;
		 XORR_V  = XORR_D;
		 XOR_V	 = XOR_D;
		 CMPR_V  = CMPR_D;
		 CMP_V	 = CMP_D;
		 MOVR_V  = MOVR_D;
		 MOV_V	 = MOV_D;
		 XCHG_V  = XCHG_D;
		 TEST_V  = TEST_D;
		 INC_V	 = INC_D;
		 DEC_V	 = DEC_D;
		 IMUL32_V= IMUL32;
		 IMULI_V = IMULI_D;
		 IDIV_V	 = IDIV_D;
		 CDQ_V	 = CDQ;
		 ROL1_V  = ROL1_D;
		 ROR1_V  = ROR1_D;
		 SHL1_V	 = SHL1_D;
		 SHLI_V	 = SHLI_D;
		 SAR1_V	 = SAR1_D;
		 SARI_V	 = SARI_D;
		 FILD_V  = FILD_I;
		 FISTP_V = FISTP_I;
		 NEG_V	 = NEG_D;
		 NOT_V	 = NOT_D;
		 MOVSXD_V= MOVSXD_D; //(=MOV_D)
		 MOVSXB_V= MOVSXB_D;
		 MOVSXW_V= MOVSXW_D;
		 SHL_V	 = SHL_D;
		 SHR_V	 = SHR_D;
		 CMPI_V	 = CMPI_D;
	 } //X86

	 init_arglist();

     gFmcp->ndlits = 0;
     gFmcp->need_iuv = FALSE;
     gFmcp->need_d = FALSE;
     gFmcp->need_m = FALSE;
     gFmcp->need_tbuf = FALSE;
     gFmcp->need_t2buf = FALSE;
	 gFmcp->need_t3buf = FALSE;
	 gFmcp->need_t4buf = FALSE;
     gFmcp->needPremiere = FALSE;
     gFmcp->needPadding = 0;
     gFmcp->bandWidth = 100;
     gFmcp->isTileable = TRUE;  //optimistic
     gParams->flags |= AFH_IS_TILEABLE;
     gParams->filterAboutText[0] = '\0';
	 if (gParams->filterCategoryText[0] == '\0') {
		//use default Category
		strcpy(gParams->filterCategoryText, "My Category");
	 }
     strcpy(gFmcp->filterCategoryText, gParams->filterCategoryText);

	 if (gParams->filterTitleText[0] == '\0') {
		 //use default Title
		 strcpy(gParams->filterTitleText, "My Filter");
	 }
     strcpy(gFmcp->filterTitleText, gParams->filterTitleText);

	 // Should we also set the default value of filterFilenameText here?
     if (gParams->filterFilenameText[0] == '\0') {
        // Supply a default filter module name...
        strcpy(gParams->filterFilenameText, "MyFilter.8bf");
     }
     strcpy(gFmcp->filterFilenameText, gParams->filterFilenameText);

     {//scope
         //Create a random number!!!
         int i;

         srand((unsigned)time(NULL));
         for (i=0;i<36;i++){
            gParams->filterUniqueID[i] = '0' + (rand() % 10); //Values from 0 to 9
         }
         gParams->filterUniqueID[36] = 0; 
     }//scope
     strcpy(gFmcp->filterUniqueID, gParams->filterUniqueID);

     gParams->embeds[0] = 0;
     gParams->cbEmbeds = 1;
     gParams->gCtlImagesSet = FALSE;
     gParams->gImagesEmbedded = 0;
     gParams->gWavesSpecified = FALSE;
     gParams->gWavesEmbedded = 0;
     gParams->gUserFunctions = 0;

     //Really only need to do the following once?  Well, it is needed now for X64!!
     X86_emit_init(&cbDivCheck, gParams->CB_divCheck, sizeof(gParams->CB_divCheck),
              "Default OnDivideCheck");
     X86_emit_divCheck();
     X86_emit_finish();

     //Default OnZeroDivide handler...
     X86_emit_init(&cbZeroDivide, gParams->CB_onZeroDivide, sizeof(gParams->CB_onZeroDivide),
                   "Default OnZeroDivide");
     X86_emit_defaultHandler();
     X86_emit_finish();

     //Default OnDivideOverflow handler...
     X86_emit_init(&cbDivideOverflow, gParams->CB_onDivideOverflow,
                   sizeof(gParams->CB_onDivideOverflow), "Default OnDivideOverflow");
     X86_emit_defaultHandler();
     X86_emit_finish();

     //Default OnCtl handler...
     X86_emit_init(&cbOnCtl, gParams->CB_onCtl, sizeof(gParams->CB_onCtl),
                   "Default OnCtl");
     X86_emit_defaultHandler();
     X86_emit_finish();

     //Default OnFilterStart handler...
     X86_emit_init(&cbOnFilterStart, gParams->CB_onFilterStart,
                   sizeof(gParams->CB_onFilterStart), "Default OnFilterStart");
     X86_emit_defaultHandler();
     X86_emit_finish();

     //Default ForEveryTile handler...
     X86_emit_init(&cbForEveryTile, gParams->CB_forEveryTile, sizeof(gParams->CB_forEveryTile),
                   "Default ForEveryTile");
     X86_emit_defaultHandler();
     X86_emit_finish();

     //Default OnFilterEnd handler...
     X86_emit_init(&cbOnFilterEnd, gParams->CB_onFilterEnd, sizeof(gParams->CB_onFilterEnd),
                   "Default OnFilterEnd");
     X86_emit_defaultHandler();
     X86_emit_finish();

     X86_emit_init(&cbUserFunctions, gParams->CB_UserFunctions, sizeof(gParams->CB_UserFunctions),
				   "UserFunctions");

     X86_emit_init(&cbChannelCode, CB_channelCode, sizeof(CB_channelCode), "Channel");

     X86_emit_init(&cbInitCode, gParams->CB_initCode, sizeof(gParams->CB_initCode),
                   "Initialization");

	 X86_emit_prolog(nTemps, cbLocals);

     tos = -1;
     nextavailablebyte = 0;
     if_depth = 0;

     init_loop_context();
     init_symbol_table();

#if 1	//for now, check to see if it matters...
	 //reset the ndp stack slot counter
	 check_ndp_stack_room(MAX_NDP_SLOTS+1);
#endif

     eosdl = (SDLITEM.kind <= 0 || SDLITEM.kind > SDKIND(LAST_VALID_SD));

     while (!eosdl)
     {
         switch (SDLITEM.kind)
         {
           /* 1 - ID item */
           case SDKIND(SD_ID):
#ifdef FML
             X86_warn("SD_ID is deprecated");
#endif
             pushnameitem(IDOBJECTKIND);
             break;

           /* 2 - literal item */
           case SDKIND(LIT_INTEGER):
             pushlititem();
             break;

           /* 3 - operator */
           case SDKIND(OP_ADD):
             performoperation();
             break;

           /* 4 - type item */
           case SDKIND(TYP_BYTE):
             settype();
             break;

           /* 5 - register item */
           case SDKIND(SD_REG):
             pushregitem();
             break;

           /* 6 - %LINE item */
           case SDKIND(SD_LINE):
             pushlinitem();
             break;

           /* 7 - %LABEL item */
           case SDKIND(SD_LABEL):
             pushlabitem();
             break;

           /* 8 - %SEGMENT item */
           case SDKIND(SD_SEGMENT):
             pushnameitem(SEGOBJECTKIND);
             break;

           /* 9 - %GROUP item */
           case SDKIND(SD_GROUP):
             pushnameitem(GRPOBJECTKIND);
             break;

           /* 10 - %COMMON item */
           case SDKIND(SD_COMMON):
             Mde_abort_command(MD386__NYI, 29, 0, 0);
             break;

           /* 11 - radix item */
           case SDKIND(SD_RADIX):
             setradix();
             break;

           /* 12 - mode item */
           case SDKIND(SD_MODE):
             setmode();
             break;

           /* 13 - file item */
           case SDKIND(SD_FILE):
             pushfilitem();
             break;

           /* 14 - current program counter ($) */
           case SDKIND(SD_PC):
             pushpcitem();
             break;

           /* 15 - last examine cell (.) */
           case SDKIND(SD_LAST):
             pushlastitem();
             break;

           /* 16 - next examine cell ( ) */
           case SDKIND(SD_NEXT):
             pushnextitem();
             break;

           /* 18 - FML declarations */
           case SDKIND(DCL_TYPEDEF):
               pushdeclitem(SDLITEM.type);
               break;

           /* 19 - Literal ID item */
           case SDKIND(SD_IDLIT):
               pushidlititem(IDOBJECTKIND);
               break;

		   /* 20 - Program type and revision */
		   case SDKIND(SD_PROG):
			   prog_type = SDLITEM.param[0];		//program type
			   nProgRevMajor = SDLITEM.param[1];	//program major revision level
			   nProgRevMinor = SDLITEM.param[2];	//program minor revision level
			   break;

           /* ??? - unknown kind of semantic descriptor */
           default:
               bug(12);
               break;
         } /*switch*/
         
         advance_sdlp();
     }

     /* Make sure then-else-if's were properly nested... */
     if (if_depth != 0) bug(14);

     X86_switch_to(&cbInitCode);    //should already be here...
     X86_emit_epilog(nTemps, cbLocals, -1);	//nArgs = 1 vpl
     X86_emit_finish();

     X86_switch_to(&cbChannelCode);
     nBytes = X86_emit_finish();

     //We do this last, after flags such as need_iuv have been set,
     //so that code generated for doOneRow can vary depending on
     //these flags.
     X86_emit_init(&cbCodeBuffer, gParams->codeBuffer, sizeof(gParams->codeBuffer), "Final");
     X86_emit_doOneRow_part1(nTemps, cbLocals);
     X86_append_cb(&cbChannelCode);
     X86_emit_doOneRow_part2(nTemps, cbLocals);
     nBytes = X86_emit_finish();
     //X86_delete(&cbChannelCode);

     reset_symbol_table();

#ifdef FULL
     if ((gParams->flags & (AFH_EDITMODE|AFH_ADVANCED)) == (AFH_EDITMODE|AFH_ADVANCED)) {
         //disassemble only in advanced edit mode...
#if 1
         //disassemble the UserFunctions code buffer
         if (gParams->flags & AFH_DISASSEMBLE)
         {
#ifdef _WIN64
#if 0
             PRINTF2("\r\nDisassembly of UserFunctions code (%d bytes):\r\n",
                     cbUserFunctions.pc);
#else //debugging
             PRINTF3("\r\nDisassembly of UserFunctions code (%d bytes at 0x%p):\r\n",
                     cbUserFunctions.pc, cbUserFunctions.pcStart);
#endif
#else //Win32
             PRINTF2("\r\nDisassembly of UserFunctions code (%d bytes):\r\n",
                     cbUserFunctions.pc - cbUserFunctions.pcStart);
#endif
             X86_disassemble(&cbUserFunctions);
         }
#endif

#if 1
         //disassemble the DIVCHK code
         if (gParams->flags & AFH_DISASSEMBLE)
         {
#ifdef _WIN64
             PRINTF2("\r\nDisassembly of DIVCHK code (%d bytes):\r\n",
                     cbDivCheck.pc);
#else //Win32
             PRINTF2("\r\nDisassembly of DIVCHK code (%d bytes):\r\n",
                     cbDivCheck.pc - cbDivCheck.pcStart);
#endif
             X86_disassemble(&cbDivCheck);
         }
#endif

#if 1
         //disassemble the OnZeroDivide handler code
         if (gParams->flags & AFH_DISASSEMBLE)
         {
#ifdef _WIN64
             PRINTF2("\r\nDisassembly of OnZeroDivide handler code (%d bytes):\r\n",
                     cbZeroDivide.pc);
#else //Win32
             PRINTF2("\r\nDisassembly of OnZeroDivide handler code (%d bytes):\r\n",
                     cbZeroDivide.pc - cbZeroDivide.pcStart);
#endif
             X86_disassemble(&cbZeroDivide);
         }
#endif

#if 1
         //disassemble the OnDivideOverflow handler code
         if (gParams->flags & AFH_DISASSEMBLE)
         {
#ifdef _WIN64
             PRINTF2("\r\nDisassembly of OnDivideOverflow handler code (%d bytes):\r\n",
                     cbDivideOverflow.pc);
#else //Win32
             PRINTF2("\r\nDisassembly of OnDivideOverflow handler code (%d bytes):\r\n",
                     cbDivideOverflow.pc - cbDivideOverflow.pcStart);
#endif
             X86_disassemble(&cbDivideOverflow);
         }
#endif

#if 1
         //disassemble the OnCtl handler code
         if (gParams->flags & AFH_DISASSEMBLE)
         {
#ifdef _WIN64
             PRINTF2("\r\nDisassembly of OnCtl handler code (%d bytes):\r\n",
                     cbOnCtl.pc);
#else //Win32
             PRINTF2("\r\nDisassembly of OnCtl handler code (%d bytes):\r\n",
                     cbOnCtl.pc - cbOnCtl.pcStart);
#endif
             X86_disassemble(&cbOnCtl);
         }
#endif

#if 1
         //disassemble the OnFilterStart handler code
         if (gParams->flags & AFH_DISASSEMBLE)
         {
#ifdef _WIN64
             PRINTF2("\r\nDisassembly of OnFilterStart handler code (%d bytes):\r\n",
                     cbOnFilterStart.pc);
#else //Win32
             PRINTF2("\r\nDisassembly of OnFilterStart handler code (%d bytes):\r\n",
                     cbOnFilterStart.pc - cbOnFilterStart.pcStart);
#endif
             X86_disassemble(&cbOnFilterStart);
         }
#endif

#if 1
         //disassemble the OnFilterEnd handler code
         if (gParams->flags & AFH_DISASSEMBLE)
         {
#ifdef _WIN64
             PRINTF2("\r\nDisassembly of OnFilterEnd handler code (%d bytes):\r\n",
                     cbOnFilterEnd.pc);
#else //Win32
             PRINTF2("\r\nDisassembly of OnFilterEnd handler code (%d bytes):\r\n",
                     cbOnFilterEnd.pc - cbOnFilterEnd.pcStart);
#endif
             X86_disassemble(&cbOnFilterEnd);
         }
#endif

#if 1
     //disassemble the init code buffer
     if (gParams->flags & AFH_DISASSEMBLE)
     {
#ifdef _WIN64
         PRINTF2("\r\nDisassembly of init code buffer (%d bytes):\r\n",
                 cbInitCode.pc);
#else //Win32
         PRINTF2("\r\nDisassembly of init code buffer (%d bytes):\r\n",
                 cbInitCode.pc - cbInitCode.pcStart);
#endif
         X86_disassemble(&cbInitCode);
     }
#endif

#if 1
     //disassemble the ForEveryTile code buffer
     if (gParams->flags & AFH_DISASSEMBLE)
     {
#ifdef _WIN64
         PRINTF2("\r\nDisassembly of ForEveryTile code buffer (%d bytes):\r\n",
                 cbForEveryTile.pc);
#else //Win32
         PRINTF2("\r\nDisassembly of ForEveryTile code buffer (%d bytes):\r\n",
                 cbForEveryTile.pc - cbForEveryTile.pcStart);
#endif
         X86_disassemble(&cbForEveryTile);
     }
#endif

#if 1
     //disassemble the channel code buffer
     if (gParams->flags & AFH_DISASSEMBLE)
     {
         PRINTF2("\r\nDisassembly of channel code buffer (%d bytes):\r\n", nBytes);
         X86_disassemble(&cbCodeBuffer);
     }
#endif
     } //if AFH_EDITMODE|AFH_ADVANCED
#endif //ifdef FULL

     // Now call release_CB_resources() to release the CGL's for each
     // code buffer.  (X86_emit_finish() couldn't do it, because the
     // CGL's were needed by the disassemblies above.)
     release_CB_resources();

     return &Estack[tos];
 } /*Mdc_evaluate_sdl*/


        /************************************/
        /*                                  */
        /*                                  */
        /*       MDC_EXECUTE_INIT_CODE      */
        /*                                  */
        /*                                  */
        /************************************/

 void Mdc_execute_init_code(void)
 {
     //execute init code
     INT_PTR res;

#ifdef X64
#ifdef _WIN64
	 if (X86gen) {
		 if ((gParams->flags & (AFH_EDITMODE|AFH_ADVANCED)) == (AFH_EDITMODE|AFH_ADVANCED)) {
			 //Only display when "Initialize" pressed in Advanced Edit mode.
			 ErrorOk("Cannot execute X32 init code on Win64.");
		 }
		 return;
	 }
#else //_WIN32
	 if (X64gen) {
		 if ((gParams->flags & (AFH_EDITMODE|AFH_ADVANCED)) == (AFH_EDITMODE|AFH_ADVANCED)) {
			 //Only display when "Initialize" pressed in Advanced Edit mode.
			 ErrorOk("Cannot execute X64 init code on Win32.");
		 }
		 return;
	 }
#endif //_WIN32
#endif //X64

     if (setjmp(Mde_error_exit_jb)) {
         //someone longjmp'ed back to us...
         return;
     }

#ifdef FULL
     if ((gParams->flags & (AFH_EDITMODE|AFH_ADVANCED)) == (AFH_EDITMODE|AFH_ADVANCED)) {
        PRINTF("\r\nExecuting init code...\r\n");
     }
#endif

	 ///Info("About to call CB_initCode from Mdc_execute_init_code");
     res = X86_call_fmf1(gParams->CB_initCode, (INT_PTR)gFmcp);

#ifdef FULL
     if ((gParams->flags & (AFH_EDITMODE|AFH_ADVANCED)) == (AFH_EDITMODE|AFH_ADVANCED)) {
         PRINTF2("result = %Id\r\n", res);
     }
#endif
 } /*Mdc_execute_init_code*/


        /************************************/
        /*                                  */
        /*                                  */
        /*         MDC_DUMP_ESTACK          */
        /*                                  */
        /*                                  */
        /************************************/

 /**procedure**/
 void Mdc_dump_estack(void *sdlp)
 {
     GENERALOBJECT  *topp;
#ifdef FULL
     INT4           tos;
     void           *objectp;
     #define OBJECT (*(GENERALOBJECT *) objectp)
     #define VAR    (*(VAROBJECT *) objectp)
     #define ID     (*(IDOBJECT  *) objectp)
     #define LIT    (*(LITOBJECT *) objectp)
     #define REG    (*(REGOBJECT *) objectp)
     #define DECL   (*(DECLOBJECT *) objectp)
     #define TYP    (*(TYPOBJECT *) objectp)
#endif

#ifdef FULL
     if ((gParams->flags & (AFH_EDITMODE|AFH_ADVANCED)) == (AFH_EDITMODE|AFH_ADVANCED)) {
        PRINTF("-- Evaluation of SDL --\r\n");
     }
#endif

     topp = Mdc_evaluate_sdl(sdlp, TRUE);

#ifdef FULL
#if 1
     if ((gParams->flags & (AFH_EDITMODE|AFH_ADVANCED|AFH_DUMP_SDL)) ==
                           (AFH_EDITMODE|AFH_ADVANCED|AFH_DUMP_SDL)) {
     for (tos = 0; &Estack[tos] <= topp; tos++)
     {
         objectp = &Estack[tos];
         switch (OBJECT.kind) {
         case IDOBJECTKIND:
            {
                PRINTF2("ID: \"%s\"", ID.charp);
            }
            break;
         case VAROBJECTKIND:
            {
             PRINTF("VAR: ");
             PRINTF2("type=%d, ", VAR.typeindex);
             PRINTF2("addr=%#x", VAR.offset);
            }
            break;
         case REGOBJECTKIND:
            {
             PRINTF("REG: ");
             PRINTF2("index=%d", REG.regindex);
            }
            break;
         case LITOBJECTKIND:
            {
             UINT1    *valuep;
             PRINTF("LIT: ");
             valuep = LIT.valuep;
             if (LIT.type == LITTYPEINTEGER)
             {
                 PRINTF2("(int*%d) ", LIT.nbytes);
                 if (LIT.nbytes <= 0) 
                 {   /* ??? nbytes <= 0 ??? */
                     PRINTF("???");
                 }
                 else
                 {
                     INT4     tempvalue;
                     int      i;
                     PRINTF("0x");
                     for (i = LIT.nbytes-1; i >= 0; i--) {
                         PRINTF2("%2.2x", valuep[i]);
                     }
                     memcpy(&tempvalue, LIT.valuep, 4);
                     /* WARNING!!!  Depending on >> to sign-extend a */
                     /* signed value is NON-PORTABLE!!!              */
                     if (LIT.nbytes == 1)
                     {
                         tempvalue = tempvalue << 24 >> 24;
                         PRINTF2(" = %d", tempvalue);
                     }
                     else if (LIT.nbytes == 2)
                     {
                         tempvalue = tempvalue << 16 >> 16;
                         PRINTF2(" = %d", tempvalue);
                     }
                     else if (LIT.nbytes == 3)
                     {
                         tempvalue = tempvalue << 8 >> 8;
                         PRINTF2(" = %d", tempvalue);
                     }
                     else if (LIT.nbytes == 4)
                     {
                         PRINTF2(" = %d", tempvalue);
                     }
                     else if (LIT.nbytes > 4)
                     {   // if the high-order bytes are filled with the sign byte
                         // then we can still convert correctly with %d
                         UINT1 signByte;
                         signByte = tempvalue < 0 ? 0xff : 0;
                         for (i = LIT.nbytes-1; i >=4 && valuep[i]==signByte; i--)
                             ;
                         if (i < 4)
                             PRINTF2(" = %d", tempvalue);
                     }
                 }
             }
             else if (LIT.type == LITTYPELOGICAL)
             {
                 PRINTF2("(bool*%d) ", LIT.nbytes);
                 if (valuep[0])
                     PRINTF("true");
                 else 
                     PRINTF("false");
             }
             else if (LIT.type == LITTYPEREAL)
             {
                 PRINTF2("(real*%d) ", LIT.nbytes);
                 if (LIT.nbytes == 4)
                     PRINTF2("%#g", (double)(*(float *)LIT.valuep));
                 else if (LIT.nbytes == 8)
                     PRINTF2("%#g", *(double *)LIT.valuep);
                 else 
                     PRINTF("illegal size for real");
             }
             else if (LIT.type == LITTYPECHAR)
             {
                 UINT4    charz;
                 INT4     i;

                 PRINTF2("(char*%d) ", LIT.nbytes);
                 PRINTF("\"");
                 for (i = 0;  i <= LIT.nbytes - 1;  i++)
                 {
                     charz = valuep[i];
                     if (charz == '\n')
                         PRINTF("\\n");
                     else if (charz == '\r')
                         PRINTF("\\r");
                     else if (charz == '\a')
                         PRINTF("\\a");
                     else if (charz == '\b')
                         PRINTF("\\b");
                     else if (charz == '\f')
                         PRINTF("\\f");
                     else if (charz == '\t')
                         PRINTF("\\t");
                     else if (charz == '\v')
                         PRINTF("\\v");
                     else if (charz == '\\')
                         PRINTF("\\\\");
                     else if (charz == '\'')
                         PRINTF("\\\'");
                     else if (charz == '\"')
                         PRINTF("\\\"");
                     else if (charz == '\?')
                         PRINTF("\\\?");
#if 0
                     else if (charz == '\0')
                         PRINTF("\\0");         //No! Can lead to ambiguities!
#endif
                     else if (charz < 0x20 || charz > 0x7E)
                     {
                         // Use octal escape: "\nnn"
                         PRINTF2("\\%3.3o", charz);
                     }
                     else 
                         PRINTF2("%c", charz);
                 }
                 PRINTF("\"");
             }
             else 
                 PRINTF2("??? unknown type=%d", LIT.type);
            }
            break;
         case LINOBJECTKIND:
             PRINTF("LIN: ");
             break;
         case SEGOBJECTKIND:
             PRINTF("SEG: ");
             break;
         case GRPOBJECTKIND:
             PRINTF("GRP: ");
             break;
         case FILOBJECTKIND:
             PRINTF("FIL: ");
             break;
         case MARKEROBJECTKIND:
             PRINTF("MARKER: ");
             break;
         case DECLOBJECTKIND:
             PRINTF("DECL: ");
             PRINTF2("type = %d ", DECL.type);
             PRINTF2("dimension = %d", DECL.dimension);
             break;
         case TYPOBJECTKIND:
             PRINTF("TYP: ");
             //PRINTF2("pType = %x", TYP.pType);
             display_OBJ(TYP.pType);
             break;
         default:
             PRINTF2("Unknown object kind: %d", OBJECT.kind);
             break;
         } /*switch*/
         PRINTF("\r\n");
     } /*for*/
    } //if AFH_EDITMODE|AFH_ADVANCED|AFH_DUMP_SDL
#endif
     if ((gParams->flags & (AFH_EDITMODE|AFH_ADVANCED)) == (AFH_EDITMODE|AFH_ADVANCED)) {
         PRINTF("--- End of Eval ---\r\n");     
     }
#endif //ifdef FULL

     #undef OBJECT
     #undef VAR
     #undef ID
     #undef LIT
     #undef REG
     #undef DECL
     #undef TYP
} /*Mdc_dump_estack*/


#if 0
        /************************************/
        /*                                  */
        /*                                  */
        /*        MDC_GET_ESTACK_VAL        */
        /*                                  */
        /*                                  */
        /************************************/

extern UINT4 SDL[];        /* The semantic descriptor list (sized elsewhere). */

 /**procedure**/
 int Mdc_get_estack_val(int index)
 {
     GENERALOBJECT  *topp;
     INT4           tos;
     void           *objectp;
     #define OBJECT (*(GENERALOBJECT *) objectp)
     #define LIT    (*(LITOBJECT *) objectp)

     topp = Mdc_evaluate_sdl(&SDL, FALSE);
     tos = topp - &Estack[0];

     if (index < tos) {
         objectp = &Estack[index];
         if (OBJECT.kind == LITOBJECTKIND && LIT.type == LITTYPEINTEGER)
         {
             if (LIT.nbytes <= 0) 
             {   /* ??? nbytes <= 0 ??? */
                 return 0;
             }
             else
             {
                 INT4     tempvalue;
                 memcpy(&tempvalue, LIT.valuep, 4);
                 /* WARNING!!!  Depending on >> to sign-extend a */
                 /* signed value is NON-PORTABLE!!!              */
                 if (LIT.nbytes == 1)
                 {
                     tempvalue = tempvalue << 24 >> 24;
                 }
                 else if (LIT.nbytes == 2)
                 {
                     tempvalue = tempvalue << 16 >> 16;
                 }
                 else if (LIT.nbytes == 3)
                 {
                     tempvalue = tempvalue << 8 >> 8;
                 }
                 return tempvalue;
             }
         }
     }
     return 0;  // not a valid value
     #undef OBJECT
     #undef LIT
 } /*Mdc_get_estack_val*/
#endif
